package com.android.launcher3.folder

import android.content.Context
import android.graphics.Rect
import android.util.AttributeSet
import android.view.View
import com.android.launcher3.BubbleTextView
import com.android.launcher3.CellLayout
import com.android.launcher3.FocusHelper
import com.android.launcher3.Folder
import com.android.launcher3.ItemInfo
import com.android.launcher3.LauncherAppState
import com.android.launcher3.ShortcutInfo
import com.teslacoilsw.launcher.NovaLauncher
import com.teslacoilsw.launcher.preferences.FolderWindowStyle
import timber.log.Timber
import java.util.*

private const val REORDER_ANIMATION_DURATION = 230
class FolderCellLayout @JvmOverloads constructor(
    context: Context, attrs: AttributeSet? = null, defStyleAttr: Int = 0
) : CellLayout(context, attrs, defStyleAttr), FolderContentView {

    @set:JvmName("_setFolder")
    private var folder: Folder? = null

    override fun setFolder(folder: Folder?) {
        this.folder = folder
    }

    override fun onFinishInflate() {
        super.onFinishInflate()
        shortcutsAndWidgets.isMotionEventSplittingEnabled = false
    }

    private fun placeInReadingOrder(items: List<ShortcutInfo>) {
        var maxH = 0
        val count = items.size
        for (i in 0 until count) {
            val item = items[i]
            val cellH = cellXToCellH(item.cellX)
            if (cellH > maxH) {
                maxH = cellH
            }
        }

        val gridComparator = GridComparator(maxH + 1)
        Collections.sort(items, gridComparator)
        val countH = countH
        for (i in 0 until count) {
            val h = i % countH
            val v = i / countH
            val item = items[i]
            item.cellX = cellHToCellX(h)
            item.cellY = cellVToCellY(v)
        }
    }

    override fun bindItems(sortedChildren: List<ShortcutInfo>) {

        placeInReadingOrder(sortedChildren)
        for (i in sortedChildren.indices) {
            val child = sortedChildren[i]
            val rank = rankFromCoordinates(cellXToCellH(child.cellX), cellYToCellV(child.cellY))
            createAndAddShortcut(child, rank)
        }
    }


    override fun createAndAddViewForRank(item: ShortcutInfo, rank: Int): View {
        val coords = coordinatesFromRank(null, rank)
        item.cellX = coords[0].toFloat()
        item.cellY = coords[1].toFloat()

        allocateSpaceForRank(rank)
        val icon = createAndAddShortcut(item, rank)
        return icon!!
    }

    fun allocateSpaceForRank(rank: Int) {
        val views = ArrayList(folder!!.itemsInReadingOrder)
        views.add(Math.min(rank, views.size), null)
        setupContentDimensions(views.size)
        arrangeChildren(views, views.size, false)
    }

    /**
     * Adds the {@param view} to the layout based on {@param rank} and updated the position
     * related attributes. It assumes that {@param item} is already attached to the view.
     */
    override fun addViewForRank(view: View, item: ShortcutInfo, rank: Int): View? {
        val pagePos = rank
        val pageNo = 0

        val (cellH, cellV) = coordinatesFromRank(null, pagePos)
        item.indexedRank = rank
        item.cellX = cellHToCellX(cellH)
        item.cellY = cellVToCellY(cellV)


        // We need to check here to verify that the given item's location isn't already occupied
        // by another item.
        if (getChildAt(item.cellX, item.cellY) != null || item.cellX < 0 || item.cellY < 0
            || item.cellX >= countH || item.cellY >= countV) {
            // This shouldn't happen, log it.
            Timber.e(
                Throwable(),
                "Folder order not properly persisted during bind %s %d %d %s [%s %s %s %s %s]",
                item, countH, countV, folder?.itemsInReadingOrder?.map { it.tag as ItemInfo }?.joinToString(),
                getChildAt(item.cellX, item.cellY),
                item.cellX < 0,
                item.cellY < 0,
                item.cellX >= countH,
                item.cellY >= countV
                )
            if (!findAndSetEmptyCells(item)) {
                return null
            }
        }

        val lp = view.layoutParams as CellLayout.LayoutParams
        lp.cellH = cellXToCellH(item.cellX)
        lp.cellV = cellYToCellV(item.cellY)
        val id = folder!!.mLauncher?.getViewIdForItem(item) ?: View.generateViewId()
        addViewToCellLayout(view, -1, id, lp, true)
        return view
    }

    private fun createAndAddShortcut(item: ShortcutInfo, rank: Int): View? {
        val launcher = folder!!.mLauncher!!

        val textView = launcher.createShortcut(this, item) as BubbleTextView

        textView.setOnClickListener(folder)
        textView.setOnLongClickListener(folder)
        textView.onFocusChangeListener = folder?.mFocusIndicatorHandler

        textView.setOnFlingViewListener(launcher as NovaLauncher)
        textView.setOnKeyListener(FocusHelper.folderKeyEventListener)

        return addViewForRank(textView, item, rank)
    }

    private fun findAndSetEmptyCells(item: ShortcutInfo): Boolean {
        val emptyCell = IntArray(2)
        if (findCellForSpan(emptyCell, 1, 1)) {
            item.cellX = cellHToCellX(emptyCell[0])
            item.cellY = cellVToCellY(emptyCell[1])
            return true
        } else {
            return false
        }
    }

    override fun setFocusOnFirstChild() {
        getChildAt(0, 0)?.requestFocus()
    }

    private fun readingOrderGreaterThan(v1: IntArray, v2: IntArray): Boolean {
        return v1[1] > v2[1] || v1[1] == v2[1] && v1[0] > v2[0]
    }


    fun coordinatesFromRank(_coords: IntArray?, rank: Int): IntArray {
        var coords = _coords
        if (coords == null)
            coords = IntArray(2)
        val h = rank % countH
        val v = rank / countH
        coords[0] = h
        coords[1] = v
        return coords
    }

    private fun rankFromCoordinates(cellH: Int, cellV: Int): Int {
        return cellH + cellV * countH
    }

    override fun realTimeReorder(emptyRank: Int, targetRank: Int) {
        realTimeReorder(coordinatesFromRank(null, emptyRank), coordinatesFromRank(null, targetRank))
    }

    private fun realTimeReorder(empty: IntArray, target: IntArray) {
        val wrap: Boolean
        var startX: Int
        var endX: Int
        val startY: Int
        var delay = 0
        var delayAmount = 30f
        if (readingOrderGreaterThan(target, empty)) {
            wrap = empty[0] >= countH - 1
            startY = if (wrap) empty[1] + 1 else empty[1]
            for (y in startY..target[1]) {
                startX = if (y == empty[1]) empty[0] + 1 else 0
                endX = if (y < target[1]) countH - 1 else target[0]
                for (x in startX..endX) {
                    val v = getChildAt(x, y)
                    if (animateChildToPosition(v, empty[0], empty[1],
                        REORDER_ANIMATION_DURATION, delay, true, true)) {
                        empty[0] = x
                        empty[1] = y
                        delay += delayAmount.toInt()
                        delayAmount *= 0.9f
                    }
                }
            }
        } else {
            wrap = empty[0] == 0
            startY = if (wrap) empty[1] - 1 else empty[1]
            for (y in startY downTo target[1]) {
                startX = if (y == empty[1]) empty[0] - 1 else getCountH() - 1
                endX = if (y > target[1]) 0 else target[0]
                for (x in startX downTo endX) {
                    val v = getChildAt(x, y)
                    if (animateChildToPosition(v, empty[0], empty[1],
                        REORDER_ANIMATION_DURATION, delay, true, true)) {
                        empty[0] = x
                        empty[1] = y
                        delay += delayAmount.toInt()
                        delayAmount *= 0.9f
                    }
                }
            }
        }
    }

    override fun arrangeChildren(list: List<View?>, count: Int, saveChanges: Boolean) {
        val folder = folder!!
        val vacant = IntArray(2)
        removeAllViews()

        for (i in list.indices) {
            val v = list[i] ?: continue
            getVacantCell(vacant, 1, 1)
            val lp = v.layoutParams as CellLayout.LayoutParams
            lp.cellH = vacant[0]
            lp.cellV = vacant[1]
            val id: Int
            if (folder.mLauncher != null) {
                val info = v.tag as ItemInfo
                id = info.id.toInt()
                if (saveChanges && (info.cellX != vacant[0].toFloat() || info.cellY != vacant[1].toFloat())) {
                    info.cellX = vacant[0].toFloat()
                    info.cellY = vacant[1].toFloat()
                    folder.mLauncher.modelWriter.addOrMoveItemInDatabase(info, folder.folderContents.id, 0, info.cellX, info.cellY)
                }
            } else {
                id = i
            }
            val insert = false
            addViewToCellLayout(v, if (insert) 0 else -1, id, lp, true)
        }
    }

    /**
     * @return the rank of the cell nearest to the provided pixel position.
     */
    override fun findNearestArea(pixelX: Int, pixelY: Int): Int {
        findNearestArea(pixelX, pixelY, 1, 1, sTmpArray)
        if (folder!!.isLayoutRtl) {
            sTmpArray[0] = countH - sTmpArray[0] - 1
        }
        return Math.min(countH * countV - 1,
            + sTmpArray[1] * countH + sTmpArray[0])
    }

    override val itemCount: Int
        get() = shortcutsAndWidgets.childCount


    override fun setupContentDimensions(list: List<View>, count: Int) {
        setupContentDimensions(count)
        arrangeChildren(list)
    }

    private fun setupContentDimensions(count: Int) {
        val folder = folder!!
        var count = count
        count = Math.max(1, count)

        var countX = countH
        var countY = countV
        var done = false

        val isLand = LauncherAppState.isScreenLandscape(context)
        val dm = resources.displayMetrics
        val dw = if (isLand) Math.max(dm.widthPixels, dm.heightPixels) else Math.min(dm.widthPixels, dm.heightPixels)

        val bgPadding = Rect()
        folder.background?.getPadding(bgPadding)
        val availableWidth = dw - bgPadding.left - bgPadding.right
        val desiredCellWidth = cellSpecs.cellWidthPx
        val mMaxCountX = availableWidth / desiredCellWidth
        val mMaxCountY = Integer.MAX_VALUE

        while (!done) {
            val oldCountX = countX
            val oldCountY = countY
            if (countX * countY < count) {
                // Current grid is too small, expand it
                if ((countX <= countY || countY == mMaxCountY) && countX < mMaxCountX) {
                    countX++
                } else if (countY < mMaxCountY) {
                    countY++
                }
                if (countY == 0) countY++
            } else if ((countY - 1) * countX >= count && countY >= countX) {
                countY = Math.max(0, countY - 1)
            } else if ((countX - 1) * countY >= count) {
                countX = Math.max(0, countX - 1)
            }
            done = countX == oldCountX && countY == oldCountY
        }

        val cellWidth = cellSpecs.cellWidthPx
        val cellHeight = cellSpecs.cellHeightPx
        setCellDimensions(cellWidth, cellHeight)
        val config = folder.config
        if (config.style == FolderWindowStyle.IMMERSIVE) {
            countX = config.cols
            countY = Math.max(countY, config.rows)
        }
        setGridSize(countX, countY)
    }

    override fun getLastItem(): View? {
        if (childCount < 1) {
            return null
        }
        val currContainer = shortcutsAndWidgets
        val lastRank = currContainer.childCount - 1
        return if (countH > 0) {
            currContainer.getChildAt(lastRank % countH, lastRank / countH)
        } else {
            currContainer.getChildAt(lastRank)
        }
    }

    override fun allocateRankForNewItem(): Int {
        val views = folder!!.itemsInReadingOrder
        val rank = views.size
        views.add(rank, null)
        arrangeChildren(views, views.size, false)
        return rank
    }

    override fun removeItem(view: View) {
        shortcutsAndWidgets.removeView(view)

    }

    override fun applyLayoutParamsToInfo(info: ItemInfo, lp: LayoutParams) {
        super.applyLayoutParamsToInfo(info, lp)
        info.indexedRank = rankFromCoordinates(lp.cellH, lp.cellV)
    }

    private inner class GridComparator(internal var mNumCols: Int) : Comparator<ShortcutInfo> {
        override fun compare(lhs: ShortcutInfo, rhs: ShortcutInfo): Int {
            val lhIndex = cellYToCellV(lhs.cellY) * mNumCols + cellXToCellH(lhs.cellX)
            val rhIndex = cellYToCellV(rhs.cellY) * mNumCols + cellXToCellH(rhs.cellX)
            return lhIndex - rhIndex
        }
    }

    companion object {
        private val sTmpArray = IntArray(2)
    }
}
