package com.android.launcher3.shortcuts

import android.content.Context
import android.util.AttributeSet
import android.view.Gravity
import android.view.View
import android.view.ViewGroup
import androidx.core.view.ViewCompat
import com.teslacoilsw.launcher.R

/**
 * A GridLayout that is much simpler than GridLayout.
 * Alternatively, a GridLayout for people too dumb (me) to use GridLayout
 *
 * Children must be fixed size and have their layoutparam size set
 */
open class DumbGridLayout(context: Context, attrs: AttributeSet) : ViewGroup(context, attrs) {

    private val canSqueezeOneExtra: Boolean
    private val columnCount: Int
    private val fillDirection: Int
    private val visibleChildrenCount: Int
        get() {
            var count = 0
            for (i in 0 until childCount) {
                if (getChildAt(i).visibility != View.GONE)
                    count++
            }
            return count
        }

    private fun effectiveColumnCount(childWidth: Int, widthSize: Int): Int {
        val cols = if (columnCount == -1)
            return widthSize/childWidth
        else
            columnCount

        return if (canSqueezeOneExtra && visibleChildrenCount % cols == 1) cols + 1 else cols
    }

    init {
        val a = context.obtainStyledAttributes(attrs, R.styleable.DumbGridLayout)
        try {
            columnCount = a.getInt(R.styleable.DumbGridLayout_android_columnCount, 1)
            canSqueezeOneExtra = a.getBoolean(R.styleable.DumbGridLayout_canSqueezeOneExtra, true)
            fillDirection = a.getInt(R.styleable.DumbGridLayout_fillDirection, Gravity.START or Gravity.TOP)
        } finally {
            a.recycle()
        }
    }

    override fun onMeasure(widthMeasureSpec: Int, heightMeasureSpec: Int) {
        var widthSize = MeasureSpec.getSize(widthMeasureSpec)
        val widthMode = MeasureSpec.getMode(widthMeasureSpec)

        var heightSize = MeasureSpec.getSize(heightMeasureSpec)
        val heightMode = MeasureSpec.getMode(heightMeasureSpec)

        if (childCount <= 0) {
            setMeasuredDimension(0, 0)
            return
        }

        val firstChild = getChildAt(0)
        var childWidth = firstChild.layoutParams.width
        var childHeight = firstChild.layoutParams.height
        if (childHeight <= 0 || childWidth <= 0) {
            firstChild.measure(MeasureSpec.makeMeasureSpec(childWidth, MeasureSpec.AT_MOST), MeasureSpec.makeMeasureSpec(childHeight, MeasureSpec.AT_MOST))
            if (childHeight <= 0)
                childHeight = firstChild.measuredHeight
            if (childWidth <= 0)
                childWidth = firstChild.measuredWidth
        }
        for (i in 0 until childCount) {
            val child = getChildAt(i)
            child.measure(MeasureSpec.makeMeasureSpec(childWidth, MeasureSpec.EXACTLY), MeasureSpec.makeMeasureSpec(childHeight, MeasureSpec.EXACTLY))
        }

        val cols = effectiveColumnCount(childWidth, widthSize)

        if (widthMode != MeasureSpec.EXACTLY)
            widthSize = childWidth * cols
        if (heightMode != MeasureSpec.EXACTLY)
            heightSize = paddingTop + paddingBottom + childHeight * Math.ceil(visibleChildrenCount/cols.toDouble()).toInt()
        setMeasuredDimension(widthSize, heightSize)

    }

    override fun onLayout(changed: Boolean, l: Int, t: Int, r: Int, b: Int) {
        if (childCount > 0)
            layoutChildren(l, t, r, b)
    }

    @Suppress("UNUSED_PARAMETER")
    private fun layoutChildren(l: Int, t: Int, r: Int, b: Int) {
        val childWidth = getChildAt(0).measuredWidth
        val childHeight = getChildAt(0).measuredHeight
        val visibleCount = visibleChildrenCount

        val thisColumnCount = effectiveColumnCount(childWidth, measuredWidth)
        val numRows = Math.ceil(visibleCount/thisColumnCount.toDouble()).toInt()

        val availableWidth = width - paddingLeft - paddingRight
        val availableHeight = height - paddingTop - paddingBottom

        val extraWidth = availableWidth - thisColumnCount * childWidth
        val numWidthGaps = thisColumnCount - 1
        val widthGapSize = extraWidth / numWidthGaps.toFloat()

        val extraHeight = availableHeight - numRows * childHeight
        val numHeightGaps = numRows - 1
        val heightGapSize = extraHeight / numHeightGaps.toFloat()


        var layoutIndex = 0
        var effectiveChildCount = visibleCount
        for (i in 0 until childCount) {
            val child = getChildAt(i)
            if (child.visibility == View.GONE)
                continue
            val lp = child.layoutParams

            /* We can allow useNextRowIfAtEnd only if we're not in the last row and we're the last
             * column, and there is still space available
             */
            if (lp is DumbGridLayoutParams && lp.useNextRowIfAtEnd &&
                (layoutIndex + 1) % thisColumnCount == 0 && layoutIndex / thisColumnCount < numRows - 1 &&
                effectiveChildCount < thisColumnCount * numRows)
            {
                layoutIndex++
                effectiveChildCount++
            }

            /* Start at bottom right and fill left then up */
            var row =
                if (fillDirection and Gravity.TOP == Gravity.TOP)
                    layoutIndex/thisColumnCount
                else
                    numRows - 1 - layoutIndex/thisColumnCount

            val col =
                if (Gravity.getAbsoluteGravity(fillDirection, ViewCompat.getLayoutDirection(this)) and Gravity.LEFT == Gravity.LEFT)
                    layoutIndex % thisColumnCount
                else
                    thisColumnCount - 1 - layoutIndex % thisColumnCount


            val childLeft = paddingLeft + ((col * (childWidth + widthGapSize))).toInt()
            val childTop = paddingTop + ((row * (childHeight + heightGapSize))).toInt()
            child.layout(childLeft, childTop, childLeft + childWidth, childTop + childHeight)
            layoutIndex++
        }
    }
}

class DumbGridLayoutParams(baseLp: ViewGroup.LayoutParams) : ViewGroup.LayoutParams(baseLp) {
    /* If you have two items next to each other you'd like to pair set this on the first one */
    var useNextRowIfAtEnd: Boolean = false
}