package com.teslacoilsw.kotlin

import android.app.Activity
import android.content.Context
import android.content.ContextWrapper
import android.content.pm.PackageManager
import android.graphics.Bitmap
import android.graphics.Canvas
import android.graphics.PointF
import android.graphics.Region
import android.graphics.drawable.Drawable
import android.os.Build
import android.util.TypedValue
import android.view.LayoutInflater
import android.view.View
import android.view.ViewGroup
import android.view.ViewParent
import android.view.inputmethod.InputMethodManager
import androidx.annotation.ColorInt
import androidx.core.content.ContextCompat
import androidx.recyclerview.widget.RecyclerView
import com.teslacoilsw.launcher.bitmaputils.ThreadLocalTypedValue
import com.teslacoilsw.launcher.daynight.DayNightContextWrapper
import com.teslacoilsw.shared.util.ReflectionHelper
import java.io.File
import java.io.FileOutputStream
import kotlin.reflect.KProperty0
import kotlin.reflect.jvm.isAccessible


fun Long.compare(that: Long): Int = if (this < that) -1 else if (this == that) 0 else 1

inline fun <T> tryCatch(defaultValue: T, body: () -> T) : T {
    return try {
        body()
    } catch (e: Exception) {
        defaultValue
    }
}

fun Bitmap.compressTo(path: String) {
    FileOutputStream(File(path)).use {
        compress(Bitmap.CompressFormat.PNG, 100, it)
    }
}

inline fun Canvas.resetAfter(body: () -> Unit) {
    val save = save()
    try {
        body()
    } finally {
        restoreToCount(save)
    }
}

@Suppress("NOTHING_TO_INLINE")
inline fun Int.toHexString(): String = Integer.toHexString(this)
fun <T> lazy1(initializer: () -> T): Lazy<T> = lazy(LazyThreadSafetyMode.NONE, initializer)

/**
 * Walk down the ContextWrapper path to find the Activity
 */
fun Context.getActivity(): Activity? {
    var context = this
    while (context is ContextWrapper) {
        if (context is Activity) {
            return context
        } else if (context is DayNightContextWrapper) {
            /* This weirdness because we do weird things to the Launcher context */
            val launcher = context.launcher
            if (launcher != null)
                return launcher
        }
        context = context.baseContext
    }
    return null
}

fun View.showKeyboard() {
    val imm = context.getSystemService(Context.INPUT_METHOD_SERVICE) as InputMethodManager
    imm.showSoftInput(this, InputMethodManager.SHOW_IMPLICIT)
}

private val typedValueCache = ThreadLocalTypedValue()

inline fun <R> TypedValue.cleanAfter(block: (TypedValue) -> R): R {
    synchronized(this) {
        return block(this).also {
            type = TypedValue.TYPE_NULL
            string = null
            data = 0
            assetCookie = 0
            resourceId = 0
            density = 0
        }
    }
}

fun Context.getThemeAttrColor(attr: Int): Int {
    typedValueCache.get().cleanAfter {
        if (!theme.resolveAttribute(attr, it, true)) {
            return 0
        }

        if (it.type < TypedValue.TYPE_FIRST_INT || it.type > TypedValue.TYPE_LAST_INT) {
            if (it.type == TypedValue.TYPE_STRING) {
                return ContextCompat.getColor(this, it.resourceId)
            }

            return 0
        }

        return it.data
    }
}

fun Context.getThemeAttrDrawable(attr: Int): Drawable? {
    typedValueCache.get().cleanAfter {
        if (!theme.resolveAttribute(attr, it, true)) {
            return null
        }

        if (it.type == TypedValue.TYPE_STRING) {
            return getDrawable(it.resourceId)
        }

        return null
    }
}

fun @receiver:ColorInt Int.toLab(): DoubleArray = DoubleArray(3).apply {  androidx.core.graphics.ColorUtils.colorToLAB(this@toLab, this) }

fun String.ucfirst(): String = Character.toUpperCase(this[0]) + this.substring(1).toLowerCase()

val KProperty0<*>.isLazyInitialized: Boolean
    get() {
        // Prevent IllegalAccessException from JVM access check
        isAccessible = true
        return (getDelegate() as Lazy<*>).isInitialized()
    }

fun Canvas.clipOutRectCompat(l: Int, t: Int, r: Int, b: Int) {
    if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.O) {
        clipOutRect(l, t, r, b)
    } else {
        @Suppress("DEPRECATION")
        clipRect(l.toFloat(), t.toFloat(), r.toFloat(), b.toFloat(), Region.Op.DIFFERENCE)
    }
}

private val _invalidateGlows by lazy { ReflectionHelper.findPrivateMethod(RecyclerView::class.java, "invalidateGlows")!!}
fun RecyclerView.invalidateGlows() {
    _invalidateGlows.invoke(this)
}

fun View.locationInParent(targetParent: ViewParent): PointF {
    val parent = this.parent
    if (parent == targetParent)
        return PointF(x, y)
    else if (parent is ViewGroup) {
        val parentPos = parent.locationInParent(targetParent)
        return PointF(x + parentPos.x, y + parentPos.y)
    }
    throw IllegalArgumentException("View.offsetFromParent must be called with an actual parent")
}

inline fun <T: View> Context.inflate(layoutId: Int, parent: ViewGroup? = null, attachToParent: Boolean = parent != null) =
        LayoutInflater.from(this).inflate(layoutId, parent, attachToParent) as T

fun <T> Array<T>.sliceExclude(excludeIndex: Int): Array<T> {
    val newIndices = indices.toMutableList()
    newIndices.remove(excludeIndex)
    return sliceArray(newIndices)
}

fun Context.hasPermission(permission: String) = ContextCompat.checkSelfPermission(this, permission) == PackageManager.PERMISSION_GRANTED


@Suppress("NOTHING_TO_INLINE")
inline fun @receiver:ColorInt Int.withAlpha(alpha: Int) = (this and 0x00ffffff) or (alpha shl 24)

inline fun <T> List<T>.toMutableList(): MutableList<T>  {
    val dst = ArrayList<T>(size)
    for (i in 0 until size) {
        dst[i] = this[i]
    }
    return dst
}
inline fun <T> List<T>.toList(): List<T> = toMutableList()