package com.teslacoilsw.launcher

import android.content.ActivityNotFoundException
import android.content.ComponentName
import android.content.Context
import android.content.Intent
import android.content.pm.ApplicationInfo
import android.content.pm.LauncherApps
import android.content.res.ColorStateList
import android.graphics.Bitmap
import android.graphics.Color
import android.graphics.PorterDuff
import android.graphics.drawable.Drawable
import android.graphics.drawable.LayerDrawable
import android.graphics.drawable.ShapeDrawable
import android.graphics.drawable.shapes.OvalShape
import android.net.Uri
import android.os.Build
import android.os.UserHandle
import android.text.TextUtils
import android.util.AttributeSet
import android.view.KeyEvent
import android.view.LayoutInflater
import android.view.View
import android.view.View.OnLongClickListener
import android.view.ViewGroup
import android.view.inputmethod.EditorInfo
import android.widget.GridLayout
import android.widget.ImageView
import android.widget.LinearLayout
import android.widget.TextView
import android.widget.Toast
import androidx.annotation.RequiresApi
import androidx.annotation.Size
import androidx.collection.SimpleArrayMap
import androidx.core.content.ContextCompat
import androidx.core.view.isVisible
import aosp.android.graphics.drawable.InsetDrawableProportional
import com.android.launcher3.AppInfo
import com.android.launcher3.AppsCustomizePagedView
import com.android.launcher3.FastBitmapDrawable
import com.android.launcher3.ItemInfo
import com.android.launcher3.Launcher
import com.android.launcher3.LauncherAppState
import com.android.launcher3.PendingAddDeepShortcutInfo
import com.android.launcher3.ShortcutInfo
import com.android.launcher3.Utilities
import com.android.launcher3.Workspace
import com.android.launcher3.compat.UserHandleCompat
import com.android.launcher3.compat.UserManagerCompat
import com.android.launcher3.shortcuts.DeepShortcutInfoCompatV25
import com.android.launcher3.shortcuts.sesame.SesameLinkContract
import com.android.launcher3.shortcuts.sesame.SesameShortcut
import com.android.launcher3.shortcuts.sesame.search
import com.android.launcher3.util.ComponentKey
import com.jakewharton.picasso.OkHttp3Downloader
import com.squareup.picasso.Picasso
import com.teslacoilsw.kotlin.getNumberExtraAsLong
import com.teslacoilsw.kotlin.isPackageInstalled
import com.teslacoilsw.kotlin.qFind
import com.teslacoilsw.launcher.deepshortcutproxy.COMPONENT_DSPDispatcherActivity
import com.teslacoilsw.launcher.drawer.AppPageStandard
import com.teslacoilsw.launcher.drawer.drawergroups.DrawerGroupModel
import com.teslacoilsw.launcher.iteminfo.DrawerItemInfo
import com.teslacoilsw.launcher.iteminfo.ItemIconUpdatableBitmap
import com.teslacoilsw.launcher.iteminfo.ItemInfoWithIcon
import com.teslacoilsw.launcher.preferences.Pref
import com.teslacoilsw.launcher.preferences.Pref2
import com.teslacoilsw.launcher.preferences.Pref3
import com.teslacoilsw.launcher.search.match
import com.teslacoilsw.launcher.shortcut.LetterTileDrawable
import com.teslacoilsw.launcher.util.DpUtil
import com.teslacoilsw.launcher.util.MaterialColors
import com.teslacoilsw.launcher.util.PackageNames
import com.teslacoilsw.launcher.util.picasso.CropCircleTransformation
import com.teslacoilsw.launcher.util.picasso.PlateTransformation
import com.teslacoilsw.launcher.util.picasso.VectorResourceRequestHandler
import com.teslacoilsw.launcher.widget.PagedViewSimple
import com.teslacoilsw.launcher.widget.RippleScrollView
import com.teslacoilsw.launcher.widget.SlidingTabIndicatorScrollView
import com.teslacoilsw.shared.colorpicker.ColorUtil
import com.teslacoilsw.shared.util.MathUtils
import org.json.JSONArray
import org.json.JSONObject
import rx.Observable
import rx.Subscription
import rx.android.schedulers.AndroidSchedulers
import timber.log.Timber
import java.io.File
import java.text.Collator
import java.util.*
import java.util.concurrent.TimeUnit
import java.util.concurrent.atomic.AtomicBoolean


private const val DEFAULT_RESULTS_FREQUENT = "FREQUENT"
private const val DEFAULT_RESULTS_RECENT = "RECENT"
private const val DEFAULT_RESULTS_NEW_UPDATED = "NEW_UPDATED"
private const val PREF_KEY_DEFAULT_TAB = "appsearch_default_tab"

private fun createActionBadge(context: Context, actionIcon: Drawable): Drawable {
    val shape = ShapeDrawable(OvalShape())
    shape.setColorFilter(ContextCompat.getColor(context, R.color.appsearch_contact_action_badge_color), PorterDuff.Mode.SRC_IN)
    val a = actionIcon.mutate()
    a.setColorFilter(Color.WHITE, PorterDuff.Mode.SRC_IN)
    return LayerDrawable(arrayOf(shape, InsetDrawableProportional(a, 8/48f)))
}
private const val MAX_LISTRESULT_RESULTS = 15

@Suppress("NOTHING_TO_INLINE")
private inline fun createListResultViewHolder(context: Context, parent: ViewGroup, colors: IntArray) =
    AppSearchView.ListResultViewHolder(LayoutInflater.from(context).inflate(R.layout.appsearch_list_item, parent, false) as LinearLayout, colors)

@Suppress("NOTHING_TO_INLINE")
private inline fun createListResultActionViewHolder(context: Context, parent: ViewGroup) =
        AppSearchView.ListResultActionViewHolder(LayoutInflater.from(context).inflate(R.layout.appsearch_action_item, parent, false) as ViewGroup)

class AppSearchView(_context: Context, attrs: AttributeSet) : DrawerSearchView(_context, attrs) {

    private val useSesame by lazy {
        Pref3.sesame_app_search.value &&
        Build.VERSION.SDK_INT >= Build.VERSION_CODES.LOLLIPOP &&
        context.isPackageInstalled(PackageNames.SESAME)
    }
    private val picassoReady = AtomicBoolean(false)
    private val picasso by lazy {
        val cache = File(context.applicationContext.cacheDir, "picasso-cache")
        if (!cache.exists()) {
            cache.mkdirs()
        }

        val out = Picasso.Builder(context)
            .downloader(OkHttp3Downloader(context))
            .addRequestHandler(VectorResourceRequestHandler(context))
            .build()
        picassoReady.set(true)
        out
    }
    private val cropCircleTransform by lazy { CropCircleTransformation(176/192f) }
    private val plateTransform by lazy { PlateTransformation(0xFFF5F5F5.toInt(), 0xFF757575.toInt()) }
    private lateinit var appGrid: AppPageStandard

    private var cellHeight: Int = 0
    private var maxDefaultApps: Int = 0
    private lateinit var searchLongClickListener: OnLongClickListener

    override fun setLauncher(launcher: Launcher) {
        searchLongClickListener = OnLongClickListener{ v: View ->
            val state = launcher.mState
            if (launcher.mAppsCustomizeContent.beginDragging(v) && !Pref.Value.desktopLocked()) {
                launcher.mAppsCustomizeContent.mCancelDragReturnsTo = if (state == Launcher.State.APPS_CUSTOMIZE) AppsCustomizePagedView.DRAG_RETURN_DRAWER_SEARCH else AppsCustomizePagedView.DRAG_RETURN_WORKSPACE_SEARCH
                launcher.mAppsCustomizeContent.mCancelDragReturnInfo = mLastQuery
            }
            if (state == Launcher.State.WORKSPACE && !Pref.Value.desktopLocked() && !launcher.dragLayer.isQuickMenuShowing() ) {
                launcher.hideAppsCustomizeHelper(Workspace.State.SPRING_LOADED, true, null)
                launcher.mState = Launcher.State.APPS_CUSTOMIZE_SPRING_LOADED
            }
            true
        }
        super.setLauncher(launcher)
        mSearchbarBackground.setCircularRevealColor(MaterialColors.PANEL_LIGHT)
        mSearchbar.setOnEditorActionListener(TextView.OnEditorActionListener { _, actionId, _ ->
            if (actionId == EditorInfo.IME_ACTION_GO) {
                /* Open the first result */
                return@OnEditorActionListener findViewById<View>(R.id.first)?.callOnClick() ?: false
            }
            false
        })
        mSearchbar.nextFocusDownId = R.id.first

    }

    override fun dispatchKeyEvent(event: KeyEvent): Boolean {
        /* Since a hardware keyboard "enter" selects the first app,
         * use hardware keyboard backspace ("DEL") to go back to the edit text
         */
        if (event.action != KeyEvent.ACTION_UP &&
            event.keyCode == KeyEvent.KEYCODE_DEL && !mSearchbar.isFocused) {
            mSearchbar.text.let {
                if (!it.isNullOrEmpty())
                    mSearchbar.setText(it.dropLast(1))
            }
            mSearchbar.setSelection(mSearchbar.text.length)
            mSearchbar.requestFocus()
            return true
        }
        val oldFocused = appGrid.shortcutsAndWidgets.focusedChild
        val superResult = super.dispatchKeyEvent(event)
        if (event.action != KeyEvent.ACTION_UP && oldFocused != null && appGrid.shortcutsAndWidgets.focusedChild == oldFocused) {
             return when (event.keyCode) {
                 KeyEvent.KEYCODE_DPAD_RIGHT,
                 KeyEvent.KEYCODE_DPAD_DOWN,
                 KeyEvent.KEYCODE_PAGE_DOWN -> {
                     findViewById<View>(R.id.first_listresult)?.let {
                         oldFocused.clearFocus()
                         it.requestFocus()
                         true
                     } ?: superResult
                 }
                 else -> superResult
             }
        }
        return superResult
    }

    private var drawerThemeSubscription: Subscription? = null
    override fun onAttachedToWindow() {
        super.onAttachedToWindow()
        drawerThemeSubscription = Pref3.appsearch_background.asObservable.skip(1).subscribe { setTheme() }
    }

    override fun onDetachedFromWindow() {
        super.onDetachedFromWindow()
        drawerThemeSubscription!!.unsubscribe()
    }

    private fun makeAppGridContainer(): RippleScrollView {
        val appGrid = AppPageStandard(NovaLauncher.getLauncher().mAppsCustomizeContent)
        appGrid.id = R.id.app_list

        val dp8 = DpUtil.getPixelSize(8)
        appGrid.setPadding(dp8, dp8, dp8, dp8)

        val grid = LauncherAppState.getInstance().dynamicGrid.deviceProfile
        maxDefaultApps = grid.allAppsNumCols * 2

        grid.drawerSearchCellSpecs.config = grid.drawerSearchCellSpecs.config.copy(labelColor = mColors[DrawerSearchView.COLOR_TEXT])
        grid.drawerSearchCellSpecs.updateForBackgroundColor(Pref2.get().drawer.appSearchBackground)

        cellHeight = MathUtils.roundInt(1.3f * grid.drawerSearchCellSpecs.cellHeightPx)
        appGrid.cellSpecs = grid.drawerSearchCellSpecs
        appGrid.setCellDimensions(grid.drawerSearchCellSpecs.cellWidthPx, cellHeight)

        val lp = LinearLayout.LayoutParams(LayoutParams.MATCH_PARENT, LayoutParams.WRAP_CONTENT)
        appGrid.layoutParams = lp

        return RippleScrollView(context).apply {
            addView(LinearLayout(context).apply { orientation = VERTICAL; addView(appGrid) })
        }
    }

    override fun setTheme() {
        mDefaultViewInitialized = false
        super.setTheme()
    }

    override fun makeResultsView(): ViewGroup {
        val sv = makeAppGridContainer()
        appGrid = sv.findViewById(R.id.app_list)
        return sv
    }

    private var mDefaultViewInitialized = false
    private lateinit var mDefaultView: View
    private lateinit var mDefaultViewFrequentApps: AppPageStandard
    private lateinit var mDefaultViewRecentApps: AppPageStandard
    private lateinit var mDefaultViewNewApps: AppPageStandard

    private fun installAndPopulateDefaultView(apps: List<DrawerItemInfo>) {
        if (!mDefaultViewInitialized) {
            val fgColor: Int
            val fgColorList: ColorStateList
            if (useStrongWhiteSearchForeground()) {
                val colorarray = searchColors
                fgColor = colorarray[0]
                fgColorList = ColorStateList(arrayOf(
                        intArrayOf(android.R.attr.state_selected),
                        intArrayOf()
                ),
                        intArrayOf(Color.WHITE, 0x80FFFFFF.toInt()))
            } else {
                val bgColor = searchBgColor
                val grayContrast = ColorUtil.calculateContrast(MaterialColors.FG_COLOR_LIAM_GRAY, bgColor)
                val grayNightContrast = ColorUtil.calculateContrast(MaterialColors.FG_COLOR_LIAM_GRAY_NIGHT, bgColor)
                if (grayContrast >= grayNightContrast) {
                    fgColor = MaterialColors.FG_COLOR_LIAM_GRAY
                    fgColorList = ContextCompat.getColorStateList(context,
                            R.color.drawer_search_tab_color)!!
                } else {
                    fgColor = MaterialColors.FG_COLOR_LIAM_GRAY_NIGHT
                    fgColorList = ContextCompat.getColorStateList(context,
                            R.color.drawer_search_tab_color_night)!!
                }
            }

            mDefaultView = LayoutInflater.from(context).inflate(R.layout.drawer_search_defaultview, mContentLayout, false)

            val pager: PagedViewSimple = mDefaultView.findViewById(R.id.pager)

            val indicator = (mDefaultView
                    .findViewById<SlidingTabIndicatorScrollView>(R.id.tab_scrollview)).indicator
            indicator.setPagedView(pager)
            pager.setPageSwitchListener { newPage, _ ->
                if (newPage != null)
                    Pref.Value.mPrefs.edit()
                            .putString(PREF_KEY_DEFAULT_TAB, newPage.tag as String).apply()
            }
            pager.setTabScrollListener(indicator)
            indicator.setTabStyle(Pref.Key.TabStyle.MATERIAL)

            val tabs: LinearLayout = mDefaultView.findViewById(android.R.id.tabs)
            val colors = IntArray(tabs.childCount)
            for (i in 0..tabs.childCount - 1) {
                val tab = tabs.getChildAt(i) as TextView
                tab.setOnClickListener(indicator)
                tab.setTextColor(fgColorList)
                colors[i] = fgColor
            }
            indicator.setColorsForTab(colors)

            val frequentContainer = makeAppGridContainer()
            frequentContainer.tag = DEFAULT_RESULTS_FREQUENT
            mDefaultViewFrequentApps = frequentContainer.findViewById(R.id.app_list)
            val recentContainer = makeAppGridContainer()
            recentContainer.tag = DEFAULT_RESULTS_RECENT
            mDefaultViewRecentApps = recentContainer.findViewById(R.id.app_list)
            val newContainer = makeAppGridContainer()
            newContainer.tag = DEFAULT_RESULTS_NEW_UPDATED
            mDefaultViewNewApps = newContainer.findViewById(R.id.app_list)

            pager.addView(frequentContainer)
            pager.addView(recentContainer)
            pager.addView(newContainer)

            val currentTab = Pref.Value.mPrefs.getString(PREF_KEY_DEFAULT_TAB, DEFAULT_RESULTS_NEW_UPDATED)
            when (currentTab) {
                DEFAULT_RESULTS_FREQUENT -> {
                    pager.currentPage = 0
                    indicator.onTabScrolled(0, 0f)
                }
                DEFAULT_RESULTS_RECENT -> {
                    pager.currentPage = 1
                    indicator.onTabScrolled(1, 0f)
                }
//                DEFAULT_RESULTS_NEW_UPDATED
                else -> {
                    pager.currentPage = 2
                    indicator.onTabScrolled(2, 0f)
                }
            }
            mDefaultViewInitialized = true
        }

        populateAppGrid(mDefaultViewFrequentApps, getFrequentApps(apps))
        populateAppGrid(mDefaultViewRecentApps, getRecentApps(apps))
        populateAppGrid(mDefaultViewNewApps, getNewApps(apps))

        if (mDefaultView.parent == null) {
            mContentLayout.removeAllViews()
            mContentLayout.addView(mDefaultView)
        }

    }

    private fun getNewApps(apps: List<DrawerItemInfo>): List<DrawerItemInfo> {
        val recentInstallTime = System.currentTimeMillis() - DrawerSearchView.RECENT_INSTALL_DURATION
        val recentUpdateTime = System.currentTimeMillis() - DrawerSearchView.RECENT_UPDATE_DURATION
        val nowPlus5Minutes = System.currentTimeMillis() + 5 * 60 * 1000
        DrawerSearchView.sRecentInstallTime = recentInstallTime
        var gridApps: MutableList<DrawerItemInfo> = ArrayList()
        for (item in apps) {
            if (item is AppInfo) {
                val app = item
                if (app.firstInstallTime > nowPlus5Minutes || app.lastUpdateTime > nowPlus5Minutes) {
                    /* Skip, installed in the future */
                } else if (app.firstInstallTime > recentInstallTime) {
                    gridApps.add(app)
                } else if (app.lastUpdateTime > recentUpdateTime) {
                    gridApps.add(app)
                }
            }
        }
        Collections.sort(gridApps, INSTALL_TIME_COMPARATOR)
        if (gridApps.size > maxDefaultApps)
            gridApps = gridApps.subList(0, maxDefaultApps)
        return gridApps
    }

    private fun getFrequentApps(apps: List<DrawerItemInfo>): List<DrawerItemInfo> {
        val predicted = ArrayList<DrawerItemInfo>()

        val ckList = mLauncher.stats.histogram
        for (ck in ckList) {
            val info: ItemInfoWithIcon? =
                if (DrawerGroupModel.get().hiddenApps.contains(ck.componentName))
                    null
                else
                    findItemInfo(apps, ck)

            if (info != null)
                predicted.add(info as DrawerItemInfo)

            if (predicted.size >= maxDefaultApps)
                break
        }

        return predicted
    }

    private fun getRecentApps(apps: List<DrawerItemInfo>): List<DrawerItemInfo> {
        val predicted = ArrayList<DrawerItemInfo>()

        val ckList = mLauncher.stats.recent
        for (ck in ckList) {
            val info: ItemInfoWithIcon? =
                if (DrawerGroupModel.get().hiddenApps.contains(ck.componentName))
                    null
                else
                    findItemInfo(apps, ck)

            if (info != null)
                predicted.add(info as DrawerItemInfo)

            if (predicted.size >= maxDefaultApps)
                break
        }

        return predicted
    }

    private var pendingSesameSubscription: Subscription? = null

    class ViewRecycler<T>(val size: Int) {
        private val buffer = ArrayList<T>(size)

        fun enqueue(item: T) {
            if (buffer.size < size)
                buffer.add(item)
        }

        fun get(): T? {
            val x = if (buffer.size > 0) buffer.removeAt(0) else null
            return x
        }
    }

    private val listResultRecycler = ViewRecycler<ListResultViewHolder>(MAX_LISTRESULT_RESULTS)
    private val listResultActionRecycler = ViewRecycler<ListResultActionViewHolder>(MAX_LISTRESULT_RESULTS * 3)

    class ListResultViewHolder(val view: ViewGroup, colors: IntArray) {
        val mainRow: LinearLayout = view.findViewById(R.id.main_row)
        val title: TextView = view.findViewById(android.R.id.title)
        val icon: ImageView = view.findViewById(android.R.id.icon)
        val actions = ArrayList<ListResultActionViewHolder>()
        val searchGrid: GridLayout = view.findViewById(R.id.search_grid)

        init {
            view.tag = this
            title.setTextColor(colors[COLOR_TEXT])
        }
    }

    class ListResultActionViewHolder(val view: ViewGroup) {
        val icon: ImageView = view.findViewById(R.id.icon)

        init {
            view.tag = this
        }
    }

    public override fun updateSearchResults(rawQuery: String, queryLc: String) {
        super.updateSearchResults(queryLc)
        val apps: List<DrawerItemInfo> = mLauncher.mAppsCustomizeContent.mAllApps

        val gridApps = mutableListOf<DrawerItemInfo>()
        if (TextUtils.isEmpty(queryLc)) {
            installAndPopulateDefaultView(apps)
            return
        } else {
            if (mResultsView.parent == null) {
                mContentLayout.removeAllViews()
                mContentLayout.addView(mResultsView)
            }

            val checkPackageName = queryLc.contains(".")

            val matchStrengthMap = SimpleArrayMap<DrawerItemInfo, Int>()
            apps.filterTo(gridApps) {
                val matchStrength = match(it.title, queryLc)
                if (matchStrength > 0) {
                    matchStrengthMap.put(it, matchStrength)
                    true
                } else {
                    checkPackageName and (match(it.component.packageName, if (queryLc[0] == '.') queryLc.substring(1) else queryLc) > 0)
                }
            }
            val comparator = mSearchMatchComparator
            comparator.strengthMap = matchStrengthMap
            Collections.sort(gridApps, comparator)
            comparator.strengthMap = null
            val cols = mLauncher.deviceProfile.allAppsNumCols
            if (useSesame) {
                while (gridApps.size > cols)
                    gridApps.removeAt(gridApps.lastIndex)
            }
            populateAppGrid(appGrid, gridApps)

            appGrid.isVisible = appGrid.shortcutsAndWidgets.childCount > 0

            if (useSesame) {
                 if (appGrid.shortcutsAndWidgets.childCount != 0) {
                    val rows = MathUtils.ceilInt(appGrid.childCount / cols.toFloat())
                    appGrid.setCellCount(cols, rows)
                    appGrid.visibility = VISIBLE
                }

                pendingSesameSubscription?.unsubscribe()
                pendingSesameSubscription = Observable.fromCallable {
                    /* Init picasso if needed on bg thread (does disk io) */
                    picasso
                    search(context, rawQuery, MAX_LISTRESULT_RESULTS)
                }
                .delay(90L, TimeUnit.MILLISECONDS)
                .observeOn(AndroidSchedulers.mainThread())
                .subscribeOn(AndroidSchedulers.from(NovaLauncher.getLauncher().bgHandler.looper))
                .subscribe({ updateSearchResultSesame(it) }, { e -> Timber.w(e) })
            } else {
                clearListResults()
                val parent = appGrid.parent as ViewGroup
                val results = mutableListOf<ListResultViewHolder>()
                val holder = listResultRecycler.get() ?: createListResultViewHolder(context, parent, mColors)
                holder.title.text = context.getString(R.string.search_for_xxx, rawQuery)
                holder.icon.setImageDrawable(ContextCompat.getDrawable(context, R.drawable.ic_qsb_m_letter_color_google))

                holder.view.setOnClickListener {
                    try {
                        context.startActivity(Intent(Intent.ACTION_WEB_SEARCH)
                                .apply {
                                    setPackage(PackageNames.GOOGLE_APP)
                                    putExtra("query", rawQuery) }
                        )
                    } catch (e: Exception) {
                        val uri = Uri.Builder().scheme("https").authority("google.com").appendPath("search").appendQueryParameter("q", rawQuery).build()
                        context.startActivity(Intent(Intent.ACTION_VIEW)
                                .apply { data = uri }
                        )
                    }
                }

                val actionButton = listResultActionRecycler.get() ?: createListResultActionViewHolder(context, holder.view)
                actionButton.icon.imageTintList = null
                actionButton.icon.setImageDrawable(ContextCompat.getDrawable(context, R.drawable.ic_search_google_play))
                val dp48 = DpUtil.getPixelSize(48)
                val lp = ViewGroup.LayoutParams(dp48, dp48)
                actionButton.view.contentDescription = "Play Store"
                actionButton.view.setOnClickListener {
                    val query = mSearchbar.text.toString()
                    var intent = Intent(Intent.ACTION_MAIN).addCategory(Intent.CATEGORY_APP_MARKET)
                    if (!TextUtils.isEmpty(query)) {
                        intent = Intent(Intent.ACTION_VIEW, Uri.parse("market://search?c=apps&q=$query"))
                    }
                    intent.addFlags(Intent.FLAG_ACTIVITY_CLEAR_TASK).addFlags(Intent.FLAG_ACTIVITY_NEW_TASK).addFlags(Intent.FLAG_ACTIVITY_CLEAR_TOP)
                    try {
                        mLauncher.startActivity(it, intent, "nova-search-market")
                    } catch (ex: ActivityNotFoundException) {
                        Toast.makeText(mLauncher, R.string.activity_not_found, Toast.LENGTH_SHORT).show()
                    }

                }
                holder.actions.add(actionButton)
                holder.mainRow.addView(actionButton.view, lp)

                results.add(holder)
                updateListlikeResults(results)
            }
        }

    }

    private fun updateSearchResultSesame(shortcuts: List<SesameShortcut>) {
        clearListResults()

        val parent = appGrid.parent as ViewGroup
        val holders = mutableListOf<ListResultViewHolder>()
        for(info in shortcuts) {
            val appIcon: Bitmap? = info.component?.let {
                LauncherAppState.getInstance().iconCache.peekCachedBitmap(it, info.user)
            }

            val holder = listResultRecycler.get() ?: createListResultViewHolder(context, parent, mColors)
            holder.title.text = info.resultLabel
            holder.icon.apply {
                when {
                    info.iconUri != null -> {
                        val iconSize = Pref.Value.desktop_icon_size.appIconSize
                        val r = picasso.load(VectorResourceRequestHandler.checkAndUpdateScheme(info.iconUri))
                                .resize(iconSize, iconSize)
                                .centerInside()
                        if ("contact" == info.type) {
                            r.transform(cropCircleTransform)
                        }
                        if ("true" == info.iconUri.getQueryParameter("commonIcon")) {
                            r.transform(plateTransform)
                        }
                        r.error(R.drawable.launcher_shortcut_generic)
                        r.placeholder(R.drawable.launcher_shortcut_generic)
                        r.into(this)
                    }
                    info.type == SesameLinkContract.TYPE_CONTACT -> {
                        val lookupKey = try {
                            info.typeData.getString("lookup")
                        } catch (e: Exception) {
                            info.displayLabel
                        }

                        val d = LetterTileDrawable(resources)
                        d.setContactDetails(info.displayLabel, lookupKey)
                        d.setScale(0.667f)
                        setImageDrawable(d)
                    }
                    appIcon != null -> setImageDrawable(FastBitmapDrawable(appIcon))
                    info.type == SesameLinkContract.TYPE_SEARCH -> {
                        var d = ContextCompat.getDrawable(context, R.drawable.launcher_shortcut_search)
                        d = InsetDrawableProportional.replaceFixedWithProportional(d, DpUtil.getPixelSize(48), DpUtil.getPixelSize(48))
                        setImageDrawable(d)
                    }
                    else -> setImageResource(R.drawable.launcher_shortcut_generic)
                }

            }

            var actions = if (info.typeData.has("actions")) try { info.typeData.getJSONArray("actions") } catch (_ : Exception) { JSONArray() } else JSONArray()
            if (info.type == SesameLinkContract.TYPE_SEARCH && actions.length() > 3) {
                var actionsPerRow = if (actions.length() % 6 == 1) 5 else 6
                holder.searchGrid.columnCount = actionsPerRow
                holder.searchGrid.rowCount = 1 //MathUtils.ceilInt(actions.length() / 6f)
                holder.searchGrid.visibility = View.VISIBLE
                for (i in 0 until actions.length()) {
                    try {
                        val actionButton = createSesameActionButtonFromJSON(holder, info, actions.getJSONObject(i))

                        val rowSpec = GridLayout.spec(GridLayout.UNDEFINED)
                        val colSpec = GridLayout.spec(GridLayout.UNDEFINED, 1, 1.0f)

                        val lp = GridLayout.LayoutParams(rowSpec, colSpec)
                        holder.actions.add(actionButton)
                        holder.searchGrid.addView(actionButton.view, lp)

                    } catch (e: Exception) {
                        e.printStackTrace()
                    }
                }
            } else {
                for (i in 0 until Math.min(3, actions.length())) {
                    try {
                        val actionButton = createSesameActionButtonFromJSON(holder, info, actions.getJSONObject(i))

                        val dp48 = DpUtil.getPixelSize(48)
                        val lp = ViewGroup.LayoutParams(dp48, dp48)
                        holder.actions.add(actionButton)
                        holder.mainRow.addView(actionButton.view, lp)
                    } catch (e: Exception) {
                        e.printStackTrace()
                    }
                }
            }

            holder.mainRow.setOnClickListener { mLauncher.startActivity(info.proxyIntent) }
            holder.mainRow.setOnLongClickListener {
                val iconView = holder.icon
                val drawable = iconView.drawable

                var bitmap: Bitmap
                if (appIcon != null && (drawable is FastBitmapDrawable && drawable.bitmap == appIcon)) {
                    val dp48 = DpUtil.getPixelSize(48)
                    val d = InsetDrawableProportional.replaceFixedWithProportional(ContextCompat.getDrawable(context, R.drawable.launcher_shortcut_generic), dp48, dp48)
                    bitmap = LauncherAppState.getInstance().iconCache.shadowGenerator.recreateAndNormalizeIcon(d)
                    bitmap = Utilities.badgeWithBitmap(bitmap, appIcon)
                } else {
                    bitmap = LauncherAppState.getInstance().iconCache.shadowGenerator.recreateAndNormalizeIcon(drawable)
                    appIcon?.let { bitmap = Utilities.badgeWithBitmap(bitmap, appIcon) }
                }

                startAppDrag(info, info.displayLabel, info.directIntent, bitmap, iconView)
                true
            }

            holders.add(holder)
        }

        updateListlikeResults(holders)
    }


    private fun clearListResults() {
        val grid = appGrid
        val parent = grid.parent as ViewGroup

        while (parent.childCount > 1) {
            val child = parent.getChildAt(1) as ViewGroup
            val tag = child.tag
            if (tag is ListResultViewHolder) {
                tag.view.setOnClickListener(null)
                if (picassoReady.get())
                    picasso.cancelRequest(tag.icon)
                tag.icon.setImageDrawable(null)
                for (i in 0 until tag.actions.size) {
                    tag.actions[i].view.setOnClickListener(null)
                    listResultActionRecycler.enqueue(tag.actions[i])
                    tag.mainRow.removeView(tag.actions[i].view)
                    tag.searchGrid.removeView(tag.actions[i].view)
                }
                tag.actions.clear()
                listResultRecycler.enqueue(tag)
            }

            parent.removeView(child)
        }
    }

    private fun updateListlikeResults(results: List<ListResultViewHolder>) {
        val grid = appGrid
        val parent = grid.parent as ViewGroup
        var first = grid.shortcutsAndWidgets.childCount == 0

        var firstListResult = true
        for(holder in results) {
            if (first || firstListResult) {
                holder.mainRow.id = if (first) R.id.first else R.id.first_listresult
                first = false
                firstListResult = false
            } else {
                holder.mainRow.id = 0
            }

            parent.addView(holder.view)
        }

        grid.ready()
    }

    private fun startAppDrag(info: SesameShortcut, label: String, intent: Intent, bitmap: Bitmap, iconView: View) {
        val tag: ItemInfo?
        if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.N_MR1 &&
                intent.component == COMPONENT_DSPDispatcherActivity) {
            tag = getPendingAddDeepShortcutInfo(info.directIntent)
        } else {
            val shortcutInfo = ShortcutInfo(intent, label, null, bitmap, info.user)
            shortcutInfo.itemIcon = ItemIconUpdatableBitmap(bitmap)
            tag = shortcutInfo
        }
        if (tag == null) {
            Toast.makeText(context, context.getString(R.string.permission_denied), Toast.LENGTH_SHORT).show()
        } else {
            iconView.tag = tag
            iconView.setTag(R.id.tag_drag_ignore_offsets, true)
            searchLongClickListener.onLongClick(iconView)
        }
    }

    @RequiresApi(Build.VERSION_CODES.N_MR1)
    private fun getPendingAddDeepShortcutInfo(intent: Intent) : PendingAddDeepShortcutInfo? {
        val packageName = intent.getStringExtra("packageName")
        val id = intent.getStringExtra("id")
        val user = intent.getParcelableExtra<UserHandle>("userHandle")
                ?: UserManagerCompat.getInstance(context)
                .getUserForSerialNumber(intent.getNumberExtraAsLong("userId", 0)).user

        val apps = context.getSystemService(Context.LAUNCHER_APPS_SERVICE) as LauncherApps
        val query = LauncherApps.ShortcutQuery()
        query.setPackage(packageName)
        query.setShortcutIds(listOf(id))
        query.setQueryFlags(LauncherApps.ShortcutQuery.FLAG_MATCH_DYNAMIC or LauncherApps.ShortcutQuery.FLAG_MATCH_MANIFEST)
        val shortcuts = apps.getShortcuts(query, user) ?: listOf()
        if (shortcuts.size == 1)
            return PendingAddDeepShortcutInfo(DeepShortcutInfoCompatV25(shortcuts.first()))

        return null

    }

    private fun populateAppGrid(grid: AppPageStandard, gridApps: List<DrawerItemInfo>) {
        val cols = mLauncher.deviceProfile.allAppsNumCols
        val rows = MathUtils.max(1, MathUtils.ceilInt(gridApps.size / cols.toFloat()))
        grid.setCellCount(cols, rows)
        grid.removeAllViewsOnPage()
        var first = true
        for (info in gridApps) {
            val icon = grid.addAppView(info)
            icon.setOnLongClickListener(searchLongClickListener)
            if (first) {
                (icon as View).id = R.id.first
                first = false
            }
        }
        grid.ready()
    }

    private fun getIsSystem(info: DrawerItemInfo): Boolean {
        if (info is AppInfo) {
            return info.pmApplicationInfoFlags and ApplicationInfo.FLAG_SYSTEM != 0
        }
        return false
    }

    public override fun onMeasure(widthMeasureSpec: Int, heightMeasureSpec: Int) {
        val width = View.MeasureSpec.getSize(widthMeasureSpec) - mLinearLayout.paddingLeft - mLinearLayout.paddingRight
        setAppGridSize(appGrid, width, 0)

        val height = Math.max(mDefaultViewFrequentApps.desiredHeight, Math.max(mDefaultViewRecentApps.desiredHeight, mDefaultViewNewApps.desiredHeight))
        setAppGridSize(mDefaultViewFrequentApps, width, height)
        setAppGridSize(mDefaultViewRecentApps, width, height)
        setAppGridSize(mDefaultViewNewApps, width, height)
        val pager: PagedViewSimple = mDefaultView.findViewById(R.id.pager)
        val lp = pager.layoutParams
        lp.height = Math.max(1, Math.max(mDefaultViewFrequentApps.desiredHeight,
                Math.max(mDefaultViewNewApps.desiredHeight, mDefaultViewNewApps.desiredHeight)))
        pager.layoutParams = lp

        super.onMeasure(widthMeasureSpec, heightMeasureSpec)
    }

    private fun setAppGridSize(appGrid: AppPageStandard, _width: Int, _height: Int) {
        var width = _width
        var height = _height
        if (height == 0)
            height = Math.max(1, appGrid.desiredHeight)

        val lp = mContentScrollviewParent.layoutParams as ViewGroup.MarginLayoutParams
        width -= lp.leftMargin + lp.rightMargin
        val innerWidth = width - appGrid.paddingLeft - appGrid.paddingRight
        val countH = appGrid.cellCountX
        appGrid.setCellDimensions(MathUtils.roundInt(1.0f * ((innerWidth - (countH - 1) * appGrid.widthGap) / countH)), cellHeight)
        appGrid.setFixedSize(width, height)
    }

    private val mSearchMatchComparator = object : Comparator<DrawerItemInfo> {
        val  collator = Collator.getInstance()
        var strengthMap: SimpleArrayMap<DrawerItemInfo, Int>? = null
        override fun compare(lhs: DrawerItemInfo, rhs: DrawerItemInfo): Int {
            val titleA = getTitle(lhs)
            val titleB = getTitle(rhs)

            val strengthA = strengthMap?.get(lhs) ?: 0
            val strengthB = strengthMap?.get(rhs) ?: 0
            return when {
                strengthA > strengthB -> -1
                strengthB > strengthA -> +1
                else -> collator.compare(titleA, titleB)
            }
        }

        private fun getTitle(a: DrawerItemInfo?): String {
            if (a != null) {
                val title = a.title
                if (title != null)
                    return title.toString()
            }
            return ""
        }
    }

    @Size(5)
    override fun getSearchColors(): IntArray {
        return getSearchColors(context)
    }

    private fun createSesameActionButtonFromJSON(sesameRow: ListResultViewHolder, info: SesameShortcut, action: JSONObject): ListResultActionViewHolder {
        val intent = Intent.parseUri(action.getString("intent"), 0)
        val directIntent = Intent.parseUri(action.optString("directIntent"), 0) ?: intent
        val description = action.optString("label", "action")
        val component = action.optString("component", "")
        val pkg = action.optString("package", "")

        val iconUri = action.optString("icon").let {
            if (it.isEmpty())
                null
            else
                Uri.parse(it)
        }

        val actionButton = listResultActionRecycler.get() ?:
                createListResultActionViewHolder(context, sesameRow.view)
        val actionIconView = actionButton.icon
        var wasAppIcon = false
        if (iconUri != null) {
            actionIconView.setImageURI(iconUri)
        } else {
            var appIcon: Bitmap? = null
            if (!component.isEmpty()) {
                appIcon = LauncherAppState.getInstance().iconCache.peekCachedBitmap(ComponentName.unflattenFromString(component), info.user)
            }
            if (appIcon == null && !pkg.isEmpty()) {
                appIcon = appIcon ?: LauncherAppState.getInstance().iconCache.getIconForPackage(pkg, false)
            }
            appIcon = appIcon ?: getIconFromIntent(directIntent, info.user)
            appIcon.let {
                actionIconView.setImageBitmap(it)
                wasAppIcon = true
            }
        }
        var isTinted = false
        if (info.type == SesameLinkContract.TYPE_SEARCH || wasAppIcon) {
            actionIconView.imageTintList = null
            actionIconView.colorFilter = null
        } else {
            actionIconView.imageTintList = ColorStateList.valueOf(ContextCompat.getColor(context, R.color.appsearch_contact_action_tint))
            actionIconView.setColorFilter(mColors[COLOR_HEADER_ICON], PorterDuff.Mode.SRC_ATOP)
            isTinted = true
        }
        actionButton.view.apply {
            contentDescription = description
            setOnClickListener { mLauncher.startActivity(intent) }
            setOnLongClickListener {
                val iconView = sesameRow.icon
                var bitmap: Bitmap
                if (isTinted) {
                    val drawable = iconView.drawable
                    bitmap = LauncherAppState.getInstance().iconCache.shadowGenerator.recreateAndNormalizeIcon(drawable)
                    bitmap = Utilities.badgeWithBitmap(bitmap, createActionBadge(context, actionIconView.drawable))
                } else {
                    val drawable = iconView.drawable
                    bitmap = LauncherAppState.getInstance().iconCache.shadowGenerator.recreateAndNormalizeIcon(drawable)
                    bitmap = Utilities.badgeWithBitmap(bitmap, actionIconView.drawable)
                }
                startAppDrag(info, description, directIntent, bitmap, iconView)
                true
            }
        }

        return actionButton
    }

    override fun hideMarketButtonWithQuery(): Boolean {
        return true || useSesame
    }

}

private fun findItemInfo(apps: List<DrawerItemInfo>, key: ComponentKey): ItemInfoWithIcon? {
    return apps.qFind { key == it.toComponentKey() }?.let {it as ItemInfoWithIcon }
}

private val INSTALL_TIME_COMPARATOR = Comparator<DrawerItemInfo> { a, b ->
    if (a !is AppInfo) {
        return@Comparator if (b is AppInfo) 1 else 0
    } else if (b !is AppInfo) {
        return@Comparator -1
    }

    val aa = a
    val ba = b

    if (aa.firstInstallTime > DrawerSearchView.sRecentInstallTime) {
        if (ba.firstInstallTime > DrawerSearchView.sRecentInstallTime) {
            if (aa.firstInstallTime < ba.firstInstallTime)
                1
            else if (aa.firstInstallTime > ba.firstInstallTime)
                -1
            else
                0
        } else {
            -1
        }
    } else if (ba.firstInstallTime > DrawerSearchView.sRecentInstallTime) {
        1
    } else {
        if (aa.lastUpdateTime < ba.lastUpdateTime)
            1
        else if (aa.lastUpdateTime > ba.lastUpdateTime)
            -1
        else
            0
    }
}

private fun getIconFromIntent(intent: Intent, user: UserHandleCompat): Bitmap {
    intent.component?.let {
        return LauncherAppState.getInstance().iconCache.peekCachedBitmap(it, user)
    }

    return LauncherAppState.getInstance().iconCache.getIconForPackage(intent.getPackage(), true)!!
}

fun useStrongWhiteSearchForeground(): Boolean {
    val bg = ColorUtil.argb(0xff, Pref2.get().drawer.appSearchBackground)
    val hsv = FloatArray(3)
    Color.colorToHSV(bg, hsv)

    val contrast = ColorUtil.calculateContrast(Color.WHITE, bg)

    return (hsv[1] >= 0.35f || (hsv[2] in 0.4f..0.6f)) && contrast > 1.65f
}

@Size(5) fun getSearchColors(context: Context): IntArray {

    val bg = ColorUtil.argb(0xff, Pref2.get().drawer.appSearchBackground)

    val lightfg = ContextCompat.getColor(context, R.color.primary_text_material_dark)
    val darkfg = ContextCompat.getColor(context, R.color.primary_text_material_light)

    val drawer = LauncherAppState.getInstance().dynamicGrid.deviceProfile.drawerCellSpecs
    val drawerBg = Pref2.get().drawer.effectiveDrawerBackgroundColor
    var preferLightFg = true
    if (drawer.config.labelVisible && Color.alpha(drawerBg) == 0xFF && drawerBg == bg)
        preferLightFg = false

    val lightfg_contrast = ColorUtil.calculateContrast(lightfg, bg)
    val darkfg_contrast = ColorUtil.calculateContrast(darkfg, bg)

    val useNightMode = preferLightFg && lightfg_contrast > 1.5f || darkfg_contrast < lightfg_contrast

    if (useStrongWhiteSearchForeground()) {
        return intArrayOf(Color.WHITE, Color.WHITE, 0x80FFFFFF.toInt(), Color.WHITE, Color.WHITE);
    }

    if (useNightMode) {
        return intArrayOf(lightfg, ContextCompat.getColor(context, R.color.appsearch_lightfg_icon), ContextCompat.getColor(context, R.color.appsearch_lightfg_hint), ContextCompat.getColor(context, R.color.appsearch_lightfg_header), ContextCompat.getColor(context, R.color.appsearch_lightfg_icon))
    } else {
        return intArrayOf(darkfg, ContextCompat.getColor(context, R.color.appsearch_darkfg_icon), ContextCompat.getColor(context, R.color.appsearch_darkfg_hint), ContextCompat.getColor(context, R.color.appsearch_darkfg_header), ContextCompat.getColor(context, R.color.appsearch_darkfg_icon))
    }
}
