package com.teslacoilsw.launcher

import android.content.Context
import android.util.DisplayMetrics
import androidx.annotation.ColorInt
import androidx.core.graphics.alpha
import com.android.launcher3.DeviceProfile
import com.teslacoilsw.cofs.Cofs
import com.teslacoilsw.cofs.CofsBase
import com.teslacoilsw.cofs.CofsIgnore
import com.teslacoilsw.cofs.ConfigObjectFromString
import com.teslacoilsw.cofs.getCofsDelegate
import com.teslacoilsw.kotlin.dp
import com.teslacoilsw.kotlin.sp
import com.teslacoilsw.kotlin.withAlpha
import com.teslacoilsw.launcher.bitmaputils.ThreadLocalPaint
import com.teslacoilsw.launcher.preferences.COLOR_AUTO
import com.teslacoilsw.launcher.preferences.Pref3
import com.teslacoilsw.launcher.util.MaterialColors
import com.teslacoilsw.launcher.widget.TypefaceHelper
import com.teslacoilsw.shared.colorpicker.ColorUtil
import rx.Observable
import rx.subjects.PublishSubject
import kotlin.math.roundToInt

@ConfigObjectFromString
data class CellSpecConfig(val rawIconScale: Float = 1.25f,
                          val labelVisible: Boolean = true,
                          val labelSizeSp: Float = 13f,
                          @ColorInt val labelColor: Int = COLOR_AUTO,
                          val labelShadow: Boolean = false,
                          val labelFont: String = TypefaceHelper.CONDENSED,
                          val labelSingleLine: Boolean = true,
                          val matchDesktopSize: Boolean = false)
{
    val iconScale: Float
        get() {
            if (matchDesktopSize && this != Pref3.desktop_cellspecs.value)
                return Pref3.desktop_cellspecs.value.rawIconScale
            return rawIconScale
        }

    companion object : Cofs<CellSpecConfig>, CofsBase<CellSpecConfig> by CellSpecConfig::class.getCofsDelegate() {
        override val DEFAULT = CellSpecConfig()
    }

    /* This is for java */
    fun copyWithLabelSettings(labelVisible: Boolean, labelShadow: Boolean, labelColor: Int) =
        copy(labelVisible = labelVisible, labelShadow = labelShadow, labelColor = labelColor)

    fun toPreferenceSummaryString(context: Context): String {
        val sb = StringBuilder()
        if (matchDesktopSize)
            sb.append(context.getString(R.string.match_desktop_size))
        else
            sb.append((iconScale * 100).roundToInt()).append("%")

        sb.append(", ")
        if (labelVisible)
            sb.append(context.getString(R.string.iconlayout_labels_on))
        else
            sb.append(context.getString(R.string.iconlayout_labels_off))

        return sb.toString()
    }
}

class CellIconSizeSpecs(
    val type: CellLayoutType,
    var config: CellSpecConfig,
    val night_mode_auto: Observable<Boolean>? = null
    )
{

    var iconSizePx: Int = 0
        private set
    var cellWidthPx: Int = 0
        private set
    var cellHeightPx: Int = 0
        private set
    var iconDrawablePaddingPx: Int = 0
        private set
    var iconTextSizePx: Int = 0
        private set
    var iconTextColor: Int = 0xFFFF00FF.toInt() /* Default value to make it obvious it hasn't been set */
        get() {
            if (is_night_mode) {
                return getForegroundColorForBackground(MaterialColors.NIGHT_THEME_BACKGROUND.withAlpha(backgroundColor.alpha))
            }
            return field
        }
        private set
    private var backgroundColor: Int = 0
    var rotate90Widgets = false

    private var is_night_mode = false

    @CofsIgnore
    private val subject = PublishSubject.create<CellIconSizeSpecs>()

    init {
        night_mode_auto?.subscribe { nightMode ->
            is_night_mode = nightMode
            subject.onNext(this@CellIconSizeSpecs)
        }
    }

    @JvmOverloads
    fun applyIconSize(dm: DisplayMetrics, baseIconSize: Int, drawablePadding: Int, scale: Float = 1.0f, isFolder: Boolean = false) {
        iconSizePx = (baseIconSize.toFloat() * config.iconScale * scale).roundToInt()
        iconTextSizePx = if (config.labelVisible) dm.sp(scale * config.labelSizeSp).roundToInt() else 0
        iconDrawablePaddingPx = drawablePadding

        val textPaint = measurePaint.get()
        textPaint.textSize = iconTextSizePx.toFloat()
        val fm = textPaint.fontMetrics
        val textLinesSize = Math.ceil((if (config.labelSingleLine) 1f else DeviceProfile.TWO_LINE_MULTIPLER) * Math.ceil((fm.bottom - fm.top).toDouble())).toInt()

        cellHeightPx = iconSizePx + (if (config.labelVisible) drawablePadding + textLinesSize else 0) +
            if (isFolder) dm.dp(16) else 0

        cellWidthPx = if (isFolder && config.labelVisible)
            (iconSizePx + drawablePadding + textLinesSize + dm.dp(8))
        else
            if (isFolder) cellHeightPx else iconSizePx

        if (isFolder) {
            val fourcolumncellwidth = Math.max(iconSizePx, (dm.widthPixels - 2 * dm.dp(8)) / 4)
            cellWidthPx = Math.min(cellWidthPx, fourcolumncellwidth)
        }

    }


    fun asObservable(): Observable<CellIconSizeSpecs> {
        return subject
    }

    fun updateForBackgroundColor(@ColorInt backgroundColor: Int) {
        this.backgroundColor = backgroundColor
        iconTextColor = if (config.labelColor == COLOR_AUTO) {
            getForegroundColorForBackground(backgroundColor)
        } else {
            config.labelColor
        }
    }

    companion object {
        private val measurePaint = ThreadLocalPaint()
    }

    override fun toString(): String {
        return "CellIconSizeSpecs(type=$type, config=$config, night_mode_auto=$night_mode_auto, iconSizePx=$iconSizePx, cellWidthPx=$cellWidthPx, cellHeightPx=$cellHeightPx, iconDrawablePaddingPx=$iconDrawablePaddingPx, iconTextSizePx=$iconTextSizePx, iconTextColor=$iconTextColor, rotate90Widgets=$rotate90Widgets, is_night_mode=$is_night_mode, subject=$subject)"
    }

    fun copyWithoutNightMode(): CellIconSizeSpecs {
        val parent = this
        return CellIconSizeSpecs(type, config, null).apply {
            iconSizePx = parent.iconSizePx
            cellWidthPx = parent.cellWidthPx
            cellHeightPx = parent.cellHeightPx
            iconDrawablePaddingPx = parent.iconDrawablePaddingPx
            iconTextSizePx = parent.iconTextSizePx
            iconTextColor = parent.iconTextColor
            backgroundColor = parent.backgroundColor
        }
    }
}


private const val AVERAGE_WALLPAPER_COLOR = 0xFF404040.toInt()
const val LABEL_FOREGROUND_LIGHT = 0xFFF0F0F0.toInt()
const val LABEL_FOREGROUND_MEDIUM_DARK = 0xFF505050.toInt()
const val LABEL_FOREGROUND_DARK = 0xFF202020.toInt()

fun getForegroundColorForBackground(@ColorInt backgroundColor: Int): Int {
    val flatBgColor = ColorUtil.compositeColors(backgroundColor, AVERAGE_WALLPAPER_COLOR)
    val lightContrast = ColorUtil.calculateContrast(LABEL_FOREGROUND_LIGHT, flatBgColor)
    val mediumDarkContrast = ColorUtil.calculateContrast(LABEL_FOREGROUND_MEDIUM_DARK, flatBgColor)
    return if (lightContrast > mediumDarkContrast) {
        LABEL_FOREGROUND_LIGHT
    } else {
        if (mediumDarkContrast < 4.5)
            LABEL_FOREGROUND_DARK
        else
            LABEL_FOREGROUND_MEDIUM_DARK
    }
}

enum class CellLayoutType {
    DESKTOP,
    DOCK,
    DRAWER,
    FOLDER
}
