package com.teslacoilsw.launcher.backup

import android.appwidget.AppWidgetHost
import android.content.ComponentName
import android.content.Context
import android.content.Intent
import android.content.SharedPreferences
import android.database.sqlite.SQLiteDatabase
import android.graphics.Color
import android.net.Uri
import android.preference.PreferenceManager
import aosp.android.internal.util.XmlUtils
import com.android.launcher3.ItemInfo
import com.android.launcher3.Launcher
import com.android.launcher3.LauncherAppState
import com.android.launcher3.LauncherAppWidgetInfo
import com.android.launcher3.LauncherModel
import com.android.launcher3.LauncherSettings
import com.android.launcher3.WorkspaceScreenInfo
import com.android.launcher3.util.ContentWriter
import com.teslacoilsw.cofs.flattenToString
import com.teslacoilsw.kotlin.tryCatch
import com.teslacoilsw.kotlin.withAlpha
import com.teslacoilsw.launcher.CellSpecConfig
import com.teslacoilsw.launcher.NovaAction
import com.teslacoilsw.launcher.NovaLauncher
import com.teslacoilsw.launcher.R
import com.teslacoilsw.launcher.drawer.drawergroups.DrawerGroup
import com.teslacoilsw.launcher.drawer.drawergroups.DrawerGroupModel
import com.teslacoilsw.launcher.icon.AdaptiveIconShape
import com.teslacoilsw.launcher.icontheme.IconPackType
import com.teslacoilsw.launcher.icontheme.IconThemeConfig
import com.teslacoilsw.launcher.importer.ApexImporter
import com.teslacoilsw.launcher.iteminfo.DrawerFolderInfo
import com.teslacoilsw.launcher.onboarding.ImportedData
import com.teslacoilsw.launcher.preferences.ActivityAnimation
import com.teslacoilsw.launcher.preferences.FolderIconConfig
import com.teslacoilsw.launcher.preferences.FolderScrollDirection
import com.teslacoilsw.launcher.preferences.FolderWindowConfig
import com.teslacoilsw.launcher.preferences.FolderWindowStyle
import com.teslacoilsw.launcher.preferences.Pref
import com.teslacoilsw.launcher.preferences.Pref3
import com.teslacoilsw.launcher.preferences.SearchBarPlacement
import com.teslacoilsw.launcher.preferences.WallpaperScrollMode
import com.teslacoilsw.launcher.preferences.initOrUpgrade
import com.teslacoilsw.launcher.quicksearchbar.SearchBarStyle
import com.teslacoilsw.launcher.util.MaterialColors
import com.teslacoilsw.launcher.util.NamedIntent
import com.teslacoilsw.launcher.widget.TypefaceHelper
import java.io.BufferedInputStream
import java.io.File
import java.io.FileInputStream
import java.io.FileOutputStream
import java.util.zip.ZipInputStream
import kotlin.math.roundToInt

/* TODO */
/*
 * drawer_icon_style KITKAT or LOLLIPOP allapps button icon
 *
 */
object ApexBackup {
    private val VALID_MARGINS = arrayOf("NONE", "SMALL", "MEDIUM", "LARGE")

    private fun apexFontToNova(apex: String): String {
        return when (apex) {
            "SYSTEM" -> TypefaceHelper.CONDENSED
            "THEME" -> TypefaceHelper.CONDENSED
            "ROBOTO_REGULAR" -> ""
            "ROBOTO_LIGHT" -> TypefaceHelper.LIGHT
            "ROBOTO_CONDENSED" -> TypefaceHelper.CONDENSED
            "ROBOTO_THIN" -> TypefaceHelper.LIGHT
            "ROBOTO_MEDIUM" -> TypefaceHelper.MEDIUM
            else -> TypefaceHelper.CONDENSED
        }
    }

    fun loadFromDirectory(context: Context, dir: File) {
        NovaLauncher.getLauncher()?.finish()
        val prefs = PreferenceManager.getDefaultSharedPreferences(context)
        val prime = if (prefs.contains(Pref.Key.NOVA_PRIME_ENABLED)) {
            prefs.getInt(Pref.Key.NOVA_PRIME_ENABLED, 512)
        } else {
            null
        }
        prefs.edit().clear().apply()
        initOrUpgrade(context, prefs)
        val editor = prefs.edit()
        if (prime != null)
            editor.putInt(Pref.Key.NOVA_PRIME_ENABLED, prime)
        editor.putBoolean("did_quickstart", true)
        val settings = tryCatch(null) { XmlUtils.readMapXml(FileInputStream(File(dir, "apex_settings.bak"))) as? Map<String, *> }
        val swipeActions = tryCatch(null) { XmlUtils.readMapXml(FileInputStream(File(dir, "SwipeActions.bak"))) as? Map<String, *> }
        val drawerGroups = tryCatch(null) { XmlUtils.readMapXml(FileInputStream(File(dir, "DrawerGroups.bak"))) as? Map<String, *> }

        if (settings != null)
            loadSettings(context, editor, settings)

        LauncherAppState.getLauncherProvider()!!.initializeMaxId()
        val db = SQLiteDatabase.openDatabase(File(dir, "apex_data.bak").path, null, 0)
        loadDatabase(context, db, settings, swipeActions)
        LauncherAppState.getLauncherProvider()!!.initializeMaxId()
        if (drawerGroups != null)
            loadDrawerGroups(context, settings, drawerGroups)
        db.close()
        editor.apply()
    }

    fun loadImportedDataFromZip(context: Context, uri: Uri): ImportedData {
        context.contentResolver.openInputStream(uri).use {
            val zip = ZipInputStream(BufferedInputStream(it))
            var apexDb: File? = null
            var settings: Map<String, *>? = null
            var swipeActions: Map<String, *>? = null
            var drawerGroup: Map<String, *>? = null
            var screens: List<WorkspaceScreenInfo>? = null
            var items: List<ItemInfo>? = null
            try {
                var ze = zip.nextEntry
                while (ze != null) {
                    when {
                        "apex_settings.bak" == ze.name -> settings = XmlUtils.readMapXml(zip) as? Map<String, *>
                        "SwipeActions.bak" == ze.name -> swipeActions = XmlUtils.readMapXml(zip) as? Map<String, *>
                        "DrawerGroups.bak" == ze.name -> drawerGroup = XmlUtils.readMapXml(zip) as? Map<String, *>
                        "apex_data.bak" == ze.name -> {
                            apexDb = File(context.cacheDir, "apex.db")
                            apexDb.deleteOnExit()
                            val os = FileOutputStream(apexDb)
                            val buffer = ByteArray(4096)
                            var bytesRead = zip.read(buffer)
                            while (bytesRead != -1) {
                                os.write(buffer, 0, bytesRead)
                                bytesRead = zip.read(buffer)
                            }
                            os.close()
                        }
                    }
                    ze = zip.nextEntry
                }
                if (apexDb != null) {
                    val db = SQLiteDatabase.openDatabase(apexDb.path, null, 0)
                    val x = loadDatabaseItems(context, db, settings, swipeActions)
                    screens = x.first
                    items = x.second
                    db.close()
                    apexDb.delete()
                }
            } catch (e: Exception) {
                e.printStackTrace()
            }

            val task = { editor: SharedPreferences.Editor ->
                if (drawerGroup != null)
                    loadDrawerGroups(context, settings, drawerGroup)

                if (settings != null)
                    loadSettings(context, editor, settings)
            }


            if (settings != null) {
                val smallestWidthDp = context.resources.configuration.smallestScreenWidthDp
                val widthDp = context.resources.configuration.screenWidthDp
                val heightDp = context.resources.configuration.screenHeightDp

                val gridCols = (settings["homescreen_portrait_grid_columns"]
                        ?: settings["home_screen_grid_columns"]) as? Int ?: when {
                    smallestWidthDp >= 768 -> 6
                    smallestWidthDp >= 720 -> 8
                    smallestWidthDp >= 600 -> 6
                    smallestWidthDp >= 410 -> 5
                    else -> 4
                }
                val gridRows = (settings["homescreen_portrait_grid_rows"]
                        ?: settings["home_screen_grid_rows"]) as? Int ?: when {
                    smallestWidthDp >= 768 -> 6
                    smallestWidthDp >= 720 -> 5
                    smallestWidthDp >= 600 -> 6
                    smallestWidthDp >= 410 -> 5
                    widthDp >= 360 && heightDp >= 644 -> 5
                    else -> 4
                }
                val dockIcons = settings["number_of_dock_icons"] as? Int ?: 5
                val dockPages = settings["number_of_dock_pages"] as? Int ?: 1
                val defaultScreen = settings["default_screen"] as? Int ?: 0
                val defaultScreenId = screens?.get(defaultScreen)?.id ?: 0

                return ImportedData(items!!, screens!!, gridRows, gridCols, defaultScreenId, dockIcons, dockPages, task)
            }
        }
        throw IllegalArgumentException()
    }

    fun loadFromZip(context: Context, uri: Uri) {
        NovaLauncher.getLauncher()?.finish()
        val prefs = PreferenceManager.getDefaultSharedPreferences(context)
        val prime = if (prefs.contains(Pref.Key.NOVA_PRIME_ENABLED)) {
            prefs.getInt(Pref.Key.NOVA_PRIME_ENABLED, 512)
        } else {
            null
        }
        prefs.edit().clear().apply()
        initOrUpgrade(context, prefs)
        val editor = prefs.edit()
        if (prime != null)
            editor.putInt(Pref.Key.NOVA_PRIME_ENABLED, prime)
        editor.putBoolean("did_quickstart", true)
        val data = loadImportedDataFromZip(context, uri)

        val cr = context.contentResolver
        val host = AppWidgetHost(context, Launcher.APPWIDGET_HOST_ID)
        AppWidgetHost.deleteAllHosts()
        LauncherAppState.getLauncherProvider().createEmptyDB()
        LauncherModel.deleteAll(context)

        LauncherModel.updateWorkspaceScreenOrder(context, data.screens)
        for (item in data.items) {
            val writer = ContentWriter(context)
            writer.put(LauncherSettings.Favorites._ID, item.id)
            item.onAddToDatabase(writer)
            cr.insert(LauncherSettings.Favorites.CONTENT_URI, writer.valuesUnsafe)
        }

        data.task?.invoke(editor)
        editor.apply()
    }

    fun loadSettings(context: Context, editor: SharedPreferences.Editor, map: Map<String, *>): Boolean {
        val apexVersion = map["previous_app_version_code"] as? Int ?: 3305
        val isApex33 = apexVersion < 4000

        val baseIconScale = (map["icon_size"] as? Int ?: 100).toFloat() / 100f
        val baseIconLabelSize = (map["icon_text_size"] as? Int ?: 100) / 100f * 13f
        val baseIconFont = (map["icon_font"] as String?) ?: "ROBOTO_CONDENSED"

        ((map["homescreen_transition_effect"]
                ?: map["home_screen_transition_effect"]) as? String).let {
            val scrollEffectName = when (it) {
                "NONE" -> Pref.Key.ScrollEffectEnum.SIMPLE.name
                "TABLET" -> Pref.Key.ScrollEffectEnum.TABLET.name
                "CARD_STACK" -> Pref.Key.ScrollEffectEnum.CARDSTACK.name
                "CUBE_IN" -> Pref.Key.ScrollEffectEnum.CUBE.name
                "CUBE" -> Pref.Key.ScrollEffectEnum.CUBE.name
                "ACCORDION" -> Pref.Key.ScrollEffectEnum.ACCORDION.name
                "CROSS" -> Pref.Key.ScrollEffectEnum.REVOLVING_DOOR.name
                "FLIP" -> Pref.Key.ScrollEffectEnum.FLIP.name
                "OVERLAP" -> Pref.Key.ScrollEffectEnum.REVOLVING_DOOR.name
                "ROTATE" -> Pref.Key.ScrollEffectEnum.THROW.name
                "SCALE" -> Pref.Key.ScrollEffectEnum.ZOOMY.name
                "WHEEL" -> Pref.Key.ScrollEffectEnum.THROW.name
                else -> Pref.Key.ScrollEffectEnum.SIMPLE.name
            }
            editor.putString(Pref.Key.DESKTOP_SCROLL_EFFECT, scrollEffectName)
        }

        (map["drawer_transition_effect"] as? String).let {
            val scrollEffectName = when (it) {
                "NONE" -> Pref.Key.ScrollEffectEnum.SIMPLE.name
                "TABLET" -> Pref.Key.ScrollEffectEnum.TABLET.name
                "CARD_STACK" -> Pref.Key.ScrollEffectEnum.CARDSTACK.name
                "CUBE_IN" -> Pref.Key.ScrollEffectEnum.CUBE.name
                "CUBE" -> Pref.Key.ScrollEffectEnum.CUBE.name
                "ACCORDION" -> Pref.Key.ScrollEffectEnum.ACCORDION.name
                "CROSS" -> Pref.Key.ScrollEffectEnum.REVOLVING_DOOR.name
                "FLIP" -> Pref.Key.ScrollEffectEnum.FLIP.name
                "OVERLAP" -> Pref.Key.ScrollEffectEnum.REVOLVING_DOOR.name
                "ROTATE" -> Pref.Key.ScrollEffectEnum.THROW.name
                "SCALE" -> Pref.Key.ScrollEffectEnum.ZOOMY.name
                "WHEEL" -> Pref.Key.ScrollEffectEnum.THROW.name
                else -> Pref.Key.ScrollEffectEnum.SIMPLE.name
            }
            editor.putString(Pref.Key.DRAWER_SCROLL_EFFECT, scrollEffectName)
        }
        // apps.new.page
        // menu_wallpaper
        // menu_theme
        // widget_padding
        // dock_elastic_scrolling
        // show_wallpaper
        // search_engine
        // hide_home_screen_crosshairs
        // enable_usage_reporting
        // widgets_in_dock
        // enable_quick_action // enable icon long-press popup menu
        // sms_app
        // drawer_touch_feedback
        // keep_in_memory
        // dock_scroll_speed
        // icon_long_press_feedback
        // resize_any_widget
        // notify_import_data
        // dock_scrolling
        // drawer_apps_sorting
        // drawer_normalize_icon_size
        // auto_merge_folders
        // previous_app_version_code
        // home_screen_scroll_speed
        // home_screen_elastic_scrolling
        // drawer_frequent_app
        // home_screen_overscroll_effect
        val folderIconScale = (map["folder_icon_scale"] ?: map["folder_icon_size"]) as? Int ?: 100
        val folderLabels = (map["folder_label_icons"]
                ?: map["folder_show_labels"]) as? Boolean ?: true
        val folderLabelShadow = map["folder_label_shadow"] as? Boolean ?: false
        val folderLabelColor = (map["folder_label_color"] as? Int
                ?: 0xFF666666.toInt()).withAlpha(255)
        val folderFont = apexFontToNova(map["folder_label_font"] as? String ?: baseIconFont)
        val folderSpecs = CellSpecConfig(folderIconScale.toFloat() / 100f * baseIconScale, folderLabels, baseIconLabelSize, folderLabelColor, folderLabelShadow, folderFont, true, false)
        editor.putString(Pref3.folder_cellspecs.key, CellSpecConfig.flattenToString(folderSpecs))

        val desktopIconScale = (map["homescreen_icon_scale"]
                ?: map["home_screen_icon_size"]) as? Int ?: 100
        val desktopLabels = (map["homescreen_label_icons"]
                ?: map["home_screen_show_labels"]) as? Boolean ?: true
        val desktopLabelShadow = (map["homescreen_label_shadows"]
                ?: map["home_screen_label_shadow"]) as? Boolean ?: true
        val desktopLabelColor = (map["homescreen_label_color"]
                ?: map["home_screen_label_color"]) as? Int ?: 0xFFFFFFFF.toInt()
        val desktopLabelFont = apexFontToNova(map["home_screen_label_font"] as? String
                ?: baseIconFont)
        val desktopSpecs = CellSpecConfig(desktopIconScale.toFloat() / 100f * baseIconScale, desktopLabels, baseIconLabelSize, desktopLabelColor, desktopLabelShadow, desktopLabelFont, true, false)
        editor.putString(Pref3.desktop_cellspecs.key, CellSpecConfig.flattenToString(desktopSpecs))

        editor.putBoolean(Pref3.drawer_searchbar.key, map["drawer_show_search_bar"] as? Boolean ?: false)
        editor.putString(Pref3.drawer_search_style.key, SearchBarStyle.DRAWER_LEGACY.flattenToString())
        editor.putBoolean(Pref.Key.DRAWER_PREDICTIVE_APPS, map["drawer_frequent_app"] as? Boolean ?: false)

        val drawerIconScale = map["drawer_icon_scale"] as? Int ?: 100
        val drawerLabels = map["drawer_label_icons"] as? Boolean ?: true
        val drawerLabelShadow = map["drawer_label_shadows"] as? Boolean ?: false
        val drawerLabelColor = map["drawer_label_color"] as? Int ?: 0xff666666.toInt()
        val drawerLabelFont = apexFontToNova(map["drawer_label_font"] as? String ?: baseIconFont)
        val drawerSpecs = CellSpecConfig(drawerIconScale.toFloat() / 100f * baseIconScale, drawerLabels, baseIconLabelSize, drawerLabelColor, drawerLabelShadow, drawerLabelFont, true, false)
        editor.putString(Pref3.drawer_cellspecs.key, CellSpecConfig.flattenToString(drawerSpecs))

        (map["drawer_tab_style"] as? String)?.let {
            when (it) {
                "NO_TABS", "SMALL_TABS"-> editor.putBoolean(Pref3.drawer_enable_tabs.key, false)
                "REGULAR_TABS" -> {
                    editor.putBoolean(Pref3.drawer_enable_tabs.key, true)
                    editor.putString(Pref.Key.DRAWER_TAB_STYLE, Pref.Key.TabStyle.CLASSIC.name)
                }
                else -> {
                    editor.putBoolean(Pref3.drawer_enable_tabs.key, true)
                    editor.putString(Pref.Key.DRAWER_TAB_STYLE, Pref.Key.TabStyle.MATERIAL.name)
                }
            }
        }

        val dockIconScale = map["dock_icon_scale"] as? Int
                ?: if (LauncherAppState.isScreenLarge(context.resources)) 90 else 100
        val dockSpecs = CellSpecConfig(dockIconScale.toFloat() / 100f * baseIconScale, false, baseIconLabelSize, desktopLabelColor, desktopLabelShadow, desktopLabelFont, true, false)
        editor.putString(Pref3.dock_cellspecs.key, CellSpecConfig.flattenToString(dockSpecs))


        (map["folder_animation"] as? String)?.let {
            val v = when (it) {
                "SCALE" -> Pref.Key.DrawerAnimation.ZOOM
                "CIRCLE" -> Pref.Key.DrawerAnimation.CIRCLE
                else -> Pref.Key.DrawerAnimation.CIRCLE
            }
            editor.putString(Pref3.folder_animation.key, v.name)
        }

        val indicator = if ((map["show_homescreen_indicator"]
                        ?: map["show_home_screen_indicator"]) as? Boolean != false) {
            when ((map["homescreen_indicator_style"]
                    ?: map["home_screen_indicator_style"]) as? String ?: "DOT") {
                "DOT" -> Pref.Key.ScrollIndicatorStyle.DOT
                "DOT_KITKAT" -> Pref.Key.ScrollIndicatorStyle.CIRCLE
                "LINE" -> Pref.Key.ScrollIndicatorStyle.LINE
                else -> Pref.Key.ScrollIndicatorStyle.DOT
            }
        } else {
            Pref.Key.ScrollIndicatorStyle.NONE
        }
        editor.putString(Pref3.desktop_page_indicator.key, indicator.name)

        ((map["homescreen_vertical_margin"]
                ?: map["home_screen_vertical_margin"]) as? String)?.let {
            if (it in VALID_MARGINS)
                editor.putString(Pref.Key.DESKTOP_HEIGHT_MARGIN, it)
        }
        ((map["homescreen_horizontal_margin"]
                ?: map["home_screen_horizontal_margin"]) as? String)?.let {
            if (it in VALID_MARGINS)
                editor.putString(Pref.Key.DESKTOP_WIDTH_MARGIN, it)
        }

        /* NONE PIXEL RING DARK_RING SQUARE IOS THEME CUSTOM */
        /* LIGHT DARK HOLO_LIGHT HOLO_DARK THEME */
        val folderPreview = when (map["folder_preview"] as? String ?: "GRID") {
            "STACK" -> Pref.Key.FolderPreview.STACK
            "FAN" -> Pref.Key.FolderPreview.FAN
            "LINE" -> Pref.Key.FolderPreview.LINE
            "GRID" -> Pref.Key.FolderPreview.GRID
            "IOS" -> Pref.Key.FolderPreview.GRID
            else -> Pref.Key.FolderPreview.GRID
        }

        val folderIconConfig = when ((map["folder_background"] ?: map["folder_icon_background"]) as? String ?: "PIXEL") {
            "NONE" -> FolderIconConfig(AdaptiveIconShape.NONE, folderPreview, false, Color.WHITE, 0)
            "PIXEL" -> FolderIconConfig(AdaptiveIconShape.CIRCLE, folderPreview, true, Color.WHITE, 160)
            "RING" -> FolderIconConfig(AdaptiveIconShape.CIRCLE, folderPreview, false, Color.WHITE, 160)
            "DARK_RING" -> FolderIconConfig(AdaptiveIconShape.CIRCLE, folderPreview, false, Color.BLACK, 0xff)
            "SQUARE" -> FolderIconConfig(AdaptiveIconShape.SQUARE, folderPreview, false, Color.BLACK, 0xFF)
            "IOS" -> FolderIconConfig(AdaptiveIconShape.SQUIRCLE, folderPreview, false, 0xFF2B2D33.toInt(), 0xFF)
            else -> FolderIconConfig.DEFAULT
        }
        editor.putString(Pref3.folder_icon_config.key, folderIconConfig.flattenToString())

        val folderBgStyle = map["folder_background_style"] as? String ?: "LIGHT"
        val folderWinAlpha = (map["folder_background_alpha"] as? Int ?: 100) * 255 / 100
        val folderShowName = !(map["hide_folder_name"] as? Boolean ?: false)
        val folderWinConfig = when (folderBgStyle) {
            "LIGHT" -> FolderWindowConfig(FolderWindowStyle.WINDOW, folderWinAlpha, 0xffF5F5F5.toInt(), 4, true, folderShowName, 4, 4, FolderScrollDirection.VERTICAL)
            "DARK" -> FolderWindowConfig(FolderWindowStyle.WINDOW, folderWinAlpha, MaterialColors.PANEL_DARK, 4, true, folderShowName, 4, 4, FolderScrollDirection.VERTICAL)
            "HOLO_LIGHT" -> FolderWindowConfig(FolderWindowStyle.WINDOW, (folderWinAlpha * .9f).roundToInt(), Color.WHITE, 2, true, folderShowName, 4, 4, FolderScrollDirection.VERTICAL)
            "HOLO_DARK" -> FolderWindowConfig(FolderWindowStyle.WINDOW, (folderWinAlpha * .9f).roundToInt(), Color.BLACK, 2, true, folderShowName, 4, 4, FolderScrollDirection.VERTICAL)
            else -> FolderWindowConfig.DEFAULT
        }
        editor.putString(Pref3.folder_window_config.key, FolderWindowConfig.flattenToString(folderWinConfig))

        (map["dock_infinite_scrolling"] as? Boolean)?.let {
            editor.putBoolean(Pref3.dock_infinite_scroll.key, it)
        }
        (map["dock_background"] as? String ?: "NONE")?.let {

        }

        (map["drawer_style"])?.let {
            val style = when (it) {
                "HORIZONTAL_PAGINATED" -> Pref.Key.DrawerStyle.HORIZONTAL_PAGINATED
                "VERTICAL_PAGINATED" -> Pref.Key.DrawerStyle.VERTICAL
                "VERTICAL_CONTINUOUS" -> Pref.Key.DrawerStyle.VERTICAL
                "VERTICAL_LIST" -> Pref.Key.DrawerStyle.VERTICAL_LIST
                else -> Pref.Key.DrawerStyle.VERTICAL
            }
            editor.putString(Pref.Key.DRAWER_STYLE, style.name)
        }
        (map["drawer_background_style"])?.let {
            val (card, color) = when (it) {
                "NO_CARD" -> false to Color.WHITE
                "LIGHT_CARD" -> true to Color.WHITE
                "DARK_CARD" -> true to MaterialColors.PANEL_DARK
                "HOLO_LIGHT_CARD" -> true to Color.WHITE.withAlpha(0xE6)
                "HOLO_DARK_CARD" -> true to Color.BLACK.withAlpha(0xE6)
                else -> false to Color.WHITE
            }
            editor.putBoolean(Pref3.drawer_cards.key, card)
            editor.putInt(Pref3.drawer_card_color.key, color)
        }
        var drawerBgAlpha = map["drawer_background_alpha"] as? Int ?: if (isApex33) 65 else 100
        if (isApex33)
            drawerBgAlpha = drawerBgAlpha * 255/100
        val drawerBgColor = map["drawer_background_color"] as? Int ?: if (isApex33) Color.BLACK else 0xFFF2F2F2.toInt()
        editor.putInt(Pref3.drawer_background_color.key, drawerBgColor.withAlpha(drawerBgAlpha))
        (map["drawer_close_on_launch"] as? Boolean)?.let {
            editor.putBoolean(Pref3.drawer_autoclose.key, it)
        }

        (map["drawer_landscape_grid_columns"] as? Int)?.let {
            editor.putInt(Pref.Key.DRAWER_APP_GRID_LAND_COLS, it)
        }
        (map["drawer_landscape_grid_rows"] as? Int)?.let {
            editor.putInt(Pref.Key.DRAWER_APP_GRID_LAND_ROWS, it)
        }

        (map["drawer_portrait_grid_columns"] as? Int)?.let {
            editor.putInt(Pref.Key.DRAWER_APP_GRID_COLS, it)
        }
        (map["drawer_portrait_grid_rows"] as? Int)?.let {
            editor.putInt(Pref.Key.DRAWER_APP_GRID_ROWS, it)
        }
        /* 3.3 ICE_CREAM_SANDWICH JELLY_BEAN KITKAT LOLLIPOP MARSHMALLOW
        *  4.0 RECTANGLE_COLORFUL RECTANGLE_GREY PILL_COLORFUL PILL_GREY  */
        val search_bar_style = map["search_bar_style"] as? String ?: "MARSHMALLOW"
        var searchbar = when (search_bar_style) {
            "ICE_CREAM_SANDWICH" -> "CLASSIC_MONO:WIRE:#FFFFFFFF:56"
            "MARSHMALLOW", "RECTANGLE_COLORFUL" -> "M_COLOR:ROUNDRECT:#FFFFFFFF:56"
            "LOLLIPOP" -> "CLASSIC_MONO:ROUNDRECT:#FFFFFFFF:56"
            "JELLY_BEAN", "KITKAT" -> "CLASSIC_MONO:ROUNDRECT:#64FFFFFF:56"
            "RECTANGLE_GREY" -> "M_LETTER_MONO:ROUNDRECT:#FFFFFFFF:56"
            "PILL_COLORFUL" -> "M_LETTER_COLOR:PILLRECT:#FFFFFFFF:56"
            "PILL_GREY" -> "M_LETTER_MONO:PILLRECT:#FFFFFFFF:56"
            else -> SearchBarStyle.DEFAULT.flattenToString()
        }
        if (map["use_tablet_ui"] == true)
            searchbar = "CLASSIC_MONO:HONEYCOMB:#FFFFFFFF:56"
        if (map["search_as_overlay"] == true)
            editor.putString(Pref3.search_window_style.key, SearchBarStyle.SearchWindowStyle.GOOGLE_OVERLAY.name)
        editor.putString(Pref3.qsb_style.key, searchbar)
        if (map["show_persistent_search_bar"] == "ALWAYS") {
            editor.putString(Pref3.searchbar_placement.key, SearchBarPlacement.PERSISTENT.name)
        }

        if (map["drawer_join_tabs"] == false) {
            editor.putBoolean(Pref3.drawer_isolate_tabs.key, true)
        }

        (map["drawer_show_swipe_indicator"] as? Boolean)?.let {
            editor.putBoolean(Pref3.drawer_swipe_indicator.key, it)
        }

        (map["drawer_infinite_scrolling"] as? Boolean)?.let {
            editor.putBoolean(Pref3.drawer_infinite_scroll.key, it)
        }

        (map["dock_horizontal_margin"] as? String)?.let {
            if (it in VALID_MARGINS) {
                editor.putString(Pref.Key.DOCK_WIDTH_MARGIN, it)
            }
        }
        (map["number_of_dock_icons"] as? Int)?.let {
            editor.putInt(Pref.Key.DOCK_GRID_COLS, it)
        }
        (map["dock_as_overlay"] as? Boolean)?.let {
            editor.putBoolean(Pref3.dock_overlay.key, it)
        }
        (map["number_of_dock_pages"] as? Int)?.let {
            editor.putInt(Pref.Key.DOCK_PAGES, it)
        }

        val smallestWidthDp = context.resources.configuration.smallestScreenWidthDp
        val widthDp = context.resources.configuration.screenWidthDp
        val heightDp = context.resources.configuration.screenHeightDp

        val gridCols = (map["homescreen_portrait_grid_columns"] ?: map["home_screen_grid_columns"]) as? Int ?: when {
            smallestWidthDp >= 768 -> 6
            smallestWidthDp >= 720 -> 8
            smallestWidthDp >= 600 -> 6
            smallestWidthDp >= 410 -> 5
            else -> 4
        }
        editor.putInt(Pref.Key.DESKTOP_GRID_COLS, gridCols)
        val gridRows = (map["homescreen_portrait_grid_rows"] ?: map["home_screen_grid_rows"]) as? Int ?: when {
            smallestWidthDp >= 768 -> 6
            smallestWidthDp >= 720 -> 5
            smallestWidthDp >= 600 -> 6
            smallestWidthDp >= 410 -> 5
            widthDp >= 360 && heightDp >= 644 -> 5
            else -> 4
        }
        editor.putInt(Pref.Key.DESKTOP_GRID_ROWS, gridRows)

        var desktopInfinite = (map["home_screen_infinite_scrolling"] ?: "JUMP") != "NONE"
        if (map["home_screen_infinite_scrolling"] == false)
            desktopInfinite = false
        editor.putBoolean(Pref3.desktop_infinite_scroll.key, desktopInfinite)
        ((map["hide_shadows"] ?: map["hide_home_screen_shadows"]) as? Boolean)?.let {
            editor.putBoolean(Pref3.desktop_shadow.key, it)
        }

        (map["default_screen"] as? Int)?.let {
            editor.putInt(Pref.Key.DESKTOP_DEFAULT_PAGE, it)
        }

        // theme_icon_type rather than hardcode "GO" type
        (map["theme_iconpack_pkg"] as? String)?.let { pkg ->
            try {
                val pm = context.applicationContext.packageManager
                val appInfo = pm.getApplicationInfo(pkg, 0)
                val label = appInfo.loadLabel(pm)?.toString() ?: pkg
                editor.putString(Pref3.icon_theme.key, IconThemeConfig.flattenToString(IconThemeConfig(label, IconPackType.GO, pkg, null)))
            } catch (e: Exception) {
                e.printStackTrace()
            }
        }

        (map["screen_orientation"] as? String)?.let {
            val v = when (it) {
                "DEFAULT" -> Pref.Key.ScreenOrientation.NOSENSOR
                "ROTATE" -> Pref.Key.ScreenOrientation.UNSPECIFIED
                "PORTRAIT" -> Pref.Key.ScreenOrientation.PORTRAIT
                "LANDSCAPE" -> Pref.Key.ScreenOrientation.LANDSCAPE
                else -> Pref.Key.ScreenOrientation.NOSENSOR
            }
            editor.putString(Pref.Key.SCREEN_ORIENTATION, v.name)

            (map["show_dock"] as? Boolean)?.let {
                editor.putBoolean(Pref.Key.DOCK_ENABLED, it)
            }

            (map["drawer_show_search_bar"] as? Boolean)?.let {
                editor.putBoolean(Pref3.drawer_searchbar.key, it)
            }

            ((map["transparent_statusbar"]
                    ?: map["transparent_notification_bar"]) as? Boolean)?.let {
                editor.putBoolean(Pref3.notification_bar_transparency.key, it)
            }
            ((map["hide_notification_bar"]) as? Boolean)?.let {
                editor.putBoolean(Pref3.show_notification_bar.key, !it)
            }


        }
        (map["overlapping_widgets"] as? Boolean)?.let {
            editor.putBoolean(Pref3.widget_overlap.key, it)
        }
        var wallpaperScroll = (map["wallpaper_mode"] ?: "MULTIPLE_SCREEN") == "MULTIPLE_SCREEN"
        if (map["wallpaper_scrolling"] == false)
            wallpaperScroll = false
        editor.putString(Pref3.wallpaper_scroll_mode.key, (if (wallpaperScroll) WallpaperScrollMode.ON else WallpaperScrollMode.DISABLED).name)
        (map["activity_animation"] as? String)?.let {
            when (it) {
                "SYSTEM" -> ActivityAnimation.SYSTEM
                "ICE_CREAM_SANDWICH" -> ActivityAnimation.ICS
                "JELLY_BEAN" -> ActivityAnimation.JELLYBEAN
                "SLIDE_LEFT" -> ActivityAnimation.SLIDE
                "SLIDE_RIGHT" -> ActivityAnimation.SLIDE
                "SLIDE_UP" -> ActivityAnimation.SLIDE_UP
                "SLIDE_DOWN" -> ActivityAnimation.SLIDE_UP
                "ZOOM" -> ActivityAnimation.JELLYBEAN
                else -> ActivityAnimation.SYSTEM
            }
            editor.putString(Pref3.app_animation.key, it)
        }
        editor.putBoolean(Pref3.drawer_swipe_to_open.key, map["drawer_all_apps_pull_up"] as? Boolean
                ?: false)
        (map["home_key_to_default_screen"] as? Boolean)?.let {
            editor.putBoolean(Pref3.gesture_home_button_only_on_default.key, !it)
        }

        for (key in arrayOf("home_key" to "SHOW_PREVIEWS", "desktop_double_tap" to "APEX_MENU", "swipe_down" to "SHOW_NOTIFICATIONS", "swipe_up" to "APP_DRAWER", "two_finger_swipe_down" to "SHOW_QUICK_SETTINGS", "two_finger_swipe_up" to "NONE", "pinch_in" to "SHOW_PREVIEWS")) {
            val ni = namedIntentFromApexKey(context, map, key.first, key.second)
            val key = apexGestureKeyToNovaKey(key.first)
            editor.putString(key, ni.toUri())
        }

        // Apex 4.0 update modifies the following settings
        // drawer_background_color ->
        // drawer_label_color ->
        // search_bar_style ->
        // badge_color (string GREEN) -> badge_color (int)
        // badge_size (string LARGE) -> badge_size (int 120)
        return true
    }


    private fun apexGestureKeyToNovaKey(key: String): String {
        return when (key) {
            "desktop_double_tap" -> Pref3.gesture_double_tap.key
            "swipe_down" -> Pref3.gesture_swipe_down.key
            "swipe_up" -> Pref3.gesture_swipe_up.key
            "two_finger_swipe_down" -> Pref3.gesture_swipe_two_down.key
            "two_finger_swipe_up" -> Pref3.gesture_swipe_two_up.key
            "pinch_in" -> Pref3.gesture_pinch_in.key
            "home_key" -> Pref3.gesture_home_button.key
            else -> throw IllegalArgumentException("key must be one of known apex keys")
        }
    }

    private fun namedIntentFromApexKey(context: Context, map: Map<String, *>, key: String, default: String): NamedIntent {
        return namedIntentFromApexKey(context, map, default, "${key}_action", "${key}_app", "${key}_intent", "${key}_shortcut_name", "${key}_shortcut_intent")
    }

    fun loadDatabaseItems(context: Context, db: SQLiteDatabase, settings: Map<String, *>?, swipeActions: Map<String, *>?): Pair<List<WorkspaceScreenInfo>, List<ItemInfo>> {
        val cr = context.contentResolver
        val dbVersion = db.version

        val importer = ApexImporter(db, settings, swipeActions)
        val numScreens = settings?.get("number_of_screens") as? Int ?: 5

        val screens = mutableListOf<WorkspaceScreenInfo>()
        if (dbVersion > 12)
            screens.addAll(importer.getScreens(context.contentResolver) ?: emptyList())
        if (screens.isEmpty()) {
            for (i in 0 until numScreens) {
                screens += WorkspaceScreenInfo(i.toLong(), i, true)
            }
        }

        val list = mutableListOf<ItemInfo>()
        importer.query(cr)?.use { c ->
            while (c.moveToNext()) {
                val info = importer.createItemInfoFromRow(c)
                if (info != null) {
                    if (info.container == LauncherSettings.Favorites.CONTAINER_DESKTOP.toLong()) {
                        screens[info.screenId.toInt()]?.allowBlank = false
                    }
                    if (info is LauncherAppWidgetInfo) {
                        info.appWidgetId = -1
                    }
                    list.add(info)
                }
            }
        }
        return screens to list

    }

    fun loadDatabase(context: Context, db: SQLiteDatabase, settings: Map<String, *>?, swipeActions: Map<String, *>?) {
        val cr = context.contentResolver
        val host = AppWidgetHost(context, Launcher.APPWIDGET_HOST_ID)
        AppWidgetHost.deleteAllHosts()
        LauncherAppState.getLauncherProvider().createEmptyDB()
        LauncherModel.deleteAll(context)

        val (screens, items) = loadDatabaseItems(context, db, settings, swipeActions)
        LauncherModel.updateWorkspaceScreenOrder(context, screens)
        for (item in items) {
            val writer = ContentWriter(context)
            writer.put(LauncherSettings.Favorites._ID, item.id)
            item.onAddToDatabase(writer)
            cr.insert(LauncherSettings.Favorites.CONTENT_URI, writer.valuesUnsafe)
        }
    }

    fun loadDrawerGroups(context: Context, settings: Map<String, *>?, drawerGroups: Map<String, *>) {
        val cr = context.contentResolver
        val dg = DrawerGroupModel.get()
        dg.load(cr)
        for (d in dg.all) {
            DrawerGroupModel.get().deleteDrawerGroup(context.contentResolver, d)
        }
        LauncherAppState.getLauncherProvider().initializeMaxId()
        val editor = dg.edit()

        if (settings != null) {
            val rawHidden = dg.get(DrawerGroupModel.GROUP_ID_RAW_HIDDEN)
            val hidden = editor.get(rawHidden)
            (settings["drawer_hidden_apps_ex"] as? String)?.split(";")
                    ?.mapNotNull {
                        tryCatch(null) {
                            hidden.add(ComponentName.unflattenFromString(it))
                        }
                    }
            (settings["hidden_apps"] as? Set<String>)?.map {
                tryCatch(null) { hidden.add(ComponentName.unflattenFromString(it)) }
            }
        }

        val folders = (drawerGroups["all_folders"] as? String)?.split(";") ?: emptyList()
        val folderParents = mutableMapOf<String, MutableSet<ComponentName>>()
        for (id in folders) {
            val name = drawerGroups["folder_name_$id"] as? String ?: ""
            val items = (drawerGroups["folder_items_$id"] as? String)?.split(";")
            val keepInApps = (drawerGroups["folder_keep_in_tab_$id"] as? Boolean) ?: false
            val parent = (drawerGroups["folder_parent_$id"] as? String) ?: "apps_all"
            if (items != null) {
                val group = dg.addDrawerGroup(cr, name, LauncherSettings.DrawerGroups.GroupType.FOLDER_APP_GROUP, !keepInApps, Color.WHITE)
                val e = editor.get(group)
                for (app in items.mapNotNull { tryCatch(null) { ComponentName.unflattenFromString(it) } }) {
                    e.add(app)
                }
                if (parent != "apps_all") {
                    val cn = DrawerFolderInfo.getComponentName((group as DrawerGroup.FolderAppGroup).folderContainerId)
                    folderParents.getOrPut(parent) { mutableSetOf() }.add(cn)
                }
            }
        }
        val tabs = (drawerGroups["all_groups"] as String?)?.split(";") ?: emptyList()
        for (id in tabs) {
            if (id == "apps_all") {
                dg.addDrawerGroup(cr, context.getString(R.string.group_applications), LauncherSettings.DrawerGroups.GroupType.TAB_APPS, false, Color.WHITE)
                continue
            }
            val name = drawerGroups["group_name_$id"] as? String ?: ""
            val items = (drawerGroups["group_items_$id"] as? String)?.split(";")
            val isapp = drawerGroups["group_isapp_$id"] as? Boolean ?: true
            if (items != null && isapp) {
                val group = dg.addDrawerGroup(cr, name, LauncherSettings.DrawerGroups.GroupType.TAB_APP_GROUP, false, Color.WHITE)
                val e = editor.get(group)
                for (app in items.mapNotNull { tryCatch(null) { ComponentName.unflattenFromString(it) } }) {
                    e.add(app)
                }
                for (folder in folderParents[id] ?: emptySet<ComponentName>()) {
                    e.add(folder)
                }
            }
        }
        editor.commit(context.contentResolver)

    }


}

private fun Context.ni(labelId: Int, action: NovaAction) =
        NamedIntent(getString(labelId), action.intent)

fun namedIntentFromApexKey(context: Context, map: Map<String, *>, default: String, keyAction: String, keyApp: String, keyIntent: String, keyShortcut: String, keyShortcutIntent: String): NamedIntent {
    val action = (map[keyAction] as? String) ?: default
    val NONE = context.ni(R.string.none, NovaAction.NONE)
    return when (action) {
        "NONE" -> NONE
        "APP_DRAWER" -> context.ni(R.string.nova_action_app_drawer, NovaAction.APP_DRAWER)
        "SHOW_PREVIEWS" -> context.ni(R.string.nova_action_show_previews, NovaAction.SHOW_PREVIEWS)
        "GOTO_DEFAULT_SCREEN" -> context.ni(R.string.nova_action_default_screen, NovaAction.GOTO_DEFAULT_SCREEN)
        "SHOW_NOTIFICATIONS" -> context.ni(R.string.nova_action_expand_notification_bar, NovaAction.EXPAND_STATUS_BAR)
        "SHOW_QUICK_SETTINGS" -> context.ni(R.string.nova_action_expand_notification_settings_bar, NovaAction.EXPAND_STATUS_SETTINGS_BAR)
        "TOGGLE_STATUS_BAR" -> context.ni(R.string.nova_action_toggle_notification_bar, NovaAction.TOGGLE_STATUS_BAR)
        "TOGGLE_DOCK" -> context.ni(R.string.nova_action_toggle_dock, NovaAction.TOGGLE_DOCK)
        "SEARCH" -> context.ni(R.string.nova_action_text_search, NovaAction.TEXT_SEARCH)
        "VOICE_SEARCH" -> context.ni(R.string.nova_action_voice_search, NovaAction.VOICE_SEARCH)
        "GOOGLE_NOW" -> context.ni(R.string.nova_action_google_now, NovaAction.GOOGLE_NOW)
        "RECENT_APPS" -> context.ni(R.string.nova_action_show_recent_apps, NovaAction.SHOW_RECENT_APPS)
        "APEX_MENU" -> NONE
        "APEX_SETTINGS" -> context.ni(R.string.nova_settings, NovaAction.NOVA_SETTINGS)
        "LOCK_DESKTOP" -> NONE
        "SCREEN_LOCK" -> context.ni(R.string.nova_action_screen_off, NovaAction.SCREEN_OFF)
        "LAUNCH_APP" -> {
            val label = (map[keyApp] as? String) ?: "App"
            val intent = (map[keyIntent] as? String)?.let { tryCatch(null) { Intent.parseUri(it, 0) } }
            if (intent != null)
                NamedIntent(label, intent)
            else
                NONE
        }
        "LAUNCH_SHORTCUT" -> {
            val label = (map[keyShortcut] as? String) ?: "Shortcut"
            val intent = (map[keyShortcutIntent] as? String)?.let { tryCatch(null) { Intent.parseUri(it, 0) } }
            if (intent != null)
                NamedIntent(label, intent)
            else
                NONE
        }
        else -> NONE
    }
}
