package com.teslacoilsw.launcher.icon

import android.annotation.SuppressLint
import android.content.ComponentName
import android.content.Context
import android.content.res.AssetManager
import android.content.res.Configuration
import android.content.res.Resources
import android.content.res.XmlResourceParser
import android.graphics.Bitmap
import android.graphics.BitmapFactory
import android.graphics.BitmapShader
import android.graphics.Canvas
import android.graphics.Color
import android.graphics.Paint
import android.graphics.Path
import android.graphics.PathMeasure
import android.graphics.Rect
import android.graphics.Shader
import android.graphics.SweepGradient
import android.graphics.drawable.AdaptiveIconDrawable
import android.graphics.drawable.ColorDrawable
import android.graphics.drawable.Drawable
import android.os.Build
import android.util.SparseIntArray
import androidx.core.content.ContextCompat
import androidx.core.graphics.ColorUtils
import androidx.core.graphics.applyCanvas
import axmlprinter2.AXmlResourceParser
import com.android.launcher3.IconCache
import com.android.launcher3.LauncherAppState
import com.android.launcher3.compat.LauncherActivityInfoCompat
import com.android.launcher3.dragndrop.AdaptiveBgAnimationMode
import com.android.launcher3.graphics.ShadowGenerator
import com.teslacoilsw.cofs.Cofs
import com.teslacoilsw.kotlin.resetAfter
import com.teslacoilsw.launcher.BuildConfig
import com.teslacoilsw.launcher.R
import com.teslacoilsw.launcher.bitmaputils.BitmapUtils
import com.teslacoilsw.launcher.bitmaputils.ReusableBufferedInputStream
import com.teslacoilsw.launcher.bitmaputils.ThreadLocalCanvas
import com.teslacoilsw.launcher.bitmaputils.ThreadLocalMatrix
import com.teslacoilsw.launcher.bitmaputils.ThreadLocalPaint
import com.teslacoilsw.launcher.bitmaputils.ThreadLocalPath
import com.teslacoilsw.launcher.bitmaputils.ThreadLocalRect
import com.teslacoilsw.launcher.preferences.Pref
import com.teslacoilsw.launcher.preferences.Pref3
import com.teslacoilsw.launcher.util.DpUtil
import com.teslacoilsw.launcher.util.palette.Palette
import com.teslacoilsw.launcher.util.palette.PaletteHelper
import com.teslacoilsw.shared.colorpicker.ColorUtil
import com.teslacoilsw.shared.util.MathUtils
import com.teslacoilsw.shared.util.ReflectionHelper
import org.xmlpull.v1.XmlPullParser
import org.xmlpull.v1.XmlPullParserException
import timber.log.Timber
import java.io.File
import java.io.FileNotFoundException
import java.io.FileOutputStream
import java.io.IOException
import java.util.*
import java.util.zip.ZipFile


private const val ANDROID_NAMESPACE = "http://schemas.android.com/apk/res/android"

private val sCanvasThreadLocal = ThreadLocalCanvas()
private val sPathThreadLocal = arrayOf(ThreadLocalPath(), ThreadLocalPath())
private val sRectThreadLocal = arrayOf(ThreadLocalRect(), ThreadLocalRect())
private val sMatrixThreadLocal = ThreadLocalMatrix()
private val sPaintThreadLocal = ThreadLocalPaint()

private var sAdaptiveMask = AdaptiveIconShape.NONE
private var sAdaptiveShadow: Bitmap? = null

private val ICONRESID_EMPTY_RESULT = Pair(0,0)

fun getIconResIdFromManifest(info: LauncherActivityInfoCompat, resources: Resources): Pair<Int, Int> {

    if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.O) {
        val zf = try { ZipFile(File(info.applicationInfo.sourceDir)) } catch (e: Exception) { return ICONRESID_EMPTY_RESULT }
        return try {
            val ze = zf.getEntry("AndroidManifest.xml")
            val xpp = AXmlResourceParser(ReusableBufferedInputStream(zf.getInputStream(ze)))

            getIconResIdFromManifest(xpp, info.componentName)

        } catch (e: IOException) {
            e.printStackTrace()
            ICONRESID_EMPTY_RESULT
        } finally {
            try {
                zf.close()
            } catch (e: Exception) {
            }
        }
    } else {
        return getIconResIdFromManifest(resources.assets.openXmlResourceParser("AndroidManifest.xml"), info.componentName)
    }
}

private fun getIconResIdFromManifest(xpp: XmlResourceParser, componentName: ComponentName): Pair<Int, Int> {
    var iconResId = 0
    var roundIconResId = 0

    try {
        while (xpp.eventType != XmlPullParser.END_DOCUMENT) {
            if (xpp.eventType == XmlPullParser.START_TAG) {
                val tagname = xpp.name
                if ("application" == tagname) {
                    iconResId = xpp.getAttributeResourceValue(ANDROID_NAMESPACE, "icon", 0)
                    roundIconResId = xpp.getAttributeResourceValue(ANDROID_NAMESPACE, "roundIcon", 0)
                } else if ("activity" == tagname) {
                    val name = xpp.getAttributeValue(ANDROID_NAMESPACE, "name")
                    if (componentName.className == name) {
                        val icon = xpp.getAttributeResourceValue(ANDROID_NAMESPACE, "icon", 0)
                        val roundIcon = xpp.getAttributeResourceValue(ANDROID_NAMESPACE, "roundIcon", 0)
                        if (icon != 0) {
                            iconResId = icon
                            roundIconResId = roundIcon
                        }
                        break
                    }
                }
            }
            xpp.next()
        }
    } catch (e: Exception) {
        if (BuildConfig.DEBUG)
            e.printStackTrace()
    } finally {
        try {
            xpp.close()
        } catch (e: Exception) {
            if (BuildConfig.DEBUG)
                e.printStackTrace()
        }
    }

    if (iconResId != 0)
        return Pair(iconResId, roundIconResId)

    return ICONRESID_EMPTY_RESULT
}

private val _setConfigurationOreo = if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.O) ReflectionHelper.findMethod(AssetManager::class.java,
        "setConfiguration", Int::class.java, Int::class.java, String::class.java,
        Int::class.java, Int::class.java, Int::class.java, Int::class.java,
        Int::class.java, Int::class.java, Int::class.java, Int::class.java, Int::class.java,
        Int::class.java, Int::class.java, Int::class.java, Int::class.java, Int::class.java,
        Int::class.java) else null

private val _setConfigurationLollipop = if (Build.VERSION.SDK_INT < Build.VERSION_CODES.O)
    ReflectionHelper.findMethod(AssetManager::class.java,
        "setConfiguration", Int::class.java, Int::class.java, String::class.java,
        Int::class.java, Int::class.java, Int::class.java, Int::class.java,
        Int::class.java, Int::class.java, Int::class.java, Int::class.java, Int::class.java,
        Int::class.java, Int::class.java, Int::class.java, Int::class.java, Int::class.java) else null

private fun AssetManager._setConfiguration(mcc: Int, mnc: Int, locale: String,
     orientation: Int, touchscreen: Int, density: Int, keyboard: Int,
     keyboardHidden: Int, navigation: Int, screenWidth: Int, screenHeight: Int,
     smallestScreenWidthDp: Int, screenWidthDp: Int, screenHeightDp: Int,
     screenLayout: Int, uiMode: Int, colorMode: Int, majorVersion: Int): Boolean
{
    if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.O) {
        _setConfigurationOreo?.let {
            try {
                it.invoke(this, mcc, mnc, locale, orientation, touchscreen, density, keyboard, keyboardHidden,
                        navigation, screenWidth, screenHeight, smallestScreenWidthDp, screenWidthDp, screenHeightDp,
                        screenLayout, uiMode, colorMode, majorVersion)
                return true
            } catch (e: Exception) {
                if (BuildConfig.DEBUG)
                    e.printStackTrace()
            }
        }
    } else {
        _setConfigurationLollipop?.let {
            try {
                it.invoke(this, mcc, mnc, locale, orientation, touchscreen, density, keyboard, keyboardHidden,
                        navigation, screenWidth, screenHeight, smallestScreenWidthDp, screenWidthDp, screenHeightDp,
                        screenLayout, uiMode, majorVersion)
                return true
            } catch (e: Exception) {
                if (BuildConfig.DEBUG)
                    e.printStackTrace()
            }
        }
    }
    return false
}

/**
 * `Locale.toLanguageTag` will transform the obsolete (and deprecated)
 * language codes "in", "ji" and "iw" to "id", "yi" and "he" respectively.
 *
 * All released versions of android prior to "L" used the deprecated language
 * tags, so we will need to support them for backwards compatibility.
 *
 * Note that this conversion needs to take place *after* the call to
 * `toLanguageTag` because that will convert all the deprecated codes to
 * the new ones, even if they're set manually.
 */
private fun adjustLanguageCode(languageCode: String): String {
    var adjusted = languageCode.toLowerCase(Locale.US)
    // Map new language codes to the obsolete language
    // codes so the correct resource bundles will be used.
    if (languageCode == "he") {
        adjusted = "iw"
    } else if (languageCode == "id") {
        adjusted = "in"
    } else if (languageCode == "yi") {
        adjusted = "ji"
    }

    return adjusted
}
/** Constant matching actual resource implementation. {@hide}  */
val KEYBOARDHIDDEN_SOFT = 3
fun Resources._setConfiguration(sdkVersion: Int): Boolean {
    val config = configuration

    val keyboardHidden = if (config.keyboardHidden == Configuration.KEYBOARDHIDDEN_NO
            && config.hardKeyboardHidden
            == Configuration.HARDKEYBOARDHIDDEN_YES) {
        KEYBOARDHIDDEN_SOFT
    } else {
        config.keyboardHidden
    }

    val dm = displayMetrics
    val width: Int
    val height: Int
    if (dm.widthPixels >= dm.heightPixels) {
        width = dm.widthPixels
        height = dm.heightPixels
    } else {
        width = dm.heightPixels
        height = dm.widthPixels
    }

    val colorMode = if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.O) config.colorMode else 0

    @Suppress("DEPRECATION")
    val out = assets._setConfiguration(config.mcc, config.mnc, adjustLanguageCode(config.locale.toLanguageTag()), config.orientation, config.touchscreen, config.densityDpi, config.keyboard, keyboardHidden,
            config.navigation, width, height, config.smallestScreenWidthDp, config.screenWidthDp, config.screenHeightDp, config.screenLayout, config.uiMode, colorMode, sdkVersion)
    if (sdkVersion == Build.VERSION.SDK_INT) {
        @Suppress("DEPRECATION")
        updateConfiguration(config, dm)
    }

    return out
}

private inline fun <T> Resources.withFakeSdkVersion(sdkVersion: Int, defaultValue: T, body: Resources.() -> T): T {
    if (sdkVersion == Build.VERSION.SDK_INT)
        return body()
    else if (_setConfiguration(sdkVersion)) {
        return try {
            body()
        } finally {
            _setConfiguration(Build.VERSION.SDK_INT)
        }
    }
    return defaultValue
}

fun getNonAdaptiveIcon(iconCache: IconCache, iconResId: Int, resources: Resources): Bitmap? {
    if (iconResId == 0) return null

    return resources.withFakeSdkVersion<Bitmap?>(Build.VERSION_CODES.N_MR1, null)
        {
//            val iconName = tryCatch("") { resources.getResourceName(iconResId) }
//            Timber.v("checking icons $iconName [$iconResId]       $roundName [$roundIconResId]")

            return BitmapUtils.getBitmap(this, iconResId, iconCache.fullResIconDpi, iconCache.fullResIconSize, iconCache.fullResIconSize, null, false, null)
        }

}

fun getAdaptiveIcon(iconCache: IconCache, iconResId: Int, resources: Resources, layers: AdaptiveDrawableLayersHolder?): Bitmap? {
    if (iconResId == 0) return null

    return resources.withFakeSdkVersion<Bitmap?>(Build.VERSION_CODES.O, null) {
        try {
            val adaptive = try {
                getXml(iconResId)
            } catch (e: Exception) {
                null
            }
            if (adaptive != null) {
                var type = adaptive.next()
                while (type != XmlPullParser.START_TAG && type != XmlPullParser.END_DOCUMENT) {
                    type = adaptive.next()
                }
                if (type != XmlPullParser.START_TAG) {
                    throw XmlPullParserException("No start tag found")
                }
                val tagName = adaptive.name;
                if (tagName == "adaptive-icon" || tagName == "adaptive-icon-compat") {
                    val d = NovaAdaptiveIconDrawable.createFromXmlInner(this, adaptive, adaptive, null)
                    d?.let {
                        val iconSize = iconCache.fullResIconSize
                        return handleAdaptiveIcon(d, iconSize, layers)
                    }
                }
            }
        } catch (e: Exception) {
        }
        null
    }

}

/**
 * This should be temporary until we overhaul icon handling for adaptive
 *
 * Since adaptive icons all share the same shape we just need to normalize them once and
 * generate a shadow for them once.
 *
 * Also let's us create a shadow for them here rather than after sanifying with the normal path
 */
fun handleAdaptiveIcon(_d: Drawable, _size: Int, layersHolder: AdaptiveDrawableLayersHolder? = null): Bitmap? {
    val d =
        if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.O && _d is AdaptiveIconDrawable)
            NovaAdaptiveIconDrawable(_d.background, _d.foreground)
        else _d as? NovaAdaptiveIconDrawable ?: return null

    val size: Int = if (_size <= 0) Pref.Value.appIconSize.appIconSize else _size

    layersHolder?.layers = d

    if (_size <= 0)
        Timber.v(Throwable(),"handleAdaptiveIcon $_d $_size")
    initAdaptiveStatics()

    val c = sCanvasThreadLocal.get()
    val out = Bitmap.createBitmap(size, size, Bitmap.Config.ARGB_8888)
    c.setBitmap(out)
    val saveCount = c.save()
    val effectiveSize = (NovaAdaptiveIconDrawable.LAYER_TO_VIEWPORT_RATIO * size * d.shape.normalizeScale).toInt()
    val padding = (effectiveSize - size)/2
    val effectiveBounds = Rect(-padding, -padding, size+padding, size +padding)
    try {
        d.background?.setBounds(effectiveBounds)
        d.background?.draw(c)

        d.foreground?.setBounds(effectiveBounds)
        d.foreground?.draw(c)
    } finally {
        c.restoreToCount(saveCount)
        c.setBitmap(null)
    }
    val x = clipIconToAdaptive(d.shape, out, 1.0f, size)
    return x
}

private fun generateAdaptiveShadow(shape: AdaptiveIconShape): Bitmap {
    val d = NovaAdaptiveIconDrawable(ColorDrawable(Color.WHITE), ColorDrawable(Color.RED))
    d.setShape(shape)
    val size = Pref.Value.appIconSize.appIconSize
    val c = Canvas() /* Don't use shared Canvas here as we could be called from someone else using it */
    val bmp = Bitmap.createBitmap(size, size, Bitmap.Config.ARGB_8888)
    c.setBitmap(bmp)
    d.setBounds(0, 0, bmp.width, bmp.height)
    d.draw(c)
    bmp.eraseColor(Color.TRANSPARENT)
    val saveCount = c.save()
    d.draw(c)
    c.restoreToCount(saveCount)
    c.setBitmap(null)

    return ShadowGenerator(size).createShadowOnly(bmp)
}
fun initAdaptiveStatics() {
    val shape = Pref3.adaptive_icon_shape.value

    if (sAdaptiveMask != shape) {
        sAdaptiveShadow = generateAdaptiveShadow(shape)
        sAdaptiveMask = shape
    }

}
fun renderAdaptiveShadow(c: Canvas, shape: AdaptiveIconShape = Pref3.adaptive_icon_shape.value) {
    if (shape != Pref3.adaptive_icon_shape.value) {
        c.drawBitmap(generateAdaptiveShadow(shape), 0f,0f, sPaintThreadLocal.get())
    } else {
        initAdaptiveStatics()
        val shadow = sAdaptiveShadow!!
        val src = sRectThreadLocal[0].get(0, 0, shadow.width, shadow.height)
        val dst = sRectThreadLocal[1].get(0, 0, c.width, c.height)
        c.drawBitmap(shadow, src, dst, sPaintThreadLocal.get())
    }
}

fun clipIconToAdaptive(shape: AdaptiveIconShape, bmp: Bitmap, scale: Float, size: Int, plateColor: Int = Color.TRANSPARENT, debugPkgName: String = ""): Bitmap {
    return clipIconToAdaptive(shape, bmp, scale, size, CornerColors(plateColor), debugPkgName)
}

fun clipIconToAdaptive(shape: AdaptiveIconShape, bmp: Bitmap, scale: Float, size: Int, plateCornerColors: CornerColors, @Suppress("UNUSED_PARAMETER") debugPkgName: String = ""): Bitmap {
    val out = Bitmap.createBitmap(size, size, Bitmap.Config.ARGB_8888)
    val mask = Bitmap.createBitmap(size, size, Bitmap.Config.ALPHA_8)

    val paint = sPaintThreadLocal.get()
    val canvas = sCanvasThreadLocal.get()
    val tmpMatrix = sMatrixThreadLocal.get()

    canvas.setBitmap(mask)
    canvas.resetAfter {
        val path = sPathThreadLocal[0].get()
        tmpMatrix.setScale(size / NovaAdaptiveIconDrawable.MASK_SIZE, size / NovaAdaptiveIconDrawable.MASK_SIZE)
        shape.path.transform(tmpMatrix, path)
        canvas.scale(shape.normalizeScale, shape.normalizeScale, size / 2f, size / 2f)
        canvas.drawPath(path, paint)
        path.reset()
        tmpMatrix.reset()
    }
    canvas.setBitmap(null)

    canvas.setBitmap(out)
    renderAdaptiveShadow(canvas, shape)

    if (plateCornerColors.isConstant()) {
        if (plateCornerColors.topLeft != Color.TRANSPARENT) {
            paint.color = plateCornerColors.topLeft
            canvas.drawBitmap(mask, 0f, 0f, paint)
            paint.color = Color.WHITE
        }
    } else {
        if (plateCornerColors.topLeft != Color.TRANSPARENT) {
            canvas.resetAfter {
                canvas.clipRect(0, 0, size / 2, size / 2)
                paint.color = plateCornerColors.topLeft
                canvas.drawBitmap(mask, 0f, 0f, paint)
                paint.color = Color.WHITE
            }
        }
        if (plateCornerColors.topRight != Color.TRANSPARENT) {
            canvas.resetAfter {
                canvas.clipRect(size / 2, 0, size, size / 2)
                paint.color = plateCornerColors.topRight
                canvas.drawBitmap(mask, 0f, 0f, paint)
                paint.color = Color.WHITE
            }
        }
        if (plateCornerColors.bottomLeft != Color.TRANSPARENT) {
            canvas.resetAfter {
                canvas.clipRect(0, size / 2, size / 2, size)
                paint.color = plateCornerColors.bottomLeft
                canvas.drawBitmap(mask, 0f, 0f, paint)
                paint.color = Color.WHITE
            }
        }
        if (plateCornerColors.bottomRight != Color.TRANSPARENT) {
            canvas.resetAfter {
                canvas.clipRect(size / 2, size / 2, size, size)
                paint.color = plateCornerColors.bottomRight
                canvas.drawBitmap(mask, 0f, 0f, paint)
                paint.color = Color.WHITE
            }
        }
    }

    val shader = BitmapShader(bmp, Shader.TileMode.CLAMP, Shader.TileMode.CLAMP)
    tmpMatrix.reset()

    val widthScale = out.width / bmp.width.toFloat()
    val heightScale = out.height / bmp.height.toFloat()
    tmpMatrix.setScale(widthScale, heightScale)
    tmpMatrix.postScale(scale, scale, out.width / 2f, out.height / 2f)
    shader.setLocalMatrix(tmpMatrix)

    paint.shader = shader
    val cx = out.width/2
    val cy = out.height/2
    val shaderWidth = out.width * scale
    val shaderHeight = out.height * scale
    val clipLeft = cx - shaderWidth/2
    val clipTop = cy - shaderHeight/2
    val clipRight = clipLeft + shaderWidth
    val clipBottom = clipTop + shaderWidth

    canvas.resetAfter {
        canvas.clipRect(clipLeft, clipTop, clipRight, clipBottom)
        canvas.drawBitmap(mask, 0f, 0f, paint)
    }
    canvas.setBitmap(null)
    tmpMatrix.reset()
    paint.shader = null
    return out
}

private val labCacheA = object: ThreadLocal<DoubleArray>() {
    override fun initialValue(): DoubleArray {
        return DoubleArray(3)
    }
}

private val labCacheB = object: ThreadLocal<DoubleArray>() {
    override fun initialValue(): DoubleArray {
        return DoubleArray(3)
    }
}

private fun DoubleArray.copyTo(other: DoubleArray) {
    forEachIndexed { i, d ->
        other[i] = d
    }
}

/*
 * Calculate Delta-E of two colors in LAB space.
 * Range is 0 to 100
 * <= 1.0 Not perceptible by human eyes
 * 1-2    Perceptible through close observation
 * 2-10   Perceptible at a glance
 * 11-49  Colors are more similar than opposite
 * 100    Colors are exact opposite
 * https://zschuessler.github.io/DeltaE/learn/
 */
private fun colorDeltaE(first: DoubleArray, second: DoubleArray): Double {
    val Ldelta = first[0] - second[0]
    val Adelta = first[1] - second[1]
    val Bdelta = first[2] - second[2]
    return Math.sqrt(Ldelta * Ldelta + Adelta * Adelta + Bdelta * Bdelta)
}
@JvmField
var DEBUG_BITMAPS = BuildConfig.DEBUG && false
/* Keep this >= 16 for the Airbnb app. Lower than 18 is preferred to handle Audax */
private val DELTA_E_THRESHOLD = 17
@SuppressLint("SdCardPath")
private fun smartClipIconToAdaptive(shape: AdaptiveIconShape, bmp: Bitmap, size: Int, normalizedResult: IconNormalizationResult, layersHolder: AdaptiveDrawableLayersHolder?, debugPkgName: String = ""): Bitmap? {
    var foundColor = Color.TRANSPARENT
    var lastColor = Color.TRANSPARENT
    val lastLab = labCacheA.get()
    val pxLab = labCacheB.get()

    var allMatch = true
    val colorForDegree = SparseIntArray(360)
    val colors: IntArray
    val positions: FloatArray

    var innerRadiusMismatchCount = 0
    val debugBitmap = if (DEBUG_BITMAPS) Bitmap.createBitmap(bmp.width, bmp.height, Bitmap.Config.ARGB_8888) else null

    val m = sMatrixThreadLocal.get()
    val bmpSize = Math.min(bmp.width, bmp.height)
    m.setScale(bmpSize / NovaAdaptiveIconDrawable.MASK_SIZE, bmpSize / NovaAdaptiveIconDrawable.MASK_SIZE)
    val normTranslateX = -normalizedResult.horizontalOffset/2f
    var normTranslateY = -normalizedResult.verticalOffset/2f
    /* Hack, add a buffer for shadows */
    if (normTranslateY < 0)
        normTranslateY -= 1
    m.postTranslate(normTranslateX, normTranslateY) // 222
    m.postScale( normalizedResult.shape.normalizeScale / normalizedResult.scale, normalizedResult.shape.normalizeScale / normalizedResult.scale, bmpSize/2f, bmpSize/2f)
    m.postScale(normalizedResult.shape.getPathScaleForSmartClip(), normalizedResult.shape.getPathScaleForSmartClip(), bmpSize/2f, bmpSize/2f)

    val originalPath = if (normalizedResult.shape == AdaptiveIconShape.SQUARE) AdaptiveIconShape.ROUNDED_SQUARE.path else normalizedResult.shape.path
    val sourcePath = sPathThreadLocal[0].get()
    originalPath.transform(m, sourcePath)

    val pathMeasure = PathMeasure(sourcePath, false)
    val pos = FloatArray(2)

    val samsungSquircle = normalizedResult.shape == AdaptiveIconShape.SAMSUNG_SQUIRCLE
    val deltaEThreshold = if (samsungSquircle) 65 else DELTA_E_THRESHOLD
    var maxDegree = 0
    var mismatchCount = 0

    /* Inset more, use to detect if we have a circle in a circle, these don't look right reshaped */
    val innerPath: Path?
    val innerMeasure: PathMeasure?
    if (normalizedResult.shape.isCircle() && !shape.isCircle() && shape != AdaptiveIconShape.TEARDROP) {
        innerPath = sPathThreadLocal[1].get()
        m.reset()
        m.setScale(0.85f, 0.85f)
        sourcePath.transform(m, innerPath)
        innerMeasure = PathMeasure(innerPath, false)
    } else {
        innerPath = null
        innerMeasure = null
    }

    for (i in 0 until 360) {
        val distance = (i * pathMeasure.length) / (360 - 1f)
        pathMeasure.getPosTan(distance, pos, null)
        val xInt = MathUtils.toRange(MathUtils.roundInt(pos[0]), 0, bmpSize-1)
        val yInt = MathUtils.toRange(MathUtils.roundInt(pos[1]), 0, bmpSize-1)

        val degrees = (360 + - 90 +  MathUtils.roundInt(Math.toDegrees(Math.atan2(xInt.toDouble() - bmpSize/2, bmpSize/2 - yInt.toDouble())))) % 360

        val px = bmp.getPixel(xInt, yInt)
        if (DEBUG_BITMAPS) {
            debugBitmap?.setPixel(xInt, yInt, if (Color.alpha(px) != 0) px else Color.BLACK)
        }

        val alpha = Color.alpha(px)
        if (alpha < 0xF9) {
            allMatch = false
            if (!DEBUG_BITMAPS)
                break
            else
                Timber.v("Breaking, found transparent (${Integer.toHexString(alpha)}) at $xInt, $yInt  ${i}º")
        } else if  (i == 0) {
            foundColor = px
            lastColor = px
            ColorUtils.colorToLAB(px, lastLab)
            colorForDegree.put(degrees, px)
            if (degrees >= maxDegree) {
                maxDegree = degrees
                colorForDegree.put(0, px)
            }

        } else if (px != lastColor) {
            val deltae = if (px == lastColor) {
                0.0
            } else {
                ColorUtils.colorToLAB(px, pxLab)
                colorDeltaE(lastLab, pxLab)
            }

            if (deltae > DELTA_E_THRESHOLD) {
                mismatchCount++
                if (mismatchCount > 45) {
                    allMatch = false
                    if (!DEBUG_BITMAPS)
                        break
                    else
                        Timber.v("Breaking, found mismatch ${Integer.toHexString(px)}   ${Integer.toHexString(foundColor)} $deltae")
                } else {
                    if (DEBUG_BITMAPS)
                        Timber.v("Mismatch ${Integer.toHexString(px)}   ${Integer.toHexString(foundColor)} $deltae, incrementing to $mismatchCount")
                }
            } else {
                if (deltae > .5f) {
                    colorForDegree.put(degrees, px)
                    if (degrees >= maxDegree) {
                        maxDegree = degrees
                        colorForDegree.put(0, px)
                    }
                    lastColor = px
                    pxLab.copyTo(lastLab)
                }
            }
        }

        if (innerPath != null && innerMeasure != null) {
            val innerDistance = (i * innerMeasure.length) / (360 - 1f)
            innerMeasure.getPosTan(innerDistance, pos, null)
            val innerX = MathUtils.toRange(MathUtils.roundInt(pos[0]), 0, bmpSize-1)
            val innerY = MathUtils.toRange(MathUtils.roundInt(pos[1]), 0, bmpSize-1)

            /* Check inner circle */
            val innerPx = bmp.getPixel(innerX, innerY)
            val deltae = if (innerPx == foundColor || innerPx == lastColor) {
                0.0
            } else {
                ColorUtils.colorToLAB(innerPx, pxLab)
                colorDeltaE(lastLab, pxLab)
            }
            if (deltae > deltaEThreshold) {
                innerRadiusMismatchCount++
                if (DEBUG_BITMAPS)
                    Timber.v("Inner mismatch @ $xInt,$yInt ($i) ${Integer.toHexString(innerPx)} $innerRadiusMismatchCount")
                if (innerRadiusMismatchCount > 180) {
                    allMatch = false
                    if (!DEBUG_BITMAPS)
                        break
                    else
                        Timber.v("Breaking, too much inner radius mismatch $debugPkgName")
                }
            }
        }

    }
    innerPath?.reset()

    if (DEBUG_BITMAPS) {
        debugBitmap?.apply {
            setPixel(0, 0, Color.BLACK)
            setPixel(0, height-1, Color.BLACK)
            setPixel(width -1, 0, Color.BLACK)
            setPixel(width -1, height -1, Color.BLACK)
            compress(Bitmap.CompressFormat.PNG, 100, FileOutputStream(File("/sdcard/nova_debugPath.png")))
            bmp.compress(Bitmap.CompressFormat.PNG, 100, FileOutputStream(File("/sdcard/nova_debugSrc.png")))
        }
    }
    if (allMatch) {
        if (DEBUG_BITMAPS)
            Timber.v("RESHAPEABLE %s %s", debugPkgName, Integer.toHexString(foundColor))

        // Can we just clip the canvas here? no antialiasing of the clip, but our arcs should match
        val p = sPaintThreadLocal.get()

        val c1 = sCanvasThreadLocal.get()
        val resultBitmap = Bitmap.createBitmap(bmp.width, bmp.height, Bitmap.Config.ARGB_8888)
        resultBitmap.eraseColor(foundColor)

        colors = IntArray(colorForDegree.size()*2)
        positions = FloatArray(colorForDegree.size()*2)
        var solidColor = Color.TRANSPARENT
        var colorAt270: Int = Color.TRANSPARENT
        try {
            c1.setBitmap(resultBitmap)

//            val slop = resultBitmap.width / 10f
//            val l = 0f - slop
//            val t = 0f - slop
//            val r = resultBitmap.width.toFloat() + slop
//            val b = resultBitmap.width.toFloat() + slop

            for (i in 0 until colorForDegree.size()) {
                val color = colorForDegree.valueAt(i)
                colors[i*2] = color
                colors[i*2+1] = color

                val thisAngle = colorForDegree.keyAt(i)
                val nextAngle = if (i + 1 < colorForDegree.size()) colorForDegree.keyAt(i + 1) else 360
//                p.color = color
//                c1.drawArc(l, t, r, b, thisAngle.toFloat() - .5f, nextAngle.toFloat() - thisAngle.toFloat() + 1f, true, p)
                positions[i*2] = thisAngle/360f
                positions[i*2+1] = nextAngle/360f
                if (thisAngle <= 270 && 270 < nextAngle) {
                    colorAt270 = color
                }
            }

            if (colors.size == 4 && colors[0] == colors[1] && colors[1] == colors[2] && colors[2] == colors[3]) {
                solidColor = colors[0]
                c1.drawColor(solidColor)
            } else {
                p.shader = SweepGradient(c1.width / 2f, c1.height / 2f, colors, positions)
                c1.drawRect(0f, 0f, c1.width.toFloat(), c1.height.toFloat(), p)
                p.shader = null
            }

            c1.resetAfter {
                c1.scale(normalizedResult.scale, normalizedResult.scale, bmp.width / 2f, bmp.height / 2f)
                c1.clipPath(sourcePath)
                if (normalizedResult.shape == AdaptiveIconShape.CIRCLE) {
                    c1.drawColor(Color.WHITE)
                }
                c1.drawBitmap(bmp, 0f, 0f, p)
            }
        } finally {
            c1.setBitmap(null)
            sourcePath.reset()
        }

        val result = clipIconToAdaptive(shape, resultBitmap, 1.0f, size, Color.TRANSPARENT, debugPkgName)
        layersHolder?.layers = DecomposedIcon(shape, if (solidColor != Color.TRANSPARENT) ColorShaderMaker(solidColor) else SweepGradientCanvasShader(colorAt270, colors, positions), null, true)

        return result
    }
    sourcePath.reset()
    return null
}

@SuppressLint("SdCardPath")
fun generateDebugPreviewSet(file: File, size: Int, c: Canvas, bigset: Bitmap, sanifyScalesMap: MutableMap<String, Float>? = null) {
    if (!BuildConfig.DEBUG)
        return
    val paint = Paint(Paint.ANTI_ALIAS_FLAG or Paint.FILTER_BITMAP_FLAG)
    val bmp = BitmapFactory.decodeFile(file.toString()) ?: return
    c.save()
    bigset.eraseColor(Color.TRANSPARENT)

    val src = sRectThreadLocal[0].get(0, 0, bmp.width, bmp.height)
    val dst = sRectThreadLocal[1].get(0, 0, size, size)
    c.drawBitmap(bmp, src, dst, paint)
    c.translate(size.toFloat(), 0f)
    val normalizationResult = arrayOf(IconNormalizationResult.EMPTY_RESULT)
    var sanifyScale = IconCache.sanifyIconComputeScale(bmp, normalizationResult)
    sanifyScalesMap?.put(file.name, sanifyScale)

    for (shape in AdaptiveIconShape.values) {
        if (shape === AdaptiveIconShape.DEFAULT || shape === AdaptiveIconShape.NONE)
            continue
        val out: Bitmap? = magicallyGiveMeGoodAdaptivePlease(shape, bmp, size, normalizationResult[0], false, Color.TRANSPARENT, file.name)
        if (out != null) {
            c.drawBitmap(out, 0f, 0f, paint)
        }
        c.translate(size.toFloat(), 0f)
    }

    try {
        bigset.compress(Bitmap.CompressFormat.PNG, 100, FileOutputStream(File("/sdcard/iconset/out/" + file.getName())))
    } catch (e: FileNotFoundException) {
        e.printStackTrace()
    }

    c.restore()
}

fun magicallyGiveMeGoodAdaptivePlease(shape: AdaptiveIconShape, bmp: Bitmap, size: Int, normalizedResult: IconNormalizationResult, includeFallback: Boolean, forcePlateColor: Int, debugName: String, layersHolder: AdaptiveDrawableLayersHolder? = null): Bitmap? {
    var out: Bitmap?
    var mustPlate = false
    /* Sanity check for Opera :-/ */
    if (Color.alpha(bmp.getPixel(bmp.width/2, bmp.height/2)) < 0xF9)
        mustPlate = true

    if (!mustPlate && shape.isCircle() && normalizedResult.shape == AdaptiveIconShape.CIRCLE) {
        out = if (layersHolder != null) {
            smartClipIconToAdaptive(shape, bmp, size, normalizedResult, layersHolder, debugName)
        } else {
            clipIconToAdaptive(shape, bmp, normalizedResult.scale, size, Color.TRANSPARENT, debugName)
        }
        /* For a circle we just assume we can radially clamp */
//        layersHolder?.layers = DecomposedIcon(shape, radiallyClamp(shape, out), null, clipIconForFg = true)
    } else {
        out = if (!mustPlate && normalizedResult.shape == AdaptiveIconShape.SQUARE && shape.canClipSquare()) {
                clipIconToAdaptive(shape, bmp, normalizedResult.scale, size, normalizedResult.cornerColors, debugName)
            } else if (!mustPlate && normalizedResult.shape != AdaptiveIconShape.NONE) {
                smartClipIconToAdaptive(shape, bmp, size, normalizedResult, layersHolder, debugName)
            } else {
                null
            }
        if (out == null && includeFallback) {
            val plateColor =
                if (forcePlateColor != Color.TRANSPARENT) {
                    forcePlateColor
                } else {
                    val palette = Palette.generate(bmp)
                    val paletteColor = PaletteHelper.getBestHighlightColorFromIconPalette(palette, false)
                    if (ColorUtil.isGray(paletteColor))
                        Color.WHITE
                    else
                        com.teslacoilsw.launcher.util.palette.ColorUtils.getMutedColor(paletteColor, 0.54f)
                }

            val scale = 0.65f * normalizedResult.scale
            out = clipIconToAdaptive(shape, bmp, scale, size, plateColor, debugName)
            layersHolder?.layers = DecomposedIcon(shape, ColorShaderMaker(plateColor), BitmapShaderMaker(bmp, bmp.width / scale))
        }
    }

    return out
}
private const val CIRCLE_PATH = "M50 0A50 50,0,1,1,50 100A50 50,0,1,1,50 0"
private val DEFAULT_ADAPTIVE_PATH =
    if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.O) {
        Resources.getSystem().getString(Resources.getSystem().getIdentifier("config_icon_mask", "string", "android")).replace("[\\r\\n\\s]+".toRegex(), " ")
    } else {
        CIRCLE_PATH
    }


private val Int.R
    get() = CornerConfig(this, CornerType.ROUND)
private val Int.S
    get() = CornerConfig(this, CornerType.SQUIRCLE)
private val Int.C
    get() = CornerConfig(this, CornerType.CROP)
class AdaptiveIconShape private constructor(private val customPath: CustomShapePath, private val alternativePaths: List<String> = emptyList(), private var _normalizeScale: Float = 0.0f, val name: String? = null) {
    companion object: Cofs<AdaptiveIconShape> {
        @JvmField
        val NONE = AdaptiveIconShape(CornerCustomShapePath(0.R), listOf(), 1.0f, "NONE")
        override val DEFAULT = AdaptiveIconShape(PathParserCustomShapePath(DEFAULT_ADAPTIVE_PATH), emptyList(),0.0f, "DEFAULT")
        @JvmField
        val CIRCLE = AdaptiveIconShape(CornerCustomShapePath(50.R), listOf(CIRCLE_PATH, "M50 0C77.6 0 100 22.4 100 50C100 77.6 77.6 100 50 100C22.4 100 0 77.6 0 50C0 22.4 22.4 0 50 0Z"), 0.91f, "CIRCLE")
        val SQUIRCLE = AdaptiveIconShape(CornerCustomShapePath(50.S), listOf("M50,0 C10,0 0,10 0,50 0,90 10,100 50,100 90,100 100,90 100,50 100,10 90,0 50,0 Z"), 0.83f, "SQUIRCLE")
        /* {30:ARC}:{30:ARC}:{30:ARC}:{30:ARC} */
        @JvmField
        val ROUNDED_SQUARE = AdaptiveIconShape(CornerCustomShapePath(30.R), listOf("M50 0h20a30 30 0 0 1 30 30v40a30 30 0 0 1-30 30H30A30 30 0 0 1 0 70V30A30 30 0 0 1 30 0z"), 0.79f, "ROUNDED_SQUARE")
        @JvmField
        val SQUARE = AdaptiveIconShape(CornerCustomShapePath(0.R), listOf("M50,0L100,0 100,100 0,100 0,0z"), 0.79f, "SQUARE")
        val TEARDROP = AdaptiveIconShape(CornerCustomShapePath(50.R, bottomRight = 15.R), listOf("M50 0a50 50 0 0 1 50 50v35a15 15 0 0 1-15 15H50A50 50 0 0 1 50 0z"), 0.91f, "TEARDROP")
        @JvmField
        val SAMSUNG_SQUIRCLE = AdaptiveIconShape(PathParserCustomShapePath("M59 0c25 0 41 16 41 41v18c0 25-16 41-41 41H41C16 100 0 84 0 59V41C0 16 16 0 41 0h18z"), emptyList(), 0.84f, "SAMSUNG_SQUIRCLE")
        private val NEXUS_SQUARE = AdaptiveIconShape(CornerCustomShapePath(10.R), listOf("M50,0L92,0C96.42,0 100,4.58 100 8L100,92C100, 96.42 96.42 100 92 100L8 100C4.58, 100 0 96.42 0 92L0 8 C 0 4.42 4.42 0 8 0L50 0Z"), 0.79f)
        val FOLDER_PLATFORM = AdaptiveIconShape(PathParserCustomShapePath("M16.8 47h66.4c.4 0 .8.3 1 .7L100 95H0l15.9-47.3a1 1 0 0 1 1-.7z"), emptyList(), 1.0f, "FOLDER_PLATFORM")
        private val LG_ROUNDED_SQUARE = AdaptiveIconShape(CornerCustomShapePath(25.R), listOf("M 22.58, 1.76 C 34.27, 0.76 65.73, 0.76 77.42, 1.76 89.11, 2.76 97.24, 10.89 98.24, 22.58 99.24, 34.27 99.24, 65.73 98.24, 77.42 97.24, 89.11 89.11, 97.24 77.42, 98.24 65.73, 99.24 34.27, 99.24 22.58, 98.24 10.89, 97.24, 2.76, 89.11 1.76, 77.42 0.76, 65.73 0.76, 34.27 1.76, 22.58 2.76, 10.89 10.89, 2.76 22.58, 1.76 Z"), 0.81f)
        @JvmField
        val CUSTOM = AdaptiveIconShape(CornerCustomShapePath(0.R), emptyList(),1.0f, "CUSTOM")

        val values = listOf(NONE, DEFAULT, CIRCLE, SQUIRCLE, ROUNDED_SQUARE, SQUARE, TEARDROP, SAMSUNG_SQUIRCLE, NEXUS_SQUARE, LG_ROUNDED_SQUARE, FOLDER_PLATFORM, CUSTOM)

        operator fun invoke(path: CustomShapePath): AdaptiveIconShape =
                AdaptiveIconShape(path)

        operator fun invoke(path: String): AdaptiveIconShape {
            val shape = path.substringAfter("SHAPE:", missingDelimiterValue = "")
            return if (shape.isNotBlank())
                AdaptiveIconShape(CornerCustomShapePath.parse(shape))
            else
                AdaptiveIconShape(PathParserCustomShapePath(path))
        }

        fun getMatchingKnownShape(path: String = DEFAULT.customPath.asPathString()): AdaptiveIconShape? {
            for (i in values) {
                if (i == NONE || i == DEFAULT || i == CUSTOM || i == FOLDER_PLATFORM)
                    continue
                if (i.pathMatch(path))
                    return i
            }
            return null
        }

        override fun flattenToString(value: AdaptiveIconShape): String {
            if (value.name != null)
                return value.name
            return value.customPath.asPathString()
        }

        override fun parse(string: String, defaultValue: AdaptiveIconShape): AdaptiveIconShape {
            if (string.isBlank())
                return defaultValue
            for (i in AdaptiveIconShape.values) {
                if (i.name == string)
                    return i
            }
            val shape = AdaptiveIconShape(string)
            return try {
                shape.path
                shape
            } catch (e: Exception) {
                defaultValue
            }
        }

    }

    val path: Path by lazy { customPath.getPath() }
    fun wantRoundIcon() = this == CIRCLE || shapeMatch(CIRCLE) ||
            (this == DEFAULT && (Build.VERSION.SDK_INT < Build.VERSION_CODES.O) ||
                    LauncherAppState.getInstance().iconCache.systemUsesRoundIcons)

    fun pathMatch(otherPath: String): Boolean {
        return customPath.asPathString() == otherPath || alternativePaths.contains(otherPath)
    }

    fun shapeMatch(otherShape: AdaptiveIconShape): Boolean {
        if (this == otherShape)
            return true
        if (customPath == otherShape.customPath)
            return true
        if (otherShape.pathMatch(customPath.asPathString()))
            return true
        if (pathMatch(otherShape.customPath.asPathString()))
            return true
        return false
    }

    fun isCircle(): Boolean {
        return this == CIRCLE || shapeMatch(CIRCLE)
    }

    fun canClipSquare(): Boolean {
        return shapeMatch(SQUARE) || shapeMatch(ROUNDED_SQUARE) || shapeMatch(NEXUS_SQUARE)
    }

    fun getPathScaleForSmartClip(): Float {
        return when {
            this == SAMSUNG_SQUIRCLE -> 0.93f
            this == ROUNDED_SQUARE -> 0.94f
            else -> 0.95f
        }
    }

    val normalizeScale by lazy { if (_normalizeScale != 0.0f)  _normalizeScale else computeScale() }

    private fun computeScale(): Float {
        /** Try to cheat, if we know of the path use that scale */
        for (s in AdaptiveIconShape.values) {
            if (s != this && s != NONE && s.normalizeScale != 0.0f && s.shapeMatch(this)) {
                return s.normalizeScale
            }
        }

        val size = 100
        val bmp = Bitmap.createBitmap(size, size, Bitmap.Config.ARGB_8888)
        bmp.applyCanvas {
            val p = Paint(Paint.ANTI_ALIAS_FLAG)
            p.color = Color.RED
            drawPath(path, p)
        }

        val scale = IconCache.sanifyIconComputeScale(bmp)
        if (BuildConfig.DEBUG) {
            Timber.tag("Nova.Adaptive").v("Computed scale for path \"${customPath.asPathString()}\"   $scale")
        }
        return scale
    }

    fun createPreviewDrawable(context: Context): PathDrawable {
        val d = PathDrawable(path, 100, 100)
        d.setColor(ContextCompat.getColorStateList(context, R.color.btn_radio_drawable_color)!!)
        d.scale = normalizeScale
        val dp48 = DpUtil.getPixelSize(48)
        d.setBounds(0, 0, dp48, dp48)
        return d
    }

    fun getCornerShape(): CornerCustomShapePath {
        if (customPath is CornerCustomShapePath)
            return customPath
        val known = getMatchingKnownShape(customPath.asPathString())
        if (known?.customPath is CornerCustomShapePath) {
            return known.customPath as CornerCustomShapePath
        }

        return SQUIRCLE.customPath as CornerCustomShapePath
//        This doesn't work like I'd like it to
//        /* If we can't find a match, just come up with a rough approximation by
//         * matching bitmaps */
//        val testSize = 16
//        val bmp = Bitmap.createBitmap(testSize, testSize, Bitmap.Config.ALPHA_8)
//        val c = Canvas(bmp)
//        val m = Matrix()
//        m.setScale(testSize/100f, testSize/100f)
//
//        val srcP = Path()
//        path.transform(m, srcP)
//
//        val paint = Paint()
//        c.drawPath(srcP, paint)
//
//        var bestDiff: Long = Long.MAX_VALUE
//        var bestShape = CornerCustomShapePath(0.R)
//        out@ for (i in 10..50 step 5) {
//            for (type in CornerType.values()) {
//                bmp.eraseColor(Color.TRANSPARENT)
//                c.drawPath(srcP, paint)
//                val shape = CornerCustomShapePath(CornerConfig(i, type))
//                val newP = shape.getPath()
//                newP.transform(m)
//                val erase = Paint()
//                erase.xfermode = PorterDuffXfermode(PorterDuff.Mode.XOR)
//                c.drawPath(newP, erase)
//
//                val pixels = IntArray(testSize * testSize)
//                bmp.getPixels(pixels, 0, testSize, 0, 0, testSize, testSize)
//                var diff = 0L
//                for (p in pixels) {
//                    diff += p.alpha
//                }
//                if (diff < bestDiff) {
//                    bestDiff = diff
//                    bestShape = shape
//                    if (diff == 0L)
//                        break@out
//                }
//            }
//        }
//
//        c.setBitmap(null)
//        bmp.recycle()
//        return bestShape
    }

    override fun toString(): String {
        return "AdaptiveIconShape{$name `${customPath.asPathString()}`}"
    }
}


interface ShaderMaker {
    fun getShader(): Shader?
    fun getViewportSize(): Float
    fun getIntrinsicSize(): Float
    fun bgColor(): Int = Color.TRANSPARENT
}

class BitmapShaderMaker @JvmOverloads constructor(val bmp: Bitmap, val _viewportSize: Float = bmp.width.toFloat(), val _bgColor: Int = Color.TRANSPARENT) : ShaderMaker {
    override fun getShader(): Shader =
        BitmapShader(bmp, Shader.TileMode.CLAMP, Shader.TileMode.CLAMP)

    override fun getViewportSize(): Float = _viewportSize

    override fun getIntrinsicSize(): Float = bmp.width.toFloat()

    override fun bgColor(): Int = _bgColor
}

class ColorShaderMaker(val color: Int) : ShaderMaker {
    override fun getViewportSize(): Float = 0f
    override fun getIntrinsicSize(): Float = 0f

    override fun getShader(): Shader? = null
//        LinearGradient(0f, 0f, 0f, 0f, color, color, Shader.TileMode.CLAMP)

    override fun bgColor(): Int = color
}

class SweepGradientCanvasShader(private val _bgColor: Int, private val colors: IntArray, private val positions: FloatArray) : ShaderMaker {
    override fun getShader(): Shader {
       return SweepGradient(100f / 2f, 100f / 2f, colors, positions)
    }

    override fun getViewportSize(): Float = 100f

    override fun getIntrinsicSize(): Float = getViewportSize()

    override fun bgColor(): Int = _bgColor
}

class DecomposedIcon(private val shape: AdaptiveIconShape, private val unclippedBg: ShaderMaker, private val unclippedFg: ShaderMaker?, private val clipIconForFg: Boolean = false, private val bgAnimationMode: AdaptiveBgAnimationMode = if (clipIconForFg) AdaptiveBgAnimationMode.WITH_FG else AdaptiveBgAnimationMode.NORMAL ) : AdaptiveDrawableLayers {
    override fun getLayers(icon: Bitmap?, bmpSize: Int): Triple<ShaderMaker, ShaderMaker, ShaderMaker?> {
        val outFg =
                unclippedFg ?: if (icon != null && clipIconForFg) {
                    val bmp = Bitmap.createBitmap(bmpSize, bmpSize, Bitmap.Config.ARGB_8888)
                    val c = sCanvasThreadLocal.get()
                    val p = sPathThreadLocal[0].get()
                    val m = sMatrixThreadLocal.get()
                    val paint = sPaintThreadLocal.get()
                    /* Inset a tad just in case of anti-aliasing in the bitmap */
                    val scale = bmpSize / NovaAdaptiveIconDrawable.MASK_SIZE
                    m.setScale(scale, scale)
                    m.postScale(0.99f * shape.normalizeScale, 0.99f * shape.normalizeScale, bmpSize/2f, bmpSize/2f)
                    shape.path.transform(m, p)
                    c.setBitmap(bmp)
                    c.resetAfter {
                        c.clipPath(p)
                        c.drawBitmap(icon, 0f, 0f, paint)
                        c.setBitmap(null)
                    }
                    p.reset()
                    m.reset()
                    BitmapShaderMaker(bmp)
                } else {
                    ColorShaderMaker(color = Color.TRANSPARENT)
                }

        return Triple(unclippedBg, outFg, null)
    }

    override fun getAnimationMode(): AdaptiveBgAnimationMode = bgAnimationMode
}

interface AdaptiveDrawableLayers {
    fun getLayers(icon: Bitmap?, bmpSize: Int) : Triple<ShaderMaker, ShaderMaker, ShaderMaker?>
    fun getAnimationMode(): AdaptiveBgAnimationMode
    fun getMask(bmpSize: Int): Bitmap? = null
}

class AdaptiveDrawableLayersHolder {
    var layers: AdaptiveDrawableLayers? = null
}