package com.teslacoilsw.launcher.importer

import android.annotation.SuppressLint
import android.content.ContentResolver
import android.content.Context
import android.content.Intent
import android.database.Cursor
import android.database.sqlite.SQLiteDatabase
import android.net.Uri
import com.android.launcher3.ItemInfo
import com.android.launcher3.LauncherAppState
import com.android.launcher3.LauncherModel
import com.android.launcher3.LauncherSettings
import com.android.launcher3.ShortcutInfo
import com.android.launcher3.WorkspaceScreenInfo
import com.teslacoilsw.launcher.NovaAction
import com.teslacoilsw.launcher.R
import com.teslacoilsw.launcher.backup.namedIntentFromApexKey
import com.teslacoilsw.launcher.iteminfo.ItemIconResource
import com.teslacoilsw.launcher.iteminfo.ItemIconUserBitmap
import com.teslacoilsw.launcher.iteminfo.ItemInfoWithIcon
import com.teslacoilsw.shared.util.MathUtils
import java.util.*

class ApexImporter(val db: SQLiteDatabase? = null, private val settingsMap: Map<String, *>? = null, private val swipeActionMap: Map<String, *>? = null) : BaseImporter() {

    override fun getContentUri(): Uri? {
        return Uri.parse("content://com.anddoes.launcher.settings" +
                "/" + BaseImporter.TABLE_FAVORITES +
                "?" + BaseImporter.PARAMETER_NOTIFY + "=true")
    }

    @SuppressLint("Recycle")
    override fun query(cr: ContentResolver?): Cursor? {
        return if (db == null)
            super.query(cr)
        else
            db.query(BaseImporter.TABLE_FAVORITES, null, null, null, null, null, null)

    }

    override fun getScreens(cr: ContentResolver?): List<WorkspaceScreenInfo>? {
        return if (db == null)
            super.getScreens(cr)
        else {
            val screenC = db.query(BaseImporter.TABLE_WORKSPACE_SCREENS, null, null, null, null, null, null)
            if (screenC != null) {
                val screens = ArrayList<WorkspaceScreenInfo>()
                val idIndex = screenC.getColumnIndexOrThrow(ID)
                val rankIndex = screenC.getColumnIndexOrThrow(SCREEN_RANK)
                while (screenC.moveToNext()) {
                    val rank = screenC.getInt(rankIndex)
                    val id = screenC.getLong(idIndex)
                    screens.add(Math.min(rank, screens.size), WorkspaceScreenInfo(id, rank, false))
                }
                screenC.close()
                screens
            }
            null
        }

    }

    override fun populateInfoFromCursor(info: ItemInfo, c: Cursor, ci: BaseImporter.ColumnIndex) {
        info.id = c.getLong(ci.id)
        val container = c.getInt(ci.container)

        info.cellX = c.getInt(ci.cellx).toFloat()
        info.cellY = c.getInt(ci.celly).toFloat()
        info.spanX = c.getInt(ci.spanx).toFloat()
        info.spanY = c.getInt(ci.spany).toFloat()

        val screen = c.getInt(ci.screen)

        if (container == BaseImporter.AOSPFavorites.CONTAINER_HOTSEAT) {
            info.container = LauncherSettings.Favorites.CONTAINER_HOTSEAT.toLong()
            /* Yay for braindead developers.
             * This is really awkward. Thanks Apex */
            val novaScreen: Int
            val novaIndex: Int
            if (screen >= 1000) {
                /* Left of center */
                novaScreen = 2 * (screen - 1000) / 100 + 2
            } else if (screen >= 100) {
                /* Right of center */
                novaScreen = 2 * (screen - 100) / 100 + 1
            } else {
                novaScreen = 0
            }
            novaIndex = screen % 100
            info.indexedRank = novaIndex
            info.screenId = novaScreen.toLong()
        } else {
            if (container == BaseImporter.AOSPFavorites.CONTAINER_DESKTOP) {
                info.container = LauncherSettings.Favorites.CONTAINER_DESKTOP.toLong()
                info.screenId = screen.toLong()
            } else {
                info.container = container.toLong()
                info.screenId = 0
                info.indexedRank = MathUtils.roundInt(info.cellY) * 10 + MathUtils.roundInt(info.cellX)
            }
        }

        if (info is ItemInfoWithIcon) {
            val iconType: Int = if (ci.iconType == -1)
                if (c.isNull(ci.iconPackage))
                    BaseImporter.AOSPFavorites.ICON_TYPE_BITMAP
                else
                    BaseImporter.AOSPFavorites.ICON_TYPE_RESOURCE
            else
                c.getInt(ci.iconType)
            when (iconType) {
                BaseImporter.AOSPFavorites.ICON_TYPE_RESOURCE -> info.itemIcon = ItemIconResource(c.getString(ci.iconPackage), c.getString(ci.iconResource))
                BaseImporter.AOSPFavorites.ICON_TYPE_BITMAP -> {
                    info.iconBitmap = LauncherModel.getIconFromCursor(c, ci.icon, LauncherAppState.getInstance().context)
                    info.itemIcon = ItemIconUserBitmap(info.iconBitmap)
                }
            }
        }
        val id = info.id
        if (info is ItemInfoWithIcon && swipeActionMap != null && swipeActionMap.containsKey("swipe_up_$id")) {
            val ni = namedIntentFromApexKey(LauncherAppState.getInstance().context, swipeActionMap, "NONE", "swipe_up_$id", "swipe_up_app_$id", "swipe_up_intent_$id", "swipe_up_shortcut_name_$id", "swipe_up_shortcut_intent_$id")
            if (ni.intent != NovaAction.NONE.intent)
                info.flingUpIntent = ni
        }
    }

    /**
     * Checks and possibly transforms a launcher specific intent into a Nova intent
     * @param info original shortcut
     */
    override fun transformInfo(info: ShortcutInfo, c: Cursor) {
        val context = LauncherAppState.getInstance().context

        val intent = info.intent

        if (intent.component != null && "com.anddoes.launcher" == intent.component!!.packageName) {
            if ("com.anddoes.launcher.ui.LuckyAppActivity" == intent.component!!.className) {
                info.intent = null
                return
            }
        }
        if (APEX_ACTION == intent.action) {
            val extra = intent.getStringExtra(LAUNCHER_ACTION)
            if ("APP_DRAWER" == extra) {
                info.intent = NovaAction.APP_DRAWER.intent
                if (info.itemIcon == null || info.itemIcon is ItemIconResource)
                    info.itemIcon = ItemIconResource(Intent.ShortcutIconResource.fromContext(context, R.drawable.ic_allapps))
            } else if ("SHOW_PREVIEWS" == extra || "MANAGE_SCREENS" == extra) {
                info.intent = NovaAction.SHOW_PREVIEWS.intent
                if (info.itemIcon == null || info.itemIcon is ItemIconResource)
                    info.itemIcon = ItemIconResource(Intent.ShortcutIconResource.fromContext(context, R.drawable.ic_nova_action_preview))
            } else if ("GOTO_DEFAULT_SCREEN" == extra) {
                info.intent = NovaAction.GOTO_DEFAULT_SCREEN.intent
                if (info.itemIcon == null || info.itemIcon is ItemIconResource)
                    info.itemIcon = ItemIconResource(Intent.ShortcutIconResource.fromContext(context, R.drawable.ic_nova_action_default_screen))
            } else if ("SHOW_NOTIFICATIONS" == extra) {
                info.intent = NovaAction.EXPAND_STATUS_BAR.intent
                if (info.itemIcon == null || info.itemIcon is ItemIconResource)
                    info.itemIcon = ItemIconResource(Intent.ShortcutIconResource.fromContext(context, R.drawable.ic_nova_action_expand_notification_bar))
            } else if ("TOGGLE_STATUS_BAR" == extra) {
                info.intent = NovaAction.TOGGLE_STATUS_BAR.intent
                if (info.itemIcon == null || info.itemIcon is ItemIconResource)
                    info.itemIcon = ItemIconResource(Intent.ShortcutIconResource.fromContext(context, R.drawable.ic_nova_action_toggle_notification_bar))
            } else if ("TOGGLE_DOCK" == extra) {
                info.intent = NovaAction.TOGGLE_DOCK.intent
                if (info.itemIcon == null || info.itemIcon is ItemIconResource)
                    info.itemIcon = ItemIconResource(Intent.ShortcutIconResource.fromContext(context, R.drawable.ic_nova_action_toggle_dock))
            } else if ("RECENT_APPS" == extra) {
                info.intent = NovaAction.SHOW_RECENT_APPS.intent
                if (info.itemIcon == null || info.itemIcon is ItemIconResource)
                    info.itemIcon = ItemIconResource(Intent.ShortcutIconResource.fromContext(context, R.drawable.ic_nova_action_recent_apps))
            } else if ("APEX_MENU" == extra) {
                info.intent = null
            } else if ("APEX_SETTINGS" == extra) {
                info.intent = NovaAction.NOVA_SETTINGS.intent
                if (info.title.toString().toLowerCase(Locale.ENGLISH).contains("apex"))
                    info.title = context.getString(R.string.nova_settings)
                if (info.itemIcon == null || info.itemIcon is ItemIconResource)
                    info.itemIcon = ItemIconResource(Intent.ShortcutIconResource.fromContext(context, R.drawable.ic_launcher_settings))
            } else if ("LOCK_DESKTOP" == extra) {
                info.intent = null
            } else if ("GOTO_SCREEN1" == extra) {
                info.intent = NovaAction.GOTO_SCREEN.intent!!.putExtra(NovaAction.INTENT_EXTRA_PAGE_NUMBER, 0)
                if (info.itemIcon == null || info.itemIcon is ItemIconResource)
                    info.itemIcon = ItemIconResource(Intent.ShortcutIconResource.fromContext(context, R.drawable.ic_nova_action_screen_1))
            } else if ("GOTO_SCREEN2" == extra) {
                info.intent = NovaAction.GOTO_SCREEN.intent!!.putExtra(NovaAction.INTENT_EXTRA_PAGE_NUMBER, 1)
                if (info.itemIcon == null || info.itemIcon is ItemIconResource)
                    info.itemIcon = ItemIconResource(Intent.ShortcutIconResource.fromContext(context, R.drawable.ic_nova_action_screen_2))
            } else if ("GOTO_SCREEN3" == extra) {
                info.intent = NovaAction.GOTO_SCREEN.intent!!.putExtra(NovaAction.INTENT_EXTRA_PAGE_NUMBER, 2)
                if (info.itemIcon == null || info.itemIcon is ItemIconResource)
                    info.itemIcon = ItemIconResource(Intent.ShortcutIconResource.fromContext(context, R.drawable.ic_nova_action_screen_3))
            } else if ("GOTO_SCREEN4" == extra) {
                info.intent = NovaAction.GOTO_SCREEN.intent!!.putExtra(NovaAction.INTENT_EXTRA_PAGE_NUMBER, 3)
                if (info.itemIcon == null || info.itemIcon is ItemIconResource)
                    info.itemIcon = ItemIconResource(Intent.ShortcutIconResource.fromContext(context, R.drawable.ic_nova_action_screen_4))
            } else if ("GOTO_SCREEN5" == extra) {
                info.intent = NovaAction.GOTO_SCREEN.intent!!.putExtra(NovaAction.INTENT_EXTRA_PAGE_NUMBER, 4)
                if (info.itemIcon == null || info.itemIcon is ItemIconResource)
                    info.itemIcon = ItemIconResource(Intent.ShortcutIconResource.fromContext(context, R.drawable.ic_nova_action_screen_5))
            } else if ("GOTO_SCREEN6" == extra) {
                info.intent = NovaAction.GOTO_SCREEN.intent!!.putExtra(NovaAction.INTENT_EXTRA_PAGE_NUMBER, 5)
                if (info.itemIcon == null || info.itemIcon is ItemIconResource)
                    info.itemIcon = ItemIconResource(Intent.ShortcutIconResource.fromContext(context, R.drawable.ic_nova_action_screen_6))
            } else if ("GOTO_SCREEN7" == extra) {
                info.intent = NovaAction.GOTO_SCREEN.intent!!.putExtra(NovaAction.INTENT_EXTRA_PAGE_NUMBER, 6)
                if (info.itemIcon == null || info.itemIcon is ItemIconResource)
                    info.itemIcon = ItemIconResource(Intent.ShortcutIconResource.fromContext(context, R.drawable.ic_nova_action_screen_7))
            } else if ("GOTO_SCREEN8" == extra) {
                info.intent = NovaAction.GOTO_SCREEN.intent!!.putExtra(NovaAction.INTENT_EXTRA_PAGE_NUMBER, 7)
                if (info.itemIcon == null || info.itemIcon is ItemIconResource)
                    info.itemIcon = ItemIconResource(Intent.ShortcutIconResource.fromContext(context, R.drawable.ic_nova_action_screen_8))
            } else if ("GOTO_SCREEN9" == extra) {
                info.intent = NovaAction.GOTO_SCREEN.intent!!.putExtra(NovaAction.INTENT_EXTRA_PAGE_NUMBER, 8)
                if (info.itemIcon == null || info.itemIcon is ItemIconResource)
                    info.itemIcon = ItemIconResource(Intent.ShortcutIconResource.fromContext(context, R.drawable.ic_nova_action_screen_9))
            }
        }
        if (info.itemType == LauncherSettings.Favorites.ITEM_TYPE_SHORTCUT) {
            info.status = info.status or ShortcutInfo.FLAG_RESTORED_ICON or ShortcutInfo.FLAG_RESTORE_STARTED
        }
        super.transformInfo(info, c)
    }

    override fun getPackageName(): String {
        return "com.anddoes.launcher"
    }

    override fun createAllApps(context: Context): ItemInfo? {
        return null
    }

    override fun needRootHelperForWidgets(): Boolean {
        return true
    }

    companion object {
        const val APEX_ACTION = "com.anddoes.launcher.ACTION"
        const val LAUNCHER_ACTION = "LAUNCHER_ACTION"
    }
}
