package com.teslacoilsw.launcher.onboarding

import android.annotation.SuppressLint
import android.content.Context
import android.database.sqlite.SQLiteDatabase
import android.graphics.Bitmap
import android.graphics.Canvas
import android.graphics.Color
import android.graphics.Paint
import android.graphics.Point
import aosp.android.internal.util.XmlUtils
import com.android.launcher3.LauncherSettings
import com.android.launcher3.WorkspaceScreenInfo
import com.teslacoilsw.launcher.BuildConfig
import com.teslacoilsw.launcher.CellSpecConfig
import com.teslacoilsw.launcher.importer.BaseImporter
import com.teslacoilsw.launcher.preferences.Pref
import com.teslacoilsw.launcher.preferences.Pref3
import com.teslacoilsw.launcher.preferences.cellSpecsConfigFromLegacyPrefs
import com.teslacoilsw.launcher.util.DpUtil
import timber.log.Timber
import java.io.BufferedInputStream
import java.io.File
import java.io.FileOutputStream
import java.io.InputStream
import java.util.*
import java.util.zip.ZipInputStream
import kotlin.math.min
import kotlin.math.roundToInt


object BackupPreviewGenerator {
    private val iconItemTypes = arrayOf(LauncherSettings.Favorites.ITEM_TYPE_APPLICATION, LauncherSettings.Favorites.ITEM_TYPE_SHORTCUT, LauncherSettings.Favorites.ITEM_TYPE_FOLDER, LauncherSettings.Favorites.ITEM_TYPE_DEEP_SHORTCUT)

    @SuppressLint("ApplySharedPref")
    fun previewFromStream(context: Context, input: InputStream, width: Int, height: Int): Bitmap {
        var prefMap: Map<String, *>? = null
        var tempDb: File? = null
        ZipInputStream(BufferedInputStream(input)).use { zip ->
            try {
                var ze = zip.nextEntry
                while (ze != null) {
                    if ("com.teslacoilsw.launcher_preferences.xml" == ze.name) {
                        try {
                            prefMap = XmlUtils.readMapXml(zip) as Map<String, *>
                        } catch (ex: Exception) {
                            ex.printStackTrace()
                        }
                    } else if ("launcher.db" == ze.name) {
                        tempDb = File.createTempFile("tempBackup", "db", context.cacheDir).apply {
                            deleteOnExit()
                        }
                        val os = FileOutputStream(tempDb)
                        os.use {
                            val buffer = ByteArray(4096)
                            var bytesRead = zip.read(buffer)
                            while (bytesRead != -1) {
                                it.write(buffer, 0, bytesRead)
                                bytesRead = zip.read(buffer)
                            }
                        }
                    }
                    zip.closeEntry()
                    ze = zip.nextEntry
                }
            } catch (e: Exception) {
                if (BuildConfig.DEBUG)
                    e.printStackTrace()
            }
        }

        val bitmap = Bitmap.createBitmap(width, height, Bitmap.Config.ARGB_8888)
        if (prefMap == null || tempDb == null || !previewFrom(bitmap, prefMap!!, tempDb!!)) {
            bitmap.eraseColor(Color.RED)
        }
        tempDb?.delete()
        return bitmap
    }

    fun previewFrom(bitmap: Bitmap, prefMap: Map<String, *>, tempDb: File): Boolean {
        val prefVersion = prefMap["preference_version"] as? Int ?: 0
        val desktopCellSpecs = if (prefVersion < 59901)
            cellSpecsConfigFromLegacyPrefs(prefMap, "desktop")
        else
            CellSpecConfig.parse(prefMap["desktop_cellspecs"] as? String
                    ?: "", Pref3.desktop_cellspecs.defaultValue)

        val dockCellSpecs = if (prefVersion < 59901)
            cellSpecsConfigFromLegacyPrefs(prefMap, "dock")
        else
            CellSpecConfig.parse(prefMap["dock_cellspecs"] as? String
                    ?: "", Pref3.dock_cellspecs.defaultValue)

        val dockIcons = mutableListOf<Int>()
        val desktopIcons = mutableListOf<DesktopItem>()
        SQLiteDatabase.openDatabase(tempDb.path, null, 0).use { db ->
            val dockC = db.query(LauncherSettings.Favorites.TABLE_NAME, null, "container = ${LauncherSettings.Favorites.CONTAINER_HOTSEAT} AND screen = 0", null, null, null, LauncherSettings.Favorites.INDEXED_RANK)
            dockC?.use {
                val rankIndex = it.getColumnIndex(LauncherSettings.Favorites.INDEXED_RANK)
                if (rankIndex != -1) {
                    while (it.moveToNext()) {
                        dockIcons.add(it.getInt(rankIndex))
                    }
                }
            }

            val screenC = db.query(LauncherSettings.WorkspaceScreens.TABLE_NAME, null, null, null, null, null, LauncherSettings.WorkspaceScreens.SCREEN_RANK)
            val defaultScreen = Pref.Value.getDesktopDefaultPage(prefMap[Pref.Key.DESKTOP_DEFAULT_PAGE] as? Int ?: -1, screenC.count, false)
            val screenId =
                if (screenC != null) {
                    val screens = ArrayList<WorkspaceScreenInfo>()
                    val idIndex = screenC.getColumnIndexOrThrow(BaseImporter.ID)
                    val rankIndex = screenC.getColumnIndexOrThrow(BaseImporter.SCREEN_RANK)
                    while (screenC.moveToNext()) {
                        val rank = screenC.getInt(rankIndex)
                        val id = screenC.getLong(idIndex)
                        screens.add(Math.min(rank, screens.size), WorkspaceScreenInfo(id, rank, false))
                    }
                    screenC.close()
                    screens[defaultScreen].id
                } else {
                    defaultScreen.toLong()
                }

            val desktopC = db.query(LauncherSettings.Favorites.TABLE_NAME, arrayOf(LauncherSettings.Favorites.ITEM_TYPE, LauncherSettings.Favorites.CELLX, LauncherSettings.Favorites.CELLY, LauncherSettings.Favorites.SPANX, LauncherSettings.Favorites.SPANY), "container = ${LauncherSettings.Favorites.CONTAINER_DESKTOP} AND screen = ?", arrayOf(screenId.toString()), null, null, null)
            desktopC?.use {
                val itemTypeIndex = it.getColumnIndexOrThrow(LauncherSettings.Favorites.ITEM_TYPE)
                val cellXIndex = it.getColumnIndexOrThrow(LauncherSettings.Favorites.CELLX)
                val cellYIndex = it.getColumnIndexOrThrow(LauncherSettings.Favorites.CELLY)
                val spanXIndex = it.getColumnIndexOrThrow(LauncherSettings.Favorites.SPANX)
                val spanYIndex = it.getColumnIndexOrThrow(LauncherSettings.Favorites.SPANY)

                while (it.moveToNext()) {
                    desktopIcons.add(DesktopItem(it.getInt(itemTypeIndex),
                            it.getFloat(cellXIndex),
                            it.getFloat(cellYIndex),
                            it.getFloat(spanXIndex),
                            it.getFloat(spanYIndex)))
                }
            }
        }

        val desktopColumns = prefMap[Pref.Key.DESKTOP_GRID_COLS] as? Int ?: 5
        val desktopRows = prefMap[Pref.Key.DESKTOP_GRID_ROWS] as? Int ?: 5
        generatePreview(bitmap, desktopIcons, desktopCellSpecs, Point(desktopColumns, desktopRows), dockIcons, dockCellSpecs, prefMap[Pref.Key.DOCK_GRID_COLS] as? Int ?: 5, prefMap[Pref.Key.DOCK_ENABLED] != false)

        return true
    }

    fun generatePreview(bitmap: Bitmap, desktopIcons: List<DesktopItem>, desktopCellSpecs: CellSpecConfig, desktopGrid: Point, dockIcons: List<Int>, dockCellSpecs: CellSpecConfig, dockColumns: Int, dockEnabled: Boolean = true) {

        // Pixel is 411x659dp, we're 104x186dp =~ 0.25
        val scale = 0.25f
        val baseIconSize = 0.25f * DpUtil.getPixelSize(48)
        val desktopIconSize = desktopCellSpecs.iconScale * baseIconSize
        val dockIconSize = if (dockCellSpecs.matchDesktopSize) desktopIconSize else (dockCellSpecs.iconScale * baseIconSize)

        var bottom = bitmap.height
        val canvas = Canvas(bitmap)
        val paint = Paint(Paint.ANTI_ALIAS_FLAG).apply {
            color = Color.WHITE
        }
        bitmap.eraseColor(0x20000000.toInt())
        if (dockEnabled) {
            val dockCellSize = dockIconSize * 1.2f
            val dockCellWidth = Math.min(dockCellSize, bitmap.width / dockColumns.toFloat())
            val effectiveDockIconSize = Math.min(dockCellSize * 0.75f, dockIconSize)
            val cy = bottom - dockCellSize / 2f
            val padding = (bitmap.width - dockCellWidth*dockColumns)/2
            Timber.v("dockIcons $dockIcons")
            for (i in dockIcons) {
                if (i in 0 until dockColumns) {
                    val cx = padding + i * dockCellWidth + dockCellWidth/2f
                    canvas.drawCircle(cx, cy, effectiveDockIconSize / 2f, paint)
                }
            }
            bottom -= dockCellSize.roundToInt()
        }

        val desktopColumns = desktopGrid.x
        val desktopRows = desktopGrid.y
        val desktopCellWidth = bitmap.width / desktopColumns.toFloat()
        val desktopCellHeight = bottom / desktopRows.toFloat()
        val effectiveDesktopIconSize = min(desktopCellWidth * 0.75f, min(desktopCellHeight * 0.75f, desktopIconSize))
        val rx = min(desktopCellWidth, desktopCellHeight) / 4
        for (i in desktopIcons) {
            if (i.cellX in 0 until desktopColumns && i.cellY in 0 until desktopRows) {
                if (i.itemType in iconItemTypes) {
                    val cx = desktopCellWidth * i.cellX + desktopCellWidth / 2f
                    val cy = desktopCellHeight * i.cellY + desktopCellHeight / 2f
                    canvas.drawCircle(cx, cy, effectiveDesktopIconSize / 2f, paint)
                } else {
                    if (i.itemType > 1000 && i.spanY == 1f) {
                        val searchbarHeight = min(desktopCellHeight, DpUtil.getPixelSize(48) * scale)
                        val cy = desktopCellHeight * i.cellY + desktopCellHeight / 2f
                        val top = cy - searchbarHeight
                        val bottom = top + searchbarHeight
                        val left = desktopCellWidth * i.cellX
                        val right = left + desktopCellWidth * i.spanX
                        canvas.drawRoundRect(left, top, right, bottom, rx, rx, paint)
                    } else {
                        val top = desktopCellHeight * i.cellY
                        val bottom = top + desktopCellHeight * i.spanY
                        val left = desktopCellWidth * i.cellX
                        val right = left + desktopCellWidth * i.spanX
                        canvas.drawRoundRect(left, top, right, bottom, rx, rx, paint)
                    }
                }
            }
        }
        canvas.setBitmap(null)
    }

    fun generatePreview(bitmap: Bitmap, importedData: ImportedData) {
        var desktopItems = importedData.items.filter { it.container == LauncherSettings.Favorites.CONTAINER_DESKTOP.toLong() && it.screenId == importedData.defaultScreen }
                .map { DesktopItem(it.itemType, it.cellX, it.cellY, it.spanX, it.spanY) }

        if (desktopItems.isEmpty()) {
            for (s in importedData.screens) {
                desktopItems = importedData.items.filter { it.container == LauncherSettings.Favorites.CONTAINER_DESKTOP.toLong() && it.screenId == s.id }
                        .map { DesktopItem(it.itemType, it.cellX, it.cellY, it.spanX, it.spanY) }
                if (desktopItems.isNotEmpty())
                    break
            }
        }
        val dockItems = importedData.items.filter { it.container == LauncherSettings.Favorites.CONTAINER_HOTSEAT.toLong() && it.screenId == 0L }
                .map { it.indexedRank }
        generatePreview(bitmap, desktopItems,
                Pref3.desktop_cellspecs.defaultValue, Point(importedData.desktopCols, importedData.desktopRows),
                dockItems, Pref3.dock_cellspecs.defaultValue, importedData.dockCols)
    }

}

public data class DesktopItem(val itemType: Int, val cellX: Float, val cellY: Float, val spanX: Float, val spanY: Float)