package com.teslacoilsw.launcher.onboarding

import android.Manifest
import android.app.Activity
import android.appwidget.AppWidgetHost
import android.content.Context
import android.content.Intent
import android.content.SharedPreferences
import android.content.pm.PackageManager
import android.graphics.Bitmap
import android.graphics.Color
import android.graphics.drawable.ColorDrawable
import android.net.Uri
import android.os.Environment
import android.widget.Toast
import androidx.core.net.toUri
import com.android.launcher3.ItemInfo
import com.android.launcher3.Launcher
import com.android.launcher3.LauncherAppState
import com.android.launcher3.LauncherAppWidgetInfo
import com.android.launcher3.LauncherModel
import com.android.launcher3.LauncherSettings
import com.android.launcher3.ShortcutInfo
import com.android.launcher3.WorkspaceScreenInfo
import com.android.launcher3.util.ContentWriter
import com.teslacoilsw.kotlin.lazy1
import com.teslacoilsw.kotlin.tryCatch
import com.teslacoilsw.launcher.NovaLauncher
import com.teslacoilsw.launcher.R
import com.teslacoilsw.launcher.backup.BackupCategoryDocStorage
import com.teslacoilsw.launcher.backup.BackupRestoreAdapter
import com.teslacoilsw.launcher.importer.Importer
import com.teslacoilsw.launcher.importer.PixelLauncherImporter
import com.teslacoilsw.launcher.preferences.Pref
import com.teslacoilsw.launcher.preferences.Pref2
import com.teslacoilsw.launcher.preferences.Pref3
import io.reactivex.Single
import io.reactivex.android.schedulers.AndroidSchedulers
import io.reactivex.disposables.CompositeDisposable
import io.reactivex.disposables.Disposable
import io.reactivex.schedulers.Schedulers
import io.reactivex.subjects.BehaviorSubject
import java.io.File
import java.util.ArrayList
import java.util.concurrent.TimeUnit
import kotlin.collections.HashSet
import kotlin.math.max
import kotlin.math.min

private const val REQUEST_CODE_BROWSE = 1000

sealed class ImportOption : WizardOption {
    abstract fun generatePreview(context: Context, width: Int, height: Int): Bitmap
    open val finishOnNext = false
    open fun doImport(context: Context, editor: SharedPreferences.Editor) {
        throw IllegalStateException("This option should never be selected $this")
    }
}

object startAfreshOption : ImportOption() {
    private val freshImportData = makeFreshImportData()

    override fun generatePreview(context: Context, width: Int, height: Int): Bitmap {
        val bmp = Bitmap.createBitmap(width, height, Bitmap.Config.ARGB_8888)
        BackupPreviewGenerator.generatePreview(bmp, freshImportData)
        return bmp
    }

    override fun getLabel(context: Context): String {
        return context.getString(R.string.preset_start_afresh)
    }

    override fun doImport(context: Context, editor: SharedPreferences.Editor) { }

    private fun makeFreshImportData(): ImportedData {
        val dockCols = 5
        val items = mutableListOf<ItemInfo>()
        for (i in 0 until dockCols) {
            val info = ShortcutInfo()
            info.indexedRank = i
            info.cellX = i.toFloat()
            info.container = LauncherSettings.Favorites.CONTAINER_HOTSEAT.toLong()
            info.screenId = 0
            items.add(info)
        }
        return ImportedData(items, listOf(WorkspaceScreenInfo(0, 0, true)), 5, 5, 0, dockCols, 1)
    }
}

object novaBackupPermissionRequestOption : ImportOption(), PermissionOption {
    override val permissionRequestCode = 1000

    override fun getLabel(context: Context): String {
        return "Nova backup"
    }

    override val permissionsRequested = arrayOf(Manifest.permission.READ_EXTERNAL_STORAGE, Manifest.permission.WRITE_EXTERNAL_STORAGE)

    override fun generatePreview(context: Context, width: Int, height: Int): Bitmap {
        val bmp = Bitmap.createBitmap(width, height, Bitmap.Config.ARGB_8888)
        bmp.eraseColor(Color.RED)
        return bmp
    }
}

object novaBackupBrowseOption : ImportOption() {
    override fun getLabel(context: Context): String {
        return context.getString(R.string.browse)
    }

    override fun generatePreview(context: Context, width: Int, height: Int): Bitmap {
        val bmp = Bitmap.createBitmap(width, height, Bitmap.Config.ARGB_8888)
        bmp.eraseColor(Color.RED)
        return bmp
    }
}

data class NovaBackupFileOption(val label: String, val file: BackupRestoreAdapter.BackupUriFile) : ImportOption() {
    override val finishOnNext = true
    override fun getLabel(context: Context) = label

    override fun generatePreview(context: Context, width: Int, height: Int): Bitmap {
        return file.generatePreview(context, width, height) ?: Bitmap.createBitmap(width, height, Bitmap.Config.ARGB_8888).apply { eraseColor(Color.MAGENTA) }
    }

    override fun doImport(context: Context, editor: SharedPreferences.Editor) {
        file.doRestore(context, editor)
    }
}

class ImportLauncherOption(val importer: Importer) : ImportOption() {
    override fun getLabel(context: Context): String {
        val pm = context.applicationContext.packageManager
        return try {
            val appInfo = pm.getApplicationInfo(importer.packageName, 0)
            appInfo.loadLabel(pm).toString()
        } catch (e: Exception) {
            importer.packageName
        }
    }

    override fun generatePreview(context: Context, width: Int, height: Int): Bitmap {
        val data = prepareImport(context, importer)
        val bmp = Bitmap.createBitmap(width, height, Bitmap.Config.ARGB_8888)
        BackupPreviewGenerator.generatePreview(bmp, data)
        return bmp
    }

    override fun doImport(context: Context, edit: SharedPreferences.Editor) {
        val cr = context.contentResolver
        val data = prepareImport(context, importer)
        NovaLauncher.getLauncher()?.finish()

        AppWidgetHost(context, Launcher.APPWIDGET_HOST_ID)
        AppWidgetHost.deleteAllHosts()
        LauncherModel.deleteAll(context)

        LauncherModel.updateWorkspaceScreenOrder(context, data.screens, false)

        val noIdItems = mutableListOf<ItemInfo>()
        for (info in data.items) {
            if (info.id == ItemInfo.NO_ID) {
                noIdItems.add(info)
            } else {
                val writer = ContentWriter(context)
                writer.put(LauncherSettings.Favorites._ID, info.id)
                info.onAddToDatabase(writer)
                cr.insert(LauncherSettings.Favorites.CONTENT_URI, writer.valuesUnsafe)
            }
        }
        LauncherAppState.getLauncherProvider()!!.initializeMaxId()
        for (info in noIdItems) {
            info.id = LauncherAppState.getLauncherProvider()!!.generateNewItemId()
            val writer = ContentWriter(context)
            writer.put(LauncherSettings.Favorites._ID, info.id)
            info.onAddToDatabase(writer)
            cr.insert(LauncherSettings.Favorites.CONTENT_URI, writer.valuesUnsafe)
        }

        if (importer is PixelLauncherImporter) {
            edit.putBoolean(Pref3.drawer_cards.key, false)
            edit.putBoolean(Pref2.get().drawer.drawer_swipe_from_dock.key(), true)
            edit.putBoolean(Pref3.drawer_swipe_indicator.key, true)
            edit.putString(Pref.Key.DESKTOP_SCROLL_INDICATOR, Pref.Key.ScrollIndicatorStyle.LINE.name)
            edit.putBoolean(Pref3.drawer_searchbar.key, true)

            edit.putInt(Pref.Key.DESKTOP_DEFAULT_PAGE, 0)

            edit.putBoolean(Pref.Key.DOCK_BACKGROUNED_ENABLED, true)
        }

        edit.putInt(Pref.Key.DESKTOP_GRID_COLS, data.desktopCols)
                .putInt(Pref.Key.DESKTOP_GRID_ROWS, data.desktopRows)
                .putInt(Pref.Key.DOCK_PAGES, data.dockPages)
                .putInt(Pref.Key.DOCK_GRID_COLS, data.dockCols)
    }
}

class ImportPage : SimpleWizardPage<ImportOption>() {
    override val finishOnNextObservable: BehaviorSubject<Boolean> = BehaviorSubject.create<Boolean>().apply {
        onNext(false)
    }
    private val previewWidth by lazy1 { requireContext().resources.getDimensionPixelSize(R.dimen.onboarding_preview_width) }
    private val previewHeight by lazy1 { requireContext().resources.getDimensionPixelSize(R.dimen.onboarding_preview_height) }
    private val backupDocStorage by lazy1 { BackupCategoryDocStorage(requireContext()).getContent }
    private val browsedOptions = mutableListOf<NovaBackupFileOption>()

    private val compositeDisposable = CompositeDisposable()
    private var previewSubscription: Disposable? = null

    override fun getTitle(context: Context): String = context.getString(R.string.preset_import)

    override fun updatePreview() {
        previewSubscription?.dispose()

        val option = currentOption
        val s = Single.create<Bitmap> {
                it.onSuccess(option.generatePreview(requireContext(), previewWidth, previewHeight))
            }
            .subscribeOn(Schedulers.io())
            .observeOn(AndroidSchedulers.mainThread())
            .doOnSuccess {
                if (option == currentOption)
                    binding.preview.setImageBitmap(it)
            }
            .subscribe()
        previewSubscription = s

        compositeDisposable.add(Single.just(ColorDrawable(0x20000000))
                .delay(200, TimeUnit.MILLISECONDS)
                .observeOn(AndroidSchedulers.mainThread())
                .doOnSuccess {
                    if (previewSubscription == s && !s.isDisposed) {
                        binding.preview.setImageDrawable(it)
                    }
                }
                .subscribe())

    }

    override val options: BehaviorSubject<WizardOptions<ImportOption>> = BehaviorSubject.create()

    override fun onAttach(context: Context) {
        super.onAttach(context)
        refreshOptions()
    }

    override fun onDestroyView() {
        super.onDestroyView()
        previewSubscription?.dispose()
    }

    override fun onOptionSelected(option: ImportOption): Boolean {
        if (!super.onOptionSelected(option))
            return false
        if (option == novaBackupBrowseOption) {
            startActivityForResult(backupDocStorage.restoreIntent, REQUEST_CODE_BROWSE)
            return false
        }
        finishOnNextObservable.onNext(option.finishOnNext)
        return true
    }

    private fun refreshOptions() {
        val list = mutableListOf<ImportOption>(startAfreshOption)
        list.addAll(browsedOptions)
        if (novaBackupPermissionRequestOption.hasPermissions(requireContext())) {
            val backupdir = File(Environment.getExternalStorageDirectory(), "data/com.teslacoilsw.launcher/backup")
            val files = backupdir.listFiles() ?: emptyArray()
            files.sortByDescending { it.lastModified() }
            for (file in files) {
                if (file.extension == "novabackup")
                    list.add(NovaBackupFileOption(file.nameWithoutExtension, BackupRestoreAdapter.BackupUriFile(file.toUri())))
            }
        } else {
            list.add(novaBackupPermissionRequestOption)
        }
        list.add(novaBackupBrowseOption)
        for (importer in Importer.IMPORTERS) {
            if (importer.exists(context))
                list.add(ImportLauncherOption(importer))
        }
        options.onNext(WizardOptions(list[0], list))
    }

    override fun save(context: Context, editor: SharedPreferences.Editor) {
        currentOption.doImport(context, editor)
    }

    override fun onRequestPermissionsResult(requestCode: Int, permissions: Array<out String>, grantResults: IntArray) {
        if (grantResults.any { it == PackageManager.PERMISSION_GRANTED })
            refreshOptions()
    }

    private val INVALID_BACKUP_FILE = BackupRestoreAdapter.BackupUriFile(Uri.fromParts("", "", ""))
    override fun onActivityResult(requestCode: Int, resultCode: Int, data: Intent?) {
        if (requestCode == REQUEST_CODE_BROWSE && resultCode == Activity.RESULT_OK) {
            compositeDisposable.add(backupDocStorage.validateResult(requireContext(), data)
                    .observeOn(AndroidSchedulers.mainThread())
                    .onErrorReturn {
                        Toast.makeText(requireContext(), "Invalid backup file", Toast.LENGTH_SHORT).show()
                        INVALID_BACKUP_FILE
                    }
                    .subscribe {  it ->
                        if (it != INVALID_BACKUP_FILE) {
                            val label = if (it.scheme == "file")
                                it.getName(requireContext())
                            else {
                                if (browsedOptions.isEmpty())
                                    "User selected"
                                else
                                    "User selected " + (browsedOptions.size + 1)
                            }
                            val option = NovaBackupFileOption(label, it)
                            browsedOptions.add(0, option)
                            setSelected(option)
                            refreshOptions()
                        }
                    }
            )

        }
    }

    override fun onDestroy() {
        super.onDestroy()
        compositeDisposable.clear()
    }
}

data class ImportedData(val items: List<ItemInfo>, val screens: List<WorkspaceScreenInfo>, val desktopRows: Int, val desktopCols: Int, val defaultScreen: Long, val dockCols: Int, val dockPages: Int, val task: ((editor: SharedPreferences.Editor) -> Unit)? = null)

private fun prepareImport(context: Context, importer: Importer): ImportedData {
    val cr = context.contentResolver
    val importerScreens = importer.getScreens(cr)
    val screens: MutableList<WorkspaceScreenInfo>
    var hasScreens = false
    if (importerScreens == null) {
        screens = mutableListOf()
    } else {
        screens = importerScreens
        screens.sortBy { it.rank }
        hasScreens = true
    }
    val items = ArrayList<ItemInfo>()

    val c = tryCatch(null) { importer.query(cr) }
    c?.use {
        while (it.moveToNext()) {
            val info = importer.createItemInfoFromRow(c)
            if (info != null) {
                items.add(info)
                if (info.container == LauncherSettings.Favorites.CONTAINER_DESKTOP.toLong()) {
                    if (!hasScreens) {
                        var needAdd = true
                        SCREEN_LOOP@ for (screen in screens) {
                            if (screen.id == info.screenId) {
                                needAdd = false
                                break@SCREEN_LOOP
                            }
                        }
                        if (needAdd) {
                            screens.add(Math.min(info.screenId.toInt(), screens.size), WorkspaceScreenInfo(info.screenId, info.screenId.toInt(), false))
                        }
                    }
                }
            }
        }
        c.close()
    }

    if (screens.isEmpty()) {
        screens.add(0, WorkspaceScreenInfo(0, 0, false))
    }

    /* Sanity check */
    var firstScreen: WorkspaceScreenInfo? = null
    val screenIds = HashSet<Long>(screens.size)
    val iter = screens.iterator()
    while (iter.hasNext()) {
        val info = iter.next()
        if (screenIds.contains(info.id))
            iter.remove()
        if (firstScreen == null)
            firstScreen = info
        if (info.rank < firstScreen.rank)
            firstScreen = info
        screenIds.add(info.id)
    }

    var desktopRows = 4
    var desktopCols = 4
    var dockIcons = 5
    var dockPages = 1

    val maxColumns = 16
    val maxRows = 16

    for (info in items) {
        if (info.container == LauncherSettings.Favorites.CONTAINER_DESKTOP.toLong()) {
            if (info.cellX < 0 || info.cellY < 0)
                continue
            val maxX = (info.cellX + max(1f, info.spanX)).toInt()
            val maxY = (info.cellY + max(1f, info.spanY)).toInt()

            desktopRows = min(max(desktopRows, maxY), maxRows)
            desktopCols = min(max(desktopCols, maxX), maxColumns)
        } else if (info.container == LauncherSettings.Favorites.CONTAINER_HOTSEAT.toLong()) {
            dockPages = min(max(dockPages.toLong(), info.screenId + 1).toInt(), 5)
            dockIcons = min(max(dockIcons, info.indexedRank + 1), 7)
        }
        if (info is LauncherAppWidgetInfo) {
            info.appWidgetId = -1
        }
//        val writer = ContentWriter(context)
//        writer.put(LauncherSettings.Favorites._ID, info.id)
//        info.onAddToDatabase(writer)
//        cr.insert(LauncherSettings.Favorites.CONTENT_URI, writer.values)
    }
    importer.createAllApps(context)?.let { items.add(it) }
//    val drawer = importer.createAllApps(context)
//    LauncherAppState.getLauncherProvider()!!.initializeMaxId()
//    if (drawer != null) {
//        items.add(drawer)
//        val writer = ContentWriter(context)
//        writer.put(LauncherSettings.Favorites._ID, LauncherAppState.getLauncherProvider()!!.generateNewItemId())
//        drawer!!.onAddToDatabase(writer)
//        cr.insert(LauncherSettings.Favorites.CONTENT_URI, writer.values)
//    }

//    @SuppressLint("CommitPrefEdits")
//    val edit = Pref.Value.mPrefs.edit()
//    edit.putInt(Pref.Key.DESKTOP_GRID_COLS, desktopCols)
//            .putInt(Pref.Key.DESKTOP_GRID_ROWS, desktopRows)
//            .putInt(Pref.Key.DOCK_PAGES, dockPages)
//            .putInt(Pref.Key.DOCK_GRID_COLS, dockIcons)

    return ImportedData(items, screens, desktopRows, desktopCols, screens[0].id, dockIcons, dockPages)
}