package com.teslacoilsw.launcher.onboarding

import android.annotation.SuppressLint
import android.content.Context
import android.content.SharedPreferences
import android.content.pm.PackageManager
import android.os.Bundle
import android.view.LayoutInflater
import android.view.View
import android.view.ViewGroup
import androidx.core.view.get
import androidx.fragment.app.Fragment
import com.google.android.material.chip.Chip
import com.teslacoilsw.kotlin.bind
import com.teslacoilsw.launcher.R
import com.teslacoilsw.launcher.databinding.OnboardingSimpleBinding
import io.reactivex.Observable
import io.reactivex.disposables.Disposable

abstract class WizardPage : Fragment() {
    open val finishOnNextObservable: Observable<Boolean> = Observable.just(false)

    abstract fun save(context: Context, editor: SharedPreferences.Editor)

    /**
     * Called when trying to progress to the next page
     * Do not save to preferences at this point
     *
     * Return true if allowed, false otherwise
     */
    open fun onNext() = true

}

abstract class SimpleWizardPage<T: WizardOption> : WizardPage() {
    protected lateinit var binding: OnboardingSimpleBinding
        private set
    abstract val options: Observable<WizardOptions<T>>
    private var isInitialized = false
    lateinit var currentOption: T
        private set
    private lateinit var optionsSubscription: Disposable
    abstract fun getTitle(context: Context): String

    override fun onCreateView(inflater: LayoutInflater, container: ViewGroup?, savedInstanceState: Bundle?): View? {
        binding = inflater.bind(R.layout.onboarding_simple, container, false)

        binding.title.text = getTitle(requireContext())
        optionsSubscription = options.subscribe {
            if (!isInitialized) {
                currentOption = it.default
                isInitialized = true
            }

            binding.chipGroup.removeAllViews()
            binding.chipGroup.setOnCheckedChangeListener(null)

            val inflater = LayoutInflater.from(context)
            for ((index, item) in it.list.withIndex()) {
                val chip: Chip = inflater.inflate(R.layout.onboarding_chip_item, binding.chipGroup, false) as Chip
                chip.text = item.getLabel(binding.chipGroup.context)
                chip.tag = item
                chip.id = index
                chip.isChecked = item == currentOption
                binding.chipGroup.addView(chip)
            }
            binding.chips.smoothScrollTo(0, 0)
            binding.chipGroup.setOnCheckedChangeListener { chipGroup, i ->
                if (i < 0 || chipGroup[i].tag != currentOption) {
                    if (i < 0 || !setSelected(chipGroup[i].tag as T)) {
                        if (chipGroup.childCount > 0) {
                            chipGroup.findViewWithTag<Chip>(currentOption).isChecked = true
                            setSelected(currentOption)
                        }
                    }
                }
            }

            updatePreview()
        }
        return binding.root
    }

    override fun onDestroyView() {
        super.onDestroyView()
        optionsSubscription.dispose()
    }

    fun setSelected(option: T): Boolean {
        if (onOptionSelected(option)) {
            currentOption = option
            updatePreview()
            return true
        }
        return false
    }

    @SuppressLint("NewApi")
    open fun onOptionSelected(selectedOption: T): Boolean {
        if (selectedOption is PermissionOption && selectedOption.permissionRequestCode >= 0) {
            val context = requireContext()
            if (selectedOption.permissionsRequested.any {
                        context.checkSelfPermission(it) != PackageManager.PERMISSION_GRANTED })
            {
                requestPermissions(selectedOption.permissionsRequested, selectedOption.permissionRequestCode)
                return !selectedOption.permissionsRequired
            }
        }
        return true
    }

    override fun onRequestPermissionsResult(requestCode: Int, permissions: Array<out String>, grantResults: IntArray) {
//        val optionForPermissions = options.find { it.permissionRequestCode == requestCode }
//        if (BuildConfig.DEBUG)
//            Timber.v("onRequestPermissionsResult: $optionForPermissions ${permissions.contentToString()} ${optionForPermissions?.permissionsRequested?.contentToString()} ${grantResults.contentToString()}")
//
//        if (optionForPermissions != null &&
//                permissions contentEquals optionForPermissions.permissionsRequested &&
//                grantResults.all { it == PackageManager.PERMISSION_GRANTED })
//        {
//            Timber.v("match")
//            selectedOption = optionForPermissions
//        }
    }

    abstract fun updatePreview()
}