package com.teslacoilsw.launcher.preferences

 import android.annotation.SuppressLint
 import android.content.ComponentName
 import android.content.SharedPreferences
 import android.graphics.Color
 import android.graphics.drawable.BitmapDrawable
 import android.graphics.drawable.ColorDrawable
 import android.graphics.drawable.Drawable
 import android.os.Build
 import android.os.Bundle
 import android.preference.PreferenceManager
 import android.util.SparseArray
 import android.view.View
 import android.widget.LinearLayout
 import android.widget.RadioButton
 import androidx.core.content.ContextCompat
 import androidx.core.graphics.createBitmap
 import androidx.core.view.isVisible
 import aosp.android.graphics.drawable.InsetDrawableProportional
 import com.afollestad.materialdialogs.MaterialDialog
 import com.android.launcher3.BubbleTextView
 import com.android.launcher3.FastBitmapDrawable
 import com.android.launcher3.IconCache
 import com.android.launcher3.LauncherAppState
 import com.android.launcher3.LauncherModel
 import com.android.launcher3.Utilities
 import com.teslacoilsw.cofs.internal.parse
 import com.teslacoilsw.kotlin.asBitmap
 import com.teslacoilsw.kotlin.bindContentView
 import com.teslacoilsw.kotlin.children
 import com.teslacoilsw.kotlin.forEachIndexed
 import com.teslacoilsw.launcher.BuildConfig
 import com.teslacoilsw.launcher.IconPackChooser
 import com.teslacoilsw.launcher.NovaLauncher
 import com.teslacoilsw.launcher.R
 import com.teslacoilsw.launcher.databinding.ActivityAdaptiveiconBinding
 import com.teslacoilsw.launcher.icon.AdaptiveIconShape
 import com.teslacoilsw.launcher.icon.CornerConfig
 import com.teslacoilsw.launcher.icon.CornerCustomShapePath
 import com.teslacoilsw.launcher.icon.CornerType
 import com.teslacoilsw.launcher.icon.NovaAdaptiveIconDrawable
 import com.teslacoilsw.launcher.icontheme.ICON_THEME_EMPTY
 import com.teslacoilsw.launcher.icontheme.IconPackType
 import com.teslacoilsw.launcher.icontheme.IconTheme
 import com.teslacoilsw.launcher.icontheme.IconThemeConfig
 import com.teslacoilsw.launcher.util.DpUtil
 import com.teslacoilsw.launcher.util.palette.ColorUtils
 import com.teslacoilsw.launcher.util.palette.Palette
 import com.teslacoilsw.launcher.util.palette.PaletteHelper
 import com.teslacoilsw.shared.colorpicker.ColorUtil
 import com.teslacoilsw.shared.poisonlollipop.activity.PoisonActionBarActivity
 import rx.Observable
 import rx.android.schedulers.AndroidSchedulers
 import rx.schedulers.Schedulers
 import rx.subscriptions.CompositeSubscription
 import timber.log.Timber

class AdaptiveIconSettingsActivity: PoisonActionBarActivity() {
    private lateinit var sharedPrefs: SharedPreferences
    private lateinit var binding: ActivityAdaptiveiconBinding
    private val adaptiveIcons = SparseArray<NovaAdaptiveIconDrawable>()
    private val iconSize = DpUtil.getPixelSize(64)
    private lateinit var selectedIconTheme: IconTheme

    private lateinit var shapeButtons: Array<RadioButton>
    private lateinit var currentShape: AdaptiveIconShape
    private lateinit var subscription: CompositeSubscription

    public override fun onCreate(savedInstanceState: Bundle?) {
        super.onCreate(savedInstanceState)
        subscription = CompositeSubscription()
        sharedPrefs = PreferenceManager.getDefaultSharedPreferences(this)
        NovaAdaptiveIconDrawable.clearMask()

        overridePendingTransition(0, 0)
        binding = bindContentView(R.layout.activity_adaptiveicon)
        if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.LOLLIPOP) {
            Utilities.setStatusAndNavTransparent(window, true, false, false)
            window.statusBarColor = 0x3d000000
        }

        binding.back.setOnClickListener {
            onBackPressed()
        }

        binding.iconTheme.setOnClickListener {
            IconPackChooser.getChooser(this, R.string.select_icon_pack, object : IconPackChooser.OnIconPackSelectedListener {
                override fun onIconPackSelected(
                        name: CharSequence, type: IconPackType, pkg: String, id: String) {

                    val n = name.toString()
                    val theme = if (n.isBlank() || (type == IconPackType.BUILTIN && id == null)) IconThemeConfig.DEFAULT else IconThemeConfig(n, type, pkg, id)

                    Pref3.icon_theme.value = theme

                    updateIconThemePref()
                    clearIconCache()
                }

                override fun showResetCustomIconDialog() {
                    val d = MaterialDialog.Builder(this@AdaptiveIconSettingsActivity)
                            .title(R.string.reset_custom_icons)
                            .content(R.string.reset_custom_icons_warning)
                            .negativeText(R.string.cancel)
                            .positiveText(R.string.ok)
                            .callback(object : MaterialDialog.ButtonCallback() {
                                override fun onPositive(dialog: MaterialDialog?) {
                                    LauncherModel.deleteCustomIcons(this@AdaptiveIconSettingsActivity)
                                    LauncherAppState.forceModelReload()
                                }
                            })
                            .show()
                    d.show()
                }
            }).show()
        }
        binding.iconThemeMasking.onUserChanged = { _, _ -> updateAllIcons() }
        currentShape = Pref3.adaptive_icon_shape.value

        binding.shapeDefault.tag = AdaptiveIconShape.DEFAULT
        binding.shapeCircle.tag = AdaptiveIconShape.CIRCLE
        binding.shapeSquircle.tag = AdaptiveIconShape.SQUIRCLE
        binding.shapeRoundedSquare.tag = AdaptiveIconShape.ROUNDED_SQUARE
        binding.shapeSquare.tag = AdaptiveIconShape.SQUARE
        binding.shapeTeardrop.tag = AdaptiveIconShape.TEARDROP
        var customShape = currentShape
        binding.shapeCustom.tag = customShape
        binding.customize.setOnClickListener {
            binding.shapeCustom.tag = currentShape
            updateCustomSliders(currentShape)
            binding.shapeCustom.isChecked = true
            binding.mainSettings.isVisible = false
            binding.customShapeSettings.isVisible = true
            updateAllIcons()
        }
        binding.customHeader.setOnClickListener {
            binding.mainSettings.isVisible = true
            binding.customShapeSettings.isVisible = false
            for (shapeButton in shapeButtons) {
                if (shapeButton != binding.shapeCustom) {
                    val tag = shapeButton.tag as AdaptiveIconShape
                    if (tag.shapeMatch(currentShape))
                        shapeButton.isChecked = true
                }
            }
            binding.shapeCustom.isVisible = binding.shapeCustom.isChecked
        }

        shapeButtons = arrayOf(binding.shapeDefault, binding.shapeCircle, binding.shapeSquircle, binding.shapeRoundedSquare, binding.shapeSquare, binding.shapeTeardrop, binding.shapeCustom)

        val updateCustom = { old: Any, new: Any ->
            val shape = CornerCustomShapePath(
                    CornerConfig(binding.customTopleftCorner.value, binding.customTopleftCornerRound.value.parse(CornerType.ROUND)),
                    CornerConfig(binding.customToprightCorner.value, binding.customToprightCornerRound.value.parse(CornerType.ROUND)),
                    CornerConfig(binding.customBottomleftCorner.value, binding.customBottomleftCornerRound.value.parse(CornerType.ROUND)),
                    CornerConfig(binding.customBottomrightCorner.value, binding.customBottomrightCornerRound.value.parse(CornerType.ROUND)))

            val adaptive = AdaptiveIconShape(shape)
            binding.shapeCustom.tag = adaptive
            binding.shapeCustom.setCompoundDrawables(null, adaptive.createPreviewDrawable(this), null, null)
            if (binding.shapeCustom.isChecked) {
                currentShape = adaptive
                adaptiveIcons.clear()
                updateAllIcons()
            }
        }
        binding.customTopleftCorner.onUserChanged = updateCustom
        binding.customToprightCorner.onUserChanged = updateCustom
        binding.customBottomleftCorner.onUserChanged = updateCustom
        binding.customBottomrightCorner.onUserChanged = updateCustom
        binding.customTopleftCornerRound.onUserChanged = updateCustom
        binding.customToprightCornerRound.onUserChanged = updateCustom
        binding.customBottomleftCornerRound.onUserChanged = updateCustom
        binding.customBottomrightCornerRound.onUserChanged = updateCustom

        val knownDefault = AdaptiveIconShape.getMatchingKnownShape()
        if (BuildConfig.DEBUG)
            Timber.v("KnownDefault $knownDefault ${AdaptiveIconShape.DEFAULT} normScale: ${AdaptiveIconShape.DEFAULT.normalizeScale}")
        var foundChecked = false
        for (shapeButton in shapeButtons) {
            val shape = shapeButton.tag as AdaptiveIconShape

            shapeButton.setCompoundDrawables(null, shape.createPreviewDrawable(this), null, null)
            if (shape != AdaptiveIconShape.DEFAULT && shape == knownDefault) {
                shapeButton.visibility = View.GONE
                @SuppressLint("SetTextI18n")
                binding.shapeDefault.text = shapeButton.text.toString() + "\n(" + getString(R.string.default_) + ")"
            }
            if (shape == Pref3.adaptive_icon_shape.value && !foundChecked) {
                foundChecked = true
                shapeButton.isChecked = true
            }
        }
        binding.shapeCustom.isVisible = binding.shapeCustom.isChecked

        binding.shapeGroup.setOnCheckedChangeListener { group, checkedId ->
            group.children().forEach {
                if (it.id != checkedId && it is RadioButton) {
                    it.isChecked = false
                    /* Workaround for 5.0/Asus items staying highlighted. TODO fix text color */
                    it.invalidate()
                }
            }
            currentShape = findViewById<RadioButton>(checkedId).tag as AdaptiveIconShape
            for (i in 0 until adaptiveIcons.size()) {
                val icon = adaptiveIcons[i]
                icon.shape = currentShape
            }
        }

        binding.reshapeLegacy.onUserChanged = { _, newValue ->
            if (newValue && binding.preferLegacy.isChecked)
                binding.preferLegacy.isChecked = false
            updateAllIcons()
        }
        binding.reshapeLegacy.isChecked = Pref2.get().adaptive_icon_plate.get() ?: false

        binding.preferLegacy.isChecked = Pref2.get().adaptive_icon_prefer_legacy.get() ?: false
        binding.preferLegacy.onUserChanged = { _, newValue ->
            if (newValue && binding.reshapeLegacy.isChecked)
                binding.reshapeLegacy.isChecked = false
            updateAllIcons()
        }

        binding.adaptiveIconAnimations.isChecked = Pref2.get().adaptive_icon_anim.get()!!

        binding.preferLegacy.isVisible = Build.VERSION.SDK_INT >= Build.VERSION_CODES.O
        binding.enableAdaptive.isVisible = Build.VERSION.SDK_INT < Build.VERSION_CODES.O
        binding.enableAdaptive.isChecked = Pref2.get().adaptive_icon_backport.get() ?: false
        binding.adaptiveHeader.isVisible = Build.VERSION.SDK_INT >= Build.VERSION_CODES.O
        binding.normalize.onUserChanged = { _, _ ->
            clearIconCache()
            updateAllIcons()
        }
    }

    override fun onBackPressed() {
        if (binding.customShapeSettings.isVisible) {
            binding.customHeader.callOnClick()
        } else {
            super.onBackPressed()
        }
    }

    private fun updateCustomSliders(adaptiveShape: AdaptiveIconShape) {
        val customShape = adaptiveShape.getCornerShape()
        binding.customTopleftCorner.value = customShape.topLeft.strength
        binding.customToprightCorner.value = customShape.topRight.strength
        binding.customBottomleftCorner.value = customShape.bottomLeft.strength
        binding.customBottomrightCorner.value = customShape.bottomRight.strength

        binding.customTopleftCornerRound.value = customShape.topLeft.type.name
        binding.customToprightCornerRound.value = customShape.topRight.type.name
        binding.customBottomleftCornerRound.value = customShape.bottomLeft.type.name
        binding.customBottomrightCornerRound.value = customShape.bottomRight.type.name
        currentShape = AdaptiveIconShape(customShape)
        binding.shapeCustom.tag = currentShape
    }

    override fun onDestroy() {
        super.onDestroy()
        subscription.unsubscribe()
    }

    override fun onResume() {
        super.onResume()
        updateIconThemePref()
    }

    private fun updateIconThemePref() {
        selectedIconTheme = Pref3.icon_theme.value.getIconTheme(this, Pref.Key.AppIconSize.calculateIconDpi(iconSize))
        binding.iconTheme.summary = Pref3.icon_theme.value.name

        if (binding.iconTheme.summary.isNullOrBlank())
            binding.iconTheme.summary = getString(R.string.system)
        binding.iconTheme.widgetIcon = null

        val callbackIconTheme = selectedIconTheme
        subscription.add(Observable.fromCallable { Pref3.icon_theme.value.getIcon(this@AdaptiveIconSettingsActivity) }
         .subscribeOn(Schedulers.io())
         .observeOn(AndroidSchedulers.mainThread())
         .subscribe {
             if (callbackIconTheme == selectedIconTheme)
                binding.iconTheme.widgetIcon = it
          })
        updateAllIcons()
    }

    private fun updateAllIcons() {
        binding.customHeader.widgetIcon = (binding.shapeCustom.tag as AdaptiveIconShape).createPreviewDrawable(this).apply { setColor(Color.WHITE) }
        binding.iconThemeMasking.isVisible = selectedIconTheme.getThemedIconFromBase(this, NovaLauncher.COMPONENT_NAME, createBitmap(1, 1)) != null
        binding.adaptiveIconThemeHint.isVisible = selectedIconTheme != ICON_THEME_EMPTY
        val x: LinearLayout = findViewById(R.id.preview_icons)
        x.forEachIndexed { i, v ->
            if (v is BubbleTextView) {
                if (v.id == R.id.legacy_icon) {
                    updateLegacyIcon()
                } else {
                    var d: Drawable
                    if (binding.preferLegacy.isChecked && i < 2) {
                        val resId = when (i) {
                            0 -> R.drawable.ic_lptheme_settings
                            else -> R.drawable.ic_lptheme_calculator
                        }
                        d = ContextCompat.getDrawable(this, resId)!!
                    } else {
                        val adaptive: NovaAdaptiveIconDrawable
                        val resId = when (i) {
                            0 -> R.drawable.adaptive_preview_settings
                            1 -> R.drawable.adaptive_preview_calculator
                            else -> R.drawable.ic_launcher_default_adaptive
                        }
                        adaptive = NovaAdaptiveIconDrawable.create(resources, resId, null)!!
                        adaptive.shape = currentShape
                        adaptiveIcons.put(i, adaptive)
                        d = adaptive
                    }

                    val cn = when (i) {
                        0 -> ComponentName("com.android.settings", "com.android.settings.Settings")
                        1 -> ComponentName("com.android.calculator2", "com.android.calculator2.Calculator")
                        2 -> ComponentName(BuildConfig.APPLICATION_ID, "ic_launcher_default_adaptive")
                        else -> ComponentName(BuildConfig.APPLICATION_ID, "ic_launcher_default")
                    }
                    val cache = LauncherAppState.getInstance().iconCache
                    if (i != 2) {
                        var themedBitmap = selectedIconTheme.getThemedIconBitmap(this, cache, cn)
                        if (themedBitmap == null && binding.iconThemeMasking.isChecked)
                            themedBitmap = selectedIconTheme.getThemedIconFromBase(this, cn, d.asBitmap(iconSize, iconSize))
                        if (binding.normalize.isChecked && themedBitmap != null)
                            themedBitmap = IconCache.sanifyIconSize(cn, themedBitmap)
                        if (themedBitmap != null)
                            d = FastBitmapDrawable(themedBitmap)
                    }
                    d.setBounds(0, 0, iconSize, iconSize)
                    v.applyCompoundDrawables(d)
                }
            }
        }
    }

    private fun updateLegacyIcon() {
        val base = (ContextCompat.getDrawable(this, R.mipmap.ic_launcher_default)!! as BitmapDrawable).bitmap
        val cn = ComponentName(BuildConfig.APPLICATION_ID, "ic_launcher_default")
        var themedBitmap = if (binding.iconThemeMasking.isChecked) selectedIconTheme.getThemedIconFromBase(this, cn, base) else null
        if (binding.normalize.isChecked && themedBitmap != null)
            themedBitmap = IconCache.sanifyIconSize(cn, themedBitmap)
        var d: Drawable = if (themedBitmap != null)
            BitmapDrawable(themedBitmap)
        else
            BitmapDrawable(base)
        if (binding.reshapeLegacy.isChecked && themedBitmap == null) {
            if (adaptiveIcons[3] == null) {
                var tintColor: Int
                val palette = Palette.generate((d as BitmapDrawable).bitmap)
                tintColor = PaletteHelper.getBestHighlightColorFromIconPalette(palette, false)
                tintColor = if (ColorUtil.isGray(tintColor))
                    Color.WHITE
                else
                    ColorUtils.getMutedColor(tintColor, 0.54f)

                d = NovaAdaptiveIconDrawable(ColorDrawable(tintColor),
                        InsetDrawableProportional(d, .25f))
                d.shape = currentShape
                adaptiveIcons.put(3, d)
            }
            d = adaptiveIcons[3]
        }
        d.setBounds(0, 0, iconSize, iconSize)
        binding.legacyIcon.applyCompoundDrawables(d)
    }

    override fun onPause() {
        super.onPause()
        var needsRefresh = false
        for (shapeButton in shapeButtons) {
            if (shapeButton.isChecked) {
                val shape = shapeButton.tag as AdaptiveIconShape
                if (Pref3.adaptive_icon_shape.value != shape) {
                    Pref3.adaptive_icon_shape.value = shape
                    needsRefresh = true
                }
                break
            }
        }

        if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.O) {
            val prefPreferLegacy = Pref2.get().adaptive_icon_prefer_legacy
            val preferLegacy = binding.preferLegacy.isChecked
            if (prefPreferLegacy.get() != preferLegacy) {
                prefPreferLegacy.set(preferLegacy)
                needsRefresh = true
            }
        } else {
            Pref2.get().adaptive_icon_backport.set(binding.enableAdaptive.isChecked)
        }

        val prefMask = Pref2.get().adaptive_icon_plate
        if (prefMask.get() != binding.reshapeLegacy.isChecked) {
            prefMask.set(binding.reshapeLegacy.isChecked)
            needsRefresh = true
        }

        Pref2.get().adaptive_icon_anim.set(binding.adaptiveIconAnimations.isChecked)

        if (needsRefresh) {
            clearIconCache()
        }
    }

    private fun clearIconCache() {
        LauncherAppState.getInstance().iconCache.flushCustomIcons()
        LauncherAppState.getInstance().iconCache.setIconThemeFromConfig()
        LauncherModel.deleteCachedIcons(this)
        LauncherAppState.forceModelReload()
    }
}