package com.teslacoilsw.launcher.preferences

import android.annotation.SuppressLint
import android.app.Dialog
import android.content.ComponentName
import android.content.Context
import android.content.Intent
import android.content.pm.ApplicationInfo
import android.content.pm.PackageManager
import android.content.pm.ResolveInfo
import android.graphics.Color
import android.graphics.drawable.ColorDrawable
import android.net.Uri
import android.os.Build
import android.os.Bundle
import android.preference.PreferenceManager
import android.text.ClipboardManager
import android.text.TextUtils
import android.text.format.DateUtils
import android.util.DisplayMetrics
import android.view.LayoutInflater
import android.widget.LinearLayout
import android.widget.ScrollView
import android.widget.TextView
import android.widget.Toast
import androidx.core.content.FileProvider
import androidx.core.view.updatePadding
import androidx.fragment.app.DialogFragment
import com.afollestad.materialdialogs.DialogAction
import com.afollestad.materialdialogs.MaterialDialog
import com.android.launcher3.LauncherAppState
import com.android.launcher3.Partner
import com.android.launcher3.Utilities
import com.android.launcher3.config.ProviderConfig
import com.google.android.libraries.gsa.launcherclient.LauncherClient
import com.google.android.material.textfield.TextInputLayout
import com.teslacoilsw.kotlin.dp
import com.teslacoilsw.kotlin.showKeyboard
import com.teslacoilsw.kotlin.tryCatch
import com.teslacoilsw.launcher.BuildConfig
import com.teslacoilsw.launcher.R
import com.teslacoilsw.launcher.drawable.FixedSizeDrawableWrapper
import com.teslacoilsw.launcher.homereset.HomeReset
import com.teslacoilsw.launcher.notificationlistener.CountlessObserver
import com.teslacoilsw.launcher.notificationlistener.NotificationListener
import com.teslacoilsw.launcher.preferences.fancyprefs.FancyPrefView
import com.teslacoilsw.launcher.preferences.widget.MatchWrapLinearLayout
import com.teslacoilsw.launcher.util.PackageNames
import com.teslacoilsw.matcompat.TintCompat
import com.teslacoilsw.shared.util.FileSlurp
import com.teslacoilsw.shared.util.Version
import com.teslacoilsw.shared.util.rewthelper.RewtHelper
import timber.log.Timber
import java.io.BufferedWriter
import java.io.File
import java.io.FileWriter
import java.io.IOException

class ContactSupportDialogFragment : DialogFragment() {
    private lateinit var subject: TextView
    private lateinit var body: TextView
    private lateinit var subjectParent: TextInputLayout
    private lateinit var bodyParent: TextInputLayout

    @SuppressLint("InflateParams")
    override fun onCreateDialog(savedInstanceState: Bundle?): Dialog {
        val view = LayoutInflater.from(activity).inflate(R.layout.contact_support, null, false)
        subject = view.findViewById(R.id.subject)
        body = view.findViewById(R.id.body)
        subjectParent = view.findViewById(R.id.til_subject)
        bodyParent = view.findViewById(R.id.til_body)


        val dialog = MaterialDialog.Builder(requireContext())
                .title(R.string.contact_support)
                .customView(view, true)
                .negativeText(R.string.cancel)
                .positiveText(R.string.email)
                .autoDismiss(false)
                .onNegative { dialog, which -> dialog.dismiss() }
                .onPositive(::submit)
                .build()
        if (savedInstanceState == null) {
            dialog.setOnShowListener {
                subject.requestFocus()
                subject.post({ subject.showKeyboard() })
            }
        }
        return dialog
    }

    private fun submit(dialog: MaterialDialog, which: DialogAction) {
        val context = requireContext()
        val pm = context.applicationContext.packageManager

        val errorFieldRequired = getString(R.string.error_field_required)

        var error = false
        subjectParent.isErrorEnabled = false
        bodyParent.isErrorEnabled = false

        if (true || !BuildConfig.DEBUG) {
            if (TextUtils.isEmpty(body.text)) {
                bodyParent.error = errorFieldRequired
                bodyParent.isErrorEnabled = true
                body.requestFocus()
                error = true
            }

            if (TextUtils.isEmpty(subject.text)) {
                subjectParent.error = errorFieldRequired
                subjectParent.isErrorEnabled = true
                subject.requestFocus()
                error = true
            }
        }

        if (error)
            return

        val debugText = StringBuilder(body.text)
        debugText.append("\n---\n")

        if (Pref.Value.nova_prime_enabled)
            debugText.append("Nova Launcher Prime: ")
        else
            debugText.append("Nova Launcher: ")

        try {
            val e = pm.getPackageInfo(BuildConfig.APPLICATION_ID, 0)
            debugText.append(e.versionName).append(' ').append(e.versionCode)
            debugText.append(' ')
                    .append(LauncherAppState.getInstance().apkMd5).append('\n')
        } catch (ex: PackageManager.NameNotFoundException) {
            debugText.append("Not installed\n")
        }

        debugText.append("Device: ").append(Build.BRAND).append(' ').append(Build.MODEL).append(' ').append(Build.DISPLAY).append('\n')
        debugText.append("Android: ").append(Build.VERSION.RELEASE).append(' ').append(Build.VERSION.SDK_INT)
        if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.M && Build.VERSION.PREVIEW_SDK_INT > 0)
            debugText.append(" Preview ").append(Build.VERSION.PREVIEW_SDK_INT)
        debugText.append('\n')
        val dm = DisplayMetrics()
        Utilities.getRealDisplayMetrics(requireActivity().windowManager.defaultDisplay, dm)
        debugText.append("Display: ").append(dm.heightPixels).append('×').append(dm.widthPixels).append(' ').append(dm.densityDpi).append('\n')
        val conf = resources.configuration
        debugText.append("Locale: ").append(conf.locale).append('\n')
        debugText.append("PPP?: ").append(PackageNames.hasLuckyPatcher(pm)).append('\n')
        debugText.append("Google Now:")

        try {
            val proxy = context.applicationContext.packageManager.getPackageInfo(LauncherClient.LAUNCHERCLIENT_PROXY_PACKAGE, 0)
            if (proxy.applicationInfo.flags and (ApplicationInfo.FLAG_SYSTEM or ApplicationInfo.FLAG_UPDATED_SYSTEM_APP or ApplicationInfo.FLAG_DEBUGGABLE) != 0)
                debugText.append(" Companion v").append(proxy.versionName).append(" (").append(proxy.versionCode).append(")")
            else
                debugText.append(" Companion invalid. Not debuggable or system app")
        } catch (e: PackageManager.NameNotFoundException) {
            e.printStackTrace()
        }

        if (LauncherAppState.isSystemApp(context))
            debugText.append(" System app")
        if (Build.VERSION.SDK_INT < Build.VERSION_CODES.LOLLIPOP)
            debugText.append(" Too old (< Lollipop)")
        try {
            val proxy = context.applicationContext.packageManager.getPackageInfo("com.google.android.googlequicksearchbox", 0)
            debugText.append(" Google v").append(proxy.versionName).append(" (").append(proxy.versionCode).append(")")
        } catch (e: PackageManager.NameNotFoundException) {
            debugText.append(" Google not installed")
        }

        debugText.append("\n")

        debugText.append("Weather setting: ")
        if (Pref.Value.mPrefs.getBoolean("labs_show_weather", false)) {
            if (Pref.Value.mPrefs.getBoolean("labs_show_weather_from_upgrade", false))
                debugText.append("Upgrade\n")
            else
                debugText.append("Labs\n")
        } else {
            debugText.append("Disabled\n")
        }
        val cn = HomeReset.getDefaultHome(pm)
        debugText.append("Default launcher: ").append(if (cn == null) cn else cn.flattenToShortString()).append('\n')

        debugText.append("Rooted: ").append(RewtHelper.hasSuperuserBin())
        @SuppressLint("SdCardPath")
        val xposed = FileSlurp.slurp("/data/data/de.robv.android.xposed.installer/conf/modules.list")
        if (!TextUtils.isEmpty(xposed))
            debugText.append("Xposed: ").append(xposed).append("\n")

        debugText.append('\n')
        debugText.append("TeslaUnread: ")
        try {
            val e = context.packageManager.getPackageInfo(Pref.Key.UnreadCountProvider.TESLA.packageName, 0)

            debugText.append(e.versionName).append(' ').append(e.versionCode).append("\n")
        } catch (ex: PackageManager.NameNotFoundException) {
            debugText.append("Not installed\n")
        }

        debugText.append("Sesame: ")
        try {
            val e = context.packageManager.getPackageInfo(PackageNames.SESAME, 0)
            debugText.append(e.versionName).append(' ').append(e.versionCode).append(' ')
        } catch (ex: PackageManager.NameNotFoundException) {
            debugText.append("Not installed").append(' ')
        }

        debugText.append(Pref3.sesame_app_search.value).append(' ')
                .append(Pref3.sesame_shortcut_menu.value).append('\n')
        debugText.append("Dynamic Badges: ")
        if (Build.VERSION.SDK_INT < Build.VERSION_CODES.JELLY_BEAN_MR2) {
            debugText.append("NO (Android version too old)\n")
        } else {
            debugText.append(Pref.Value.unread_count).append(' ')
            debugText.append(Pref.Value.unread_count_provider.name)
            debugText.append(' ')
            val nl = NotificationListener.getInstance()
            debugText.append(nl).append('\n')
            if (Pref.Value.unread_count && Pref.Value.unread_count_provider.isCountless && nl != null) {
                nl.debugCurrentNotifications(activity, debugText)
                val co = CountlessObserver.getInstance()
                co.debugCurrentBadges(activity, debugText)
                Pref.Value.countless_blacklist.append(debugText.append("Badge Blacklist: "))
            }
        }

        //                        debugText.append("\n\nInstalled Apps:\n");
        //                        List<PackageInfo> packages = getActivity().getApplicationContext().getPackageManager().getInstalledPackages(0);
        //                        for (int i = 0; i < packages.size(); i++) {
        //                            debugText.append(packages.get(i).packageName).append(' ').append(packages.get(i).versionName).append(' ').append(packages.get(i).versionCode).append('\n');
        //                        }

        debugText.append("\n\nCurrent preferences:\n")
        Pref.dumpForDebug(PreferenceManager.getDefaultSharedPreferences(context), debugText)
        debugText.append("\n\n")

        val tree = LauncherAppState.getInMemoryTree()

        tree.append(debugText)

        var fcLog: String? = null
        try {
            val crashlog = context.getFileStreamPath("crashlog.txt")
            val time = System.currentTimeMillis()
            if (crashlog.lastModified() < time && time - DateUtils.DAY_IN_MILLIS < crashlog.lastModified()) {
                fcLog = FileSlurp.slurp(crashlog.absolutePath)
            }
        } catch (ex: Throwable) {

        }

        if (!TextUtils.isEmpty(fcLog))
            debugText.append("\n\n\n    ----\n").append(fcLog).append("\n")


        val supportDetailsTxt = File(context.cacheDir, "fileprovider_temp/supportDetails.txt")
        supportDetailsTxt.parentFile.mkdirs()
        try {
            val writer = BufferedWriter(FileWriter(supportDetailsTxt))
            writer.write(debugText.toString())
            writer.close()
        } catch (e1: IOException) {
            e1.printStackTrace()
        }

        val uri = FileProvider.getUriForFile(context, ProviderConfig.FILEPROVIDER_AUTHORITY, supportDetailsTxt)

        val emailIntent = Intent(Intent.ACTION_SEND)
        emailIntent.type = "message/rfc822"
        emailIntent.putExtra(Intent.EXTRA_EMAIL, arrayOf<String>(getSupportEmail()))
        emailIntent.putExtra(Intent.EXTRA_SUBJECT, "Nova Launcher " + (if (Pref.Value.nova_prime_enabled) "Prime " else "") + Version.getVersionString(context) + ": " + subject.text)
        emailIntent.putExtra(Intent.EXTRA_TEXT, body.text.toString())
        emailIntent.putExtra(Intent.EXTRA_STREAM, uri)

        val emailSendToIntent = Intent(Intent.ACTION_SENDTO)
        emailSendToIntent.data = Uri.parse("mailto:support@teslacoilsw.com")


        val emailApps = getEmailApps(pm)
        when {
            emailApps.size == 1 -> startEmailIntent(context, ComponentName(emailApps[0].activityInfo.packageName, emailApps[0].activityInfo.name), uri)
            emailApps.isNotEmpty() -> {
                val view = MatchWrapLinearLayout(context)
                view.orientation = LinearLayout.VERTICAL

                val sv = ScrollView(context).apply {
                    addView(view)
                    val dp8 = context.dp(8)
                    updatePadding(left = dp8, right = dp8)
                }
                val builder = MaterialDialog.Builder(context)
                        .title(R.string.contact_support_send_email_using)
                        .customView(sv, false)

                if (TintCompat.getThemeAttrColor(activity, android.R.attr.windowBackground) == Color.WHITE)
                    builder.backgroundColor(0xFFF0F0F0.toInt())

                val emailAppDialog = builder.build()
                emailApps.map {
                    val row = FancyPrefView<Unit>(context)
                    val iconSize = context.dp(48)
                    row.icon = FixedSizeDrawableWrapper(tryCatch(ColorDrawable(Color.TRANSPARENT)) { it.loadIcon(pm) }, iconSize, iconSize)
                    row.title = tryCatch (it.activityInfo.packageName) { it.loadLabel(pm) }
                    row.iconView?.elevation = context.dp(8).toFloat()
                    row.setOnClickListener { _ ->
                        startEmailIntent(context, ComponentName(it.activityInfo.packageName, it.activityInfo.name), uri)
                        emailAppDialog.dismiss()
                    }
                    view.addView(row)
                }
                emailAppDialog.show()

            }
            else -> showNoEmailAppToast(context, body.text.toString())
        }

        if (BuildConfig.DEBUG)
            Timber.v("Debug Log\n\n%s", debugText)
        dismiss()

    }


    private fun getSupportEmail(): String {
        val p = Partner.get(activity)
        if (p != null) {
            val email = p.supportEmail
            if (!TextUtils.isEmpty(email))
                return email
        }
        return if (BuildConfig.DEBUG) "kevin@teslacoilsw.com" else "support@teslacoilsw.com"
    }

    private fun getEmailApps(pm: PackageManager): List<ResolveInfo> {
        val emailSendToIntent = Intent(Intent.ACTION_SENDTO)
        emailSendToIntent.data = Uri.parse("mailto:support@teslacoilsw.com")
        val sendTos = pm.queryIntentActivities(emailSendToIntent, 0)?.map { it.activityInfo.packageName }?.toSet() ?: emptySet()

        return if (sendTos.isEmpty()) {
            /* Bah */
            emptyList()
        } else {
            val emailIntent = Intent(Intent.ACTION_SEND)
            emailIntent.type = "message/rfc822"

            pm.queryIntentActivities(emailIntent, 0)?.filter { sendTos.contains(it.activityInfo.packageName) } ?: emptyList()
        }

    }

    private fun startEmailIntent(context: Context, cn: ComponentName, uri: Uri) {
        val emailIntent = Intent(Intent.ACTION_SEND)
        emailIntent.type = "message/rfc822"
        emailIntent.putExtra(Intent.EXTRA_EMAIL, arrayOf<String>(getSupportEmail()))
        emailIntent.putExtra(Intent.EXTRA_SUBJECT, "Nova Launcher " + (if (Pref.Value.nova_prime_enabled) "Prime " else "") + Version.getVersionString(context) + ": " + subject.text)
        emailIntent.putExtra(Intent.EXTRA_TEXT, body.text.toString())
        emailIntent.putExtra(Intent.EXTRA_STREAM, uri)
        emailIntent.component = cn
        try {
            context.startActivity(Intent.createChooser(emailIntent, "Email"))
            true
        } catch (ex: Exception) {
            showNoEmailAppToast(context, body.text.toString());
        }
    }

    private fun showNoEmailAppToast(context: Context, body: String) {
        val clip = context.getSystemService(Context.CLIPBOARD_SERVICE) as ClipboardManager
        clip.text = body
        Toast.makeText(context, "No activity found for email. Bug report copied to clipboard.", Toast.LENGTH_LONG).show()
        Timber.tag("BugReport").w(body)
    }


}