package com.teslacoilsw.launcher.preferences

import android.appwidget.AppWidgetHost
import android.content.ContentValues
import android.content.Context
import android.content.SharedPreferences
import android.content.pm.PackageManager
import android.graphics.Bitmap
import android.graphics.BitmapFactory
import android.graphics.Color
import android.graphics.drawable.BitmapDrawable
import android.os.Build
import android.util.DisplayMetrics
import androidx.annotation.RequiresApi
import androidx.core.graphics.alpha
import com.android.launcher3.AutoInstallsLayout
import com.android.launcher3.IconCache
import com.android.launcher3.Launcher
import com.android.launcher3.LauncherAppState
import com.android.launcher3.LauncherModel
import com.android.launcher3.LauncherSettings
import com.teslacoilsw.cofs.flattenToString
import com.teslacoilsw.cofs.internal.parse
import com.teslacoilsw.kotlin.withAlpha
import com.teslacoilsw.launcher.CellSpecConfig
import com.teslacoilsw.launcher.NewAppShortcutReceiver
import com.teslacoilsw.launcher.NovaAction
import com.teslacoilsw.launcher.R
import com.teslacoilsw.launcher.drawer.WidgetPageFactory
import com.teslacoilsw.launcher.icon.AdaptiveIconShape
import com.teslacoilsw.launcher.preferences.Pref.Key
import com.teslacoilsw.launcher.preferences.Pref.Key.DesktopMargin
import com.teslacoilsw.launcher.quicksearchbar.SearchBarStyle
import com.teslacoilsw.launcher.screenoff.ScreenOff
import com.teslacoilsw.launcher.util.NamedIntent
import com.teslacoilsw.launcher.util.OEMUtils
import com.teslacoilsw.launcher.weather.WeatherUnits
import com.teslacoilsw.launcher.widget.TypefaceHelper
import com.teslacoilsw.shared.colorpicker.ColorUtil
import com.teslacoilsw.shared.util.DeviceSpecifics
import com.teslacoilsw.shared.util.FileSlurp
import com.teslacoilsw.shared.util.Log
import com.teslacoilsw.shared.util.MD5
import com.teslacoilsw.shared.util.MathUtils
import com.teslacoilsw.shared.util.Version
import com.teslacoilsw.shared.util.rewthelper.RewtHelper
import com.teslacoilsw.tesladirect.MarketHandler
import com.teslacoilsw.tesladirect.TeslaDirectPrefs
import com.teslacoilsw.tesladirect.TeslaDirectPrefs.UpdateMethod
import timber.log.Timber
import java.io.File
import java.io.FileOutputStream
import java.math.BigInteger
import java.security.MessageDigest
import java.security.NoSuchAlgorithmException
import java.util.*

private val FIRST_RUN = 0

/* Returns list of runnables to execute after initing */
fun initOrUpgrade(context: Context, prefs: SharedPreferences): List<() -> Unit> {
    val runAfter = mutableListOf<() -> Unit>()

    val version = Version.getVersionCode(context)
    if (version == 0) {
        /* This can happen if we load while updating. Just get out of here */
        Log.e("Nova", "Attempting to load while updating")
        System.exit(0)
    }
    val fromVersion = prefs.getInt("preference_version", 0)

    /* Verify the device hasn't changed out from under us (oh the joys of roms) */
    var needCacheInvalidate = false
    val densityDpi = context.resources.displayMetrics.densityDpi
    if (densityDpi != prefs.getInt(Key.SYSTEM_DPI, -1)) {
        needCacheInvalidate = true
    }

    var buildpropdigest = ""
    if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.N_MR1 && OEMUtils.isAtLeast(Build.VERSION_CODES.N_MR1, 1)) {
        buildpropdigest = generateODigest(context)
    } else {
        val buildprop = File("/system/build.prop")
        if (buildprop.exists()) {
            buildpropdigest = MD5.digestFromFile(buildprop)
        }
    }
    if (prefs.getString(Key.SYSTEM_BUILD_PROP, "") != buildpropdigest) {
        needCacheInvalidate = true
    }

    if (needCacheInvalidate && fromVersion != FIRST_RUN) {
        Timber.w("Something changed, need to invalidate icon cache")
        val editor = prefs.edit()
        editor.putInt(Key.SYSTEM_DPI, densityDpi)
        editor.putString(Key.SYSTEM_BUILD_PROP, buildpropdigest)
        editor.apply()
        runAfter.add { LauncherModel.invalidateCachedIcons(context) }
        FileSlurp.deleteDirectory(WidgetPageFactory.getWidgetPreviewCacheDir(context))
        RewtHelper.getInstance(context).rewtHelperFile.delete()
        //            FirebaseAnalytics.getInstance(context).setAnalyticsCollectionEnabled(prefs.getBoolean(Key.ANALYTICS_ENABLED, true));
    }

    if (version == fromVersion)
        return runAfter

    val editor = prefs.edit()

    if (fromVersion == FIRST_RUN) {
        editor.putInt(Key.SYSTEM_DPI, densityDpi)
        editor.putString(Key.SYSTEM_BUILD_PROP, buildpropdigest)
        val isTablet = context.resources.getBoolean(R.bool.is_tablet)
        val isTabletXL = context.resources.getBoolean(R.bool.is_large_tablet)

        editor.putInt("first_install_version", version)

        val res = context.resources
        val grid = getDefaultGridSize(res.displayMetrics)
        val cellCountX = grid[0]
        val cellCountY = grid[1]

        if (isTablet) {
            editor.putInt(Pref.Key.DOCK_GRID_COLS, 7)
        }
        editor.putInt(Pref.Key.DESKTOP_GRID_COLS, cellCountX)
        editor.putInt(Pref.Key.DESKTOP_GRID_ROWS, cellCountY)

        setDefaultDrawerGrid(context, editor)

        editor.putString(Pref.Key.SCREEN_ORIENTATION, if (isTablet) Pref.Key.ScreenOrientation.UNSPECIFIED.name else Pref.Key.ScreenOrientation.NOSENSOR.name)
        if (!isTabletXL) {
            val margin = DesktopMargin.NONE
            editor.putString(Pref.Key.DOCK_WIDTH_MARGIN, margin.name)
            editor.putString(Pref.Key.DESKTOP_WIDTH_MARGIN, DesktopMargin.NONE.name)
        } else {
            editor.putString(Key.DESKTOP_WIDTH_MARGIN, DesktopMargin.AUTO.name)
            editor.putString(Key.DESKTOP_HEIGHT_MARGIN, DesktopMargin.AUTO.name)
        }

        editor.putString(Pref.Key.GESTURE_PINCH, NamedIntent(res.getString(R.string.nova_action_show_previews), NovaAction.SHOW_PREVIEWS.intent).toUri())
        editor.putString(Pref.Key.GESTURE_PINCH_OUT, NamedIntent(res.getString(R.string.nova_action_show_recent_apps), NovaAction.SHOW_RECENT_APPS.intent).toUri())
        editor.putString(Pref.Key.GESTURE_SWIPE_UP, null)
        editor.putString(Pref.Key.GESTURE_SWIPE_DOWN, NamedIntent(res.getString(R.string.nova_action_expand_notification_bar), NovaAction.EXPAND_STATUS_BAR.intent).toUri())
        editor.putString(Pref.Key.GESTURE_TWO_FINGER_SCROLL_DOWN, NamedIntent(res.getString(R.string.nova_action_app_drawer), NovaAction.APP_DRAWER.intent).toUri())
        editor.putString(Pref.Key.GESTURE_TWO_FINGER_SCROLL_UP, NamedIntent(res.getString(R.string.nova_settings), NovaAction.NOVA_SETTINGS.intent).toUri())
        // editor.putString(Pref.Key.GESTURE_DOUBLE_TAP, null);

        editor.putString(Pref.Key.HOME_BUTTON_INTENT, NamedIntent(res.getString(R.string.nova_action_app_search), NovaAction.APP_SEARCH.intent).toUri())
        editor.putString("long_menu_intent", NamedIntent(res.getString(R.string.nova_action_text_search), NovaAction.TEXT_SEARCH.intent).toUri())
        if (MarketHandler.isBeta(context)) {
            val p = TeslaDirectPrefs.getPrefs(context)
            p.method = UpdateMethod.BETA
        }

        if (DeviceSpecifics.SUPPORTS_KITKAT) {
            editor.putBoolean(Key.NOTIFICATION_BAR_ROM_TRANSPARENT, true)
        }

        editor.putInt("drawer_transparency", 128)

        if (context.resources.displayMetrics.densityDpi <= DisplayMetrics.DENSITY_XHIGH) {
            for (prefix in arrayOf("desktop", "dock", "drawer", "folder")) {
                editor.putString(Pref.Key.getIconSpecKey(prefix, Key.ICONSPEC_LABEL_FONT_FAMILY), "")
            }
        }

        if (LauncherAppState.isScreenLarge(res))
            editor.putInt(Pref.Key.getIconSpecKey("dock", Key.ICONSPEC_ICON_SIZE), DOCK_SMALL_ICON_SCALE)

        setDefaultWeatherUnits(context, editor)

        if (LauncherAppState.isSystemApp(context)) {
            editor.putBoolean("did_quickstart", true)
            /* We're a system app, check if we have an OEM config */
            val info = AutoInstallsLayout.populateDefaultPreferences(context, editor)
            if (info != null) {
                /* Save this and if the partner apk changes out from under us before first run
                 * of the launcher, re-init.
                 * This happens during auto-installs if somehow our process is started too
                 * early, such as a receiver being called */
                editor.putString("firstrun_oem_partner_apkinfo", info.key)
            }
        }

    } else if (fromVersion < version) {
        onUpgrade(context, prefs, editor, fromVersion, runAfter)
    } else {
        onDowngrade(context, prefs, editor, fromVersion)
    }

    editor.putLong("0", System.currentTimeMillis())

    editor.putInt("preference_version", version)

    editor.apply()

    return runAfter
}

private fun setDefaultDrawerGrid(context: Context, editor: SharedPreferences.Editor) {
    val drawerGrid = calculateDefaultDrawerGrid(context)
    editor.putInt(Pref.Key.DRAWER_APP_GRID_COLS, drawerGrid[0])
    editor.putInt(Pref.Key.DRAWER_APP_GRID_ROWS, drawerGrid[1])
    editor.putInt(Pref.Key.DRAWER_APP_GRID_LAND_COLS, drawerGrid[2])
    editor.putInt(Pref.Key.DRAWER_APP_GRID_LAND_ROWS, drawerGrid[3])
}

private fun calculateDefaultDrawerGrid(context: Context): IntArray {
    val grid = getDefaultGridSize(context.resources.displayMetrics)

    return intArrayOf(grid[0], grid[1], grid[1], grid[0])
}

private fun onDowngrade(context: Context, prefs: SharedPreferences, editor: SharedPreferences.Editor, fromVersion: Int) {
}

private fun onUpgrade(context: Context, prefs: SharedPreferences, editor: SharedPreferences.Editor, fromVersion: Int, runAfter: MutableList<() -> Unit>) {
    val pm = context.applicationContext.packageManager

    if (fromVersion < 30101) {
        if ("lge" == Build.BRAND && "g3" == Build.DEVICE) {
            runAfter.add { LauncherModel.deleteCachedIcons(context) }
        }
    }

    if (fromVersion < 30201) {
        val menuActionSet = HashSet<String>()
        val action = prefs.getString(Key.DRAWER_MENU_ACTION, Key.DrawerMenuAction.MENU.name)
        if ("NONE" != action) {
            menuActionSet.add(action)
        }

        editor.putStringSet(Key.DRAWER_MENU_ACTION_SET, menuActionSet)
    }

    if (fromVersion < 32004 && DeviceSpecifics.SUPPORTS_LOLLIPOP) {
        RewtHelper.getInstance(context).rewtHelperFile.delete()
    }

    if (fromVersion < 39000) {
        if ("SMALL" != prefs.getString(Key.DRAWER_TAB_STYLE, Key.TabStyle.COLORBLOCK.name)) {
            editor.putBoolean("drawer_enable_tabs", true)
        } else {
            editor.putString(Key.DRAWER_TAB_STYLE, Key.TabStyle.COLORBLOCK.name)
        }

        val font = prefs.getString("desktop_icon_font", "DEFAULT")
        var size = prefs.getInt("app_icon_size", -1)
        for (prefix in arrayOf("desktop", "dock", "drawer", "folder")) {
            if ("DEFAULT" != font)
                editor.putString(Pref.Key.getIconSpecKey(prefix, Key.ICONSPEC_LABEL_FONT_FAMILY), font)
            if (size != -1)
                editor.putInt(Pref.Key.getIconSpecKey(prefix, Key.ICONSPEC_ICON_SIZE), size)
        }
        var dockSmallIcons = false
        if (prefs.getBoolean("dock_use_small_icons", false)) {
            dockSmallIcons = true
            if (size == -1)
                size = 100
            editor.putInt(Pref.Key.getIconSpecKey("dock", Key.ICONSPEC_ICON_SIZE), 5 * (.5f + size * .8f / 5f).toInt())
        }

        val show = prefs.getBoolean("dock_show_label", false)
        if (show) {
            editor.putBoolean(Pref.Key.getIconSpecKey("dock", Key.ICONSPEC_SHOW_LABELS), true)
            if (dockSmallIcons)
                editor.putString(Pref.Key.getIconSpecKey("dock", Key.ICONSPEC_LABEL_SIZE), "SMALL")
        }

        if (!prefs.contains("drawer_cards")) {
            editor.putBoolean("drawer_cards", false)
            if (!prefs.contains(Pref.Key.getIconSpecKey("drawer", Key.ICONSPEC_LABEL_COLOR))) {
                editor.putInt(Pref.Key.getIconSpecKey("drawer", Key.ICONSPEC_LABEL_COLOR), Color.WHITE)
            }
        }
    }

    if (fromVersion > 39000 && fromVersion < 39004) {
        val cr = context.contentResolver
        val cv = ContentValues()
        cv.put(LauncherSettings.Favorites.PROFILE_ID, -1L)
        cr.update(LauncherSettings.Favorites.CONTENT_URI, cv, "profileId = 0", null)
    }

    if (fromVersion < 39006) {
        var show = prefs.getBoolean("favorites_show_label", true)
        if (!show) {
            editor.putBoolean(Pref.Key.getIconSpecKey("desktop", Key.ICONSPEC_SHOW_LABELS), false)
        }

        show = prefs.getBoolean("folder_favorites_show_label", true)
        if (!show) {
            editor.putBoolean(Pref.Key.getIconSpecKey("folder", Key.ICONSPEC_SHOW_LABELS), false)
        }

        show = prefs.getBoolean("folder_favorites_show_label", true)
        if (!show) {
            editor.putBoolean(Pref.Key.getIconSpecKey("folder", Key.ICONSPEC_SHOW_LABELS), false)
        }

    }

    if (fromVersion < 39008) {
        if (prefs.getBoolean("auto_add_shortcuts", false)) {
            pm.setComponentEnabledSetting(NewAppShortcutReceiver.COMPONENT_NAME, PackageManager.COMPONENT_ENABLED_STATE_ENABLED, PackageManager.DONT_KILL_APP)
        }
    }

    if (fromVersion < 39013) {
        if (Key.ScrollIndicatorStyle.LINE.name == prefs.getString(Key.DESKTOP_SCROLL_INDICATOR, Key.ScrollIndicatorStyle.CIRCLE.name)) {
            editor.putInt(Key.DESKTOP_SCROLL_INDICATOR_COLOR, prefs.getInt(Key.THEME_HOLO_COLOR, Color.WHITE))
        }
    }

    if (fromVersion > 39000 && fromVersion < 39024) {
        for (key in arrayOf("drawer", "dock", "desktop", "folder")) {
            val oldSize = prefs.getString(Pref.Key.getIconSpecKey(key, Key.ICONSPEC_LABEL_SIZE), "MEDIUM")
            var sizeSp = 13f
            if ("SMALL" == oldSize) {
                sizeSp = 11f
            } else if ("MEDIUM" == oldSize) {
                sizeSp = 14.4f
            } else if ("LARGE" == oldSize) {
                sizeSp = 18f
            } else if ("XLARGE" == oldSize) {
                sizeSp = 21f
            }
            editor.putFloat(Pref.Key.getIconSpecKey(key, Key.ICONSPEC_LABEL_SIZE_SP), sizeSp)
        }
    }

    if (fromVersion <= 39024) {
        File(context.cacheDir, "widgetpreviews.db").delete()
        File(context.cacheDir, "widgetpreviews.db-journal").delete()
        FileSlurp.deleteDirectory(WidgetPageFactory.getWidgetPreviewCacheDir(context))
    }

    if (fromVersion <= 39027) {
        FileSlurp.deleteDirectory(WidgetPageFactory.getWidgetPreviewCacheDir(context))
    }

    if (fromVersion <= 39028) {
        if (prefs.getBoolean(Pref3.drawer_enable_tabs.key, false) && !prefs.contains(Key.DRAWER_TAB_STYLE))
            editor.putString(Key.DRAWER_TAB_STYLE, Key.TabStyle.COLORBLOCK.name)
    }

    if (fromVersion <= 40200) {
        editor.putBoolean("sanify_icon_size", false)
    }

    if (fromVersion <= 41000) {
        if (!prefs.contains(Key.DRAWER_PREDICTIVE_APPS))
            editor.putBoolean(Key.DRAWER_PREDICTIVE_APPS, false)
        if (!prefs.contains(Key.DRAWER_STYLE))
            editor.putString(Key.DRAWER_STYLE, Key.DrawerStyle.HORIZONTAL_PAGINATED.name)
        editor.remove("search_overlay")
    }

    if (fromVersion <= 42200) {
        val appWidgetId = prefs.getInt("googleSearchAppWidgetId", -1)
        if (appWidgetId != -1) {
            try {
                val host = AppWidgetHost(context, Launcher.APPWIDGET_HOST_ID)
                host.deleteAppWidgetId(appWidgetId)
            } catch (ex: Throwable) {
                ex.printStackTrace()
            }

        }
        editor.remove("googleSearchAppWidgetId")
    }

    if (fromVersion <= 42900) {
        var oldPersistentQsb = prefs.getString("qsb_style", "NONE")
        var qsb_enabled = true
        if ("NONE" == oldPersistentQsb) {
            qsb_enabled = false
            oldPersistentQsb = prefs.getString("novawidget_qsb_style", "MARSHMALLOW_GRAY")
        }

        val newStyle: String
        when (oldPersistentQsb) {
            "MARSHMALLOW" -> newStyle = "M_COLOR:ROUNDRECT:#FFFFFFFF:56"
            "L" -> newStyle = "CLASSIC_MONO:ROUNDRECT:#FFFFFFFF:56"
            "JELLYBEAN", "KITKAT" -> newStyle = "CLASSIC_MONO:ROUNDRECT:#64FFFFFF:56"
            "ICS" -> newStyle = "CLASSIC_MONO:WIRE:#FFFFFFFF:56"
            "HONEYCOMB" -> newStyle = "CLASSIC_MONO:HONEYCOMB:#FFFFFFFF:56"
            "MARSHMALLOW_GRAY" -> newStyle = SearchBarStyle.DEFAULT.flattenToString()
            else -> newStyle = SearchBarStyle.DEFAULT.flattenToString()
        }
        editor.putString("qsb_style", newStyle)
        editor.putBoolean(Key.QSB_ENABLED, qsb_enabled)
    }

    if (fromVersion < 49901) {
        if (prefs.getBoolean(Pref.Key.ROOT_HELPER_SCREEN_OFF, false)) {
            editor.putString("screen_off_mode", ScreenOff.ROOT.name)
        } else if (ScreenOff.DEVICE_ADMIN.hasPermission(context)) {
            editor.putString("screen_off_mode", ScreenOff.DEVICE_ADMIN.name)
        }

        if ("DEFAULT" == prefs.getString("night_mode", "DEFAULT")) {
            editor.putString("night_mode", NightMode.NEVER.name)
        }

        val dockFile = context.getFileStreamPath(Key.DockBackgroundPrefs.DOCK_BG_FILENAME)
        val opts = BitmapFactory.Options()
        val bmp = BitmapFactory.decodeFile(context.getFileStreamPath(Key.DockBackgroundPrefs.DOCK_BG_FILENAME).absolutePath, opts)
        if (bmp != null && bmp.width <= 16 && bmp.height <= 16) {
            /* Nova prior to 5.0 used a small white bitmap in repeating mode with a colorfilter
               instead of just drawing a color directly. Kill it.
             */
            dockFile.delete()
        }

        if (!prefs.getBoolean("drawer_cards", true) && !prefs.contains("drawer_background_color"))
            editor.putInt("drawer_background_color", -0x73f7f7f8)
    }

    if (fromVersion <= 49907) {
        setDefaultWeatherUnits(context, editor)
    }

    if (fromVersion <= 49908) {
        val style = SearchBarStyle.fromString(prefs.getString("qsb_style", ""))
        if (style.hasWeather()) {
            editor.putBoolean("labs_show_weather", true)
            editor.putBoolean("labs_show_weather_from_upgrade", true)
        }

        if (prefs.getBoolean("drawer_swipe_from_dock", false) && !prefs.contains("desktop_indicator_drawerdock_swipe"))
            editor.putBoolean("desktop_indicator_drawerdock_swipe", false)
    }

    if (fromVersion <= 49911) {
        if (!prefs.contains(Pref3.drawer_searchbar.key))
            editor.putBoolean(Pref3.drawer_searchbar.key, false)
        editor.putBoolean("did_quickstart", true)
    }

    if (fromVersion <= 50903) {
        if (prefs.getInt(Key.DRAWER_ICON_SIZE, 100) > prefs.getInt(Key.DESKTOP_ICON_SIZE, 100))
            runAfter.add { LauncherModel.deleteCachedIcons(context) }
    }

    if (fromVersion <= 50905) {
        if (IconCache.checkHasSystemTheme(context)) {
            runAfter.add { LauncherModel.deleteCachedIcons(context) }
        }
    }

    if (fromVersion <= 51000) {
        if (prefs.getBoolean(Key.UNREAD_COUNT, false) && !prefs.contains(Key.UNREAD_COUNT_PROVIDER)) {
            editor.putString(Key.UNREAD_COUNT_PROVIDER, Key.UnreadCountProvider.TESLA.name)
        }
    }

    if (fromVersion <= 53004) {
        editor.putInt("appsearch_background", ColorUtil.argb(0xFF, prefs.getInt("drawer_card_color", Color.WHITE)))
    }

    if (fromVersion <= 54000) {
        editor.putInt("appsearch_background", ColorUtil.argb(0xFF, prefs.getInt("appsearch_background", Color.WHITE)))
    }

    if (fromVersion == 54901) {
        if (prefs.getBoolean("adaptive_icon_backport", false) || Build.VERSION.SDK_INT >= Build.VERSION_CODES.O) {
            if (prefs.getBoolean("adaptive_icon_plate", false)) {
                runAfter.add { LauncherModel.deleteCachedIcons(context) }
            }
        }
    }

    if (fromVersion < 54904 || fromVersion >= 55900 && fromVersion < 56000) {
        if (!prefs.contains("qsb_style"))
            editor.putString("qsb_style", SearchBarStyle.OLD_DEFAULT.flattenToString())
        runAfter.add { LauncherModel.invalidateCachedIcons(context) }
        editor.putBoolean("show_drop_targets", true)
        if (!(Build.VERSION.SDK_INT >= Build.VERSION_CODES.LOLLIPOP && Pref.Value.unread_count && Pref.Value.unread_count_provider.isCountless))
            editor.putString("popup_menu_style", PopupMenuStyle.NOUGAT.name)

        if (!prefs.contains("folder_background"))
            editor.putString("folder_background", Key.FolderBg.CIRCLE_WHITE.name)
        if (!prefs.contains("folder_preview"))
            editor.putString("folder_preview", Key.FolderPreview.LINE.name)
    }

    if (fromVersion <= 54905) {
        runAfter.add { LauncherModel.invalidateCachedIcons(context) }
    }

    if (fromVersion == 55900 && !prefs.contains("workspace_screen_count")) {
        /* Upgrade from Nova 5.4.7-razer without the user having seen the launcher yet,
         * Let's incorrectly seed the partner info to force a reinit */
        editor.putString("firstrun_oem_partner_apkinfo", "com.teslacoilsw.launcher:55900")
    }

    if (fromVersion in 54901..54906) {
        editor.putBoolean(Key.WIDGET_CLEANUP_IDS, true)
    }

    if (fromVersion <= 59400) {
        if (!prefs.contains("drawer_swipe_from_dock"))
            editor.putBoolean("drawer_swipe_from_dock", false)
        if (!prefs.contains("desktop_scroll_indicator"))
            editor.putString("desktop_scroll_indicator", Key.ScrollIndicatorStyle.CIRCLE.name)

        editor.putString(Pref3.drawer_cellspecs.key, CellSpecConfig.flattenToString(cellSpecConfigFromLegacyPrefs(prefs, "drawer")))
        editor.putString(Pref3.desktop_cellspecs.key, CellSpecConfig.flattenToString(cellSpecConfigFromLegacyPrefs(prefs, "desktop")))
        editor.putString(Pref3.dock_cellspecs.key, CellSpecConfig.flattenToString(cellSpecConfigFromLegacyPrefs(prefs, "dock")))
        editor.putString(Pref3.folder_cellspecs.key, CellSpecConfig.flattenToString(cellSpecConfigFromLegacyPrefs(prefs, "folder")))

        if (prefs.getBoolean("search_overlay", false))
            editor.putString("search_window_style", SearchBarStyle.SearchWindowStyle.GOOGLE_OVERLAY.name)

        var newValue = SearchBarPlacement.WIDGET
        if (prefs.getBoolean("qsb_enabled", false))
            newValue = SearchBarPlacement.PERSISTENT
        else {
            val dockQsb = prefs.getString("dock_qsb_pos", "NONE")
            if (dockQsb == "TOP") {
                newValue = SearchBarPlacement.DOCK_ABOVE_ICONS
            } else if (dockQsb == "BOTTOM") {
                newValue = SearchBarPlacement.DOCK_BELOW_ICONS
            }
        }
        editor.putString(Pref3.searchbar_placement.key, newValue.name)

        if (prefs.getString(Pref3.wallpaper_scroll_mode.key, "ON") == "MANUAL") {
            editor.putString(Pref3.wallpaper_scroll_mode.key, "ON")
            editor.putBoolean(Pref3.wallpaper_scrolling_force.key, true)
        }

        if (Build.VERSION.SDK_INT < Build.VERSION_CODES.O) {
            editor.putBoolean(Pref.Key.AUTO_ADD_SHORTCUTS, pm.getComponentEnabledSetting(NewAppShortcutReceiver.COMPONENT_NAME) == PackageManager.COMPONENT_ENABLED_STATE_ENABLED)
        }

        if (!prefs.getBoolean("drawer_join_tabs", true))
            editor.putBoolean("drawer_isolate_tabs", true)

        if (prefs.getString("drawer_fastscroll_side", "END") == "NONE")
            editor.putBoolean("drawer_fastscroll", false)

        val folderWindowColor = prefs.getInt("folder_window_color", Color.WHITE)
        val config = FolderWindowConfig(FolderWindowStyle.WINDOW, folderWindowColor.alpha, folderWindowColor.withAlpha(0xff), 2, true,  true, 4, 4, FolderScrollDirection.VERTICAL)
        editor.putString(Pref3.folder_window_config.key, FolderWindowConfig.flattenToString(config))

        val folderIconPreview = prefs.getString("folder_preview", Key.FolderPreview.GRID.name).parse(Key.FolderPreview.GRID)
        val folderIconBg = prefs.getString("folder_background", Key.FolderBg.PEEPHOLE.name).parse(Key.FolderBg.PEEPHOLE)
        val folderShape: AdaptiveIconShape
        val folderIconColor: Int
        val folderIconAlpha: Int
        when (folderIconBg) {
            Key.FolderBg.PEEPHOLE -> {
                folderShape = AdaptiveIconShape.CIRCLE
                folderIconColor = COLOR_AUTO
                folderIconAlpha = 160
            }
            Key.FolderBg.CIRCLE -> {
                folderShape = AdaptiveIconShape.CIRCLE
                folderIconColor = Color.BLACK
                folderIconAlpha = 160
            }
            Key.FolderBg.CIRCLE_WHITE -> {
                folderShape = AdaptiveIconShape.CIRCLE
                folderIconColor = Color.WHITE
                folderIconAlpha = 160
            }
            Key.FolderBg.PLATFORM -> {
                folderShape = AdaptiveIconShape.FOLDER_PLATFORM
                folderIconColor = Color.BLACK
                folderIconAlpha = 200
            }
            Key.FolderBg.DISABLED -> {
                folderShape = AdaptiveIconShape.CIRCLE
                folderIconColor = COLOR_AUTO
                folderIconAlpha = 0
            }
            Key.FolderBg.SQUARE -> {
                folderShape = AdaptiveIconShape.SQUARE
                folderIconColor = Color.BLACK
                folderIconAlpha = 200
            }
            Key.FolderBg.DISC -> {

                val outFile = File(Pref.Key.FolderBg.getCustomImagePath(context))
                try {
                    val bitmap = (context.getDrawable(R.drawable.portal_disc_inner_holo) as BitmapDrawable).bitmap
                    outFile.parentFile.mkdirs()
                    val outputStream = FileOutputStream(outFile)
                    outputStream.use {
                        bitmap.compress(Bitmap.CompressFormat.PNG, 100, outputStream)
                    }
                    Pref.Key.FolderBg.CUSTOM.resetImageDrawable()
                } catch (e: Exception) {
                    e.printStackTrace()
                }
                folderShape = AdaptiveIconShape.CUSTOM
                folderIconColor = COLOR_AUTO
                folderIconAlpha = 255
            }
            Key.FolderBg.CUSTOM -> {
                folderShape = AdaptiveIconShape.CUSTOM
                folderIconAlpha = 255
                folderIconColor = COLOR_AUTO
            }
        }
        val folderIconConfig = FolderIconConfig(folderShape, folderIconPreview, folderIconBg == Key.FolderBg.PEEPHOLE, folderIconColor, folderIconAlpha)
        editor.putString(Pref3.folder_icon_config.key, folderIconConfig.flattenToString())
    }

    if (fromVersion == 59901) {
        fun remigratePref(prefix: String) {
            val newKey = prefix + "_cellspecs"
            if (prefs.contains(newKey)) {
                val legacy = cellSpecConfigFromLegacyPrefs(prefs, prefix)
                val cur = CellSpecConfig.parse(prefs.getString(newKey, ""))
                if (cur.labelVisible && !legacy.labelVisible && cur == legacy.copy(labelVisible = true)) {
                    editor.putString(newKey, legacy.flattenToString())
                }
            }
        }

        remigratePref("drawer")
        remigratePref("desktop")
        remigratePref("dock")
        remigratePref("folder")
    }

}

private fun cellSpecConfigFromLegacyPrefs(prefs: SharedPreferences, prefix: String): CellSpecConfig =
        cellSpecsConfigFromLegacyPrefs(prefs.all, prefix)

/* Used by onboard backup preview */
public fun cellSpecsConfigFromLegacyPrefs(prefs: Map<String, *>, prefix: String): CellSpecConfig {
    val isDesktopOrDock = "desktop" == prefix || "dock" == prefix

    val iconTextVisible = prefs[prefix + "_" + Pref.Key.ICONSPEC_SHOW_LABELS] as? Boolean ?: (prefix != "dock")
    val iconTextColorDay = prefs[prefix + "_" + Pref.Key.ICONSPEC_LABEL_COLOR] as? Int ?: if (isDesktopOrDock) Color.WHITE else Color.BLACK

    val iconTextShadow = prefs[prefix + "_" + Pref.Key.ICONSPEC_LABEL_SHADOW] as? Boolean ?: isDesktopOrDock
    val iconFontFamily = prefs[prefix + "_" + Pref.Key.ICONSPEC_LABEL_FONT_FAMILY] as? String ?: TypefaceHelper.CONDENSED
    val iconTextSingleLine = prefs[prefix + "_" + Pref.Key.ICONSPEC_SINGLE_LINE] as? Boolean ?: true

    val iconScaleInt = prefs[prefix + "_" + Pref.Key.ICONSPEC_ICON_SIZE] as? Int ?: 100
    val rawIconScale = iconScaleInt / 100f
    val iconTextSizeSp = prefs[prefix + "_" + Pref.Key.ICONSPEC_LABEL_SIZE_SP] as? Float? ?: 13f

    return CellSpecConfig(rawIconScale, iconTextVisible, iconTextSizeSp, iconTextColorDay, iconTextShadow, iconFontFamily!!, iconTextSingleLine, false)
}

@RequiresApi(23)
private fun generateODigest(context: Context): String {
    val s = StringBuilder()
    s.append(Build.BOARD).append(Build.BRAND).append(Build.CPU_ABI).append(Build.DEVICE)
            .append(Build.DISPLAY).append(Build.HOST).append(Build.ID).append(Build.MANUFACTURER)
            .append(Build.MODEL).append(Build.PRODUCT).append(Build.TAGS).append(Build.TYPE)
            .append(Build.USER).append(Build.VERSION.BASE_OS).append(Build.VERSION.CODENAME)
            .append(Build.VERSION.INCREMENTAL).append(Build.VERSION.PREVIEW_SDK_INT)
            .append(Build.VERSION.SDK_INT).append(Build.VERSION.RELEASE)
            .append(Build.VERSION.SECURITY_PATCH)
            .append(context.resources.displayMetrics)
    try {
        val md5 = MessageDigest.getInstance("MD5")
        md5.update(s.toString().toByteArray())
        val hash = BigInteger(1, md5.digest())
        return hash.toString(16)
    } catch (e: NoSuchAlgorithmException) {
        throw RuntimeException(e)
    }

}

private fun setDefaultWeatherUnits(context: Context, editor: SharedPreferences.Editor) {
    val countryCode = context.resources.configuration.locale.country
    if (usesFahrenheit(countryCode))
        editor.putString("weather_units", WeatherUnits.FAHRENHEIT.name)
}

private fun usesFahrenheit(countryCode: String): Boolean {
    return when (countryCode) {
        "US", "PR" /* Puerto Rico */, "VI" /* US Virgin Islands */, "UM" /* US Outlying Islands */, "PW" /* Palau */, "FM" /* Micronesia */, "MH" /* Marshall Islands */, "BS" /* Bahamas */, "BZ" /* Belize */, "KY" /* Cayman Islands */ -> true
        else -> false
    }
}

fun getDefaultGridSize(dm: DisplayMetrics): IntArray {
    var cellCountY = 4
    var cellCountX = 4

    val width = MathUtils.roundInt(Math.min(dm.widthPixels, dm.heightPixels) / dm.density)
    val height = MathUtils.roundInt(Math.max(dm.widthPixels, dm.heightPixels) / dm.density)


    if (height >= 300)
        cellCountY = 3
    if (height >= 400)
        cellCountY = 4
    if (height >= 491)
        cellCountY = 5
    if (height >= 694)
        cellCountY = 6
    if (height >= 2527)
        cellCountY = 7

    if (width >= 255)
        cellCountX = 4
    if (width >= 406)
        cellCountX = 5
    if (width >= 575)
        cellCountX = 6
    if (width >= 1527)
        cellCountX = 7

    return intArrayOf(cellCountX, cellCountY)
}