package com.teslacoilsw.launcher.preferences

import android.content.SharedPreferences
import android.content.SharedPreferences.OnSharedPreferenceChangeListener
import android.graphics.Color
import android.os.Build
import androidx.annotation.ColorInt
import com.teslacoilsw.cofs.Cofs
import com.teslacoilsw.cofs.CofsBase
import com.teslacoilsw.cofs.ConfigObjectFromString
import com.teslacoilsw.cofs.getCofsDelegate
import com.teslacoilsw.cofs.internal.parse
import com.teslacoilsw.launcher.CellSpecConfig
import com.teslacoilsw.launcher.daynight.DayNightManager
import com.teslacoilsw.launcher.icon.AdaptiveIconShape
import com.teslacoilsw.launcher.icontheme.IconThemeConfig
import com.teslacoilsw.launcher.preferences.fancyprefs.PrefHolder
import com.teslacoilsw.launcher.preferences.fancyprefs.TransformPrefHolder
import com.teslacoilsw.launcher.quicksearchbar.SearchBarStyle
import com.teslacoilsw.launcher.util.MaterialColors
import com.teslacoilsw.launcher.util.NamedIntent
import com.teslacoilsw.launcher.util.OEMUtils
import com.teslacoilsw.launcher.util.StatusBar.HideClock
import com.teslacoilsw.shared.colorpicker.ColorUtil
import rx.Emitter
import rx.Observable
import rx.functions.Action1
import kotlin.properties.ReadOnlyProperty
import kotlin.reflect.KProperty


const val COLOR_AUTO = 0x00040302

@ConfigObjectFromString(paramsRequired = 5)
data class FolderIconConfig(val _shape: AdaptiveIconShape, val preview: Pref.Key.FolderPreview, val _clipToBounds: Boolean, @ColorInt @get:ColorInt val _iconColor: Int, val iconAlpha: Int = 255) {
    companion object : Cofs<FolderIconConfig>, CofsBase<FolderIconConfig> by FolderIconConfig::class.getCofsDelegate() {
        @JvmStatic
        override val DEFAULT = FolderIconConfig(AdaptiveIconShape.NONE, Pref.Key.FolderPreview.GRID, true, COLOR_AUTO, 160)
    }

    val background = Pref.Key.FolderBg.PEEPHOLE
    val shape: AdaptiveIconShape
        get() = if (_shape != AdaptiveIconShape.NONE) _shape else Pref3.adaptive_icon_shape.value
    val iconColor: Int @ColorInt
        get() = if (_iconColor == COLOR_AUTO) ColorUtil.argb(0xFF, Pref3.folder_window_config.value.backgroundColor) else ColorUtil.argb(0xFF, _iconColor)

    fun usePlatformStyle() = (background == Pref.Key.FolderBg.PEEPHOLE && shape == AdaptiveIconShape.FOLDER_PLATFORM) || background == Pref.Key.FolderBg.PLATFORM || background == Pref.Key.FolderBg.DISC
    fun canClipToBounds() = shape != AdaptiveIconShape.FOLDER_PLATFORM && shape != AdaptiveIconShape.CUSTOM
    fun clipToBounds() = canClipToBounds() && _clipToBounds
}

enum class FolderWindowStyle { WINDOW, IMMERSIVE }
enum class FolderScrollDirection { HORIZONTAL, VERTICAL }
enum class SearchBarPlacement {
    NONE, WIDGET, DOCK_ABOVE_ICONS, DOCK_BELOW_ICONS, PERSISTENT;

    val isDock: Boolean
        get() = this == DOCK_ABOVE_ICONS || this == DOCK_BELOW_ICONS

    val requireRestart: Boolean
        get() = this != SearchBarPlacement.NONE && this != SearchBarPlacement.WIDGET
}

enum class WallpaperScrollMode {
    DISABLED, ON, REVERSE
}

@ConfigObjectFromString
data class FolderWindowConfig(val style: FolderWindowStyle, val rawBackgroundAlpha: Int, @field:ColorInt @get:ColorInt val rawBackgroundColor: Int, val cornerRadiusDp: Int, val rawCardBackground: Boolean, val showFolderName: Boolean, val rows: Int, val cols: Int, val scrollDirection: FolderScrollDirection) {

    companion object: Cofs<FolderWindowConfig>, CofsBase<FolderWindowConfig> by FolderWindowConfig::class.getCofsDelegate() {
        @JvmStatic
        override val DEFAULT = FolderWindowConfig(FolderWindowStyle.WINDOW, 0xAC, Color.WHITE,24, true, true, 3, 4, FolderScrollDirection.VERTICAL)
    }

    var ignoreNightMode = false

    val cardBackground: Boolean
        get() = style == FolderWindowStyle.WINDOW || rawCardBackground

    val backgroundColor: Int
        get() {
            if (!ignoreNightMode && Pref3.night_mode_auto_folder.value && DayNightManager.get().isNightMode) {
                return MaterialColors.NIGHT_THEME_BACKGROUND
            }
            return rawBackgroundColor
        }

    val backgroundAlpha: Int
        get() {
            if (!ignoreNightMode && Pref3.night_mode_auto_folder.value && DayNightManager.get().isNightMode) {
                return Math.max(0x80, rawBackgroundAlpha)
            }
            return rawBackgroundAlpha
        }
}

object Pref3 {
    private lateinit var sharedPreferences: SharedPreferences
    private lateinit var sharedPreferencesObservable: Observable<String>
    private val prefMap = mutableMapOf<String, Pref<out Any>>()

    fun get(propertyName: String) : Pref3.Pref<out Any> {
        return prefMap[propertyName]!!
    }

    val folder_icon_config by PrefCofs(FolderIconConfig)
    val folder_window_config by PrefCofs(FolderWindowConfig)
    val adaptive_icon_shape by PrefCofs(AdaptiveIconShape)

    val desktop_cellspecs by PrefCofs(CellSpecConfig, defaultValue = CellSpecConfig(labelColor = Color.WHITE, labelShadow = true))
    val dock_cellspecs by PrefCofs(CellSpecConfig, defaultValue = CellSpecConfig(labelColor = Color.WHITE, labelShadow = true, labelVisible = false, matchDesktopSize = true))
    val drawer_cellspecs by PrefCofs(CellSpecConfig, defaultValue = CellSpecConfig(matchDesktopSize = true))
    val folder_cellspecs by PrefCofs(CellSpecConfig, defaultValue = CellSpecConfig(matchDesktopSize = true))

    val search_window_style by PrefEnum(SearchBarStyle.SearchWindowStyle.GOOGLE)
    val qsb_style by Pref.searchbarstyle(SearchBarStyle.DEFAULT)
    val drawer_search_style by Pref.searchbarstyle(SearchBarStyle.DRAWER_LEGACY, { _ -> !drawer_cards.value })
    val desktop_page_indicator by PrefEnum(com.teslacoilsw.launcher.preferences.Pref.Key.ScrollIndicatorStyle.LINE, key = "desktop_scroll_indicator")
    val desktop_page_indicator_color by Pref(Color.WHITE, key = "desktop_scroll_indicator_color")
    val desktop_lock by Pref(false)
    val desktop_shadow by Pref(false, key = "dock_show_shadow")

    val desktop_infinite_scroll by Pref(false)
    val desktop_scroll_effect by PrefEnum(com.teslacoilsw.launcher.preferences.Pref.Key.ScrollEffectEnum.SIMPLE, key = "scroll_effect")
    val desktop_subgrid by Pref(true, key = "desktop_grid_subgrid")
    val wallpaper_scroll_mode by PrefEnum(WallpaperScrollMode.ON, key = "wallpaper_scrolling")
    val wallpaper_scrolling_force by Pref(false)

    val dock_infinite_scroll by Pref(false)
    val dock_overlay by Pref(false)
    val dock_autoclose by Pref(false, isValid = { dock_overlay.value })
    val dock_remember_position by Pref(false)

    val drawer_background_color by Pref(Color.WHITE)
    val drawer_card_color by Pref(Color.WHITE)
    val drawer_cards by Pref(false)
    val drawer_enable_tabs by Pref(false)
    val drawer_searchbar by Pref(true, key = "drawer_persistent_search")
    val drawer_isolate_tabs by Pref(false, isValid = { drawer_enable_tabs.value })
    val drawer_swipe_to_open by Pref(true, key = "drawer_swipe_from_dock")
    val drawer_swipe_indicator by Pref(false, key ="desktop_indicator_drawerdock_swipe", isValid = { Pref2.get().drawer.drawer_swipe_from_dock.get()!! })
    val drawer_pull_to_search by Pref(false, isValid = { !Pref2.get().drawer.drawer_swipe_from_dock.get()!! })
    val drawer_scroll_effect by PrefEnum(com.teslacoilsw.launcher.preferences.Pref.Key.ScrollEffectEnum.SIMPLE)
    val drawer_infinite_scroll by Pref(false)
    val drawer_autoclose by Pref(false)
    val drawer_remember_position by Pref.invBoolean(false, key = "drawer_user_reset")
    val drawer_fastscroll by Pref(true)
    val drawer_fastscroll_color by Pref(Color.TRANSPARENT)
    val drawer_folders_first by Pref(true, key = "drawer_folders_before_apps")
    val appsearch_background by Pref(Color.WHITE)

    val folder_animation by PrefEnum(com.teslacoilsw.launcher.preferences.Pref.Key.DrawerAnimation.CIRCLE)

    val searchbar_placement by PrefEnum(SearchBarPlacement.NONE)
    val searchbar_widget_ensure_onresume by Pref(0)

    val widget_overlap by Pref(false)
    val widget_overlap_when_placing by Pref(false, { _ -> widget_overlap.value })

    val popup_menu_style by PrefEnum(PopupMenuStyle.default, isValid = { it.isSupported })
    val popup_menu_include_notifications by Pref(true, isValid = { Build.VERSION.SDK_INT >= Build.VERSION_CODES.LOLLIPOP && com.teslacoilsw.launcher.preferences.Pref.Value.unread_count && com.teslacoilsw.launcher.preferences.Pref.Value.unread_count_provider.isCountless })
    val popup_menu_item_visibility by PrefCofs(PopupItems)

    val show_drop_targets by Pref(false)

    val gesture_home_button by Pref.intent(NamedIntent.NONE, key = "home_button_intent")
    val gesture_home_button_only_on_default by Pref(true, key = "home_button_action_default_page_only")
    val gesture_long_menu by Pref.intent(NamedIntent.NONE, key = "long_menu_intent")
    val gesture_swipe_up by Pref.intent(NamedIntent.NONE)
    val gesture_swipe_down by Pref.intent(NamedIntent.NONE)
    val gesture_pinch_in by Pref.intent(NamedIntent.NONE, key = "gesture_pinch")
    val gesture_pinch_out by Pref.intent(NamedIntent.NONE)
    val gesture_swipe_two_up by Pref.intent(NamedIntent.NONE, key = "gesture_two_finger_scroll_up")
    val gesture_swipe_two_down by Pref.intent(NamedIntent.NONE, key = "gesture_two_finger_scroll_down")
    val gesture_double_tap by Pref.intent(NamedIntent.NONE)
    val gesture_rotate_ccw by Pref.intent(NamedIntent.NONE)
    val gesture_rotate_cw by Pref.intent(NamedIntent.NONE)
    val gesture_double_tap_swipe_up by Pref.intent(NamedIntent.NONE)
    val gesture_double_tap_swipe_down by Pref.intent(NamedIntent.NONE)

    val icon_theme by PrefCofs(IconThemeConfig.Companion, key = "theme_icon_pack")
    val icon_theme_masking by Pref(true)
    val icon_normalization by Pref.boolean(true, key = "sanify_icon_size")

    val app_animation by PrefEnum(ActivityAnimation.SYSTEM, key = "activity_animation")
    val animation_speed by PrefEnum(com.teslacoilsw.launcher.preferences.Pref.Key.AnimationSpeed.NOVA)
    val scroll_speed by PrefEnum(com.teslacoilsw.launcher.preferences.Pref.Key.ScrollSpeed.NOVA)
    val screen_orientation by PrefEnum(com.teslacoilsw.launcher.preferences.Pref.Key.ScreenOrientation.NOSENSOR)
    val show_notification_bar by Pref.boolean(true)
    val notification_bar_transparency by Pref.boolean(true)
    val notification_bar_dark_icons by Pref.boolean(false)
    val samsung_hide_navbar by Pref.boolean(false, isValid = { OEMUtils.hasSamsungS8NavBar() })
    val navbar_hide_back by Pref.boolean(false, isValid = { Build.VERSION.SDK_INT >= Build.VERSION_CODES.P })
    val notification_bar_hide_clock by PrefEnum(HideClock.DISABLED)
    val night_mode by PrefEnum(NightMode.DEFAULT)
    val night_mode_location by Pref(false)
    val night_mode_auto_searchbar by Pref(false)
    val night_mode_auto_drawer by Pref(false)
    val night_mode_auto_drawer_icon by Pref(false)
    val night_mode_auto_folder by Pref(false)

    val google_now_page by Pref(true)
    val google_now_bezel_swipe by Pref(false)
    val google_now_transition by PrefEnum(GoogleNowTransition.PAGE)
    val google_now_okgoogle by Pref(false)

    val sesame_app_search by Pref(false)
    val sesame_shortcut_menu by Pref(false)

    val widget_restart_on_resume by Pref(false, key = "widget_restart_on_start")
    val edit_dialog_show_cn by Pref(false)
    val debug_desktopmenu_export_png by Pref(false)

    fun init(sharedPreferences: SharedPreferences) {
        if (!this::sharedPreferences.isInitialized)
            this.sharedPreferences = sharedPreferences

        sharedPreferencesObservable = Observable.create(Action1<Emitter<String>> { emitter ->
            val listener = OnSharedPreferenceChangeListener { _, key ->
                emitter.onNext(key)
            }

            emitter.setCancellation { sharedPreferences.unregisterOnSharedPreferenceChangeListener(listener) }
            sharedPreferences.registerOnSharedPreferenceChangeListener(listener)
        }, Emitter.BackpressureMode.BUFFER).share()
    }


    @Suppress("UNUSED_PARAMETER")
    private fun defaultIsValid(unused: Any) = true

    open class Pref<T: Any>(@get:JvmName("defaultValue") val defaultValue: T,
                    private val isValid: (value: T) -> Boolean = ::defaultIsValid, key: String? = null)
        : PrefHolder<T>, ReadOnlyProperty<Pref3, Pref<T>>
    {
        private val overrideKey = key

        @get:JvmName("key")
        lateinit var key: String
            private set

        operator fun provideDelegate(thisRef: Pref3, property: KProperty<*>): ReadOnlyProperty<Pref3, Pref<T>> {
            key = overrideKey ?: property.name
            prefMap[property.name] = this
            return this
        }

        override fun getValue(thisRef: Pref3, property: KProperty<*>): Pref<T> = this

        private lateinit var _value: T
        private val baseObservable by lazy { sharedPreferencesObservable
                .filter{ it == this.key }.share() }

        override val asObservable
            get() = baseObservable.startWith(key).map { value }

        override var storedValue: T
            get() = value
            set(value) { this.value = value }

        @get:JvmName("value")
        var value: T
            get() {
                if (!::_value.isInitialized) {
                    reload()
                    baseObservable.subscribe {
                        reload()
                    }
                }

                return if (isValid(_value))
                    _value
                else
                    defaultValue
            }
            set(x) {
                _value = x
                val edit = sharedPreferences.edit()
                storeValue(edit, x)
                edit.apply()
            }

        private fun reload() {
            _value = loadValue(key, defaultValue)
        }

        fun setIfChanged(newValue: T): Boolean {
            if (::_value.isInitialized && value != newValue) {
                value = newValue
                return true
            }
            return false
        }


        protected open fun loadValue(key: String, defaultValue: T): T {
            @Suppress("IMPLICIT_CAST_TO_ANY", "UNCHECKED_CAST")
            return when (defaultValue) {
                is Boolean -> sharedPreferences.getBoolean(key, defaultValue)
                is Int -> sharedPreferences.getInt(key, defaultValue)
                is Long -> sharedPreferences.getLong(key, defaultValue)
                is Float -> sharedPreferences.getFloat(key, defaultValue)
                is String -> sharedPreferences.getString(key, defaultValue)
                is Set<*> -> sharedPreferences.getStringSet(key, defaultValue as Set<String>)
                else -> throw IllegalStateException("Unable to load pref '$key' with type '${defaultValue.javaClass}'")
            } as T
        }

        open fun storeValue(editor: SharedPreferences.Editor, value: T): SharedPreferences.Editor {
            return when (value) {
                is Boolean -> editor.putBoolean(key, value)
                is Int -> editor.putInt(key, value)
                is Long -> editor.putLong(key, value)
                is Float -> editor.putFloat(key, value)
                is String -> editor.putString(key, value)
                is Set<*> -> editor.putStringSet(key, value as Set<String>)
                else -> throw IllegalStateException("Unable to save pref '$key' with type '${defaultValue.javaClass}'")
            }
        }

        open fun asStringPrefHolder(): PrefHolder<String> {
            @Suppress("UNCHECKED_CAST")
            return this as PrefHolder<String>
        }

        fun delete() {
            sharedPreferences.edit().remove(key).apply()
        }

        fun isSet() = sharedPreferences.contains(key)

        override fun toString(): String {
            return "Pref($key, $value, $defaultValue)"
        }

        companion object {
            @Deprecated("", ReplaceWith("Pref(defaultValue, isValid, key)"))
            fun string(defaultValue: String, isValid: (value: String) -> Boolean = ::defaultIsValid, key: String? = null) =
                    Pref(defaultValue, isValid, key)

            @Deprecated("", ReplaceWith("Pref(defaultValue, isValid, key)"))
            fun boolean(defaultValue: Boolean, isValid: (value: Boolean) -> Boolean = ::defaultIsValid, key: String? = null) =
                    Pref(defaultValue, isValid, key)

            // TODO
            fun invBoolean(defaultValue: Boolean, isValid: (value: Boolean) -> Boolean = ::defaultIsValid, key: String? = null) =
                    Pref(defaultValue, isValid, key)

            fun searchbarstyle(defaultValue: SearchBarStyle, isValid: (value: SearchBarStyle) -> Boolean = ::defaultIsValid, key: String? = null) =
                    object: Pref<SearchBarStyle>(defaultValue, isValid, key) {
                        override fun loadValue(key: String, defaultValue: SearchBarStyle): SearchBarStyle {
                            val valueString = sharedPreferences.getString(key, "")!!
                            return if (valueString.isBlank())
                                defaultValue
                            else
                                SearchBarStyle.fromString(valueString)
                        }

                        override fun storeValue(editor: SharedPreferences.Editor, value: SearchBarStyle): SharedPreferences.Editor {
                            return editor.putString(this.key, value.flattenToString())
                        }
                    }

            fun intent(defaultValue: NamedIntent, isValid: (value: NamedIntent) -> Boolean = ::defaultIsValid, key: String? = null) =
                    object: Pref<NamedIntent>(defaultValue, isValid, key)  {

                        override fun loadValue(key: String, defaultValue: NamedIntent): NamedIntent {
                            val valueString = sharedPreferences.getString(key, "")!!
                            return if (valueString.isBlank())
                                defaultValue
                            else
                                NamedIntent.parse(valueString)                        }

                        override fun storeValue(editor: SharedPreferences.Editor, value: NamedIntent): SharedPreferences.Editor {
                            return editor.putString(this.key, value.toUri())
                        }
                    }
        }
    }

    class PrefCofs<T: Any>(val companion: Cofs<T>,
                           defaultValue: T = companion.DEFAULT,
                           isValid: (value: T) -> Boolean = ::defaultIsValid, key: String? = null) :
            Pref<T>(defaultValue, isValid, key)
    {
        override fun loadValue(key: String, defaultValue: T): T =
                companion.parse(sharedPreferences.getString(key, ""), defaultValue)

        override fun storeValue(editor: SharedPreferences.Editor, value: T): SharedPreferences.Editor =
                editor.putString(key, companion.flattenToString(value))

        override fun asStringPrefHolder(): PrefHolder<String> {
            val src = this
            return TransformPrefHolder(this,
                { companion.flattenToString(this) },
                { companion.parse(this, src.defaultValue) })
        }
    }

    class PrefEnum<T: Enum<T>>(defaultValue: T,
                           isValid: (value: T) -> Boolean = ::defaultIsValid, key: String? = null) :
            Pref<T>(defaultValue, isValid, key)
    {
        override fun loadValue(key: String, defaultValue: T): T {
            val x =
            sharedPreferences.getString(key, null)
            return x.parse(defaultValue)
        }

        override fun storeValue(editor: SharedPreferences.Editor, value: T): SharedPreferences.Editor =
                editor.putString(key, value.name)

        override fun asStringPrefHolder(): PrefHolder<String> {
            val src = this
            return TransformPrefHolder(this, { this.name }, { this.parse(src.defaultValue) })
        }
    }
}

fun <T : Any> SharedPreferences.Editor.put(pref: Pref3.Pref<T>, value: T): SharedPreferences.Editor =
    pref.storeValue(this, value)

