package com.teslacoilsw.launcher.preferences.fancyprefs

import android.animation.Animator
import android.animation.AnimatorListenerAdapter
import android.animation.AnimatorSet
import android.animation.ObjectAnimator
import android.animation.ValueAnimator
import android.content.Context
import android.util.AttributeSet
import android.view.View
import android.view.animation.DecelerateInterpolator
import com.android.launcher3.LauncherAnimUtils
import com.android.launcher3.LogDecelerateInterpolator
import com.android.launcher3.Utilities
import com.teslacoilsw.launcher.R
import com.teslacoilsw.launcher.anim.CircularRevealFrameLayout
import com.teslacoilsw.launcher.preferences.Pref
import com.teslacoilsw.launcher.preferences.widget.DrawerAnimationPreference
import com.teslacoilsw.launcher.util.DpUtil
import com.teslacoilsw.shared.util.MathUtils
import kotlin.math.roundToLong

class FancyPrefDrawerAnimDialog @JvmOverloads
constructor(context: Context, attrs: AttributeSet? = null, defStyleAttr: Int = 0)
    : FancyPrefSummaryListView(context, attrs, defStyleAttr)
{

    private lateinit var mWorkspace: DrawerAnimationPreference.WorkspaceView
    private lateinit var mDesktop: View
    private lateinit var mDrawer: View
    private lateinit var mDrawerCard: CircularRevealFrameLayout

    private var mAnimationMode: Pref.Key.DrawerAnimation = Pref.Key.DrawerAnimation.SLIDE

    init {
        dialogLayoutId = R.layout.dialog_preference_draweranimation
    }

    override fun bindDialogView(dialogView: View): (() -> String)? {
        val result = super.bindDialogView(dialogView)

        mWorkspace = dialogView.findViewById(R.id.workspace)
        mDesktop = dialogView.findViewById(R.id.desktop)
        mDrawer = dialogView.findViewById(R.id.drawer)
        mDrawerCard = dialogView.findViewById(R.id.drawer_card)
        mDrawerCard.setCircleAnimationProgress(0f)
        mDrawerCard.alpha = 0f
        mDrawerCard.setCircularRevealColor(-0xa0a0b)
        val p = DpUtil.getPixelSize(10)
        mDrawerCard.setClipRect(p, p, p, p)

        mWorkspace.setOnClickListener { anim() }

        return result
    }

    override fun onSelectedInDialog(dialogView: View, entry: SummaryListRowEntry) {
        super.onSelectedInDialog(dialogView, entry)
        mAnimationMode = Pref.Key.DrawerAnimation.valueOf(entry.value)
        anim()
    }

    private var mIsAnimating = false
    private fun anim() {
        if (!mIsAnimating) {
            animToDrawer(mAnimationMode)
        }
    }

    private fun animToDrawer(mode: Pref.Key.DrawerAnimation) {

        mIsAnimating = true
        val duration = Pref.Key.AnimationSpeed.SLOW.getDuration(mDesktop.resources, R.integer.config_appsCustomizeZoomInTime)

        val set = AnimatorSet()

        if (mode == Pref.Key.DrawerAnimation.ZOOM) {
            mDesktop.pivotX = (mDesktop.width / 2).toFloat()
            mDesktop.pivotY = (mDesktop.height / 2).toFloat()

            mDrawer.pivotX = (mDrawer.width / 2).toFloat()
            mDrawer.pivotY = (mDrawer.height / 2).toFloat()

            val desktopAlpha = ObjectAnimator.ofFloat(mDesktop, "alpha", 1.0f, 0f).setDuration((duration * 2f / 3).roundToLong())
            val desktopScaleX = ObjectAnimator.ofFloat(mDesktop, "scaleX", 1.0f, .3f).setDuration(duration)
            val desktopScaleY = ObjectAnimator.ofFloat(mDesktop, "scaleY", 1.0f, .3f).setDuration(duration)

            val drawerAlpha = ObjectAnimator.ofFloat(mDrawer, "alpha", 0f, 1.0f).setDuration((duration * 2f / 3).roundToLong())
            val drawerScaleX = ObjectAnimator.ofFloat(mDrawer, "scaleX", 5f, 1.0f).setDuration(duration)
            val drawerScaleY = ObjectAnimator.ofFloat(mDrawer, "scaleY", 5f, 1.0f).setDuration(duration)


            set.playTogether(desktopAlpha, desktopScaleX, desktopScaleY, drawerAlpha, drawerScaleX, drawerScaleY)

            set.addListener(object : AnimatorListenerAdapter() {
                override fun onAnimationStart(animation: Animator) {
                    mDrawer.setLayerType(View.LAYER_TYPE_HARDWARE, Utilities.sEmptyPaint)
                    mDesktop.setLayerType(View.LAYER_TYPE_HARDWARE, Utilities.sEmptyPaint)
                }
            })
        } else if (mode == Pref.Key.DrawerAnimation.CIRCLE) {
            mWorkspace.setCircleAnchor(mDesktop.width / 2, mDesktop.height - DpUtil.getPixelSize(11))
            val c = ValueAnimator()
            c.setFloatValues(0.02f, 1.0f)
            c.addUpdateListener { animation -> mWorkspace.setCircleRevealRadius(animation.animatedValue as Float) }
            c.addListener(object : AnimatorListenerAdapter() {
                override fun onAnimationStart(animation: Animator) {
                    mDesktop.alpha = 1.0f
                    mDrawer.alpha = 1.0f
                }

                override fun onAnimationEnd(animation: Animator) {
                    mDesktop.alpha = 0.0f
                }
            })
            c.duration = duration
            set.play(c)
        } else if (mode == Pref.Key.DrawerAnimation.CIRCLE_CARD) {
            val c = mDrawerCard.createCircularReveal(mDesktop.width / 2, mDesktop.height / 2, DpUtil.getPixelSize(5).toFloat(), DpUtil.getPixelSize(132).toFloat())
            mDrawerCard.setCircleAnimationProgress(0f)
            mDrawerCard.alpha = 1.0f
            c.interpolator = LogDecelerateInterpolator(30, 0)
            c.duration = duration

            val yDrift = mDesktop.height / 2 - DpUtil.getPixelSize(11)
            mDrawerCard.translationY = yDrift.toFloat()
            val trans = LauncherAnimUtils.ofFloat(mDrawerCard, "translationY", false, yDrift.toFloat(), 0f)
            trans.interpolator = DecelerateInterpolator(1.5f)
            trans.duration = duration

            val anim = LauncherAnimUtils.ofFloat(mDesktop, "alpha", false, 1.0f, 0.0f)
            anim.interpolator = DecelerateInterpolator(1.5f)
            anim.duration = duration

            set.playTogether(c, trans, anim)
        } else if (mode == Pref.Key.DrawerAnimation.SLIDE) {
            mDesktop.pivotX = (mDesktop.width / 2).toFloat()
            mDesktop.pivotY = (mDesktop.height / 2).toFloat()
            val desktopAlpha = ObjectAnimator.ofFloat(mDesktop, "alpha", 1.0f, 0f).setDuration(duration)
            val desktopScaleX = ObjectAnimator.ofFloat(mDesktop, "scaleX", 1.0f, .5f).setDuration(duration)
            val desktopScaleY = ObjectAnimator.ofFloat(mDesktop, "scaleY", 1.0f, .5f).setDuration(duration)

            val drawerAlpha = ObjectAnimator.ofFloat(mDrawer, "alpha", 0.0f, 1.0f).setDuration((duration * 2f / 3f).roundToLong())
            val drawerTrans = ObjectAnimator.ofFloat(mDrawer, "translationY", mDesktop.height.toFloat(), 0f).setDuration(duration)

            set.playTogether(desktopAlpha, desktopScaleX, desktopScaleY, drawerAlpha, drawerTrans)
            //        } else if (mode == Pref.Key.DrawerAnimation.FLIP) {
            //
            //            mDesktop.setPivotX(mDesktop.getWidth()/2);
            //            mDesktop.setPivotY(mDesktop.getHeight());
            //
            //            mDrawer.setPivotX(mDrawer.getWidth()/2);
            //            mDrawer.setPivotY(mDrawer.getHeight());
            //
            //            mDesktop.setCameraDistance(3250);
            //
            //            Animator desktopAlpha = ObjectAnimator.ofFloat(mDesktop, "alpha", 1.0f, 0).setDuration(duration);
            //            Animator desktopRot = ObjectAnimator.ofFloat(mDesktop, "rotationX", 0, 90).setDuration(duration);
            //
            //            Animator drawerAlpha = ObjectAnimator.ofFloat(mDrawer, "alpha", 0, 1.0f).setDuration(duration);
            //            Animator drawerRot = ObjectAnimator.ofFloat(mDrawer, "rotationX", -90, 0).setDuration(duration);
            //
            //            set.playTogether(desktopAlpha, desktopRot, drawerAlpha, drawerRot);

        } else if (mode == Pref.Key.DrawerAnimation.FADE) {
            val desktopAlpha = ObjectAnimator.ofFloat(mDesktop, "alpha", 1.0f, 0f).setDuration(duration)
            val drawerAlpha = ObjectAnimator.ofFloat(mDrawer, "alpha", 0f, 1.0f).setDuration(duration)
            set.playTogether(desktopAlpha, drawerAlpha)
        }

        set.addListener(object : AnimatorListenerAdapter() {
            override fun onAnimationEnd(animation: Animator) {
                animToDesktop(mode)
            }
        })
        set.start()

    }

    private fun animToDesktop(mode: Pref.Key.DrawerAnimation) {

        val duration = Pref.Key.AnimationSpeed.SLOW.getDuration(mDesktop.resources, R.integer.config_appsCustomizeZoomInTime)

        val set = AnimatorSet()

        if (mode == Pref.Key.DrawerAnimation.ZOOM) {
            val desktopAlpha = ObjectAnimator.ofFloat(mDesktop, "alpha", 0f, 1.0f).setDuration((duration * 2f / 3).roundToLong())
            val desktopScaleX = ObjectAnimator.ofFloat(mDesktop, "scaleX", .5f, 1.0f).setDuration(duration)
            val desktopScaleY = ObjectAnimator.ofFloat(mDesktop, "scaleY", .5f, 1.0f).setDuration(duration)

            val drawerAlpha = ObjectAnimator.ofFloat(mDrawer, "alpha", 1.0f, 0f).setDuration((duration * 2f / 3).roundToLong())
            val drawerScaleX = ObjectAnimator.ofFloat(mDrawer, "scaleX", 1.0f, 1.5f).setDuration(duration)
            val drawerScaleY = ObjectAnimator.ofFloat(mDrawer, "scaleY", 1.0f, 1.5f).setDuration(duration)


            set.playTogether(desktopAlpha, desktopScaleX, desktopScaleY, drawerAlpha, drawerScaleX, drawerScaleY)

            set.addListener(object : AnimatorListenerAdapter() {
                override fun onAnimationEnd(animation: Animator) {
                    mDrawer.setLayerType(View.LAYER_TYPE_NONE, null)
                    mDesktop.setLayerType(View.LAYER_TYPE_NONE, null)
                    mDrawer.scaleX = 1.0f
                    mDrawer.scaleY = 1.0f
                }
            })
        } else if (mode == Pref.Key.DrawerAnimation.CIRCLE) {
            mWorkspace.setCircleAnchor(mDesktop.width / 2, mDesktop.height - DpUtil.getPixelSize(11))
            val c = ValueAnimator()
            c.setFloatValues(.98f, 0.0f)
            c.addUpdateListener { animation -> mWorkspace.setCircleRevealRadius(animation.animatedValue as Float) }
            c.addListener(object : AnimatorListenerAdapter() {
                override fun onAnimationStart(animation: Animator) {
                    mDesktop.alpha = 1.0f
                }

                override fun onAnimationEnd(animation: Animator) {
                    mDrawer.alpha = 0.0f
                    mWorkspace.setCircleRevealRadius(1.0f)
                }
            })
            c.duration = duration
            set.play(c)
        } else if (mode == Pref.Key.DrawerAnimation.CIRCLE_CARD) {
            val c = mDrawerCard.createCircularReveal(mDesktop.width / 2, mDesktop.height / 2, DpUtil.getPixelSize(132).toFloat(), DpUtil.getPixelSize(9).toFloat())
            mDrawerCard.setCircleAnimationProgress(0f)
            c.addListener(object : AnimatorListenerAdapter() {
                override fun onAnimationEnd(animation: Animator) {
                    mDrawerCard.alpha = 0.0f
                }
            })
            c.interpolator = DecelerateInterpolator(1.5f)
            c.duration = duration

            val yDrift = mDesktop.height / 2 - DpUtil.getPixelSize(11)
            val trans = LauncherAnimUtils.ofFloat(mDrawerCard, "translationY", false, 0f, yDrift.toFloat())
            trans.interpolator = DecelerateInterpolator(1.5f)
            trans.duration = duration

            val anim = LauncherAnimUtils.ofFloat(mDesktop, "alpha", false, 0f, 1f)
            anim.duration = duration

            set.playTogether(c, trans, anim)
        } else if (mode == Pref.Key.DrawerAnimation.SLIDE) {
            val desktopAlpha = ObjectAnimator.ofFloat(mDesktop, "alpha", 0f, 1.0f).setDuration(MathUtils.roundInt(duration * 2f / 3).toLong())
            val desktopScaleX = ObjectAnimator.ofFloat(mDesktop, "scaleX", .5f, 1.0f).setDuration(duration)
            val desktopScaleY = ObjectAnimator.ofFloat(mDesktop, "scaleY", .5f, 1.0f).setDuration(duration)

            val drawerAlpha = ObjectAnimator.ofFloat(mDrawer, "alpha", 1.0f, 0f).setDuration(MathUtils.roundInt(duration * 2f / 3).toLong())
            val drawerTrans = ObjectAnimator.ofFloat(mDrawer, "translationY", 0f, mDesktop.height.toFloat()).setDuration(duration)

            set.playTogether(desktopAlpha, desktopScaleX, desktopScaleY, drawerAlpha, drawerTrans)

            set.addListener(object : AnimatorListenerAdapter() {
                override fun onAnimationEnd(animation: Animator) {
                    mDrawer.setLayerType(View.LAYER_TYPE_NONE, null)
                    mDesktop.setLayerType(View.LAYER_TYPE_NONE, null)
                    mDrawer.translationY = 0f
                }
            })
            //        } else if (mode == Pref.Key.DrawerAnimation.FLIP) {
            //            Animator desktopAlpha = ObjectAnimator.ofFloat(mDesktop, "alpha", 0.0f, 1.0f).setDuration(duration);
            //            Animator desktopRot = ObjectAnimator.ofFloat(mDesktop, "rotationX", 90, 0).setDuration(duration);
            //
            //            Animator drawerAlpha = ObjectAnimator.ofFloat(mDrawer, "alpha", 1.0f, 0.0f).setDuration(duration);
            //            Animator drawerRot = ObjectAnimator.ofFloat(mDrawer, "rotationX", 0, -90).setDuration(duration);
            //            set.addListener(new AnimatorListenerAdapter() {
            //                @Override
            //                public void onAnimationEnd(Animator animation) {
            //                    mDrawer.setRotationX(0);
            //                }
            //            });
            //
            //            set.playTogether(desktopAlpha, desktopRot, drawerAlpha, drawerRot);
        } else if (mode == Pref.Key.DrawerAnimation.FADE) {
            val desktopAlpha = ObjectAnimator.ofFloat(mDesktop, "alpha", 0f, 1f).setDuration(duration)
            val drawerAlpha = ObjectAnimator.ofFloat(mDrawer, "alpha", 1f, 0f).setDuration(duration)

            set.playTogether(desktopAlpha, drawerAlpha)
        }

        set.addListener(object : AnimatorListenerAdapter() {
            override fun onAnimationEnd(animation: Animator) {
                mIsAnimating = false
                if (mAnimationMode != mode) {
                    anim()
                }
            }
        })
        set.startDelay = 200
        set.start()
    }
}