package com.teslacoilsw.launcher.preferences.fancyprefs

import android.animation.Animator
import android.animation.AnimatorListenerAdapter
import android.animation.AnimatorSet
import android.animation.ObjectAnimator
import android.animation.ValueAnimator
import android.content.Context
import android.graphics.Canvas
import android.graphics.Color
import android.graphics.Paint
import android.util.AttributeSet
import android.view.View
import android.view.animation.AccelerateInterpolator
import com.android.launcher3.LauncherAnimatorUpdateListener
import com.android.launcher3.Utilities
import com.teslacoilsw.kotlin.bind
import com.teslacoilsw.launcher.R
import com.teslacoilsw.launcher.databinding.DialogPreferenceFolderanimationBinding
import com.teslacoilsw.launcher.preferences.Pref
import com.teslacoilsw.shared.util.MathUtils
import kotlin.math.roundToLong

class FancyPrefFolderAnimDialog @JvmOverloads
constructor(context: Context, attrs: AttributeSet? = null, defStyleAttr: Int = 0)
    : FancyPrefSummaryListView(context, attrs, defStyleAttr)
{
    private lateinit var binding: DialogPreferenceFolderanimationBinding

    private var animationMode: Pref.Key.DrawerAnimation = Pref.Key.DrawerAnimation.SLIDE

    init {
        dialogLayoutId = R.layout.dialog_preference_folderanimation
    }

    override fun bindDialogView(dialogView: View): (() -> String)? {
        val result = super.bindDialogView(dialogView)
        binding = dialogView.bind()
        binding.folder.alpha = 0f
        binding.workspace.setOnClickListener {
            anim()
        }

        return result
    }

    override fun onSelectedInDialog(dialogView: View, entry: SummaryListRowEntry) {
        super.onSelectedInDialog(dialogView, entry)
        animationMode = Pref.Key.DrawerAnimation.valueOf(entry.value)
        anim()
    }

    private var isAnimating = false
    private fun anim() {
        if (!isAnimating) {
            animToFolder(animationMode)
        }
    }

    private fun animToFolder(mode: Pref.Key.DrawerAnimation) {
        isAnimating = true
        val duration = 350L

        val set = AnimatorSet()

        val folder = binding.folder
        folder.pivotX = folder.width * 61 / 100f
        folder.pivotY = folder.height * 85 / 100f

        if (mode == Pref.Key.DrawerAnimation.ZOOM) {
            val folderAlpha = ObjectAnimator.ofFloat(folder, "alpha", 0f, 1.0f).setDuration((duration * 2f / 3f).roundToLong())
            val folderScaleX = ObjectAnimator.ofFloat(folder, "scaleX", .3f, 1.0f).setDuration(duration)
            val folderScaleY = ObjectAnimator.ofFloat(folder, "scaleY", .3f, 1.0f).setDuration(duration)

            set.playTogether(folderAlpha, folderScaleX, folderScaleY)

            set.addListener(object : AnimatorListenerAdapter() {
                override fun onAnimationStart(animation: Animator) {
                    folder.setLayerType(View.LAYER_TYPE_HARDWARE, Utilities.sEmptyPaint)
                }
            })
        } else if (mode == Pref.Key.DrawerAnimation.CIRCLE) {
            folder.alpha = 1f
            val va = ValueAnimator.ofFloat(0f, 1.0f)

            /* We create a slight offset to give a make translation animation */
            va.addUpdateListener(object : LauncherAnimatorUpdateListener() {
                public override fun onAnimationUpdate(a: Float, b: Float) {
                    folder.setCircleAnimationProgress(b)
                }
            })
            va.addListener(object : AnimatorListenerAdapter() {
                override fun onAnimationEnd(animation: Animator) {
                    folder.setCircleAnimationProgress(1f)
                }
            })
            va.interpolator = AccelerateInterpolator()
            va.duration = duration
            set.play(va)
        }

        set.addListener(object : AnimatorListenerAdapter() {
            override fun onAnimationEnd(animation: Animator) {
                animToDesktop(mode)
            }
        })
        set.start()

    }

    private fun animToDesktop(mode: Pref.Key.DrawerAnimation) {

        val duration = 200L

        val set = AnimatorSet()

        val folder = binding.folder
        if (mode == Pref.Key.DrawerAnimation.ZOOM || mode == Pref.Key.DrawerAnimation.CIRCLE) {
            folder.setCircleAnimationProgress(1f)
            val folderAlpha = ObjectAnimator.ofFloat(folder, "alpha", 1.0f, 0f).setDuration((duration * 2f / 3f).roundToLong())
            val folderScaleX = ObjectAnimator.ofFloat(folder, "scaleX", 1.0f, .3f).setDuration(duration)
            val folderScaleY = ObjectAnimator.ofFloat(folder, "scaleY", 1.0f, .3f).setDuration(duration)

            set.playTogether(folderAlpha, folderScaleX, folderScaleY)

            set.addListener(object : AnimatorListenerAdapter() {
                override fun onAnimationEnd(animation: Animator) {
                    folder.setLayerType(View.LAYER_TYPE_NONE, Utilities.sEmptyPaint)
                    folder.scaleX = 1f
                    folder.scaleY = 1f
                }
            })
        } else if (mode == Pref.Key.DrawerAnimation.CIRCLE) {

        }

        set.addListener(object : AnimatorListenerAdapter() {
            override fun onAnimationEnd(animation: Animator) {
                isAnimating = false
                if (animationMode != mode) {
                    anim()
                }
            }
        })
        set.startDelay = 400
        set.start()
    }

    override fun requireConfirmation(): Boolean = true
}

class FolderView : View {
    private val paint = Paint()

    private var circleAnimationProgress = 1f

    init {
        paint.isAntiAlias = true
        paint.color = Color.WHITE
    }

    constructor(context: Context, attrs: AttributeSet) : super(context, attrs) {}

    constructor(context: Context, attrs: AttributeSet, defStyleAttr: Int) : super(context, attrs, defStyleAttr) {}

    fun setCircleAnimationProgress(p: Float) {
        circleAnimationProgress = p
        invalidate()
    }

    override fun draw(canvas: Canvas) {
        super.draw(canvas)
        val width = width / 2
        val height = MathUtils.roundInt(height * 55 / 100f)

        val cx = MathUtils.roundInt(getWidth() * 61 / 100f)
        val cy = MathUtils.roundInt(getHeight() * 40 / 100f)
        val count = canvas.save()
        canvas.clipRect(cx - width / 2, cy - height / 2, cx + width / 2, cy + height / 2)
        if (circleAnimationProgress < 1f) {
            val px = pivotX
            val py = pivotY

            val ltDist = MathUtils.dist(0f, 0f, px, py)
            val lbDist = MathUtils.dist(0f, getHeight().toFloat(), px, py)
            val rtDist = MathUtils.dist(getWidth().toFloat(), 0f, px, py)
            val rbDist = MathUtils.dist(getWidth().toFloat(), getHeight().toFloat(), px, py)

            val size = Math.max(Math.max(ltDist, lbDist), Math.max(rtDist, rbDist))
            val radius = circleAnimationProgress * size
            if (radius <= .005) {
                return
            }

            canvas.drawCircle(pivotX, pivotY, radius, paint)
        } else {
            canvas.drawColor(Color.WHITE)
        }
        canvas.restoreToCount(count)
    }
}