package com.teslacoilsw.launcher.preferences.fancyprefs

import android.content.Context
import android.graphics.Canvas
import android.util.AttributeSet
import android.view.View
import android.widget.SeekBar
import androidx.core.graphics.withSave
import com.teslacoilsw.kotlin.clipOutRectCompat
import com.teslacoilsw.launcher.R
import com.teslacoilsw.launcher.preferences.Pref
import com.teslacoilsw.launcher.preferences.widget.RoundRectView
import com.teslacoilsw.launcher.util.DpUtil

class FancyPrefPaddingView @JvmOverloads
    constructor(context: Context, attrs: AttributeSet? = null, defStyleAttr: Int = 0)
    : FancyPrefDialogView<Pair<Pref.Key.DesktopMargin, Pref.Key.DesktopMargin>>(context, attrs, defStyleAttr)
{
    var horizontalPadding = Pref.Key.DesktopMargin.NONE
        private set
    var verticalPadding = Pref.Key.DesktopMargin.NONE
        private set
    var canAuto: Boolean

    init {
        if (dialogLayoutId == 0)
            dialogLayoutId =  R.layout.preference_padding
        if (summary == null)
            summary = "%s ✕ %s"

        val a = context.obtainStyledAttributes(attrs, R.styleable.FancyPrefPaddingView)
        try {
            canAuto = a.getBoolean(R.styleable.FancyPrefPaddingView_canAuto, false)
        } finally {
            a.recycle()
        }
    }

    override fun bindDialogView(dialogView: View): (() -> Pair<Pref.Key.DesktopMargin, Pref.Key.DesktopMargin>)? {
        val hBar: SeekBar = dialogView.findViewById(R.id.horizontal_padding)!!
        val vBar: SeekBar = dialogView.findViewById(R.id.vertical_padding)!!
        val preview: PaddingPreviewView = dialogView.findViewById(R.id.preview)
        val auto: FancyPrefCheckableView? = dialogView.findViewById(R.id.auto)
        if (canAuto)
            requireNotNull(auto)

        hBar.progress = horizontalPadding.ordinal - 1
        vBar.progress = verticalPadding.ordinal - 1

        val dp4 = DpUtil.getPixelSize(4)
        val seekBarListener = object: SeekBar.OnSeekBarChangeListener {
            override fun onStartTrackingTouch(seekBar: SeekBar) = Unit
            override fun onStopTrackingTouch(seekBar: SeekBar) = Unit

            override fun onProgressChanged(seekBar: SeekBar, progress: Int, fromUser: Boolean) {
                preview.hPadding = hBar.progress * dp4
                preview.vPadding = vBar.progress * dp4
                preview.invalidate()
            }
        }
        hBar.progress = Math.max(horizontalPadding.ordinal - 1, 0)
        vBar.progress = Math.max(verticalPadding.ordinal - 1, 0)


        seekBarListener.onProgressChanged(hBar, 0, true)
        hBar.setOnSeekBarChangeListener(seekBarListener)
        vBar.setOnSeekBarChangeListener(seekBarListener)
        auto?.onUserChanged = { _, _ ->
            hBar.isEnabled = !auto!!.isChecked
            vBar.isEnabled = !auto.isChecked
        }
        if (canAuto) {
            auto!!.visibility = View.VISIBLE
            auto.isChecked = horizontalPadding == Pref.Key.DesktopMargin.AUTO || verticalPadding == Pref.Key.DesktopMargin.AUTO
        } else {
            auto?.visibility = View.GONE
            auto?.isChecked = false
        }

        return {
            if (canAuto && auto?.isChecked == true)
            Pref.Key.DesktopMargin.AUTO to Pref.Key.DesktopMargin.AUTO
        else
            Pref.Key.DesktopMargin.values()[hBar.progress + 1] to Pref.Key.DesktopMargin.values()[vBar.progress + 1]
        }
    }

    override fun summaryFormat(text: String, value: Pair<Pref.Key.DesktopMargin, Pref.Key.DesktopMargin>) =
            String.format(text, value.first.getLabel(context), value.second.getLabel(context))

    override fun updateSummary() {
        val text =
        if (horizontalPadding == Pref.Key.DesktopMargin.AUTO || verticalPadding == Pref.Key.DesktopMargin.AUTO) {
            horizontalPadding.getLabel(context)
        } else {
            val s = summary
            if (s != null && s.contains('%') && valueFieldInitialized) {
                summaryFormat(s.toString(), value)
            } else {
                s
            }
        }
        updateSummaryViews(text)
    }

    override fun onChange(value: Pair<Pref.Key.DesktopMargin, Pref.Key.DesktopMargin>) {
        super.onChange(value)
        if (value.first == Pref.Key.DesktopMargin.AUTO || value.second == Pref.Key.DesktopMargin.AUTO) {
            if (canAuto) {
                horizontalPadding = Pref.Key.DesktopMargin.AUTO
                verticalPadding = Pref.Key.DesktopMargin.AUTO
            } else {
                this.value = Pref.Key.DesktopMargin.NONE to Pref.Key.DesktopMargin.NONE
            }
        } else {
            horizontalPadding = value.first
            verticalPadding = value.second
        }
    }
}

class PaddingPreviewView @JvmOverloads constructor(context: Context, attrs: AttributeSet? = null) : RoundRectView(context, attrs) {
    private val scrimColor = 0x80000000.toInt()

    var hPadding = 0
    var vPadding = 0

    init {
        setWillNotDraw(false)
    }

    override fun onDraw(canvas: Canvas) {
        super.onDraw(canvas)
        canvas.withSave {
            canvas.clipOutRectCompat(hPadding, vPadding, width - hPadding, height - vPadding)
            canvas.drawColor(scrimColor)
        }
    }
}