package com.teslacoilsw.launcher.preferences.fancyprefs

import android.annotation.SuppressLint
import android.content.Context
import android.graphics.Color
import android.graphics.Rect
import android.graphics.drawable.Drawable
import android.graphics.drawable.GradientDrawable
import android.graphics.drawable.InsetDrawable
import android.util.AttributeSet
import android.view.View
import android.view.inputmethod.EditorInfo
import android.widget.LinearLayout
import android.widget.SeekBar
import androidx.annotation.ColorInt
import com.afollestad.materialdialogs.DialogAction
import com.afollestad.materialdialogs.MaterialDialog
import com.teslacoilsw.kotlin.lazy1
import com.teslacoilsw.kotlin.showKeyboard
import com.teslacoilsw.launcher.R
import com.teslacoilsw.launcher.util.DpUtil
import com.teslacoilsw.launcher.widget.DisabledSegmentSeekBar
import com.teslacoilsw.shared.util.MathUtils
import kotlinx.android.synthetic.main.fancypref_seekbar_percent_input_dialog.*
import kotlinx.android.synthetic.main.fancypref_seekbar_row.view.*

open class FancyPrefSeekBarView @JvmOverloads
    constructor(context: Context, attrs: AttributeSet? = null, defStyleAttr: Int = 0)
    : FancyPrefView<Int>(context, attrs, defStyleAttr)
{
    override fun getPreferenceLayoutId(attrs: AttributeSet?): Int {
        val a = context.obtainStyledAttributes(attrs, R.styleable.FancyPrefSeekBarView)
        try {
            val titleText: String? = a.getString(R.styleable.FancyPrefSeekBarView_android_text)
            val orientation = a.getInt(R.styleable.FancyPrefSeekBarView_android_orientation, LinearLayout.HORIZONTAL)

            if (orientation == LinearLayout.VERTICAL)
                return R.layout.fancypref_seekbar_row_vertical_notitle
            return if (titleText != null) R.layout.fancypref_seekbar_row else R.layout.fancypref_seekbar_row_notitle
        } finally {
            a.recycle()
        }
    }
    internal val seekBar by lazy1 { SeekBarWrapper(findViewById(R.id.seekBar)) }

    var format: String = "PERCENT"
        set(value) {
            field = value
            updateText(progress)
        }

    var min: Int = 0
        set(value) {
            val oldValue = progress
            field = value
            progress = oldValue
        }

    var max: Int = 100
        set(value) {
            val oldValue = progress
            field = value
            seekBar.max = (max - min)/stepSize
            progress = oldValue
        }
    val stepSize: Int

    var progress: Int
        get() = Math.min(max, seekBar.progress*stepSize + min)
        set(x) {
            seekBar.progress = (androidx.core.math.MathUtils.clamp(x, min, max) - min)/stepSize
            value = progress
            updateText(progress)
        }

    var percentInt: Int
        get() = (100 * progress.toFloat() / max.toFloat() + 0.5f).toInt()
        set(x) { progress = Math.max(0, Math.min(100,x))*max/100 }

    init {
        val a = context.obtainStyledAttributes(attrs, R.styleable.FancyPrefSeekBarView)
        try {
            stepSize = a.getInt(R.styleable.FancyPrefSeekBarView_stepSize, 1)
            min = a.getInt(R.styleable.FancyPrefSeekBarView_min, 0)
            max = a.getInt(R.styleable.FancyPrefSeekBarView_max, 100)
            format = a.getString(R.styleable.FancyPrefSeekBarView_format) ?: "PERCENT"
            a.getColorStateList(R.styleable.FancyPrefSeekBarView_android_textColor)?.let { titleView?.setTextColor(it) }
        } finally {
            a.recycle()
        }

        seekBar.max = (max - min)/stepSize
        seekBar.progressChangeListener = { rawProgress: Int, fromUser: Boolean ->
            val newProgress = Math.min(max, rawProgress*stepSize + min)
            updateText(newProgress)
            if (fromUser)
                setFromUser(progress)
            else
                value = progress
        }
        seekBar.progressChangeListener!!.invoke(seekBar.progress, false)

        valueText.setOnClickListener {
            fun saveResult(text: String) {
                if (!text.isBlank()) {
                    try {
                        val x = text.toInt()
                        if (format == "PERCENT")
                            percentInt = x
                        else
                            progress = x
                    } catch (e: NumberFormatException) { }
                } }
            val md = MaterialDialog.Builder(context)
                    .customView(R.layout.fancypref_seekbar_percent_input_dialog, false)
                    .negativeText(R.string.cancel)
                    .positiveText(R.string.set)
                    .onAny { dialog, which ->
                        if (which == DialogAction.POSITIVE) {
                            val text = dialog.editText.text?.toString() ?: ""
                            saveResult(text)
                        }
                    }
                    .build()

            val editText = md.editText
            md.editTextTitle.text = titleView?.text
            md.format_units.text = if (format == "PERCENT") "%"
                else format.replace("%s", "").trim()

            if (format == "PERCENT")
                editText.setText(percentInt.toString())
            else
                editText.setText(progress.toString())

            editText.setOnEditorActionListener { _, actionId, _ ->
                if (actionId == EditorInfo.IME_ACTION_DONE) {
                    saveResult(editText.text?.toString() ?: "")
                    md.dismiss()
                    true
                } else {
                    false
                }
            }


            md.show()
            editText.apply { post {
                requestFocus()
                selectAll()
                showKeyboard()
            } }
        }
    }

    override fun setEnabled(enabled: Boolean) {
        super.setEnabled(enabled)
        seekBar.isEnabled = enabled
        valueText.isEnabled = enabled
    }

    @SuppressLint("SetTextI18n")
    open fun updateText(newProgress: Int) {
        if (format == "PERCENT")
            valueText.text = "$percentInt%"
        else if (!format.isBlank())
            valueText.text = String.format(format, newProgress)
    }

    override fun onChange(value: Int) {
        super.onChange(value)
        progress = value
    }
}

class FancyPrefDisabledSegmentSeekBarView @JvmOverloads
constructor(context: Context, attrs: AttributeSet? = null, defStyleAttr: Int = 0)
    : FancyPrefSeekBarView(context, attrs, defStyleAttr)
{

    override fun getPreferenceLayoutId(attrs: AttributeSet?): Int = R.layout.fancypref_disabledsegment_seekbar_row

    var effectiveMax: Int = -1
        set(x) {
            val disableAfter = MathUtils.ceilInt((x - min) / stepSize.toFloat())
            (seekBar.view as DisabledSegmentSeekBar).setDisableAfter(disableAfter)
            field = x
            updateText(progress)
        }

    override fun updateText(newProgress: Int) {
        if (newProgress in 0 until effectiveMax) {
            super.updateText(newProgress)
        } else {
            super.updateText(effectiveMax)
        }
    }
}

class FancyPrefGradientSeekBarView @JvmOverloads
constructor(context: Context, attrs: AttributeSet? = null, defStyleAttr: Int = 0)
    : FancyPrefSeekBarView(context, attrs, defStyleAttr)
{
    private val gradientBarDrawable: GradientDrawable
    private val hsv = FloatArray(3)

    init {
        val dp8 = DpUtil.getPixelSize(8)
        val dp16 = DpUtil.getPixelSize(16)
        seekBar.view.setPadding(seekBar.view.paddingLeft, seekBar.view.paddingTop + dp8, seekBar.view.paddingRight, seekBar.view.paddingBottom + dp8)

        var thumb = GradientDrawable()
        thumb.setColor(Color.WHITE)
        thumb.setSize(dp16, dp16)
        thumb.shape = GradientDrawable.OVAL
        thumb.setStroke(DpUtil.getPixelSize(1), 0xFF757575.toInt())

        seekBar.thumb = thumb
        seekBar.progressDrawable = null
        gradientBarDrawable = GradientDrawable()
        gradientBarDrawable.orientation = GradientDrawable.Orientation.LEFT_RIGHT

        val barHeight = dp8
        seekBar.view.background = NoPaddingInsetDrawable(gradientBarDrawable, dp16, seekBar.view.paddingTop - barHeight/2, dp16, seekBar.view.paddingBottom - barHeight/2)

    }

    fun setColors(@ColorInt color0: Int, @ColorInt color1: Int) {
        gradientBarDrawable.colors = intArrayOf(color0, color1)
    }

    fun setColors(colors: IntArray) {
        gradientBarDrawable.colors = colors
    }

    fun setThumbColor(@ColorInt color: Int) {
        val thumb = seekBar.thumb as GradientDrawable
        thumb.setColor(color)
        Color.colorToHSV(color, hsv)
        hsv[1] = 0.39f
        val lightHue = Color.HSVToColor(hsv)
        hsv[1] = 0.12f
        hsv[2] = 0.5f
        val darkHue = Color.HSVToColor(hsv)
        if (200 < hsv[0] && hsv[0] < 300)
            thumb.setStroke(DpUtil.getPixelSize(1), lightHue)
        else
            thumb.setStroke(DpUtil.getPixelSize(1), darkHue)
    }
}

internal class SeekBarWrapper(val view: View) {
    private val listener: SeekBar.OnSeekBarChangeListener = object: SeekBar.OnSeekBarChangeListener, DisabledSegmentSeekBar.OnSeekBarChangeListener {
        override fun onProgressChanged(seekBar: SeekBar?, progress: Int, fromUser: Boolean) {
            progressChangeListener?.invoke(progress, fromUser)
        }

        override fun onProgressChanged(seekBar: DisabledSegmentSeekBar?, progress: Int, fromUser: Boolean) {
            progressChangeListener?.invoke(progress, fromUser)
        }

        override fun onStartTrackingTouch(seekBar: SeekBar?) { }
        override fun onStartTrackingTouch(seekBar: DisabledSegmentSeekBar?) { }
        override fun onStopTrackingTouch(seekBar: SeekBar?) { }
        override fun onStopTrackingTouch(seekBar: DisabledSegmentSeekBar?) { }
    }

    init {
        assert(view is SeekBar || view is DisabledSegmentSeekBar)
        asSeekBar()?.setOnSeekBarChangeListener(listener)
        asDisabledSegmentSeekBar()?.setOnSeekBarChangeListener(listener as DisabledSegmentSeekBar.OnSeekBarChangeListener)
    }

    var progress: Int
        get() = asSeekBar()?.progress ?: asDisabledSegmentSeekBar()!!.progress
        set(x) {
            asSeekBar()?.progress = x
            asDisabledSegmentSeekBar()?.progress = x
        }

    var max: Int
        get() = asSeekBar()?.max ?: asDisabledSegmentSeekBar()!!.max
        set(x) {
            asSeekBar()?.max = x
            asDisabledSegmentSeekBar()?.max = x
        }

    var isEnabled: Boolean
        get() = asSeekBar()?.isEnabled ?: asDisabledSegmentSeekBar()!!.isEnabled
        set(x) {
            asSeekBar()?.isEnabled = x
            asDisabledSegmentSeekBar()?.isEnabled = x
        }

    var thumb: Drawable?
        get() = asSeekBar()?.thumb ?: asDisabledSegmentSeekBar()!!.thumb
        set(x) {
            asSeekBar()?.thumb = x
            asDisabledSegmentSeekBar()?.thumb = x
        }

    var progressDrawable: Drawable?
        get() = asSeekBar()?.progressDrawable ?: asDisabledSegmentSeekBar()!!.progressDrawable
        set(x) {
            asSeekBar()?.progressDrawable = x
            asDisabledSegmentSeekBar()?.progressDrawable = x
        }

    var progressChangeListener: ((progress: Int, fromUser: Boolean) -> Unit)? = null


    private fun asSeekBar(): SeekBar? {
        if (view is SeekBar)
            return view
        return null
    }

    private fun asDisabledSegmentSeekBar(): DisabledSegmentSeekBar? {
        if (view is DisabledSegmentSeekBar)
            return view
        return null
    }

}


private class NoPaddingInsetDrawable(drawable: Drawable, insetLeft: Int, insetTop: Int, insetRight: Int, insetBottom: Int) : InsetDrawable(drawable, insetLeft, insetTop, insetRight, insetBottom) {
    override fun getPadding(padding: Rect): Boolean {
        return false
    }
}
