package com.teslacoilsw.launcher.preferences.fancyprefs

import android.content.Context
import android.content.DialogInterface
import android.graphics.Color
import android.graphics.drawable.ColorDrawable
import android.graphics.drawable.InsetDrawable
import android.util.AttributeSet
import android.view.LayoutInflater
import android.view.MotionEvent
import android.view.View
import android.view.ViewGroup
import android.widget.FrameLayout
import android.widget.PopupWindow
import android.widget.TextView
import androidx.core.content.ContextCompat
import androidx.core.view.isVisible
import androidx.core.view.updateLayoutParams
import androidx.core.view.updateMarginsRelative
import androidx.core.widget.PopupWindowCompat
import com.teslacoilsw.kotlin.dp
import com.teslacoilsw.launcher.R
import com.teslacoilsw.launcher.util.palette.ColorUtils
import timber.log.Timber

class FancyPrefSpinnerView @JvmOverloads
constructor(context: Context, attrs: AttributeSet? = null, defStyleAttr: Int = 0)
    : FancyPrefSummaryListView(context, attrs, defStyleAttr) {

    private var isInitialized = false
    private val spinner: TextView

    init {
        setWidget(R.layout.fancypref_summaryspinner_row)
        widgetFrame!!.updateLayoutParams<MarginLayoutParams> {
            marginEnd = context.dp(8)
        }
        spinner = findViewById(R.id.spinner)
        updateSpinnerSize()

        setOnClickListener {
            showDialog()
        }

        isInitialized = true
    }

    override fun onEntriesUpdated(entries: List<SummaryListRowEntry>) {
        if (isInitialized)
            updateSpinnerSize()
    }

    private fun updateSpinnerSize() {
        val x = spinner.text
        val maxSize = entries.map {
            spinner.text = it.title
            spinner.measure(MeasureSpec.UNSPECIFIED, MeasureSpec.UNSPECIFIED)
            spinner.measuredWidth
        }.max() ?: 0
        spinner.text = x
        spinner.minWidth = maxSize
    }

    override fun onChange(value: String) {
        super.onChange(value)
        spinner.text = entryForValue(value)?.title ?: value
    }

    override fun showDialog(): DialogInterface {
        val v = if (dialogLayoutId != 0)
            LayoutInflater.from(context).inflate(dialogLayoutId, null, false)
        else
            FrameLayout(context)
        val resultCallback = bindDialogView(v)

        val dm = resources.displayMetrics
        onBindDialog?.invoke(v)

        val dialog = PopupDialog(v)
        val inset = context.dp(16)
        var bgColor = ContextCompat.getColor(context, R.color.windowBackground)
        val contrast = ColorUtils.calculateContrast(bgColor, Color.BLACK)
        if (contrast < 2)
           bgColor = ColorUtils.compositeColors(0x10FFFFFF, bgColor)
        v.background = InsetDrawable(ColorDrawable(bgColor), inset, 0, inset, 0)
        dialog.elevation = context.dp(8).toFloat()
        v.measure(MeasureSpec.makeMeasureSpec(dm.widthPixels, MeasureSpec.AT_MOST), MeasureSpec.makeMeasureSpec(dialog.getMaxAvailableHeight(spinner), MeasureSpec.AT_MOST))
        dialog.width = v.measuredWidth
        dialog.height = v.measuredHeight
        val loc = IntArray(2)
        spinner.getLocationInWindow(loc)
        val minXOffset = context.dp(16)
        Timber.v("loc ${loc.contentToString()} ${dm.widthPixels} ${spinner.width} $minXOffset")
        var xOff = 0
        var yOff = 0
        if (loc[0] < minXOffset) {
            xOff = minXOffset - loc[0]
        } else if (loc[0] + spinner.width + minXOffset >= dm.widthPixels) {
            xOff = dm.widthPixels - loc[0]  - minXOffset
        }
        xOff += inset
        PopupWindowCompat.setOverlapAnchor(dialog, true)
        dialog.showAsDropDown(spinner, xOff, yOff)
        dialog.setOnDismissListener {
            val result = resultCallback?.invoke()
            if (result != null)
                setFromUser(result)
            this.dialog = null
        }
        this.dialog = dialog
        return dialog
    }

    override fun onCreateViewHolder(parent: ViewGroup, viewType: Int, clickListener: OnClickListener): SummaryListViewHolder {
        val holder = super.onCreateViewHolder(parent, viewType, clickListener)
        holder.fancyPref.findViewById<TextView>(R.id.title).updateLayoutParams<MarginLayoutParams> {
            updateMarginsRelative(start = 0, end = 0)
        }
        holder.fancyPref.findViewById<View>(R.id.widget_frame).isVisible = false
        return holder
    }

}

class PopupDialog(view: View) : PopupWindow(view), DialogInterface {
    init {
        isFocusable = true
        isOutsideTouchable = true
        setTouchInterceptor { v, event ->
            if (event.action == MotionEvent.ACTION_OUTSIDE) {
                dismiss()
                true
            }
            false
        }

    }

    override fun cancel() {
        dismiss()
    }

}