package com.teslacoilsw.launcher.preferences.fancyprefs

import android.annotation.SuppressLint
import android.content.Context
import android.graphics.drawable.Drawable
import android.util.AttributeSet
import android.view.LayoutInflater
import android.view.View
import android.view.View.OnClickListener
import android.view.ViewGroup
import android.widget.ImageView
import androidx.core.content.res.use
import androidx.core.view.updatePadding
import androidx.recyclerview.widget.LinearLayoutManager
import androidx.recyclerview.widget.RecyclerView
import com.teslacoilsw.kotlin.dp
import com.teslacoilsw.kotlin.forEachIndexed
import com.teslacoilsw.kotlin.getThemeAttrDrawable
import com.teslacoilsw.launcher.R

@SuppressLint("CustomViewStyleable")
abstract class FancyPrefSummaryListBaseView<T: Any> @JvmOverloads
constructor(context: Context, attrs: AttributeSet? = null, defStyleAttr: Int = 0)
    : FancyPrefDialogView<T>(context, attrs, defStyleAttr)
{
    var rowLayoutId: Int = R.layout.dialog_summary_list_row
    open var entries: List<SummaryListRowEntry> = emptyList()
        set(value) {
            val old = value
            field = value
            onEntriesUpdated(value)
        }
    var onSelectedInDialog: ((dialogContent: View, entry: SummaryListRowEntry) -> Unit)? = null
    private val inflater = LayoutInflater.from(context)

    init {
        if (dialogLayoutId == 0)
            dialogLayoutId = R.layout.dialog_summary_list

        val a = context.obtainStyledAttributes(attrs, R.styleable.FancyPrefSummaryListView)
        try {
            if (!requireConfirmation() && !a.getBoolean(R.styleable.FancyPrefSummaryListView_requireConfirmation, false))
                positiveButtonLabel = null

            val textEntries = a.getTextArray(R.styleable.FancyPrefSummaryListView_entries)
            val textSummaries = a.getTextArray(R.styleable.FancyPrefSummaryListView_entrySummaries)
            val values = a.getTextArray(R.styleable.FancyPrefSummaryListView_entryValues)

            val iconArrayResId = a.getResourceId(R.styleable.FancyPrefSummaryListView_entryIcons, 0)
            val icons: IntArray? = if (iconArrayResId != 0) {
                //noinspection Recycle
                a.resources.obtainTypedArray(iconArrayResId).use {
                    val arr = IntArray(it.length())
                    for (i in 0 until it.length()) {
                        arr[i] = it.getResourceId(i, 0)
                    }
                    arr
                }
            } else {
                null
            }

            val list = mutableListOf<SummaryListRowEntry>()
            for (i in 0 until textEntries.size) {
                val value = values[i].toString()
                val entry = textEntries[i]
                var summary = textSummaries?.get(i)
                val icon = icons?.get(i) ?: 0
                list.add(SummaryListRowEntry(value, entry, summary, icon))
            }
            entries = list
        } finally {
            a.recycle()
        }
    }

    fun entryForValue(value: String) = entries.find { it.value == value }

    override fun bindDialogView(dialogView: View): (() -> T)? {
        val recycler = dialogView.findViewById<RecyclerView>(R.id.recycler)
        recycler.layoutManager = LinearLayoutManager(context)

        recycler.adapter = object: RecyclerView.Adapter<SummaryListViewHolder>() {
            private val clickListener = OnClickListener { view ->
                val itemPosition = recycler.getChildLayoutPosition(view)
                val entry = entries[itemPosition]
                onSelectedInDialog(dialogView, entry)
                onSelectedInDialog?.invoke(dialogView, entry)
                updateChecked(recycler)

                if (positiveButtonLabel == null)
                    dismiss()
            }

            override fun onCreateViewHolder(parent: ViewGroup, viewType: Int): SummaryListViewHolder {
                return this@FancyPrefSummaryListBaseView.onCreateViewHolder(parent, viewType, clickListener)
            }

            override fun getItemCount(): Int = entries.size

            override fun onBindViewHolder(holder: SummaryListViewHolder, position: Int) {
                this@FancyPrefSummaryListBaseView.onBindViewHolder(holder, position)
            }

        }

        return null
    }

    open fun onEntriesUpdated(entries: List<SummaryListRowEntry>) { }

    abstract fun onSelectedInDialog(dialogView: View, entry: SummaryListRowEntry)

    open fun onCreateViewHolder(parent: ViewGroup, viewType: Int, clickListener: OnClickListener): SummaryListViewHolder {
        val view = inflater.inflate(rowLayoutId, parent, false)
        view.setOnClickListener(clickListener)
        val holder = SummaryListViewHolder(view)
        holder.fancyPref.checkMarkDrawable = context.getThemeAttrDrawable(android.R.attr.listChoiceIndicatorSingle)
        val dp8 = context.dp(8)
        holder.fancyPref.updatePadding(left = dp8, right = dp8)
        return holder
    }

    fun onBindViewHolder(holder: SummaryListViewHolder, position: Int) {
        val entry = entries[position]
        holder.fancyPref.title = entry.title
        holder.fancyPref.summary = entry.summary
        val drawable = entry.iconDrawable ?: if (entry.iconResId != 0) context.getDrawable(entry.iconResId) else null
        holder.fancyPref.icon = drawable
        holder.fancyPref.isChecked = isChecked(entry.value)
    }

    protected abstract fun isChecked(value: String): Boolean

    private fun updateChecked(recycler: RecyclerView) {
        recycler.forEachIndexed { index, view ->
            (recycler.getChildViewHolder(view) as SummaryListViewHolder).fancyPref.isChecked = isChecked(entries[index].value)
        }
    }

    open fun requireConfirmation() = false

}

data class SummaryListRowEntry(val value: String, val title: CharSequence, val summary: CharSequence? = null, val iconResId: Int = 0, val iconDrawable: Drawable? = null) {
    val hasIcon = iconDrawable != null || iconResId != 0
}
class SummaryListViewHolder(itemView: View) : RecyclerView.ViewHolder(itemView) {
    val fancyPref = itemView as FancyPrefCheckableView
    val icon = itemView.findViewById<ImageView>(R.id.icon)
}