package com.teslacoilsw.launcher.preferences.fancyprefs

import android.content.Context
import android.content.Intent
import android.content.res.ColorStateList
import android.content.res.TypedArray
import android.graphics.drawable.Drawable
import android.text.Html
import android.text.SpannableString
import android.text.SpannableStringBuilder
import android.text.method.LinkMovementMethod
import android.text.style.ClickableSpan
import android.util.AttributeSet
import android.view.LayoutInflater
import android.view.View
import android.view.ViewGroup
import android.widget.FrameLayout
import android.widget.ImageView
import android.widget.ScrollView
import android.widget.TextView
import androidx.annotation.DrawableRes
import androidx.constraintlayout.widget.ConstraintLayout
import androidx.core.content.res.use
import androidx.core.view.isVisible
import androidx.core.view.updatePadding
import androidx.core.widget.TextViewCompat
import com.afollestad.materialdialogs.MaterialDialog
import com.teslacoilsw.kotlin.dp
import com.teslacoilsw.kotlin.getActivity
import com.teslacoilsw.kotlin.getThemeAttrColor
import com.teslacoilsw.kotlin.getThemeAttrDrawable
import com.teslacoilsw.kotlin.lazy1
import com.teslacoilsw.launcher.R
import com.teslacoilsw.launcher.preferences.Pref
import com.teslacoilsw.launcher.preferences.Pref3
import com.teslacoilsw.launcher.preferences.SettingsActivity
import com.teslacoilsw.launcher.util.DpUtil
import com.teslacoilsw.licensing.ProKeyGoogleLicenseChecker
import com.teslacoilsw.tesladirect.DirectLicensingEnterCode
import com.teslacoilsw.tesladirect.MarketHandler
import rx.Observable
import rx.subscriptions.CompositeSubscription

enum class IncludeInSearch {
    NEVER,
    VISIBLE,
    ALWAYS
}
open class FancyPrefView<T: Any> @JvmOverloads
    constructor(context: Context, attrs: AttributeSet? = null, defStyleAttr: Int = 0)
    : ConstraintLayout(context, attrs, defStyleAttr)
{
    val iconView: ImageView?
    protected val titleView: TextView?
    protected val summaryView: TextView?
    private val info: ImageView?
    val widgetFrame: FrameLayout?
    private val requirePrime: Boolean
    private val includeInSearch: IncludeInSearch
    val searchKeywords: String
    var requireRestart: Boolean

    private val primeBanner by lazy1 { findViewById<View>(R.id.prime_banner) }

    var icon: Drawable?
        get() = iconView?.drawable
        set(value) {
            iconRes = -1
            if (value == null) {
                iconView?.setImageDrawable(null)
                iconView?.visibility = View.GONE
            } else {
                iconView?.setImageDrawable(value)
                iconView?.visibility = View.VISIBLE
            }
        }

    @DrawableRes
    var iconRes: Int = 0
        set(value) {
            if (value == -1) {
                field = 0
            } else {
                field = iconRes
                if (value == 0) {
                    icon = null
                } else {
                    iconView?.setImageResource(value)
                    iconView?.visibility = View.VISIBLE
                }
            }
        }

    open var title: CharSequence?
        get() = titleView?.text
        set(value) {
            titleView?.text = value
            titleView?.isVisible = value != null
        }

    fun setTitle(titleResId: Int) {
        title = context.getString(titleResId)
    }

    open var summary: CharSequence? = null
        set(value) {
            field = value
            updateSummary()
        }

    val summaryActualText: CharSequence?
        get() = summaryView?.text

    var onUserChanged: ((oldValue: T, newValue: T) -> Unit)? = null
    var infoMessage: String = ""

    val valueFieldInitialized: Boolean
        get() = ::valueField.isInitialized
    open lateinit var valueField: T
        protected set

    open var value: T
        get() = valueField
        set(value) {
            if (!valueFieldInitialized || value != valueField) {
                valueField = value
                onChange(value)
            }
        }
    private var pref: PrefHolder<T>? = null
    private var subscription: CompositeSubscription? = null
    private var enableCondition: Observable<Boolean>? = null
    private val attrBindKey: String?

    open fun getPreferenceLayoutId(attrs: AttributeSet?): Int {
        //noinspection Recycle
        return context.obtainStyledAttributes(attrs, R.styleable.FancyPrefView).use {
            it.getResourceId(R.styleable.FancyPrefView_preferenceRowLayout, R.layout.fancypref_row)
        }
    }

    init {
        if (background == null)
            background = context.getThemeAttrDrawable(R.attr.selectableItemBackground)
        minimumHeight = context.dp(64)

        LayoutInflater.from(context).inflate(getPreferenceLayoutId(attrs), this, true)
        iconView = findViewById(R.id.icon)
        titleView = findViewById(R.id.title)
        summaryView = findViewById(R.id.summary)
        info = findViewById(R.id.info)
        widgetFrame = findViewById(R.id.widget_frame)

        val a = context.obtainStyledAttributes(attrs, R.styleable.FancyPrefView)
        try {
            title = a.getString(R.styleable.FancyPrefView_android_text)
            summary = a.getString(R.styleable.FancyPrefView_android_summary)
            val styleId = a.getResourceId(R.styleable.FancyPrefView_labelStyle, 0)
            if (styleId != 0)
                TextViewCompat.setTextAppearance(titleView!!, styleId)
            infoMessage = a.getString(R.styleable.FancyPrefView_info) ?: ""
            iconRes = a.getResourceId(R.styleable.FancyPrefView_android_icon, 0)
            if (!a.hasPadding()) {
                val dp8 = context.dp(8)
                updatePadding(top = dp8, bottom = dp8)
            }
            requirePrime = a.getBoolean(R.styleable.FancyPrefView_requirePrime, false);
            if (!Pref.Value.nova_prime_enabled && requirePrime) {
                refreshPrimeState()
            }
            includeInSearch = IncludeInSearch.values()[a.getInt(R.styleable.FancyPrefView_includeInSearch, 1)]
            searchKeywords = a.getString(R.styleable.FancyPrefView_searchKeywords) ?: ""
            attrBindKey = a.getString(R.styleable.FancyPrefView_bind)
            requireRestart = a.getBoolean(R.styleable.FancyPrefView_requireRestart, false)
        } finally {
            a.recycle()
        }

        if (!infoMessage.isBlank()) {
            info!!.imageAlpha = 192
            info.visibility = View.VISIBLE
            info.setOnClickListener {
                MaterialDialog.Builder(context)
                        .content(infoMessage)
                        .show()
            }
            setOnLongClickListener {
                info.callOnClick()
            }
        }

        refreshPrimeState()
    }

    fun refreshPrimeState() {
        if (requirePrime) {
            val hasPrime = Pref.Value.nova_prime_enabled
            primeBanner.isVisible = !hasPrime
            titleView?.isDuplicateParentStateEnabled = hasPrime
            titleView?.isEnabled = hasPrime
            summaryView?.isDuplicateParentStateEnabled = hasPrime
            summaryView?.isEnabled = hasPrime
            iconView?.imageTintList = if (!hasPrime) ColorStateList.valueOf(0xFFA5A5A5.toInt()) else ColorStateList.valueOf(context.getThemeAttrColor(R.attr.colorControlActivated))
            widgetFrame?.isDuplicateParentStateEnabled = hasPrime
            widgetFrame?.isEnabled = hasPrime
        }
    }

    override fun onFinishInflate() {
        super.onFinishInflate()
        if (minimumHeight == 0)
            minimumHeight = context.dp(56)
        if (attrBindKey != null)
            bindFromAttr(attrBindKey)
    }

    open fun summaryFormat(text: String, value: T) =
        String.format(text, value)

    open fun updateSummary() {
        val s = summary
        val text = if (s != null && s.contains('%') && valueFieldInitialized) {
            summaryFormat(s.toString(), value)
        } else {
            s
        }
        updateSummaryViews(text)
    }

    protected fun updateSummaryViews(text: CharSequence?) {
        summaryView?.text = text
        summaryView?.visibility = if (text == null) View.GONE else View.VISIBLE
        widgetFrame?.minimumHeight = if (text == null) context.dp(32) else context.dp(48)
    }

    fun setWidget(layoutId: Int) {
        widgetFrame!!.removeAllViews()
        LayoutInflater.from(context).inflate(layoutId, widgetFrame, true)
    }


    override fun onAttachedToWindow() {
        super.onAttachedToWindow()
        updateSubscription()
    }

    override fun onDetachedFromWindow() {
        super.onDetachedFromWindow()
        unsubscribe()
    }

    private fun unsubscribe() {
        subscription?.unsubscribe()
        subscription = null
    }

    private fun updateSubscription() {
        if (isAttachedToWindow) {
            val s = subscription ?: CompositeSubscription()
            subscription = s
            if (pref != null)
                s.add(pref!!.asObservable!!.subscribe {
                    value = it
                })
            if (enableCondition != null)
                s.add(enableCondition!!.subscribe {
                    isEnabled = it
                })
        }
    }

    override fun setOnClickListener(l: OnClickListener?) {
        if (requirePrime) {
            super.setOnClickListener(PrimeClickListenerWrapper(l))
        } else {
            super.setOnClickListener(l)
        }
    }

    fun setEnableCondition(observable: Observable<Boolean>) {
        if (enableCondition != null)
            throw IllegalStateException("FancyPref enableCondition may only be set once")

        enableCondition = observable
        updateSubscription()
    }

    fun bind(pref: PrefHolder<T>?) {
        if (this.pref != null)
            throw IllegalStateException("FancyPref Pref may only be set once")
        unsubscribe()
        this.pref = pref
        if (pref != null)
            value = pref.storedValue
        updateSubscription()
    }

    protected open fun bindFromAttr(propName: String) {
        val x = Pref3.get(propName)
        when (x) {
            is Pref3.PrefEnum -> bind(x.asStringPrefHolder() as PrefHolder<T>)
            is Pref3.PrefCofs -> bind(x.asStringPrefHolder() as PrefHolder<T>)
            else -> bind(x as PrefHolder<T>)
        }
    }

    open fun onChange(value: T) {
        pref?.storedValue = value
        if (summary?.contains('%') == true)
            updateSummary()
    }

    open fun setFromUser(newValue: T) {
        val oldValue = value
        value = newValue
        onUserChanged?.invoke(oldValue, newValue)
        if (requireRestart)
            (context.getActivity() as SettingsActivity).mNeedLauncherRestart = true
    }

    fun includeInSearch() = includeInSearch == IncludeInSearch.ALWAYS ||
            includeInSearch == IncludeInSearch.VISIBLE && visibility == View.VISIBLE && parentVisible()

    private fun View.parentVisible(): Boolean {
        val parent = this.parent ?: return true
        if (parent is ScrollView)
            return true
        else if (parent is ViewGroup) {
            return parent.visibility == View.VISIBLE && parent.parentVisible()
        }
        throw IllegalArgumentException("View.parentVisible must be called with an actual parent")
    }

    private fun TypedArray.hasPadding() =
        hasValue(R.styleable.FancyPrefView_android_padding) ||
        hasValue(R.styleable.FancyPrefView_android_paddingTop) ||
        hasValue(R.styleable.FancyPrefView_android_paddingBottom) ||
        hasValue(R.styleable.FancyPrefView_android_paddingLeft) ||
        hasValue(R.styleable.FancyPrefView_android_paddingRight) ||
        hasValue(R.styleable.FancyPrefView_android_paddingStart) ||
        hasValue(R.styleable.FancyPrefView_android_paddingEnd)
}

private class PrimeClickListenerWrapper(private val listener: View.OnClickListener?) : View.OnClickListener {
    override fun onClick(v: View) {
        if (!Pref.Value.nova_prime_enabled) {
            val context = v.context
            val messageView = TextView(context)
            val dp8 = DpUtil.getPixelSize(8)
            messageView.setPadding(dp8, dp8, dp8, dp8)
            messageView.movementMethod = LinkMovementMethod.getInstance()
            val d = MaterialDialog.Builder(context)
                    .title(R.string.nova_prime)
                    .customView(messageView, true)
                    .positiveText(R.string.market)
                    .negativeText(R.string.cancel)
                    .callback(object : MaterialDialog.ButtonCallback() {
                        override fun onPositive(materialDialog: MaterialDialog?) {
                            val intent = MarketHandler.getMarketLink(MarketHandler.GOOGLE, ProKeyGoogleLicenseChecker.PACKAGE)
                            intent.addFlags(Intent.FLAG_ACTIVITY_NEW_TASK)
                            context.startActivity(intent)
                        }
                    })
                    .build()

            val directLicense = SpannableString("If you have a direct license enter the authorization code here.")
            val directLicenseClickable = object : ClickableSpan() {
                override fun onClick(widget: View) {
                    val intent = Intent(context, DirectLicensingEnterCode::class.java)
                    context.startActivity(intent)
                    d.dismiss()
                }
            }
            directLicense.setSpan(directLicenseClickable, 29, directLicense.length - 1, 0)
            val message = SpannableStringBuilder().append("Upgrade to Nova Launcher Prime for more features including:\n\n" +
                    "- Gestures. Pinch, Swipe, Two Finger Swipe, Rotate and Double-tap.\n\n" +
                    "- Group and Hide Apps. Avoid a cluttered App Drawer, create as many folders or drawer tabs as you'd like to group your apps. Or hide them completely.\n\n" +
                    "- Dock swipe. Swipe up on dock icons for alternative action.\n\n" +
                    "- More Scroll Effects.\n\n")
                    .append(Html.fromHtml("<b>Previously purchased Nova Launcher Prime from the Play Store?</b><br /> Reinstall by pressing the Play Store button below and clicking the Install button. If it asks you to purchase again make sure you are using the same Google account that you purchased with.<br /><br />")
                    ).append(directLicense)

            messageView.text = message

            d.show()
        } else {
            listener?.onClick(v)
        }
    }

}