package com.teslacoilsw.launcher.preferences.fancyprefs

import android.content.Context
import android.graphics.Canvas
import android.util.AttributeSet
import android.view.View
import android.view.ViewGroup
import android.view.WindowInsets
import com.teslacoilsw.launcher.R
import com.teslacoilsw.launcher.util.DpUtil
import com.teslacoilsw.launcher.widget.RippleScrollView
import kotterknife.bindView

class FancyPreviewLayout @JvmOverloads
constructor(context: Context, attrs: AttributeSet? = null, defStyleAttr: Int = 0)
    : ViewGroup(context, attrs, defStyleAttr)
{
    private val preview by bindView<View>(R.id.previewFrame)
    private val mainScrollView by bindView<RippleScrollView>(R.id.mainView)
    private val shadowDrawable by lazy { context.getDrawable(R.drawable.divider_shadow_gradient_top) }
    private var topInset = 0
    private var bottomInset = 0
    private var previewExtra = 0
    private var recomputeExtra = true

    override fun onApplyWindowInsets(insets: WindowInsets): WindowInsets {
        topInset = insets.systemWindowInsetTop
        bottomInset = insets.systemWindowInsetBottom
        updateMeasure()
        return super.onApplyWindowInsets(insets)
    }

    override fun onMeasure(widthMeasureSpec: Int, heightMeasureSpec: Int) {
        val oldMeasureHeight = measuredHeight
        super.onMeasure(widthMeasureSpec, heightMeasureSpec)
        if (measuredHeight != oldMeasureHeight)
            recomputeExtra = true
        updateMeasure()
    }

    private fun updateMeasure() {
        val widthSpec = MeasureSpec.makeMeasureSpec(measuredWidth, MeasureSpec.EXACTLY)
        val peekHeight = DpUtil.getPixelOffset(24)
        var availableHeight = measuredHeight - preview.minimumHeight - mainScrollView.paddingTop - peekHeight

        mainScrollView.measure(MeasureSpec.makeMeasureSpec(measuredWidth, MeasureSpec.EXACTLY), MeasureSpec.makeMeasureSpec(availableHeight, MeasureSpec.EXACTLY))
        val mainLayout = mainScrollView.getChildAt(0) as ViewGroup

        /* To make scrolling obvious we make sure the last view is not fully visible.
         * Also don't want to bisect because some rows might line up there.
         * Instead ensure that either only the top 24dp are showing or the bottom 24dp are not
         * showing.
         * For example on Nextbit Robin Folder Window Style should show a peek of "Corner Radius"
         * and Folder Icon Appearance cuts off the bottom text of Background Shape
         */
        if (recomputeExtra) {
            for (i in 0 until mainLayout.childCount) {
                val v = mainLayout.getChildAt(i)
                val lp = v.layoutParams
                val viewHeight = when {
                    v.visibility == View.GONE -> 0
                    lp is MarginLayoutParams -> v.measuredHeight + lp.topMargin + lp.bottomMargin
                    else -> v.measuredHeight
                }
                if (availableHeight - viewHeight <= 0) {
                    previewExtra = if (viewHeight - peekHeight - availableHeight < peekHeight) {
                        peekHeight - (viewHeight - peekHeight - availableHeight)
                    } else {
                        availableHeight
                    }
                    break
                } else {
                    availableHeight -= viewHeight
                }
            }
        }

        val previewHeight = preview.minimumHeight + previewExtra

        preview.measure(widthSpec, MeasureSpec.makeMeasureSpec(previewHeight, MeasureSpec.EXACTLY))
        mainScrollView.measure(widthSpec, MeasureSpec.makeMeasureSpec(measuredHeight - previewHeight, MeasureSpec.EXACTLY))
        shadowDrawable.setBounds(0, previewHeight, measuredWidth, previewHeight + DpUtil.getPixelSize(8))
        recomputeExtra = false
    }

    override fun onLayout(changed: Boolean, l: Int, t: Int, r: Int, b: Int) {
        val w = r - l
        val h = b - t
        preview.layout(0, 0, w, preview.minimumHeight + previewExtra)
        mainScrollView.layout(0, preview.minimumHeight + previewExtra, w, h)
    }

    override fun drawChild(canvas: Canvas?, child: View?, drawingTime: Long): Boolean {
        val out = super.drawChild(canvas, child, drawingTime)
        if (child == mainScrollView) {
            shadowDrawable.draw(canvas)
        }
        return out
    }

    fun forceRecompute() {
        recomputeExtra = true
        updateMeasure()
    }
}