/*
 * Copyright (C) 2013 The Android Open Source Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.teslacoilsw.launcher.preferences.fancyprefs.color

import android.annotation.SuppressLint
import android.app.Dialog
import android.content.DialogInterface
import android.content.res.ColorStateList
import android.graphics.Color
import android.graphics.drawable.LayerDrawable
import android.graphics.drawable.RippleDrawable
import android.graphics.drawable.ShapeDrawable
import android.graphics.drawable.shapes.OvalShape
import android.os.Bundle
import android.preference.PreferenceManager
import android.text.Spannable
import android.text.SpannableString
import android.text.style.ImageSpan
import android.view.LayoutInflater
import android.view.View
import android.view.ViewGroup
import android.view.inputmethod.EditorInfo
import android.widget.TextView
import androidx.annotation.ColorInt
import androidx.appcompat.app.AppCompatActivity
import aosp.android.graphics.drawable.InsetDrawableProportional
import com.afollestad.materialdialogs.MaterialDialog
import com.google.android.material.bottomsheet.BottomSheetDialogFragment
import com.teslacoilsw.colornames.ColorNames
import com.teslacoilsw.kotlin.get
import com.teslacoilsw.kotlin.lazy1
import com.teslacoilsw.kotlin.showKeyboard
import com.teslacoilsw.kotlin.size
import com.teslacoilsw.launcher.R
import com.teslacoilsw.launcher.preferences.COLOR_AUTO
import com.teslacoilsw.launcher.preferences.fancyprefs.FancyPrefCheckableView
import com.teslacoilsw.launcher.preferences.fancyprefs.FancyPrefGradientSeekBarView
import com.teslacoilsw.launcher.preferences.fancyprefs.FancyPrefIconView
import com.teslacoilsw.launcher.preferences.fancyprefs.FancyPrefSeekBarView
import com.teslacoilsw.launcher.util.palette.ColorUtils
import com.teslacoilsw.launcher.widget.DumbRadioGrid
import com.teslacoilsw.shared.colorpicker.ColorUtil
import kotlinx.android.synthetic.main.fancypref_color_hex_input_dialog.*
import kotlinx.android.synthetic.main.fancypref_color_picker_bottomsheet.*
import kotlinx.android.synthetic.main.fancypref_color_picker_dialog.view.*
import kotlin.math.roundToInt

/**
 * A dialog which takes in as input an array of colors and creates a palette allowing the user to
 * select a specific color swatch, which invokes a listener.
 */
class ColorPickerBottomSheet : BottomSheetDialogFragment() {

    private lateinit var alertDialog: Dialog

    private lateinit var recentColors: IntArray
    private var title: String? = null
    private var columns: Int = 0
    private var size: Int = 0
    private var showAlpha: Boolean = false
    private var showManage: Boolean = false
    private var defaultColor: Int = 0

    private lateinit var mainView: View
    private val paletteGrid by lazy1 { mainView.findViewById<DumbRadioGrid>(R.id.palette_radiogrid) }
    private val recentGrid by lazy1 { mainView.findViewById<DumbRadioGrid>(R.id.recent_radiogrid) }
    private val recentLabel by lazy1 { mainView.findViewById<TextView>(R.id.color_picker_recent_label) }
    private val transparency by lazy1 { mainView.findViewById<FancyPrefSeekBarView>(R.id.transparency) }

    private val gotoAdvanced by lazy1 { mainView.findViewById<FancyPrefIconView>(R.id.goto_advanced) }

    private val paletteFrame by lazy1 { mainView.findViewById<ViewGroup>(R.id.palette_frame) }
    private val advancedFrame by lazy1 { mainView.findViewById<ViewGroup>(R.id.advanced_frame) }

    private val colorPreviewFrame by lazy1 { mainView.findViewById<View>(R.id.color_preview_frame) }
    private val colorPreviewHex by lazy1 { mainView.findViewById<TextView>(R.id.color_preview_hex) }
    private val colorPreviewName by lazy1 { mainView.findViewById<TextView>(R.id.color_preview_name) }
    private var recentSwatchFromAdvanced: ColorSwatchRadioButton? = null

    private val managedByNova by lazy1 { mainView.findViewById<FancyPrefCheckableView>(R.id.managed_by_nova) }

    private val previewDrawable by lazy1 { ColorStateDrawable(resources, Color.WHITE, 0f, PreviewShape.Rect) }

    private val tempHsv = FloatArray(3)

    private var inColorSetter = false
    private var color: Int = 0
        set(@ColorInt newColor) {
            if (inColorSetter)
                return
            inColorSetter = true
            try {
                field = newColor
                if (::mainView.isInitialized) {
                    updatePreview()
                }
            } finally {
                inColorSetter = false
            }
        }

    private val componentHue by lazy1 { mainView.findViewById<FancyPrefGradientSeekBarView>(R.id.component_hue) }
    private val componentSaturation by lazy1 { mainView.findViewById<FancyPrefGradientSeekBarView>(R.id.component_saturation) }
    private val componentValue by lazy1 { mainView.findViewById<FancyPrefGradientSeekBarView>(R.id.component_value) }

    private var listener: ((Int) -> Unit)? = null

    lateinit var colors: IntArray

    private var discardOnDismiss = false

    fun initialize(title: String?, colors: IntArray, selectedColor: Int, columns: Int, showAlpha: Boolean, showManage: Boolean, defaultColor: Int) {
        setArguments(title, columns, size, showAlpha, showManage, defaultColor)
        setColors(colors, selectedColor)
    }

    fun show(activity: AppCompatActivity) {
        show(activity.supportFragmentManager, "colorPickerDialog")
    }

    fun setArguments(title: String?, columns: Int, size: Int, showAlpha: Boolean, showManage: Boolean, defaultColor: Int) {
        val bundle = Bundle()
        if (title != null)
            bundle.putString(KEY_TITLE, title)
        bundle.putInt(KEY_COLUMNS, columns)
        bundle.putInt(KEY_SIZE, size)
        bundle.putBoolean(KEY_SHOW_ALPHA, showAlpha)
        bundle.putBoolean(KEY_SHOW_MANAGE, showManage)
        bundle.putInt(KEY_DEFAULT_COLOR, defaultColor)
        arguments = bundle
    }

    fun setOnColorSelectedListener(listener: ((Int) -> Unit)?) {
        this.listener = listener
    }

    override fun onCreate(savedInstanceState: Bundle?) {
        super.onCreate(savedInstanceState)

        arguments?.let {
            title = it.getString(KEY_TITLE)
            columns = it.getInt(KEY_COLUMNS)
            size = it.getInt(KEY_SIZE)
            showAlpha = it.getBoolean(KEY_SHOW_ALPHA)
            showManage = it.getBoolean(KEY_SHOW_MANAGE)
            defaultColor = it.getInt(KEY_DEFAULT_COLOR)
        }

        if (savedInstanceState != null) {
            colors = savedInstanceState.getIntArray(KEY_COLORS)
            color = savedInstanceState.getSerializable(KEY_SELECTED_COLOR) as Int
        }

        val recentColorStr = PreferenceManager.getDefaultSharedPreferences(activity).getString(PREF_KEY_RECENT_COLORS, "")!!
        if (!recentColorStr.isEmpty()) {
            val split = recentColorStr.split(',').dropLastWhile { it.isEmpty() }.toTypedArray()
            val size = Math.min(RECENT_COLOR_COUNT, split.size)
            recentColors = IntArray(size)
            var i = 0
            for (color in split) {
                try {
                    recentColors[i] = Integer.parseInt(color)
                } catch (ex: NumberFormatException) {
                    /* Invalid value, use red */
                    recentColors[i] = Color.RED
                }

                i++
                if (i >= size) break
            }
        } else {
            recentColors = IntArray(0)
        }
    }

    private fun ViewGroup.getViewIdForSwatchColor(@ColorInt color: Int): Int {
        for (i in 0 until size) {
            val v = this[i]
            if (v is ColorSwatchRadioButton) {
                if (ColorUtil.argb(0xFF, v.color) == ColorUtil.argb(0xFF, color))
                    return v.id
            }
        }
        return -1
    }

    private @ColorInt fun getColorFromViewId(viewGroup: ViewGroup, viewId: Int): Int {
        viewGroup.findViewById<View>(viewId)?.let {
            if (it is ColorSwatchRadioButton)
                return it.color
        }
        throw IllegalStateException("Trying to getColorFromViewId for a viewId that does not exist or is not a ColorSwatchRadioButton")
    }

    private @ColorInt fun getColorFromHsvSliders(): Int {
        val hsv = tempHsv
        hsv[0] = componentHue.progress.toFloat()
        hsv[1] = componentSaturation.progress/ 100f
        hsv[2] = componentValue.progress/100f
        return Color.HSVToColor(hsv)
    }

    @SuppressLint("SetTextI18n")

    override fun onCreateDialog(savedInstanceState: Bundle?): Dialog {
        val activity = activity
        val inflater = LayoutInflater.from(getActivity())

        @SuppressLint("InflateParams")
        mainView = inflater.inflate(R.layout.fancypref_color_picker_bottomsheet, null)

        mainView.title.text = title
        for (c in colors) {
            val swatch = ColorSwatchRadioButton(activity!!, null)
            swatch.id = View.generateViewId()
            swatch.color = ColorUtil.argb(0xFF, c)
            paletteGrid.addView(swatch)
        }

        for (c in recentColors) {
            val swatch = ColorSwatchRadioButton(activity!!, null)
            swatch.id = View.generateViewId()
            swatch.color = ColorUtil.argb(0xFF,c)
            recentGrid.addView(swatch)
        }

        val gridListener = object: DumbRadioGrid.OnCheckedChangeListener {
            override fun onCheckedChanged(group: DumbRadioGrid, checkedId: Int) {
                if (checkedId != -1)
                    color = getColorFromViewId(group, checkedId)
            }
        }
        paletteGrid.setOnCheckedChangeListener(gridListener)
        recentGrid.setOnCheckedChangeListener(gridListener)

        if (showAlpha) {
            transparency.progress = 255 - Color.alpha(color)
            transparency.onUserChanged = ::updatePreview
        } else {
            transparency.visibility = View.GONE
            transparency.progress = 0
        }

        showPaletteView()

        colorPreviewName.setOnClickListener {
            color = ColorNames.getBestName(color).rgb
        }
        colorPreviewFrame.background = previewDrawable
        colorPreviewHex.setOnClickListener {
            fun saveResult(text: String) {
                if (!text.isBlank()) {
                    try {
                        @ColorInt val x = ColorUtil.convertToColorInt(text)
                        color = x
                    } catch (e: NumberFormatException) { }
                } }
            val md = MaterialDialog.Builder(activity!!)
                    .customView(R.layout.fancypref_color_hex_input_dialog, false)
                    .negativeText(R.string.cancel)
                    .positiveText(R.string.set)
                    .onPositive { dialog, _ ->
                        saveResult(dialog.editText.text?.toString() ?: "")
                    }.build()

            val editText = md.editText
            editText.imeOptions = EditorInfo.IME_ACTION_DONE
            editText.setText(ColorUtil.convertToRGB(color))

            editText.setOnEditorActionListener { _, actionId, _ ->
                if (actionId == EditorInfo.IME_ACTION_DONE) {
                    saveResult(editText.text?.toString() ?: "")
                    md.dismiss()
                    true
                } else {
                    false
                }
            }

            md.show()
            editText.apply { post {
                requestFocus()
                selectAll()
                showKeyboard()
            } }
        }


        gotoAdvanced.setOnClickListener {
            paletteFrame.visibility = View.INVISIBLE
            advancedFrame.visibility = View.VISIBLE
        }

//        colorPreviewBack.setOnClickListener {
//            if (!managedByNova.isChecked && paletteGrid.checkedRadioButtonId == -1 && recentGrid.checkedRadioButtonId == -1) {
//                if (recentSwatchFromAdvanced == null) {
//                    /* Switching from advanced to palette, need to add a fake recent to indicate the
//                     * selected color
//                     */
//                    if (recentGrid.size >= RECENT_COLOR_COUNT)
//                        recentGrid.removeViewAt(RECENT_COLOR_COUNT - 1)
//
//                    val swatch = ColorSwatchRadioButton(activity!!, null)
//                    swatch.id = View.generateViewId()
//                    recentGrid.addView(swatch, 0)
//                    recentSwatchFromAdvanced = swatch
//                }
//                recentSwatchFromAdvanced!!.let {
//                    it.color = color
//                    recentGrid.check(if (managedByNova.isChecked) -1 else it.id)
//                }
//            }
//            paletteFrame.visibility = View.VISIBLE
//            advancedFrame.visibility = View.GONE
//        }

        val componentListener = { _: Int , _: Int -> color = getColorFromHsvSliders() }

        val hsv = tempHsv
        val hueColors = IntArray(361)
        for (i in 0 until hueColors.size) {
            hsv[0] = i.toFloat()
            hsv[1] = 1.0f
            hsv[2] = 1.0f
            hueColors[i] = Color.HSVToColor(hsv)
        }

        componentHue.setColors(hueColors)
        componentHue.onUserChanged = componentListener
        componentSaturation.onUserChanged = componentListener
        componentValue.onUserChanged = componentListener

        if (showManage) {
            managedByNova.isChecked = color == COLOR_AUTO
            managedByNova.visibility = View.VISIBLE
            managedByNova.onUserChanged = ::updatePreview
        } else {
            managedByNova.isChecked = false
        }

        updatePreview()


        alertDialog = super.onCreateDialog(savedInstanceState)
        alertDialog.setContentView(mainView)
//        alertDialog = MaterialDialog.Builder(activity!!)
//                .customView(mainView, false)
//                .negativeText(R.string.cancel)
//                .positiveText(R.string.set)
//                .onAny { _, which ->
//                    var finalColor = ColorUtil.argb(255 - transparency.progress, color)
//                    if (which == DialogAction.POSITIVE) {
//                        if (managedByNova.isChecked) {
//                            listener?.invoke(COLOR_AUTO)
//                        } else {
//                            if (finalColor == COLOR_AUTO)
//                                finalColor--
//                            appendAndSaveRecent(finalColor)
//                            listener?.invoke(finalColor)
//                        }
//                    }
//                }
//                .build()


        alertDialog.done.setOnClickListener { dismiss() }
        alertDialog.discard.setOnClickListener {
            if (advancedFrame.visibility == View.VISIBLE) {
                if (!managedByNova.isChecked && paletteGrid.checkedRadioButtonId == -1 && recentGrid.checkedRadioButtonId == -1) {
                    if (recentSwatchFromAdvanced == null) {
                        /* Switching from advanced to palette, need to add a fake recent to indicate the
                         * selected color
                         */
                        if (recentGrid.size >= RECENT_COLOR_COUNT)
                            recentGrid.removeViewAt(RECENT_COLOR_COUNT - 1)

                        val swatch = ColorSwatchRadioButton(activity!!, null)
                        swatch.id = View.generateViewId()
                        recentGrid.addView(swatch, 0)
                        recentSwatchFromAdvanced = swatch
                    }
                    recentSwatchFromAdvanced!!.let {
                        it.color = color
                        recentGrid.check(if (managedByNova.isChecked) -1 else it.id)
                    }
                }
                paletteFrame.visibility = View.VISIBLE
                advancedFrame.visibility = View.INVISIBLE
            } else {
                discardOnDismiss = true
                dismiss()
            }
        }
        alertDialog.setOnShowListener { updatePreview() }

        return alertDialog
    }

    override fun onDismiss(dialog: DialogInterface?) {
        super.onDismiss(dialog)
        if (!discardOnDismiss) {
            if (managedByNova.isChecked) {
                listener?.invoke(COLOR_AUTO)
            } else {
                var finalColor = ColorUtil.argb(255 - transparency.progress, color)
                if (finalColor == COLOR_AUTO)
                    finalColor--
                appendAndSaveRecent(finalColor)
                listener?.invoke(finalColor)
            }
        }
    }

    private fun updatePreview(old: Any = Unit, new: Any = Unit) {
        componentHue.isEnabled = !managedByNova.isChecked
        componentSaturation.isEnabled = !managedByNova.isChecked
        componentValue.isEnabled = !managedByNova.isChecked
        paletteGrid.isEnabled = !managedByNova.isChecked
        recentGrid.isEnabled = !managedByNova.isChecked
        colorPreviewHex.isEnabled = !managedByNova.isChecked
        gotoAdvanced.isEnabled = !managedByNova.isChecked

        @ColorInt val previewColor =  if (managedByNova.isChecked) Color.WHITE else color
//        val (_, r, g, b) = previewColor
        if (getColorFromHsvSliders() != previewColor) {
            val hsv = tempHsv
            Color.colorToHSV(previewColor, hsv)
            if (hsv[2] != 0f) {
                if (hsv[1] != 0f)
                    componentHue.progress = hsv[0].roundToInt()
                componentSaturation.progress = (100f * hsv[1]).roundToInt()
            }
            componentValue.progress = (100f * hsv[2]).roundToInt()
        }

        val hsv = tempHsv
        hsv[0] = componentHue.progress.toFloat()
        hsv[2] = (componentValue.progress / 100f)
        hsv[1] = 0f
        @ColorInt val sat0 = Color.HSVToColor(hsv)
        hsv[1] = 1f
        @ColorInt val sat1 = Color.HSVToColor(hsv)

        hsv[1] = (componentSaturation.progress/100f)
        hsv[2] = 0f
        @ColorInt val val0 = Color.HSVToColor(hsv)
        hsv[2] = 1f
        @ColorInt val val1 = Color.HSVToColor(hsv)

        hsv[0] = componentHue.progress.toFloat()
        hsv[1] = 1f
        hsv[2] = 1f
        componentHue.setThumbColor(Color.HSVToColor(hsv))
        componentSaturation.setThumbColor(previewColor)
        componentValue.setThumbColor(previewColor)

        componentSaturation.setColors(sat0, sat1)
        componentValue.setColors(val0, val1)

        val alpha = if (transparency.visibility == View.VISIBLE)
            255 - transparency.progress
        else
            255

        if (managedByNova.isChecked) {
            paletteGrid.check(-1)
            recentGrid.check(-1)
        } else {
            paletteGrid.check(paletteGrid.getViewIdForSwatchColor(previewColor))
            recentGrid.check(recentGrid.getViewIdForSwatchColor(previewColor))
            for (i in 0 until paletteGrid.size) {
                val child = paletteGrid[i]
                if (child is ColorSwatchRadioButton) {
                    child.color = ColorUtil.argb(alpha, child.color)
                    child.invalidate()
                }
            }

            for (i in 0 until recentGrid.size) {
                val child = recentGrid[i]
                if (child is ColorSwatchRadioButton) {
                    child.color = ColorUtil.argb(alpha, child.color)
                    child.invalidate()
                }
            }
        }

        @ColorInt val opaqueColor = ColorUtil.compositeColors(ColorUtil.argb(alpha, previewColor), Color.WHITE)
        val rippleColor: ColorStateList
        @ColorInt val colorPreviewForegroundColor: Int

        val darkFg = 0xFF212121.toInt()
        if (ColorUtils.calculateContrast(Color.WHITE, opaqueColor) >= ColorUtil.calculateContrast(darkFg, opaqueColor)) {
            colorPreviewForegroundColor = Color.WHITE
            rippleColor = ColorStateList.valueOf(ColorUtil.argb(0x40, Color.WHITE))
        } else {
            colorPreviewForegroundColor = darkFg
            rippleColor = ColorStateList.valueOf(ColorUtil.argb(0x40, darkFg))
        }

        colorPreviewName.setTextColor(colorPreviewForegroundColor)
        colorPreviewHex.setTextColor(colorPreviewForegroundColor)
//        colorPreviewBack.imageTintList = ColorStateList.valueOf(colorPreviewForegroundColor)
        ((colorPreviewHex.background as LayerDrawable).getDrawable(0) as RippleDrawable).setColor(rippleColor)
        (colorPreviewHex.background as LayerDrawable).getDrawable(1).setTint(colorPreviewForegroundColor)
//        (colorPreviewBack.background as RippleDrawable).setColor(rippleColor)
        (colorPreviewName.background as RippleDrawable).setColor(rippleColor)

        @SuppressLint("SetTextI18n")
        colorPreviewHex.text = "#${ColorUtil.convertToRGB(previewColor)}"
        val colorName = ColorNames.getBestName(previewColor)

        val inset = 0.15f
        val size = (colorPreviewName.textSize * (1f+inset)).roundToInt()
        var str = "${colorName.name}${if (colorName.deltaE > 1) "*" else ""} "
        if (colorPreviewName.paint.measureText(str) + size + colorPreviewName.paddingLeft + colorPreviewName.paddingRight > colorPreviewName.measuredWidth) {
           str = str.replaceFirst(' ', '\n')
        }

        colorPreviewName.text = SpannableString(str + "#").apply {
                setSpan(ImageSpan(InsetDrawableProportional(ShapeDrawable(OvalShape()).apply {
                paint.color = colorName.rgb
            }, inset, 0f, 0f, inset).apply {
                    setBounds(0, 0, size, size)
                }, ImageSpan.ALIGN_BOTTOM), length-1, length, Spannable.SPAN_EXCLUSIVE_EXCLUSIVE)
        }

        previewDrawable.color = ColorUtil.argb(alpha, previewColor)
    }

    private fun colorNoAlphaEqual(colorA: Int, colorB: Int): Boolean {
        return colorA and 0xFFFFFF == colorB and 0xFFFFFF
    }

    private fun appendAndSaveRecent(_color: Int) {
        var colorToSave = _color
        colorToSave = Color.argb(255, Color.red(colorToSave), Color.green(colorToSave), Color.blue(colorToSave))
        val sb = StringBuilder()
        sb.append(colorToSave)
        var count = 1
        for (i in recentColors.indices) {
            if (!colorNoAlphaEqual(recentColors[i], colorToSave)) {
                sb.append(",").append(recentColors[i])
                count++
                if (count > RECENT_COLOR_COUNT) {
                    break
                }
            }
        }

        val context = activity
        if (context != null) {
            PreferenceManager.getDefaultSharedPreferences(context).edit()
                    .putString(PREF_KEY_RECENT_COLORS, sb.toString()).apply()
        }
    }

    private fun showPaletteView() {
        paletteGrid.visibility = View.VISIBLE
        if (recentColors.isNotEmpty()) {
            recentGrid.visibility = View.VISIBLE
            recentLabel.visibility = View.VISIBLE
        }

        if (showAlpha) {
            transparency.visibility = View.VISIBLE
        }
    }

    fun setColors(_colors: IntArray, selectedColor: Int) {
        if (!::colors.isInitialized || !colors.contentEquals(_colors) || selectedColor != color) {
            colors = _colors
            color = selectedColor
        }
    }

    override fun onSaveInstanceState(outState: Bundle) {
        super.onSaveInstanceState(outState)
        outState.putIntArray(KEY_COLORS, colors)
        outState.putSerializable(KEY_SELECTED_COLOR, color)
    }

    companion object {
        private const val KEY_TITLE = "title_text"
        private const val KEY_COLORS = "colors"
        private const val KEY_SELECTED_COLOR = "selected_color"
        private const val KEY_COLUMNS = "columns"
        private const val KEY_SIZE = "size"
        private const val KEY_SHOW_ALPHA = "showAlpha"
        private const val KEY_SHOW_MANAGE = "showManage"
        private const val KEY_DEFAULT_COLOR = "defaultColor"

        private const val PREF_KEY_RECENT_COLORS = "colorpickeropt.recent_colors"
        private const val RECENT_COLOR_COUNT = 4

        /* Default color palette */
        private val DEFAULT_COLORS = intArrayOf(
                0xFFFFFF, /* White */
                0x607D8B, /* Blue Grey */
                0x263238, /* Quantum Dark */
                0x0A0A0A, /* Black */

                0x3F51B5, /* Indigo */
                0x2196F3, /* Blue */
//                0x03A9F4, /* Light Blue */
                0x00BCD4, /* Cyan */

                0x009688, /* Teal */
                0x4CAF50, /* Green */
                0x8BC34A, /* Light Green */
                0xCDDC39, /* Lime */

                0xFFEB3B, /* Yellow */
                0xFFC107, /* Amber */
                0xFF9800, /* Orange */
//                0xFF5722, /* Deep Orange */

                0xF44336, /* Red */
                0xD32F2F, /* Dark Red */
                0xC2185B, /* Dark Pink */
//                0xE91E63, /* Pink */

                0x9C27B0, /* Purple */
                0x7B1FA2, /* Dark Purple */
                0x673AB7 /* Deep Purple */
//                0x512DA8 /* Dark Deep Purple */
        )

        fun newInstance(title: String?, selectedColor: Int, showAlpha: Boolean, showManage: Boolean, defaultColor: Int): ColorPickerBottomSheet {
            return newInstance(title, DEFAULT_COLORS, selectedColor, 4, showAlpha, showManage, defaultColor)
        }

        fun newInstance(title: String?, colors: IntArray, selectedColor: Int,
                        columns: Int, showAlpha: Boolean, showManage: Boolean, defaultColor: Int): ColorPickerBottomSheet {
            val ret = ColorPickerBottomSheet()
            ret.initialize(title, colors, selectedColor, columns, showAlpha, showManage, defaultColor)
            return ret
        }

        private val ALPHA_TO_PROGRESS_LOOKUP_TABLE = intArrayOf(255, 253, 251, 247, 241, 236, 229, 222, 216, 205, 196, 187, 178, 166, 154, 141, 128, 116, 103, 90, 77, 63, 47, 31, 14, 0)

        private fun alphaToProgress(alpha: Int): Int {
            var progress = 0
            var bestMatch = 0
            for (i in ALPHA_TO_PROGRESS_LOOKUP_TABLE.indices) {
                val testAlpha = ALPHA_TO_PROGRESS_LOOKUP_TABLE[i]
                if (testAlpha > alpha) {
                    bestMatch = testAlpha
                    progress = i
                } else {
                    if (Math.abs(testAlpha - alpha) < Math.abs(bestMatch - alpha)) {
                        bestMatch = testAlpha
                        progress = i
                    }
                    break
                }
            }
            return progress
        }

        private fun progressToAlpha(progress: Int): Int {
            return ALPHA_TO_PROGRESS_LOOKUP_TABLE[progress]
        }
    }
}