/*
 * Copyright (C) 2013 The Android Open Source Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.teslacoilsw.launcher.preferences.fancyprefs.color

import android.content.res.Resources
import android.graphics.BitmapShader
import android.graphics.Canvas
import android.graphics.Color
import android.graphics.ColorFilter
import android.graphics.Paint
import android.graphics.PixelFormat
import android.graphics.Rect
import android.graphics.Shader
import android.graphics.drawable.BitmapDrawable
import android.graphics.drawable.Drawable
import com.android.launcher3.LauncherAppState
import com.teslacoilsw.kotlin.resetAfter
import com.teslacoilsw.launcher.preferences.COLOR_AUTO
import com.teslacoilsw.launcher.util.MaterialColors
import com.teslacoilsw.launcher.widget.TypefaceHelper
import com.teslacoilsw.shared.colorpicker.ColorUtil
import com.teslacoilsw.shared.colorpicker.R


enum class PreviewShape {
    Circle,
    Rect
}
/**
 * A drawable which sets its color filter to a color specified by the user, and changes to a
 * slightly darker color when pressed or focused.
 */
class ColorStateDrawable(res: Resources, initialColor: Int, strokeWidthDp: Float = 2f, private val previewShape: PreviewShape = PreviewShape.Circle) : Drawable() {
    var color: Int = Color.TRANSPARENT
        set(color) {
            field = color
            pressedColor = getPressedColor(color)

            setColorForState(state)
        }
    private var pressedColor: Int = 0
    private var strokePaint = Paint(Paint.ANTI_ALIAS_FLAG).apply {
        style = Paint.Style.STROKE
        strokeWidth = res.displayMetrics.density * strokeWidthDp
    }
    private var paint: Paint = Paint(Paint.ANTI_ALIAS_FLAG).apply {
        textSize = res.displayMetrics.density * 16f
        typeface = TypefaceHelper.get(LauncherAppState.getInstance().context, TypefaceHelper.CONDENSED)
    }
    private var alphaPatternPaint: Paint = Paint(Paint.ANTI_ALIAS_FLAG or Paint.FILTER_BITMAP_FLAG)
            .apply {
                val alphaPattern = (res.getDrawable(R.drawable.colorpicker_alpha) as BitmapDrawable).bitmap
                shader = BitmapShader(alphaPattern, Shader.TileMode.REPEAT, Shader.TileMode.REPEAT)
            }
    private var isDisabled = false
    private val checkmark = res.getDrawable(R.drawable.ic_colorpicker_swatch_selected)

    var strokeWidth: Float
        get() = strokePaint.strokeWidth
        set(x) { strokePaint.strokeWidth = x }

    init {
        color = initialColor
    }

    private fun setColorForState(states: IntArray): Boolean {
        var disabled = states.isNotEmpty()
        var pressedOrFocused = false
        var checked = false
        for (s in states) {
            when (s) {
                android.R.attr.state_enabled ->
                    disabled = false
                android.R.attr.state_pressed,
                android.R.attr.state_focused -> pressedOrFocused = true
                android.R.attr.state_checked -> checked = true
            }
        }

        isDisabled = disabled
        val strokeColor = when {
            color == COLOR_AUTO -> ColorUtil.argb(if (disabled) 0x40 else 0xff, 0xA0A0A0)
            Color.alpha(color) < 0xCC -> ColorUtil.argb(0xFF, color)
            ColorUtil.isGray(color) && ColorUtil.isLightColor(color) ->
                ColorUtil.argb(if (disabled) 0x40 else 0x80, 0x808080)
            else -> ColorUtil.argb(if (disabled) 0x40 else 0x80, 0xFFFFFF)
        }

        val paintColor = when {
            disabled -> ColorUtil.multiplyAlpha(0x40, color)
            pressedOrFocused -> pressedColor
            else -> color
        }
        state = states

        if (paint.color != paintColor || strokePaint.color != strokeColor) {
            paint.color = paintColor
            strokePaint.color = strokeColor
            invalidateSelf()
            return true
        }

        return true
    }

    override fun onStateChange(states: IntArray): Boolean {
        return setColorForState(states)
    }

    /**
     * Given a particular color, adjusts its value by overlaying a white (or black) color
     */
    private fun getPressedColor(color: Int): Int {
        val pressedColor = ColorUtil.compositeColors(0x80FFFFFF.toInt(), color)
        val darkPressedColor = ColorUtil.compositeColors(0x0A000000, color)
        val isPressedGood = ColorUtil.calculateContrast(pressedColor, ColorUtil.argb(0xFF, color)) > 1.05f
        return if (isPressedGood) pressedColor else darkPressedColor
    }

    override fun isStateful(): Boolean {
        return true
    }

    override fun getOpacity(): Int {
        return PixelFormat.TRANSLUCENT
    }

    override fun setBounds(left: Int, top: Int, right: Int, bottom: Int) {
        super.setBounds(left, top, right, bottom)
        updateCheckmarkBounds()
    }

    private fun updateCheckmarkBounds() {
        val w = checkmark.intrinsicWidth
        val h = checkmark.intrinsicHeight
        val l = bounds.centerX() - w / 2
        val t = bounds.centerY() - h / 2
        val r = l + w
        val b = l + h
        checkmark.setBounds(l, t, r, b)
    }

    private fun drawShape(canvas: Canvas, bounds: Rect, paint: Paint, inset: Float = 0f) {
        val cx = bounds.centerX()
        val cy = bounds.centerY()
        val r = Math.min(bounds.width(), bounds.height()) / 2

        if (previewShape == PreviewShape.Circle) {
            canvas.drawCircle(cx.toFloat(), cy.toFloat(), r.toFloat() - inset, paint)
        } else {
            canvas.drawRect(bounds.left.toFloat() + inset, bounds.top.toFloat() + inset, bounds.right.toFloat() - inset, bounds.bottom.toFloat() - inset, paint)
        }
    }

    override fun draw(canvas: Canvas) {
        val cx = bounds.centerX()
        val cy = bounds.centerY()

        if (color == COLOR_AUTO) {
            val save = canvas.save()
            canvas.rotate(45f, bounds.centerX().toFloat(), bounds.centerY().toFloat())
            val origColor = paint.color
            val alpha = if (isDisabled) 0x40 else 0xFF
            canvas.resetAfter {
                canvas.clipRect(bounds.left, bounds.top, bounds.right, bounds.height()/2)
                paint.color = ColorUtil.argb(alpha, Color.WHITE)
                drawShape(canvas, bounds, paint, strokePaint.strokeWidth/2)
            }
            canvas.clipRect(bounds.left, bounds.height()/2, bounds.right, bounds.height())
            paint.color = ColorUtil.argb(alpha, MaterialColors.FG_COLOR_LIAM_GRAY_NIGHT)
            drawShape(canvas, bounds, paint, strokePaint.strokeWidth/2)

            canvas.restoreToCount(save)

            paint.color = ColorUtil.argb(alpha, Color.BLACK)
            canvas.drawText("A", cx.toFloat() - paint.measureText("A")/2, cy.toFloat() + paint.measureText("A")/2, paint)
            if (strokePaint.strokeWidth > 0)
                drawShape(canvas, bounds, strokePaint, strokePaint.strokeWidth/2)

            paint.color = origColor
        } else {
            if (Color.alpha(color) < 0xFF)
                drawShape(canvas, bounds, alphaPatternPaint, strokePaint.strokeWidth/2 + 0.6f)
            drawShape(canvas, bounds, paint)
            if (strokePaint.strokeWidth > 0)
                drawShape(canvas, bounds, strokePaint,strokePaint.strokeWidth/2)
        }

        if (state.contains(android.R.attr.state_checked)) {
            checkmark.draw(canvas)
        }
    }

    override fun setAlpha(alpha: Int) {
        paint.alpha = alpha
        alphaPatternPaint.alpha = alpha
        strokePaint.alpha = alpha
    }

    override fun setColorFilter(cf: ColorFilter?) {
        paint.colorFilter = cf
        alphaPatternPaint.colorFilter = cf
        strokePaint.colorFilter = cf
    }

}
