/*
 * Copyright (C) 2013 The Android Open Source Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.teslacoilsw.launcher.preferences.fancyprefs.color

import android.content.Context
import android.graphics.Canvas
import android.graphics.Color
import android.graphics.drawable.Drawable
import androidx.annotation.ColorInt
import androidx.annotation.NonNull
import android.util.AttributeSet
import android.widget.RadioButton
import com.teslacoilsw.launcher.R

/**
 * Creates a circular swatch of a specified color.  Adds a checkmark if marked as checked.
 */
class ColorSwatchRadioButton(context: Context, attrs: AttributeSet?) : RadioButton(context, attrs, 0, R.style.ColorSwatchRadioButton) {

    private val colorStateDrawable: ColorStateDrawable = ColorStateDrawable(resources, Color.WHITE)
    var color: Int
        @ColorInt get() = colorStateDrawable.color
        set(x) {
            colorStateDrawable.color = x
        }

    init {
        setWillNotDraw(false)
        val a = context.obtainStyledAttributes(attrs, R.styleable.ColorSwatchRadioButton)
        try {
            color = a.getColor(R.styleable.ColorSwatchRadioButton_android_color, Color.WHITE)
        } finally {
            a.recycle()
        }
        textSize = 0f
        compoundDrawablePadding = 0
    }

    override fun onMeasure(widthMeasureSpec: Int, heightMeasureSpec:Int) {
        super.onMeasure(widthMeasureSpec, heightMeasureSpec)

        val availableWidth = measuredWidth - paddingLeft - paddingRight
        val availableHeight = measuredHeight - paddingTop - paddingBottom

        val centerX = availableWidth/2 + paddingLeft
        val centerY = availableHeight/2 + paddingTop
        val size = Math.min(availableWidth, availableHeight)

        val left = centerX - size/2
        val top = centerY - size/2
        val right = left + size
        val bottom = top + size
        colorStateDrawable.setBounds(left, top, right, bottom)
    }

    override fun draw(canvas: Canvas) {
        super.draw(canvas)
        colorStateDrawable.draw(canvas)
    }

    override fun drawableStateChanged() {
        super.drawableStateChanged()

        if (colorStateDrawable.setState(drawableState)) {
            invalidate()
//            invalidateDrawable(colorStateDrawable)
        }
    }

    override fun verifyDrawable(@NonNull who: Drawable): Boolean {
        return super.verifyDrawable(who) || who === colorStateDrawable
    }

}
