package com.teslacoilsw.launcher.preferences.fragments

import android.Manifest
import android.annotation.SuppressLint
import android.annotation.TargetApi
import android.app.Activity
import android.app.Dialog
import android.content.ActivityNotFoundException
import android.content.Context
import android.content.Intent
import android.content.pm.PackageManager
import android.os.Build
import android.os.Bundle
import android.os.SystemClock
import android.text.TextUtils
import android.text.format.DateFormat
import android.view.LayoutInflater
import android.view.View
import android.view.ViewGroup
import android.widget.AdapterView
import android.widget.EditText
import android.widget.Spinner
import android.widget.Toast
import androidx.annotation.NonNull
import androidx.core.content.ContextCompat
import com.afollestad.materialdialogs.MaterialDialog
import com.android.launcher3.LauncherAppState
import com.android.launcher3.Utilities
import com.teslacoilsw.kotlin.bind
import com.teslacoilsw.launcher.R
import com.teslacoilsw.launcher.backup.BackupCategory
import com.teslacoilsw.launcher.backup.BackupCategoryLocal
import com.teslacoilsw.launcher.backup.BackupHelper
import com.teslacoilsw.launcher.backup.BackupLocationAdapter
import com.teslacoilsw.launcher.backup.BackupRestoreAdapter
import com.teslacoilsw.launcher.databinding.SettingsBackupBinding
import com.teslacoilsw.launcher.importer.LauncherImportActivity
import com.teslacoilsw.launcher.onboarding.OnboardingActivity
import com.teslacoilsw.launcher.preferences.Pref
import com.teslacoilsw.launcher.util.OnActivityResultHandler
import com.teslacoilsw.launcher.util.OnPermissionGrantedHandler
import com.teslacoilsw.launcher.widget.IsMeasuringListView
import timber.log.Timber
import java.util.*

class SettingsBackup : NovaSettingsFragment(), OnActivityResultHandler.ActivityForResultDispatcher, OnPermissionGrantedHandler.RequestPermissionDispatcher {
    private val PERMISSION_REQUEST_WRITE_STORAGE = 0
    private val PERMISSION_REQUEST_WRITE_STORAGE_FOR_ACTIVITY_RESULT = 1

    private var pendingBackupCategory: BackupCategory? = null
    private var pendingName: String? = null
    private var pendingData: Intent? = null
    private lateinit var binding: SettingsBackupBinding

    private var activityResultHandler: OnActivityResultHandler? = null
    private var requestPermissionHandler: OnPermissionGrantedHandler? = null

    override fun onCreateView(inflater: LayoutInflater, container: ViewGroup, savedInstanceState: Bundle?): View {
        binding = inflater.bind(R.layout.settings_backup, container, false)

        binding.backup.setOnClickListener {
            doBackup(it.context)
        }
        binding.importLauncher.setOnClickListener { doImport() }
        binding.restore.setOnClickListener { showRestoreDialog() }

        binding.onboarding.setOnClickListener {
            val intent = Intent(activity, OnboardingActivity::class.java) //PresetsActivity.makeIntent()
            intent.putExtra("allow_skipping_steps", true)
            startActivity(intent)
        }
        binding.resetToDefaults.setOnClickListener {
            val md = MaterialDialog.Builder(activity)
                    .title(R.string.delete_current_layout)
                    .content(R.string.replace_with_default)
                    .positiveText(R.string.reset)
                    .negativeText(R.string.cancel)
                    .callback(object : MaterialDialog.ButtonCallback() {
                        override fun onPositive(dialog: MaterialDialog?) {
                            val lp = LauncherAppState.getLauncherProvider()
                            Pref.Value.mPrefs.unregisterOnSharedPreferenceChangeListener(Pref.Value)
                            lp?.closeHard()
                            Pref.Value.mPrefs.edit().clear().apply()

                            activity.getDatabasePath("launcher.db").delete()
                            activity.getDatabasePath("launcher.db-journal").delete()
                            SystemClock.sleep(200)
                            System.exit(0)
                        }
                    }).build()
            md.show()
        }
        return binding.root
    }

    private fun doImport() {
        val d = MaterialDialog.Builder(activity)
                .title(R.string.warning)
                .content(getString(R.string.backup_erase_layout))
                .negativeText(R.string.cancel)
                .positiveText(R.string.ok)
                .onPositive { _, _ ->
                    val intent = Intent(activity, LauncherImportActivity::class.java)
                    intent.flags = Intent.FLAG_ACTIVITY_CLEAR_TASK or Intent.FLAG_ACTIVITY_NEW_TASK
                    startActivity(intent)
                }.build()
        d.show()
    }

    private fun doBackup(context: Context) {
        val filename = DateFormat.format("yyyy-MM-dd_kk-mm", Calendar.getInstance())

        @SuppressLint("InflateParams")
        val view = LayoutInflater.from(context).inflate(R.layout.dialog_create_backup, null)

        val spinner = view.findViewById<View>(R.id.type_spinner) as Spinner
        val adapter = BackupLocationAdapter(context)
        spinner.adapter = adapter
        val pos = adapter.getPosForEnum(Pref.Value.mPrefs.getString("backup_location", ""))
        spinner.setSelection(pos)

        val input = view.findViewById<View>(R.id.title) as EditText
        input.setText(filename)
        val d = MaterialDialog.Builder(context)
                .title(R.string.preference_backup)
                .customView(view, false)
                .positiveText(R.string.ok)
                .negativeText(R.string.cancel)
                .callback(object : MaterialDialog.ButtonCallback() {
                    override fun onPositive(materialDialog: MaterialDialog?) {
                        val name = input.text.toString()
                        if (!TextUtils.isEmpty(name)) {
                            val bc = adapter.getItem(spinner.selectedItemPosition)
                            if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.M && bc is BackupCategoryLocal) {
                                if (ContextCompat.checkSelfPermission(getContext(), Manifest.permission.WRITE_EXTERNAL_STORAGE) != PackageManager.PERMISSION_GRANTED) {
                                    pendingBackupCategory = bc
                                    pendingName = name
                                    requestPermissions(arrayOf(Manifest.permission.WRITE_EXTERNAL_STORAGE), PERMISSION_REQUEST_WRITE_STORAGE)
                                    return
                                }
                            }
                            Pref.Value.mPrefs.edit().putString("backup_location", bc.enumName).apply()

                            bc.createBackup(this@SettingsBackup, name)
                            Toast.makeText(activity, "Backup created: $name", Toast.LENGTH_SHORT).show()
                        }
                    }
                }).build()
        d.show()
    }

    @TargetApi(Build.VERSION_CODES.M)
    override fun requestPermissions(@NonNull permissions: Array<String>, handler: OnPermissionGrantedHandler) {
        requestPermissions(permissions, OnPermissionGrantedHandler.REQUEST_CODE)
        requestPermissionHandler = handler
    }

    override fun onRequestPermissionsResult(requestCode: Int, permissions: Array<String>,
                                            grantResults: IntArray) {
        Timber.tag(NovaPreferenceFragment.TAG).v(
                "onRequestPermissionResult %d %s %s", requestCode,
                Arrays.toString(permissions), Arrays.toString(grantResults))
        if (requestCode == OnPermissionGrantedHandler.REQUEST_CODE) {
            if (requestPermissionHandler?.onRequestPermissionsResult(permissions, grantResults) == true) {
                requestPermissionHandler = null
            }
        } else if (requestCode == PERMISSION_REQUEST_WRITE_STORAGE) {
            if (grantResults[0] == PackageManager.PERMISSION_GRANTED) {
                Pref.Value.mPrefs.edit().putString("backup_location", pendingBackupCategory?.enumName).apply()
                pendingBackupCategory!!.createBackup(this@SettingsBackup, pendingName)
                Toast.makeText(activity, "Backup created: $pendingName", Toast.LENGTH_SHORT).show()
            } else {
                Toast.makeText(activity, "Permission denied", Toast.LENGTH_SHORT).show()
            }
        } else if (requestCode == PERMISSION_REQUEST_WRITE_STORAGE_FOR_ACTIVITY_RESULT) {
            if (activityResultHandler?.onActivityResult(activity, Activity.RESULT_OK, pendingData) == true) {
                activityResultHandler = null
            }
            pendingData = null
        }
    }


    override fun startActivityForResult(data: Intent, handler: OnActivityResultHandler) {
        try {
            activityResultHandler = handler
            startActivityForResult(data, OnActivityResultHandler.REQUEST_CODE)
        } catch (ex: ActivityNotFoundException) {
            Toast.makeText(activity, R.string.activity_not_found, Toast.LENGTH_SHORT).show()
            handler.onActivityResult(activity, Activity.RESULT_CANCELED, null)
        } catch (ex: Throwable) {
            Toast.makeText(activity, R.string.activity_not_found, Toast.LENGTH_SHORT).show()
            handler.onActivityResult(activity, Activity.RESULT_CANCELED, null)
        }

    }


    @SuppressLint("NewApi")
    override fun onActivityResult(requestCode: Int, resultCode: Int, data: Intent?) {
        super.onActivityResult(requestCode, resultCode, data)

        if ((resultCode == Activity.RESULT_OK && data?.data != null &&
             Build.VERSION.SDK_INT >= Build.VERSION_CODES.M &&
             Utilities.isFileOrSdcardUri(data.data) &&
             ContextCompat.checkSelfPermission(activity,
                    Manifest.permission.WRITE_EXTERNAL_STORAGE) != PackageManager.PERMISSION_GRANTED))
        {
            requestPermissions(arrayOf(Manifest.permission.WRITE_EXTERNAL_STORAGE),
                    PERMISSION_REQUEST_WRITE_STORAGE_FOR_ACTIVITY_RESULT)
            pendingData = data
            return
        }

        if (activityResultHandler?.onActivityResult(activity, resultCode, data) == true) {
            activityResultHandler = null
        }
    }

    fun showRestoreDialog() {
        showRestoreDialog(this)
    }

    companion object {
        @JvmStatic
        fun showRestoreDialog(dispatcher: OnPermissionGrantedHandler.RequestPermissionDispatcher) {
            val context = dispatcher.context
            val lv = IsMeasuringListView(context, null)
            val adapter = BackupRestoreAdapter(dispatcher)
            lv.adapter = adapter
            lv.onItemClickListener = AdapterView.OnItemClickListener { parent, view, position, id ->
                val item = adapter.getItem(position)
                if (item is BackupRestoreAdapter.BackupIntent) {
                    val intent = item.restoreIntent
                    dispatcher.startActivityForResult(intent, item.resultHandler)
                } else if (item is BackupRestoreAdapter.BackupFile) {
                    getRestoreConfirmDialog(context, item).show()
                }
            }
            val maindialog = MaterialDialog.Builder(context)
                    .title(R.string.preference_restore_backups)
                    .customView(lv, false)
                    .build()
            maindialog.show()
        }

        @JvmStatic
        fun getRestoreConfirmDialog(context: Context, file: BackupRestoreAdapter.BackupFile): Dialog {
            return MaterialDialog.Builder(context)
                    .title(R.string.delete_current_layout)
                    .content(R.string.replace_from_backup)
                    .negativeText(R.string.cancel)
                    .positiveText(R.string.ok)
                    .callback(object : MaterialDialog.ButtonCallback() {
                        override fun onPositive(dialog: MaterialDialog?) {
                            file.doRestore(context)
                            BackupHelper.restartLauncher(context)
                        }
                    }).build()
        }
    }
}