package com.teslacoilsw.launcher.search

import java.text.Normalizer
import java.util.*


private val Int.C: Char
    get() = this.toChar()

/* Based on https://en.wikipedia.org/wiki/Korean_language_and_computers#Hangul_in_Unicode */

private const val JAMO_FIRST = 0x1100.toChar() /* ᄀ */
private const val JAMO_LAST = 0x1112.toChar() /* ᄒ	 */
private const val JAMO_STANDALONE_FIRST = 0x3131.toChar() /* ㄱ */
private const val JAMO_STANDALONE_LAST = 0x314E.toChar() /* ㅎ */
private const val HANGUL_SYLLABLE_FIRST = 0xAC00.toChar() /* 가 */
private const val HANGUL_SYLLABLE_LAST = 0xAE4B.toChar() /* 깋 */
private const val HANGUL_SYLLABLES_PER_JAMO = HANGUL_SYLLABLE_LAST - HANGUL_SYLLABLE_FIRST + 1
private val KOREAN_LETTER_TO_JAMO = mapOf(
        /* ㄱ */ 0x3131.C to 0x1100.C,
        /* ㄴ */ 0x3134.C to 0x1102.C,
        /* ㄷ */ 0x3137.C to 0x1103.C,
        /* ㄹ */ 0x3139.C to 0x1105.C,
        /* ㅁ */ 0x3141.C to 0x1106.C,
        /* ㅂ */ 0x3142.C to 0x1107.C,
        /* ㅅ */ 0x3145.C to 0x1109.C,
        /* ㅇ */ 0x3147.C to 0x110B.C,
        /* ㅈ */ 0x3148.C to 0x110C.C,
        /* ㅊ */ 0x314A.C to 0x110E.C,
        /* ㅋ */ 0x314B.C to 0x110F.C,
        /* ㅌ */ 0x314C.C to 0x1110.C,
        /* ㅍ */ 0x314D.C to 0x1111.C,
        /* ㅎ */ 0x314E.C to 0x1112.C
)


private fun Char.jamoAwareMatch(other: Char): Boolean {
    if (this == other) return true

    /* Handle åçcéñt chäřêçtërs */
    val normalized = Normalizer.normalize(other.toString(), Normalizer.Form.NFD)
    if (!normalized.isEmpty() && this == normalized[0])
        return true

    val jamo = when {
        this in JAMO_FIRST..JAMO_LAST -> this
        this in JAMO_STANDALONE_FIRST..JAMO_STANDALONE_LAST -> KOREAN_LETTER_TO_JAMO[this] ?: 0.C
        else -> 0.C
    }

    if (jamo != 0.C) {
        return ((other - HANGUL_SYLLABLE_FIRST) / HANGUL_SYLLABLES_PER_JAMO) + JAMO_FIRST.toInt() == jamo.toInt()
    }
    return false
}

data class MatchState(var titleIndex: Int = -1, var queryIndex: Int = 0, var lastWasNotLetter: Boolean = true, var lastWasUpper: Int = 0, var lastWasMatch: Boolean = false, var nonUpperSinceLastUpper: Boolean = false, var matchStrength: Int = Int.MAX_VALUE)
/* _N_ova _L_auncher
 * _N_ova_L_auncher
 * _N_OVA_L_auncher
 * _N_OVA_l_auncher
 * _N_ova3_l_auncher
 * _N_ova._l_auncher
 * _n_ova_L_auncher
 * _N__l_auncher (okay for lower consonant, not vowel?)
 * _
 */

/* Returns match strength
 * 0 No match
 * 1 Weak match
 * 2 less weak match
 * ...
 *
 * Used for sorting
*/
fun match(title: CharSequence, query: String): Int {
    val matches = ArrayList<MatchState>()

    /* First find all possible starting matches (To handle cases like
     * Amazon Alexa where we possibly need to match either "Amazon" or "Alexa" */
    // TODO Fix "Amal" should match "Amazon Alexa"
    var initialState = MatchState()
    title.forEachIndexed {i, c ->
        if (matchChar(initialState, query, c, i == 0)) {
            initialState.titleIndex = i
            matches.add(initialState)
            initialState = MatchState()
        }
    }

    if (matches.size == 0)
        return 0

    title.forEachIndexed { i, c ->
        for (m in matches) {
            if (m.queryIndex >= query.length || m.titleIndex >= i) {
                continue
            }

            matchChar(m, query, c, i == 0)
        }
    }

    return matches.filter { it.queryIndex >= query.length }.maxBy { initialState.matchStrength }?.matchStrength ?: 0
}

/**
 * Return true if should increment queryIndex
 */
private fun matchChar(state: MatchState, query: String, c: Char, first: Boolean): Boolean {
    if (state.queryIndex >= query.length)
        return false

    if (c.isWhitespace()) {
        state.nonUpperSinceLastUpper = true
        state.lastWasNotLetter = true
        state.lastWasUpper = 0
        return false
    }

    val isNotLower = !c.isLowerCase()
    val isNotLetter = !c.isLetter()
    state.nonUpperSinceLastUpper = state.nonUpperSinceLastUpper || !c.isUpperCase()

    /* Query: no, queryIndex 1 char o, matching Nova Launcher */
    val isContinuedWord = state.lastWasMatch
    /* Query: nl, queryIndex 1, matching Nova Launcher */
    val isNewWord = first || (isNotLower && state.nonUpperSinceLastUpper) || isNotLetter || state.lastWasNotLetter
    /* Query nl, queryIndex 1, matching NOVAlauncher, query nv, doesn't match NOVA */
    val isLowerAfterUpper = state.nonUpperSinceLastUpper and (state.lastWasUpper > 1)
    /* I can imagine a case of NOVALauncher however I don't see many apps doing that */

    var hasMatch = false
    if (isContinuedWord || isNewWord || isLowerAfterUpper) {
        if (query[state.queryIndex].jamoAwareMatch(c.toLowerCase())) {
            state.matchStrength -=
                    if (isContinuedWord) {
                        0
                    } else if (isNewWord) {
                        if (first) 0 else 1
                    } else {
                        8
                    }
            state.lastWasMatch = true
            hasMatch = true
        }
    }

    if (c.isUpperCase()) {
        state.lastWasUpper++
        state.nonUpperSinceLastUpper = false
    } else {
        state.lastWasUpper = 0
    }
    state.lastWasNotLetter = isNotLetter

    if (!hasMatch) {
        if (state.lastWasMatch && query[state.queryIndex].jamoAwareMatch(c.toLowerCase())) {
            state.lastWasMatch = true
            hasMatch = true
        } else {
            state.lastWasMatch = false
        }
    }

    if (hasMatch)
        state.queryIndex++

    return hasMatch
}
