package com.teslacoilsw.launcher.theme

import android.app.Activity
import android.content.Context
import android.content.Intent
import android.content.Loader
import android.graphics.Bitmap
import android.os.Bundle
import android.view.View
import android.widget.AdapterView
import com.afollestad.materialdialogs.MaterialDialog
import com.android.launcher3.FastBitmapDrawable
import com.android.launcher3.IconCache
import com.teslacoilsw.kotlin.asBitmap
import com.teslacoilsw.kotlin.isMainIntent
import com.teslacoilsw.kotlin.lazy1
import com.teslacoilsw.launcher.EditShortcutDialogFragment
import com.teslacoilsw.launcher.NovaLauncher
import com.teslacoilsw.launcher.R
import com.teslacoilsw.launcher.UiLock2
import com.teslacoilsw.launcher.bitmaputils.BitmapManager
import com.teslacoilsw.launcher.bitmaputils.BitmapUtils
import com.teslacoilsw.launcher.preferences.Pref
import com.teslacoilsw.launcher.theme.ThemeImagePicker.*
import com.teslacoilsw.launcher.theme.element.ThemableElement
import org.icontheme.api.IconSource
import org.icontheme.api.ThemeConfig
import org.icontheme.api.ThemeProvider
import java.util.concurrent.atomic.AtomicBoolean

class IconThemeOrgIconPicker : AbstractThumbnailPicker() {

    private val themePackage by lazy1 { arguments.getString(ARG_PACKAGE_NAME) ?: "" }
    private val targetIntent: Intent? by lazy1 { arguments.getParcelable<Intent>(EditShortcutDialogFragment.KEY_DEFAULT_ICON_INTENT) }
    private val iconSize by lazy1 { arguments.getInt(PickerActivity.EXTRA_ICON_SIZE, ThemableElement.ICON.getPreviewSize(resources)) }
    private val iconDpi by lazy1 { Pref.Key.AppIconSize.calculateIconDpi(iconSize) }
    private var themeHeaderInfo = ThemeHeaderInfo()
    private val themeConfig by lazy1 { arguments.getParcelable<ThemeConfig>("themeConfig")!! }
    private val densityDpi by lazy1 { activity.resources.displayMetrics.densityDpi }

    override fun onCreate(savedInstanceState: Bundle?) {
        super.onCreate(savedInstanceState)

        themeHeaderInfo.name = arguments.getString(AbstractThumbnailPicker.ARG_THEME_TITLE)
        themeHeaderInfo.iconResId = arguments.getInt(AbstractThumbnailPicker.ARG_THEME_ICON_RESID)
        mInternalPicker = arguments.getParcelable(ARG_INTERNAL_PICKER)
        mPreviewSize = iconSize

        if (mInternalPicker != null && iconSize != 0) {
            mInternalPicker.putExtra(PickerActivity.EXTRA_ICON_SIZE, iconSize)
        }
    }

    override fun onCreateLoader(id: Int, args: Bundle?): Loader<MutableList<out ThumbnailInfo>> {
        return IconThemeOrgLoader(activity, themeConfig, iconDpi, iconSize, targetIntent)
    }

    private val defaultBitmaps = mutableListOf<Bitmap>()
    override fun onLoadFinished(loader: Loader<List<AbstractThumbnailPicker.ThumbnailInfo>>, data: List<AbstractThumbnailPicker.ThumbnailInfo>) {
        defaultBitmaps.clear()
        if (data is SectionedThumbnailList) {
            themeHeaderInfo = data.headerInfo
            defaultBitmaps.addAll(data.defaultBitmaps)
        }

        super.onLoadFinished(loader, data)
        bindHeader()
    }

    override fun canReleaseBitmap(bmp: Bitmap): Boolean {
        return if (!super.canReleaseBitmap(bmp)) false else !defaultBitmaps.contains(bmp)
    }

    override fun shouldShowColorChanger(): Boolean = false

    override fun loadBitmap(manager: BitmapManager, info: ThumbnailInfo, canceled: AtomicBoolean): Bitmap? {
        if (info is IconSourceThumbnailInfo) {
            val d = info.iconSource.loadDrawable(activity)
            if (!canceled.get())
                return d.asBitmap(iconSize)
        }
        return null
    }

    companion object {
        @JvmStatic
        fun newInstance(info: ThemeListFragment.ThemePackageInfo, themeConfig: ThemeConfig, targetIntent: Intent): IconThemeOrgIconPicker {
            val frag = IconThemeOrgIconPicker()

            val args = Bundle()
            args.putString(ARG_PACKAGE_NAME, info.packageName)
            args.putInt(ARG_THEME_TYPE, ThemableElement.ICON.ordinal)
            args.putString(AbstractThumbnailPicker.ARG_THEME_TITLE, info.displayLabel.toString())
            args.putInt(AbstractThumbnailPicker.ARG_THEME_ICON_RESID, info.displayIconResId)
            args.putParcelable(AbstractThumbnailPicker.ARG_THEME_PACKAGE_INFO, info)
            args.putParcelable("themeConfig", themeConfig)
            args.putParcelable(EditShortcutDialogFragment.KEY_DEFAULT_ICON_INTENT, targetIntent)
            frag.arguments = args

            return frag
        }
    }


    override fun onItemClick(parent: AdapterView<*>, view: View, position: Int, id: Long) {
        val item = mAdapter.getItem(position)
        if (item is BitmapThumbnailInfo) {
            val intent = Intent()
            intent.data = AbstractThumbnailPicker.returnBitmapAsUri(activity, item.bmp)
            activity.setResult(Activity.RESULT_OK, intent)
            activity.finish()
        } else if (item is IconSourceThumbnailInfo) {
            picker.onResult(item.iconSource, "IconThemeOrg")
        }
    }
}

private class IconThemeOrgLoader(context: Context, private val themeConfig: ThemeConfig, private val iconDpi: Int, private val iconSize: Int, private val targetIntent: Intent?) : AbstractThumbnailPicker.ThumbnailListLoader(context) {

    override fun loadInBackground(): MutableList<AbstractThumbnailPicker.ThumbnailInfo>? {

        val cr = context.contentResolver

        var preloadDrawable: FastBitmapDrawable? = null
        val iconCache = IconCache(mContext, IconCache.IconConfigStatic(iconDpi, iconSize, "XX:ICONTHEMEORG:${themeConfig.id}"), null)
        val iconTheme = iconCache.iconTheme

        if (iconTheme.canGenerateIcon()) {
            val input = BitmapUtils.getBitmap(mContext.resources, R.drawable.ic_iconpack_preload_base, iconDpi, iconSize, iconSize, null)
            val iconback = iconCache.iconTheme.getThemedIconFromBase(mContext, NovaLauncher.COMPONENT_NAME, input)
            if (iconback != null && iconback != input) {
                preloadDrawable = FastBitmapDrawable(iconback)
                preloadDrawable.alpha = 128
                preloadDrawable.setIsDisabled(true)
            }
        }

        val list = ThemeImagePicker.SectionedThumbnailList()

        if (targetIntent?.isMainIntent() == true) {
            val defaultIcons = themeConfig.getIcons(cr, targetIntent.component)
            if (!defaultIcons.isEmpty()) {
                list.addSection("Default")
                list.addAll(defaultIcons.map { IconSourceThumbnailInfo(it) })
            }
        }
        list.preloadDrawable = preloadDrawable
        val pickerIcons = themeConfig.getPickerIcons(cr)
        pickerIcons.forEach {
            list.addSection(it.key)
            list.addAll(it.value.map { IconSourceThumbnailInfo(it) })
        }

        /* Wait for the animation before returning, as when we return we'll load bitmaps */
        try {
            UiLock2.waitForIdle("ThemePicker loadInBackground")
        } catch (e1: InterruptedException) {
            return null
        }

        return list
    }

}

private class IconSourceThumbnailInfo(val iconSource: IconSource) : AbstractThumbnailPicker.ThumbnailInfo() {
    override fun getTitle(): CharSequence? = null
}

fun ThemeProvider.showSelectDialog(context: Context, listener: (ThemeConfig?) -> Unit) {
    val subthemes = getThemes(context.contentResolver)
    val label = getLabel(context.applicationContext.packageManager)

    when {
        subthemes.size > 1 -> {
            val labels = arrayOfNulls<String>(subthemes.size)
            for (i in labels.indices) {
                labels[i] = subthemes[i].label
            }
            MaterialDialog.Builder(context)
                    .title(label.toString() + " Themes")
                    .items(*labels)
                    .itemsCallback { dialog, _, which, _ ->
                        dialog.dismiss()
                        listener.invoke(subthemes[which])
                    }.show()
        }
        subthemes.size == 1 -> listener.invoke(subthemes[0])
        else -> listener.invoke(null)
    }
}